/*
 * daphne.cpp
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// DAPHNE : Dragon's Lair / Space Ace Emulator
// Started by Matt Ownby, many people have contributed since then
// Begun August 3rd, 1999

// ORIGINAL GOAL (we've obviously done much more than this now hehe):

// To emulate the original Dragon's Lair / Space Ace game such that the original ROM's are executed
// properly, the joystick and buttons are emulated, and the Laserdisc Player is properly
// controlled
// In short, so that the original game may be played with the following requirements:
// - Original Dragon's Lair Laserdisc
// - Original Dragon's Lair ROM's (disk files, not the physical ROM's heh)
// - A laserdisc player
// - A common modern computer (i386 Win32, i386 Linux most likely)
// No other original hardware (motherboard, power supply, harness) required

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL_main.h>

#ifdef LINUX
#include <unistd.h>
//#include <signal.h>
#endif

#include "io/input.h"
#include "daphne.h"
#include "timer/timer.h"
#include "io/serial.h"
#include "sound/sound.h"
#include "io/conout.h"
#include "io/conin.h"
#include "io/cmdline.h"
#include "video/video.h"
#include "video/led.h"
#include "ldp-out/ldp.h"
#include "video/SDL_Console.h"
#include "io/error.h"
#include "cpu/cpu-debug.h"
#include "game/game.h"

#include "globals.h"
// some global data is stored in this file

// -------------------------------------------------------------------------------------------------

// adds a string to a log file (creates the logfile if it does not exist)
void addlog(char *s)
{

	FILE *F = 0;
	unsigned int written = 0;	// # of unsigned chars written

	F = fopen(LOGNAME, "ab");	// open for appending
	if (F)
	{
		written = fprintf(F, "%s", s);
		if (written != strlen(s))
		{
#ifdef LINUX
			printf("Cannot write to logfile, do you have write permissions?\n");
#endif
		}
		fclose(F);	// close each time in case we get in an endless loop ...
	}
	// else directory is read-only so we will just ignore for now

	/*
	else
	{
		printline("Could not open log file!");
	}
	*/
}





////////////////////////////////////////

// loads size # of bytes from filename into buf
// returns true if successful, or false if there was an error
bool load_rom(char *filename, Uint8 *buf, Uint32 size)
{
	FILE *F = NULL;
	Uint32 bytes_read = 0;
	bool result = false;
	char fullpath[81] = { "roms/" }; // pathname to roms directory
	char s[81] = { 0 };

	strcat(fullpath, filename);	// append filename to roms directory	
	outstr("Loading ");
	outstr(fullpath);
	printline(" ...");
	F = fopen(fullpath, "rb");
	
	// if file was opened successfully
	if (F)
	{
		bytes_read = fread(buf, 1, size, F);
		
		// if we read in the # of bytes we expected to
		if (bytes_read == size)
		{
			result = true;
		}
		// notify the user what the problem is
		else
		{
			sprintf(s, "error in rom_load: expected %u bytes but only read %u", size, bytes_read);
			printline(s);
		}
		fclose(F);
	}
	
	sprintf(s, "%d bytes read into memory", bytes_read);
	printline(s);
	
	return(result);
	
}

unsigned char get_filename(char *s, unsigned char n)
// prepares a filename using any wildcards we may have
// returns filename in s
// returns 1 if this filename contains wild characters
// returns 0 if this filename does not contain any wild characters
{

	unsigned int i = 0;
	unsigned char result = 0;

	for (i = 0; i < strlen(s); i++)
	{
		if (s[i] == '+')
		{
			result = 1;
			s[i] = (char) (n + 0x30);	// convert wildcard to number
		}
	}

	return(result);

}

void set_quitflag()
// sets the quit flag
{

	quitflag = 1;

}

unsigned char get_quitflag()
// returns the quit flag
{

	return(quitflag);

}




// the main function for both Windows and Linux <grin>
int main(int argc, char **argv)
{

	char cmd_line[160] = { 0 };	// maximum size of command line we'll accept
	int i = 0;

	delete_logfile();

	// put command line into 'cmd_line' so we can process it
	for (i = 1; i < argc; i++)
	{
		// if adding the next command line option won't overflow our buffer
		if (strlen(cmd_line) + strlen(argv[i]) < 160)
		{
			strcat(cmd_line, argv[i]);
			strcat(cmd_line, " ");
		}
		else
		{
			printerror("Command line is too long!");
			break;
		}
	}

	// parse the command line (which allocates game and ldp) and continue if no errors
	// this is important!  if game_type or ldp_type fails to allocate g_game and g_ldp,
	// then the program will segfault and daphne must NEVER segfault!  hehe
	if (parse_game_type(cmd_line) && parse_ldp_type(cmd_line) && parse_cmd_line(cmd_line))
	{

		// MATT : we have to wait until after the command line is parsed before we do anything with the LEDs in case the
		// user does not enable them
		remember_leds(); // memorizes the status of keyboard leds
		change_led(false, false, false); // turns all keyboard leds off

  	  // if the display initialized properly
	  if (load_bmps() && init_display())
	  {
		if (sound_init())
		{
			if (SDL_input_init())
			{
				// if the roms were loaded successfully
				if (g_game->load_roms())
				{
				  // if the video was initialized successfully
				  if (g_game->video_init())
				  {

					// if the game has some problems, notify the user before the user plays the game
					if (g_game->get_issues())
					{
						printnowookin(g_game->get_issues());
					}

				  	// if the laserdisc player was initialized properly
					if (g_ldp->init())
					{
						g_game->cpu_init();	// initalize the cpu
						printline("Booting ROM ...");
						g_game->cpu_execute();	// HERE IS THE MAIN LOOP RIGHT HERE
						g_game->cpu_shutdown();
						g_ldp->shutdown();
					}
					else
					{
						printerror("Could not initialized laserdisc player!");
					}
				  }
				  else
				  {
						printerror("Game-specific video initialization failed!");
				  }
				}
				else
				{
					printerror("Could not load ROM images! Did you download them and name them correctly?");
				}
				SDL_input_shutdown();
			}
			else
			{
				printerror("Could not initialize keyboard!");
			}
			sound_shutdown();
		}
		else
		{
			printerror("Sound initialization failed!");
		}
		shutdown_display();	// shut down the display
	  } // end init display
	  else
	  {
		printerror("Video initialization failed!");
	  }
	} // end game class allocated properly

	// if command line was bogus, quit
	else
	{
		printline("ERROR: Bad command line or could not allocate game classes.");
		printline("To run DAPHNE, you must specify which game to run and which player you are using.");
		printline("For example, try 'daphne lair noldp' to run Dragon's Lair without a laserdisc player.");
	}

	// if our g_game class was allocated
	if (g_game)
	{
		delete(g_game);
	}

	// if g_ldp was allocated, de-allocate it
	if (g_ldp)
	{
		delete(g_ldp);
	}

	free_bmps();	// always do this no matter what
    
	restore_leds();  // sets keyboard leds back how they were (this is safe even if we have the led's disabled)
	
	SDL_Quit();
	exit(0);

}


// sets the serial port to be used to control LDP
void set_serial_port(unsigned char i)
{

	serial_port = i;

}


unsigned char get_serial_port()
{

	return(serial_port);

}

void set_baud_rate(int i)
{

	baud_rate = i;

}

int get_baud_rate()
{

	return(baud_rate);

}

void set_search_offset(int i)
{

	search_offset = i;

}

int get_search_offset()
{

	return(search_offset);

}






// we call this if we have detected a joystick
void set_joystick_present()
{

	joystick_detected = 1;

}

unsigned char get_joystick_present()
{

	return(joystick_detected);

}

void set_sound_enabled (unsigned char value)
{
	sound_enabled = value;
}

unsigned char get_sound_enabled()
{
	return sound_enabled;
}

unsigned char get_frame_modifier()
{
	return(frame_modifier);
}

void set_frame_modifier(unsigned char value)
{
	frame_modifier = value;
}

void set_scoreboard(unsigned char value)
{
	realscoreboard = value;
}

unsigned char get_scoreboard()
{
	return (realscoreboard);
}

void set_scoreboard_port(unsigned char value)
{
	rsb_port = value;
}

unsigned char get_scoreboard_port()
{

	return(rsb_port);

}

void delete_logfile()
{

	if (unlink(LOGNAME) == -1)
	{
#ifdef LINUX
		printf("Could not delete the old logfile! (it might not exist if this is your first time running DAPHNE)\n");
		// assume unix if we can't delete logfile
#endif
	}
}

// computes a checksum using a generic, weak algorithm
// buf is where in memory to compute the checksum, buf_length is how long the piece of memory is
unsigned int compute_checksum(unsigned char *buf, unsigned int buf_length)
{
	Uint32 checksum = 0;
	Uint32 index = 0;

	while (index < buf_length)
	{
		checksum ^= (buf[index] << (index % 24));
		index++;
	}

	return checksum;
}