/*
 * astron.cpp
 *
 * Copyright (C) 2001 Mark Broadhead
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// astron.cpp
// by Mark Broadhead
//
// Pioneer Astron Belt Hardware
//
// There are 2 different versions of this hardware. One that uses the Pioneer
// LD-V1000 and another that uses the Sega / Hitachi VIP9500SG.
//
// Games include:
//
// Astron Belt   (Hitachi and Pioneer)
// Galaxy Ranger/Star Blazer (Galaxy Ranger Pioneer and Star Blazer Hitachi are 
//			undumped if they exist)
// Albegas/Cybernaught (unverified - discs exist (I have one) but to date no 
//			collector that I know has found a romset, rumored to be a prototype, 
//			might also have been a Japanese only release)
// Cobra Command (Special version that runs on Hitachi AB hardware)

#include <string.h>
#include "astron.h"
#include "../daphne.h"
#include "../io/conout.h"
#include "../ldp-in/ldv1000.h"
#include "../ldp-in/vip9500sg.h"
#include "../ldp-out/ldp.h"
#include "../video/video.h"
#include "../sound/sound.h"
#include Z80_INCLUDE

extern ldp *g_ldp;

astron::astron()
{
	memset(banks, 0xFF, 4);	// fill banks with 0xFF's
	memset(used_sprite_color, 0x00, 256);	// fill banks with 0x00's
	video_overlay = NULL;
	vldp_overlay = NULL;
	screen_needs_update = true;
	palette_modified = true;
	disc_fps = 29.97; 
	game_type = GAME_ASTRON;
	cpu_type = CPU_Z80;
	cpu_initialized = false;
	cpu_hz = 5000000; // the schematics seem to indicate that its 20MHz / 4
	irq_period = 16.6666; // interrupt from hblank (60hz)
	nmi_period = 16.6666; // nmi from LD-V1000 command strobe
	current_bank = 0;
	m_transparent_color = 0;
	ldp_output_latch = 0xff;
	has_issues = true;
	strcpy(game_issues, "No in game sounds, due to lack of samples");

	num_sounds = 7;
	sound_name[S_AB_PLAYER_SHIP] = "ab_ship.wav";
	sound_name[S_AB_PLAYER_FIRE] = "ab_fire.wav";
	sound_name[S_AB_ENEMY_FIRE] = "ab_enemy.wav";
	sound_name[S_AB_ALARM1] = "ab_alarm1.wav";
	sound_name[S_AB_ALARM2] = "ab_alarm2.wav";
	sound_name[S_AB_ALARM3] = "ab_alarm3.wav";
	sound_name[S_AB_ALARM4] = "ab_alarm4.wav";

}

blazer::blazer()
{
	game_type = GAME_GALAXY;
}

astronh::astronh()
{
	nmi_period = 2; // we'll use this to clock the 8251
	rxrdy = txrdy = false;
}

galaxy::galaxy()
{
	game_type = GAME_GALAXY;

	num_sounds = 8;
	sound_name[S_GR_FIRE] = "gr_fire.wav";
	sound_name[S_GR_ENEMY_CANNON] = "gr_cannon.wav";
	sound_name[S_GR_ENEMY_MINION] = "gr_mineon.wav";
	sound_name[S_GR_ENEMY_ATTACK] = "gr_attack.wav";
	sound_name[S_GR_ALARM1] = "gr_alarm1.wav";
	sound_name[S_GR_ALARM2] = "gr_alarm2.wav";
	sound_name[S_GR_ALARM3] = "gr_alarm3.wav";
	sound_name[S_GR_ALARM4] = "gr_alarm4.wav";
}

cobraab::cobraab()
{
	banks[3] = 0xfb;
	has_issues = true;
	strcpy(game_issues, "No in game sounds, and missing graphics due to an undumped prom");
	banks[3] = 0xfb;
}

// does anything special needed to send an IRQ
void astron::do_irq()
{
	// Redraws the screen (if needed) on interrupt
	recalc_palette();
	display_update();
	Z80_ASSERT_IRQ;
//	z80_set_irq_line(0, HOLD_LINE);	// assert IRQ line
}

// does anything special needed to send an NMI
void astron::do_nmi()
{
	// only do an nmi if nmie is enabled
	if (cpumem[0xd801] & 0x40)
	{
		write_ldv1000(ldp_output_latch);
		ldp_input_latch = read_ldv1000();
//		z80_set_nmi_line(ASSERT_LINE);
//		z80_execute(1);	// execute a quick instruction
//		z80_set_nmi_line(CLEAR_LINE);
		Z80_ASSERT_NMI;
	}
}

// we'll use this for clocking the 8251 since it generates the nmi's
void astronh::do_nmi()
{
	clock_8251();
}

// does anything special needed to send an NMI
void astronh::astronh_nmi()
{
	// only do an nmi if nmie is enabled
	if (cpumem[0xd801] & 0x40)
	{
//		printline("doin' the NMI");
		
//		z80_set_nmi_line(ASSERT_LINE);
//		z80_execute(1);	// execute a quick instruction
//		z80_set_nmi_line(CLEAR_LINE);
		Z80_ASSERT_NMI;
	}
}

// reads a byte from the cpu's memory
Uint8 astron::cpu_mem_read(Uint16 addr)
{
	Uint8 result = cpumem[addr];
	char s[81] = { 0 };
	
	// main rom
	if (addr <= 0x7fff)
	{
	}

	// 0x8000 to 0xbfff is the bankswitching memory
	else if (addr >= 0x8000 && addr <= 0xbfff)
	{
//		sprintf(s, "Bank rom read at %x, rombank %x, current bank %x", addr, (addr - 0x8000) + (0x4000 * current_bank), current_bank);
//		printline(s);
		result = rombank[(addr - 0x8000) + (0x4000 * current_bank)];
	}

	else if (addr >= 0xc800 && addr <= 0xcfff)
	{
		result = read_ldp(addr);
//		sprintf(s, "LDP read %x", result);
//		printline(s);
	}
	else if (addr == 0xd000)
	{
		result = banks[2];
	}
	else if (addr == 0xd001)
	{
		result = banks[3];
	}
	else if (addr == 0xd002)
	{
		result = banks[0];
	}
	else if (addr == 0xd003)
	{
		result = banks[1];
	}
	
	// obj
	else if (addr >= 0xc000 && addr <= 0xc320)
	{
	}

	// out
	else if (addr >= 0xd800 && addr <= 0xdfff)
	{
	}

	// color
	else if (addr >= 0xe000 && addr <= 0xe1ff)
	{
	}

	// work ram
	else if (addr >= 0xf800 )
	{
	}

	else
	{
		char s[81] = { 0 };
//		sprintf(s, "Unmapped read from %x", addr);
//		printline(s);
	}

	return result;
}

// writes a byte to the cpu's memory
void astron::cpu_mem_write(Uint16 addr, Uint8 value)
{
	cpumem[addr] = value;
	char s[81] = { 0 };

	// main rom
	if (addr <= 0x7fff)
	{
		sprintf(s, "Attempted write to main ROM! at %x with value %x", addr, value);
		printline(s);
	}

	// bank rom
	else if (addr >= 8000 && addr <= 0xbfff)
	{
		sprintf(s, "Attempted write to bank ROM! at %x with value %x", addr, value);
		printline(s);
	}

	// disc
	else if (addr >= 0xc800 && addr <= 0xcfff)
	{
//		sprintf(s, "LDP write %x", value);
//		printline(s);
		write_ldp(value, addr);
	}


	// obj (sprite)
	else if (addr >= 0xc000 && addr <= 0xc320)
	{
		cpumem[addr] = value;
		screen_needs_update = true;
	}
	
	// out
	else if (addr == 0xd800)
	{
//		sprintf(s, "ASTRON: Write to D800 with %x", value);
//		printline(s);

		if (game_type == GAME_ASTRON)
		{
			if (((value & 0x0f) == 0x00) && !(value & 0x20))
			{
				sound_play(S_AB_PLAYER_SHIP);
			}
			else if ((value & 0x0f) == 0xb)
			{
				sound_play(S_AB_PLAYER_FIRE);
			}
			else if ((value & 0x0f) == 0xa)
			{
				sound_play(S_AB_ENEMY_FIRE);
			}
			else if ((value & 0x0f) == 0xf)
			{
				sound_play(S_AB_ALARM1);
			}
			else if ((value & 0x0f) == 0xe)
			{
				sound_play(S_AB_ALARM2);
			}
			else if ((value & 0x0f)== 0xd)
				{
				sound_play(S_AB_ALARM3);
			}
			else if ((value & 0x0f) == 0xc)
			{
				sound_play(S_AB_ALARM4);
			}
		}
		else if (game_type == GAME_GALAXY)
		{
			if ((value & 0x0f) == 0xb)
			{
				sound_play(S_GR_FIRE);
			}
			else if ((value & 0x0f) == 0x9)
			{
				sound_play(S_GR_ENEMY_CANNON);
			}
			else if ((value & 0x0f) == 0x8)
			{
				sound_play(S_GR_ENEMY_MINION);
			}
			else if ((value & 0x0f) == 0xa)
			{
				sound_play(S_GR_ENEMY_ATTACK);
			}
			else if ((value & 0x0f) == 0xf)
			{
				sound_play(S_GR_ALARM1);
			}
			else if ((value & 0x0f) == 0xe)
			{
				sound_play(S_GR_ALARM2);
			}
			else if ((value & 0x0f) == 0xd)
			{
				sound_play(S_GR_ALARM3);
			}
			else if ((value & 0x0f) == 0xc)
			{
				sound_play(S_GR_ALARM4);
			}
		}

		cpumem[addr] = value;
	}
	else if (addr == 0xd801)
	{
		cpumem[addr] = value;
	}
	else if (addr == 0xd802)
	{
		cpumem[addr] = value;
	}
	else if (addr == 0xd803)
	{
		cpumem[addr] = value;
	}

	// color
	else if (addr >= 0xe000 && addr <= 0xe1ff)
	{
		cpumem[addr] = value;
		palette_modified = true;
	}

	// fix (character)
	else if (addr >= BASE_VID_MEM_ADDRESS && addr <= BASE_VID_MEM_ADDRESS + 0x7ff)
	{
		if (addr < (BASE_VID_MEM_ADDRESS + 0x400))
		{
			cpumem[addr] = value;
		}
		else
		{
			cpumem[addr-0x400] = value;
		}
		screen_needs_update = true;
	}

	// work ram
	else if (addr >= 0xf800 && addr <= 0xffff)
	{
		cpumem[addr] = value;
	}

	else
	{
			cpumem[addr] = value;
			char s[81] = { 0 };
//			sprintf(s, "Unmapped write to %x with value %x", addr, value);
//			printline(s);
	}
}

unsigned char astron::read_ldp(unsigned char addr)
{
	char s[81] = {0};
		
		unsigned char result = ldp_input_latch;
//		sprintf(s, "Read from player %x at pc: %x", result, Z80_GET_PC);
//		printline(s);

		return result;
}

unsigned char astronh::read_ldp(unsigned char addr)
{
	unsigned char result = 0x00;
	
	switch (addr & 0x01)
	{
	case 0x00:
		result = read_8251_data();
		break;
	case 0x01:
		result = read_8251_status();
		break;
	default:
		printline("ERROR: ASTRONH read_ldp()! check disc address");
		break;
	}

	return result;
}

void astron::write_ldp(unsigned char value, unsigned char addr)
{
	char s[81] = {0};
	
//	sprintf(s, "Write to player %x at pc %x", value, Z80_GET_PC);
//	printline(s);
	ldp_output_latch = value;
}

void astronh::write_ldp(unsigned char value, unsigned char addr)
{
	switch (addr & 0x01)
	{
	case 0x00:
		write_8251_data(value);
		break;
	case 0x01:
		write_8251_control(value);
		break;
	default:
		printline("ERROR: ASTRONH write_ldp()! check disc address");
		break;
	}
}

// reads a byte from the cpu's port
Uint8 astron::port_read(Uint16 port)
{
	char s[81] = { 0 };

	port &= 0xFF;	
	sprintf(s, "ERROR: CPU port %x read requested, but this function is unimplemented!", port);
	printline(s);

	return(0);
}

// writes a byte to the cpu's port
void astron::port_write(Uint16 port, Uint8 value)
{
	char s[81] = { 0 };

	port &= 0xFF;
	
	switch (port & 0xFF)
	{
	case 0x00:	// astron switches rom banks with the D0 bit here
	case 0x01:	// at 0x01 too?
		current_bank = value & 0x01;
		break;
	default:
		sprintf(s, "ERROR: CPU port %x write requested (value %x) but this function is unimplemented!", port, value);
		printline(s);
		break;
	}
}

// does any video initialization we might need
// returns 'true' if the initialization was successful, false if it failed
bool astron::video_init()
{
	video_overlay = SDL_CreateRGBSurface(SDL_SWSURFACE, ASTRON_OVERLAY_W, ASTRON_OVERLAY_H, 8, 0, 0, 0, 0); // create an 8-bit surface
	vldp_overlay = SDL_CreateRGBSurface(SDL_SWSURFACE, ASTRON_OVERLAY_W, ASTRON_OVERLAY_H, 8, 0, 0, 0, 0); // create an 8-bit surface
	
	if (video_overlay)
	{
		// check how many colors are used in total to see if we can cheat and only use one palette
		bool used_tile_color[256] = {0};
		int total_sprite_colors = 0;
		int total_tile_colors = 0;
		int x, y = 0;
		
	
		// first, check the sprites
		for (x = 0x0000; x < 0x4000; x++)
		{
			unsigned char data_lo = sprite[x];
			unsigned char data_high = sprite[x + 0x4000];
			
			used_sprite_color[((data_lo & 0x0f) | (data_high << 0x04) & 0xff)] = true;
			used_sprite_color[((data_lo >> 0x04) | (data_high & 0xf0) & 0xff)] = true;
		}
		
		// now, the tiles
		for (x = 0x000; x < 0x200; x++)
		{
			used_tile_color[bankprom[x]] = true;
		}

		// count up the total colors		
		for (x = 0; x < 256; x++)
		{
			if (used_sprite_color[x] == true)
			{
				total_sprite_colors++;
			}
			if (used_tile_color[x] == true)
			{
				total_tile_colors++;
			}
		}
		
		//	color conversion for the tiles 8 bit
		//	Convert palette rom into a useable palette
		//  THIS IS PROBABLY WRONG!
		for (int i = 0; i < 256; i++)
		{
			Uint8 bit0,bit1,bit2,bit3;	

			/* red component */
			bit0 = static_cast<Uint8>((i >> 0) & 0x01);
			bit1 = static_cast<Uint8>((i >> 1) & 0x01);
			bit2 = static_cast<Uint8>((i >> 2) & 0x01);
			bit3 = 0;
			colors[i].r = static_cast<Uint8>((0x90 * bit0) + (0x41 * bit1) + (0x1f * bit2) + (0x0f * bit3));
			/* green component */
			bit0 = static_cast<Uint8>((i >> 3) & 0x01);
			bit1 = static_cast<Uint8>((i >> 4) & 0x01);
			bit2 = static_cast<Uint8>((i >> 5) & 0x01);
			bit3 = 0;
			colors[i].g = static_cast<Uint8>((0x90 * bit0) + (0x41 * bit1) + (0x1f * bit2) + (0x0f * bit3));
			/* blue component */
			bit0 = static_cast<Uint8>((i >> 6) & 0x01);
			bit1 = static_cast<Uint8>((i >> 7) & 0x01);
			bit2 = 0;
			bit3 = 0;
			colors[i].b = static_cast<Uint8>((0x90 * bit0) + (0x41 * bit1) + (0x1f * bit2) + (0x0f * bit3));
		}
		
		// if the total number of colors is under 0xff we can cheat and have them all in one palette
		if ((total_tile_colors + total_sprite_colors) < 0xff)
		{
			char s[81] = {0};
			sprintf(s, "total used colors 0x%x - compressing palette", total_tile_colors + total_sprite_colors);
			printline(s);
			compress_palette = true;

			// check each tile color and if we need it remap it to a sprite color
			for (x = 0; x < 256; x++)
			{
				if (used_tile_color[x] == true)
				{
					while (used_sprite_color[y] == true)
					{
						y++;
					}
				
					mapped_tile_color[x] = y;
					colors[y].r = colors[x].r;
					colors[y].g = colors[x].g;
					colors[y].b = colors[x].b;
					// since color 0 is moved we need to find our new transparent color
					if (x == 0)
					{
						m_transparent_color = y;
					}
					y++;
				}
			}
		}
		else
		{
			char s[81] = {0};
			sprintf(s, "total used colors 0x%x - cannot compress palette!", total_tile_colors + total_sprite_colors);
			printline(s);
			compress_palette = false;
		}
		

		SDL_SetColors(video_overlay, colors, 0, 256);
		SDL_SetColors(vldp_overlay, colors, 0, 256);
	}
	else
	{
		printline("ERROR: Could not create astron overlay");
	}

	return true;

}

// shuts down any video we might have initialized
void astron::video_shutdown()
{
	SDL_FreeSurface(video_overlay);
	SDL_FreeSurface(vldp_overlay);
}

// redraws our video on the screen (in case the screen needs to be redrawn)
void astron::video_repaint()
{
	screen_needs_update = true;
	display_update();
}

bool astron::load_roms()
{
	return (
		load_rom("astronp/abp-ic03.bin", &cpumem[0x0000], 0x4000) &&
		load_rom("astronp/abp-ic10.bin", &cpumem[0x4000], 0x4000) &&
	
		load_rom("astron/5284", &rombank[0x0000], 0x4000) &&
		load_rom("astron/5284", &cpumem[0x8000], 0x4000) &&
		load_rom("astron/5285", &rombank[0x4000], 0x4000) &&
	
		load_rom("astron/5280", &character[0x0000], 0x0800) &&
		load_rom("astron/5281", &character[0x0800], 0x0800) &&

		load_rom("astron/5286", &sprite[0x0000], 0x4000) &&
		load_rom("astron/5338", &sprite[0x4000], 0x4000) &&

		load_rom("astron/5279", &bankprom[0x000], 0x200) 
);
}

bool blazer::load_roms()
{
	return (
		load_rom("blazer/epr-5590.bin", &cpumem[0x0000], 0x4000) &&
		load_rom("blazer/epr-5591.bin", &cpumem[0x4000], 0x4000) &&
	
		load_rom("galaxy/gr5592.bin", &rombank[0x0000], 0x4000) &&
		load_rom("galaxy/gr5592.bin", &cpumem[0x8000], 0x4000) &&
		load_rom("galaxy/gr5593.bin", &rombank[0x4000], 0x4000) &&
	
		load_rom("astron/5280", &character[0x0000], 0x0800) &&
		load_rom("astron/5281", &character[0x0800], 0x0800) &&

		load_rom("galaxy/gr5594.bin", &sprite[0x0000], 0x4000) &&
		load_rom("galaxy/gr5595.bin", &sprite[0x4000], 0x4000) &&
		load_rom("galaxy/gr5611.bin", &sprite[0x8000], 0x4000) &&
		load_rom("galaxy/gr5612.bin", &sprite[0xc000], 0x4000) &&

		load_rom("astron/5279", &bankprom[0x000], 0x200) 
);
}

bool astronh::load_roms()
{
	return (
		load_rom("astron/5473c", &cpumem[0x0000], 0x4000) &&
		load_rom("astron/5474a", &cpumem[0x4000], 0x4000) &&
	
		load_rom("astron/5284", &rombank[0x0000], 0x4000) &&
		load_rom("astron/5284", &cpumem[0x8000], 0x4000) &&
		load_rom("astron/5285", &rombank[0x4000], 0x4000) &&
	
		load_rom("astron/5280", &character[0x0000], 0x0800) &&
		load_rom("astron/5281", &character[0x0800], 0x0800) &&

		load_rom("astron/5286", &sprite[0x0000], 0x4000) &&
		load_rom("astron/5338", &sprite[0x4000], 0x4000) &&

		load_rom("astron/5279", &bankprom[0x000], 0x200) 
);
}

bool galaxy::load_roms()
{
	return (
		load_rom("galaxy/gr5633.bin", &cpumem[0x0000], 0x4000) &&
		load_rom("galaxy/gr5634.bin", &cpumem[0x4000], 0x4000) &&
	
		load_rom("galaxy/gr5592.bin", &rombank[0x0000], 0x4000) &&
		load_rom("galaxy/gr5592.bin", &cpumem[0x8000], 0x4000) &&
		load_rom("galaxy/gr5593.bin", &rombank[0x4000], 0x4000) &&
	
		load_rom("astron/5280", &character[0x0000], 0x0800) &&
		load_rom("astron/5281", &character[0x0800], 0x0800) &&

		load_rom("galaxy/gr5594.bin", &sprite[0x0000], 0x4000) &&
		load_rom("galaxy/gr5595.bin", &sprite[0x4000], 0x4000) &&
		load_rom("galaxy/gr5611.bin", &sprite[0x8000], 0x4000) &&
		load_rom("galaxy/gr5612.bin", &sprite[0xc000], 0x4000) &&

		load_rom("astron/5279", &bankprom[0x000], 0x200) 
);
}

bool cobraab::load_roms()
{
	return (
		load_rom("cobraab/ic-1.bin", &cpumem[0x0000], 0x4000) &&
		load_rom("cobraab/ic-2.bin", &cpumem[0x4000], 0x2000) &&
	
		load_rom("cobraab/ic-11.bin", &character[0x0000], 0x0800) &&
		load_rom("cobraab/ic-12.bin", &character[0x0800], 0x0800) &&

		load_rom("cobraab/ic-7.bin", &sprite[0x0000], 0x2000) &&
		load_rom("cobraab/ic-8.bin", &sprite[0x4000], 0x2000) &&

		// this rom is undumped...
		load_rom("astron/5279", &bankprom[0x000], 0x200) 
);
}

void astron::recalc_palette()
{
	if (palette_modified)
	{
		//	color conversion for the sprites from color memory
		//	Convert palette ram into a useable palette
		//  THIS IS PROBABLY WRONG!
		for (int i = 0; i < 256; i++)
		{
			int j;
			j = 0xe000 + (i * 0x02);
			Uint8 bit0,bit1,bit2,bit3;	
			
			// only update the sprite colors if they are used
			if (used_sprite_color[i])
			{
				// color is 3 nibbles, each nibble corresponding to R, G, and B
				// B is the most significant nibble, R is the least significant

				// MATT's new code for calculating color palette ..

				colors[i].r = (cpumem[j] & 0x0F) << 4;
				colors[i].g = (cpumem[j] & 0xF0);
				colors[i].b = (cpumem[j+1] & 0x0F) << 4;

				// here's the old reference code for color palette calculation

				/*
				// red component
				bit3 = static_cast<Uint8>((cpumem[j] >> 0) & 0x01);
				bit2 = static_cast<Uint8>((cpumem[j] >> 1) & 0x01);
				bit1 = static_cast<Uint8>((cpumem[j] >> 2) & 0x01);
				bit0 = static_cast<Uint8>((cpumem[j] >> 3) & 0x01);
				colors[i].r = static_cast<Uint8>((0x90 * bit0) + (0x41 * bit1) + (0x1f * bit2) + (0x0f * bit3));
				// green component
				bit3 = static_cast<Uint8>((cpumem[j] >> 4) & 0x01);
				bit2 = static_cast<Uint8>((cpumem[j] >> 5) & 0x01);
				bit1 = static_cast<Uint8>((cpumem[j] >> 6) & 0x01);
				bit0 = static_cast<Uint8>((cpumem[j] >> 7) & 0x01);
				colors[i].g = static_cast<Uint8>((0x90 * bit0) + (0x41 * bit1) + (0x1f * bit2) + (0x0f * bit3));
				// blue component
				bit3 = static_cast<Uint8>((cpumem[j + 1] >> 0) & 0x01);
				bit2 = static_cast<Uint8>((cpumem[j + 1] >> 1) & 0x01);
				bit1 = static_cast<Uint8>((cpumem[j + 1] >> 2) & 0x01);
				bit0 = static_cast<Uint8>((cpumem[j + 1] >> 3) & 0x01);
				colors[i].b = static_cast<Uint8>((0x90 * bit0) + (0x41 * bit1) + (0x1f * bit2) + (0x0f * bit3));
				*/
			}
		}

		SDL_SetColors(video_overlay, colors, 0, 256);
		SDL_SetColors(vldp_overlay, colors, 0, 256);
	}
	palette_modified = false;
}


// updates astron's video
void astron::display_update()
{
	if (screen_needs_update)
	{
		SDL_Rect dest;	// setup place on screen to blit astron overlay
		dest.x = 0;	// upper left corner
		dest.y = 0;
		dest.w = (unsigned short) video_overlay->w;
		dest.h = (unsigned short) video_overlay->h;

		// we need to black the overlay before we draw
		for (int x = 0; x < 256; x++)
		{
			for (int y = 0; y <256; y++)
			{
				*((Uint8 *) video_overlay->pixels + y * ASTRON_OVERLAY_W + x) = static_cast<Uint8>(m_transparent_color);
			}
		}

		// The sprites are bottom priority so we draw them first
		// START modified Mame code
		// check if sprites need to be drawn
		int spr_number,sprite_bottom_y,sprite_top_y;
		unsigned char *sprite_base;

		for (spr_number = 0;spr_number < 32;spr_number++)
		{
			sprite_base = cpumem + BASE_SPRITE_MEM_ADDRESS + 0x10 * spr_number;
			sprite_top_y = sprite_base[SPR_Y_TOP];
			sprite_bottom_y = sprite_base[SPR_Y_BOTTOM];
			if (sprite_bottom_y && (sprite_bottom_y-sprite_top_y > 0))
			{
				draw_sprite(spr_number);
			}
		}
		// END modified Mame code


		
		Uint8 pixel[8];

		// loop through video memory and draw characters
		for (int charx = 0; charx < 32; charx++)
		{
			for (int chary = 0; chary < 32; chary++)
			{
				for (int y = 0; y < 8; y++)
				{
					int current_character = chary * 32 + charx + BASE_VID_MEM_ADDRESS;
					
					pixel[0] = static_cast<Uint8>(bankprom[((character[cpumem[current_character]*8+y] & 0x80) >> 7) | ((character[cpumem[current_character]*8+y+0x800] & 0x80) >> 6) | (((cpumem[current_character]) & 0xf8) >> 0x01) | ((cpumem[0xd801] & 0x20) << 2)]);
					pixel[1] = static_cast<Uint8>(bankprom[((character[cpumem[current_character]*8+y] & 0x40) >> 6) | ((character[cpumem[current_character]*8+y+0x800] & 0x40) >> 5) | (((cpumem[current_character]) & 0xf8) >> 0x01) | ((cpumem[0xd801] & 0x20) << 2)]);
					pixel[2] = static_cast<Uint8>(bankprom[((character[cpumem[current_character]*8+y] & 0x20) >> 5) | ((character[cpumem[current_character]*8+y+0x800] & 0x20) >> 4) | (((cpumem[current_character]) & 0xf8) >> 0x01) | ((cpumem[0xd801] & 0x20) << 2)]);
					pixel[3] = static_cast<Uint8>(bankprom[((character[cpumem[current_character]*8+y] & 0x10) >> 4) | ((character[cpumem[current_character]*8+y+0x800] & 0x10) >> 3) | (((cpumem[current_character]) & 0xf8) >> 0x01) | ((cpumem[0xd801] & 0x20) << 2)]);
					pixel[4] = static_cast<Uint8>(bankprom[((character[cpumem[current_character]*8+y] & 0x08) >> 3) | ((character[cpumem[current_character]*8+y+0x800] & 0x08) >> 2) | (((cpumem[current_character]) & 0xf8) >> 0x01) | ((cpumem[0xd801] & 0x20) << 2)]);
					pixel[5] = static_cast<Uint8>(bankprom[((character[cpumem[current_character]*8+y] & 0x04) >> 2) | ((character[cpumem[current_character]*8+y+0x800] & 0x04) >> 1) | (((cpumem[current_character]) & 0xf8) >> 0x01) | ((cpumem[0xd801] & 0x20) << 2)]);
					pixel[6] = static_cast<Uint8>(bankprom[((character[cpumem[current_character]*8+y] & 0x02) >> 1) | ((character[cpumem[current_character]*8+y+0x800] & 0x02) >> 0) | (((cpumem[current_character]) & 0xf8) >> 0x01) | ((cpumem[0xd801] & 0x20) << 2)]);
					pixel[7] = static_cast<Uint8>(bankprom[((character[cpumem[current_character]*8+y] & 0x01) >> 0) | ((character[cpumem[current_character]*8+y+0x800] & 0x01) << 1) | (((cpumem[current_character]) & 0xf8) >> 0x01) | ((cpumem[0xd801] & 0x20) << 2)]);
			
					// draw the remapped colors instead of the regular if our palette is compressed
					if (compress_palette)
					{
						for (int x = 0; x < 8; x++)
						{
							if (pixel[x])
							{
								*((Uint8 *) video_overlay->pixels + ((chary * 8 + y) * ASTRON_OVERLAY_W) + (charx * 8 + x)) = mapped_tile_color[pixel[x]];
								
								//char s[81] = {0};
								//sprintf(s, "drawing color %x", mapped_tile_color[pixel[x]]);
								//printline(s);
							}
						}
					}
					else
					{
						for (int x = 0; x < 8; x++)
						{
							if (pixel[x])
							{
								*((Uint8 *) video_overlay->pixels + ((chary * 8 + y) * ASTRON_OVERLAY_W) + (charx * 8 + x)) = pixel[x];
							}
						}
					}					
				}
			}
		}

//		SDL_BlitSurface(video_overlay, NULL, 

		dest.x = 0;	// upper left corner
		dest.y = 0;
		dest.w = (unsigned short) video_overlay->w;
		dest.h = (unsigned short) video_overlay->h;

		// SMPEG doesn't allow blitting, which is why we have to check
		if (g_ldp->is_blitting_allowed())
		{
			SDL_BlitSurface(video_overlay, NULL, get_screen(), &dest);
			SDL_Flip(get_screen());
		//	SDL_UpdateRects(get_screen(), 1, &dest);
		}
		else
		{
			SDL_BlitSurface(video_overlay, NULL, vldp_overlay, &dest);
		}

		screen_needs_update = false;
	}
	
	return;
}

// this gets called when the user presses a key or moves the joystick
void astron::input_enable(Uint8 move)
{
	switch (move)
	{
	case SWITCH_UP:
		banks[1] &= ~0x08; 
		break;
	case SWITCH_LEFT:
		banks[1] &= ~0x02;
		break;
	case SWITCH_RIGHT:
		banks[1] &= ~0x01;
		break;
	case SWITCH_DOWN:
		banks[1] &= ~0x04; 
		break;
	case SWITCH_BUTTON1: // '1' on keyboard
		banks[0] &= ~0x10; 
		break;
	case SWITCH_BUTTON2: // '2' on keyboard
		banks[0] &= ~0x40;
		break;
	case SWITCH_BUTTON3: // space on keyboard
		banks[1] &= ~0x10;
		break;
	case SWITCH_COIN1: 
		banks[0] &= ~0x01;
		break;
	case SWITCH_COIN2: 
		banks[0] &= ~0x02;
		break;
	case SWITCH_SERVICE: 
		banks[0] &= ~0x08;
		break;
	case SWITCH_TEST: 
		banks[0] &= ~0x04;
		break;
	}
}  

// this gets called when the user releases a key or moves the joystick back to center position
void astron::input_disable(Uint8 move)
{
	switch (move)
	{
	case SWITCH_UP:
		banks[1] |= 0x08; 
		break;
	case SWITCH_LEFT:
		banks[1] |= 0x02;
		break;
	case SWITCH_RIGHT:
		banks[1] |= 0x01;
		break;
	case SWITCH_DOWN:
		banks[1] |= 0x04; 
		break;
	case SWITCH_BUTTON1: // '1' on keyboard
		banks[0] |= 0x10; 
		break;
	case SWITCH_BUTTON2: // '2' on keyboard
		banks[0] |= 0x40;
		break;
	case SWITCH_BUTTON3: // space on keyboard
		banks[1] |= 0x10;
		break;
	case SWITCH_COIN1: 
		banks[0] |= 0x01;
		break;
	case SWITCH_COIN2: 
		banks[0] |= 0x02;
		break;
	case SWITCH_SERVICE: 
		banks[0] |= 0x08;
		break;
	case SWITCH_TEST: 
		banks[0] |= 0x04;
		break;
	}
}

// used to set dip switch values
bool astron::set_bank(unsigned char which_bank, unsigned char value)
{
	bool result = true;
	
	switch (which_bank)
	{
	case 0:	// bank A
		banks[2] = (unsigned char) (value ^ 0xFF);	// dip switches are active low
		break;
	case 1:	// bank B
		banks[3] = (unsigned char) (value ^ 0xFF);	// switches are active low
		break;
	default:
		printline("ERROR: Bank specified is out of range!");
		result = false;
		break;
	}
	
	return result;
}

// returns a pointer to Astron's video overlay surface
// also returns width, height, and bits per pixel of surface
SDL_Surface *astron::get_video_overlay()
{
	return(vldp_overlay);
}

Uint8 astron::get_transparent_color()
{
	return m_transparent_color;
}

SDL_Color *astron::get_video_colors()
{
	return colors;
}

// START modified Mame code
void astron::draw_sprite(int spr_number)
{
	int sy,row,height,src,bank;
	unsigned char *sprite_base;
	int skip;	/* bytes to skip before drawing each row (can be negative) */
	unsigned char *gfx;


	sprite_base	= cpumem + BASE_SPRITE_MEM_ADDRESS + 0x10 * spr_number;

	src = sprite_base[SPR_GFXOFS_LO] + (sprite_base[SPR_GFXOFS_HI] << 8);

	// astron only has sprite data 0x0000-0x7fff, but galaxy has 0x0000-0xffff
	bank = 0x8000 * (((sprite_base[SPR_X_HI] & 0x80) >> 7) + ((sprite_base[SPR_X_HI] & 0x40) >> 5));
//	bank &= (memory_region_length(REGION_GFX2)-1);	/* limit to the range of available ROMs */
	skip = sprite_base[SPR_SKIP_LO] + (sprite_base[SPR_SKIP_HI] << 8);

	height = sprite_base[SPR_Y_BOTTOM] - sprite_base[SPR_Y_TOP];
//	sprite_palette = Machine->remapped_colortable + 0x10 * spr_number;

	sy = sprite_base[SPR_Y_TOP];// + 1;

	/* sprite[] contains the packed sprite data */
	gfx = &sprite[bank];

	for (row = 0;row < height;row++)
	{
		int x, y;
		int src2;

		src = src2 = src + skip;

		x = sprite_base[SPR_X_LO] + ((sprite_base[SPR_X_HI] & 0x01) << 8);
		y = sy+row;

//		char s[81] = { 0 };
	//	sprintf(s, "Drawing sprite %x, pixel at %x, %x", spr_number, x, y);
	//	printline(s);

		while (1)
		{
			int data_lo, data_high, data;

			data = gfx[src2 & 0x7fff];
			data_lo = gfx[src2 & 0x7fff];
			data_high = gfx[(src2 + 0x4000) & 0x7fff];

			// stop drawing when the sprite data is 0xff
			if ((data_lo == 0xff) && (data_high == 0xff)) 
			{
				break;
			}
			
			unsigned char pixel1 = (data_lo >> 0x04) | (data_high & 0xf0);
			unsigned char pixel2 = (data_lo & 0x0f) | (data_high << 0x04);

			// draw these guys backwards
			if (src & 0x8000)
			{
				unsigned char temp_pixel;
				temp_pixel = pixel1;
				pixel1 = pixel2;
				pixel2 = temp_pixel;
				
				src2--;
			}
			else
			{
				src2++;
			}

			// draw the pixel - don't draw if its black as to not cover up any tiles
			if (pixel1)
			{
				*((Uint8 *) video_overlay->pixels + (y * ASTRON_OVERLAY_W) + x) = pixel1;
			}
			x++;
			if (pixel2)
			{
				*((Uint8 *) video_overlay->pixels + (y * ASTRON_OVERLAY_W) + x) = pixel2;
			}
			x++;
		}
	}
}
// END modified Mame code

// very minimal 8251 implementation, just to get Astron working
void astronh::write_8251_data(unsigned char data)
{
	char s[81] = { 0 };
	sprintf(s, "ASTRONH: 8251_write_data() with %x", data);
//	printline(s);
	// writing to the 8251 resets txrdy
	txrdy = false;
	write_vip9500sg(data);
}

unsigned char astronh::read_8251_data(void)
{
	char s[81] = { 0 };
	unsigned char result = read_vip9500sg();
	sprintf(s, "ASTRONH: 8251_read_data with %x", result);
//	printline(s);
	// reading from the 8251 resets rxrdy
	rxrdy = false;
	return result;
}

void astronh::write_8251_control(unsigned char control)
{
	char s[81] = { 0 };
	static bool init = false;

	// initialize chip with first control instruction after reset
	if (!init)
	{
		printline("8251 Reset!");
		if (control == 0x8f)
		{
			printline("8251 mode selected as 1.5 stop bits/parity disabled/8 bit characters/64x baud factor");
		}
		else if (control == 0x4e)
		{
			printline("8251 mode selected as 1 stop bit/parity disabled/8 bit characters/16x baud factor");
		}
		else
		{
			sprintf(s, "8251 attempted to initialize with mode %x - unsupported", control);
			printline(s);
		}
		init = true;
	}
	// otherwise its a command
	else
	{
		if (control & 0x01)
		{
		//	printline("8251: Transmit Enable");
			if (!transmit_enable)
			{
				// when transmit is enabled it makes txrdy true
				transmit_enable = true;
				txrdy = true;
				astronh_nmi();
			}
		}
		else
		{
			//printline("8251: Transmit Disable");
			transmit_enable = false;
			txrdy = false;
		}
		
		// DTR pin is not connected on ASTRON
		//if (control & 0x02)
		//{
		//	printline("8251: /DTR = 0");
		//}
		//else
		//{
		//	printline("8251: /DTR = 1");
		//}

		if (control & 0x04)
		{
		//	printline("8251: Recieve Enable");
			recieve_enable = true;
		}
		else
		{
		//	printline("8251: Recieve Disable");
			recieve_enable = false;
		}
		
		// I don't think ASTRON uses the Break Character... we'll see
		if (control & 0x08)
		{
			printline("8251: Sent Break Character!");
		}
		else
		{
			//printline("8251: Break Character Normal Operation");
		}
		
		// I haven't seen astron reset the error flag yet
		if (control & 0x10)
		{
			printline("8251: Reset Error Flag!");
		}
		else
		{
//			printline("8251: Error Flag Normal Operation");
		}
		
		// not sure what this does
		if (control & 0x20)
		{
//			printline("8251: RTS = 0");
		}
		else
		{
			printline("8251: RTS = 1");
		}
		
		if (control & 0x40)
		{
			printline("8251: Internal Reset");
			init = false;
		}
		else
		{
		//	printline("8251: Reset Normal Operation");
		}
		
		// Hunt mode should only be enabled in Sync mode
		if (control & 0x80)
		{
			printline("8251 ERROR: Hunt Mode!");
		}
		else
		{
		}
	}	
}

unsigned char astronh::read_8251_status(void)
{
	// with astron all we care about are the bottom two bits of status (rxrdy, txrdy)
	unsigned char result = (0x00 | txrdy | (rxrdy << 0x01));

	return result;
}

void astronh::clock_8251(void)
{
	// periodically check the status of the 8251 

	// if we have data ready from the player inform the 8251
	if ((vip9500sg_result_ready()) && (!rxrdy) && (recieve_enable))
	{
		rxrdy = true;
		astronh_nmi();
	}
	// or if neither rxrdy or txrdy are active turn txrdy on
	else if ((!txrdy) && (transmit_enable))
	{
		txrdy = true;
		astronh_nmi();
	}
}
		
// returns how many pixels down to shift the video
int astron::get_video_row_offset()
{
	return -32;
}

