/*
 * badlands.cpp
 *
 * Copyright (C) 2001 Mark Broadhead
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// badlands.cpp
// by Mark Broadhead
//

#include <string.h>
#include "badlands.h"
#include "../daphne.h"
#include "../cpu/mc6809.h"
#include "../io/conout.h"
#include "../ldp-in/ldv1000.h"
#include "../ldp-out/ldp.h"
#include "../video/video.h"

extern ldp *g_ldp;

////////////////

badlands::badlands()
{
	memset(banks, 0xFF, 3);	// fill banks with 0xFF's
	game_type = GAME_BADLANDS;
	disc_fps = 29.97;
	video_overlay = NULL;
	screen_needs_update = true;
	cpu_type = CPU_M6809;
	cpu_hz = 14318180 / 4;	
	nmi_period = (1000.0 / 60.0); // nmi from ?? strobe
	irq_period = (1000.0 / (10.20 * 60.0)); // firq 8 * hblank + irq + nmi
	m_transparent_color = 0;
	firq_on = false;
	irq_on = false;
	nmi_on = false;
}

// we cheat here... we call do_irq 10 times per hblank and do 8 firq's, 1 irq, and 1 nmi
void badlands::do_irq()
{
	static int int_count = 0;


	if (int_count < 8)
	{
		int_count++;
		
		if (firq_on)
		{
			mc6809_firq = 1;
		}
	}
	else if (int_count == 8)
	{
		int_count++;
		
		if (irq_on)
		{
			mc6809_irq = 1;
		}
	}
	else if (int_count == 9)
	{		
		int_count = 0;

		if (nmi_on)
		{
			mc6809_nmi = 1;
		}
	}
}

// we cheat here... we call this once per hblank to update the display
void badlands::do_nmi()
{
	if (screen_needs_update)
	{
		display_update();
	}
}

// does anything special needed to send an FIRQ
void badlands::do_firq()
{
}

Uint8 badlands::cpu_mem_read(Uint16 addr)
{
	char s[81] = {0};

	Uint8 result = cpumem[addr];

	// Dipswitch 2
	if (addr == 0x0000)
	{
		result = banks[2];
	}
	// Dipswitch 1
	else if (addr == 0x0800)
	{
		result = banks[1];
	}

	// Laserdisc (in)
	else if (addr == 0x1000)
	{
		result = read_ldv1000();
//		sprintf(s, "LD-V1000 command read with %x", result);
//		printline(s);
	}

	// controls
	else if (addr == 0x1800)
	{
		result = banks[0];
	}


	return result;
}

void badlands::cpu_mem_write(Uint16 addr, Uint8 value)
{
	char s[81] = {0};
	
	// shoot led
	if (addr == 0x1000)
	{
		if (value)
		{
//			shoot_led = true;
		}
		else
		{
			shoot_led = false;		
		}
	}
	// nmi on
	if (addr == 0x1004)
	{
		if (value)
		{
			nmi_on = true;
		}
		else
		{
			nmi_on = false;		
		}
	}
	// irq on
	else if (addr == 0x1006)
	{
		if (value)
		{
			irq_on = true;
		}
		else
		{
			irq_on = false;		
		}
	}
	// firq on
	else if (addr == 0x1007)
	{
		if (value)
		{
			firq_on = true;
		}
		else
		{
			firq_on = false;		
		}
	}

	// Laserdisc (out)
	else if (addr == 0x0800)
	{
		write_ldv1000(value);
//		sprintf(s, "LD-V1000 command written with %x", value);
//		printline(s);
	}

	// video memory
	else if (addr >= 0x4000 && addr <= 0x47ff)
	{
		screen_needs_update = true;
//		sprintf(s, "Memory being written at %x with %x", addr & 0xFFFF, value & 0xff);
//		printline(s);
	}

	cpumem[addr] = value;
}

// does any video initialization we might need
// returns 'true' if the initialization was successful, false if it failed
bool badlands::video_init()
{
	bool result = false;

	video_overlay = SDL_CreateRGBSurface(SDL_SWSURFACE, BADLANDS_OVERLAY_W, BADLANDS_OVERLAY_H, 8, 0, 0, 0, 0); // create an 8-bit surface
	
	if (video_overlay)
	{
		colors[0].r = (Uint8) (0);
		colors[0].g = (Uint8) (0);
		colors[0].b = (Uint8) (0);

		colors[1].r = (Uint8) (255);
		colors[1].g = (Uint8) (0);
		colors[1].b = (Uint8) (0);

		colors[2].r = (Uint8) (0);
		colors[2].g = (Uint8) (255);
		colors[2].b = (Uint8) (0);

		colors[3].r = (Uint8) (0);
		colors[3].g = (Uint8) (0);
		colors[3].b = (Uint8) (255);

		colors[4].r = (Uint8) (255);
		colors[4].g = (Uint8) (255);
		colors[4].b = (Uint8) (0);

		colors[5].r = (Uint8) (0);
		colors[5].g = (Uint8) (255);
		colors[5].b = (Uint8) (255);

		colors[6].r = (Uint8) (255);
		colors[6].g = (Uint8) (0);
		colors[6].b = (Uint8) (255);

		colors[7].r = (Uint8) (0);
		colors[7].g = (Uint8) (255);
		colors[7].b = (Uint8) (0);

		colors[8].r = (Uint8) (255);
		colors[8].g = (Uint8) (0);
		colors[8].b = (Uint8) (0);

		colors[9].r = (Uint8) (0);
		colors[9].g = (Uint8) (0);
		colors[9].b = (Uint8) (255);

		colors[10].r = (Uint8) (255);
		colors[10].g = (Uint8) (255);
		colors[10].b = (Uint8) (0);

		colors[11].r = (Uint8) (0);
		colors[11].g = (Uint8) (255);
		colors[11].b = (Uint8) (255);

		colors[12].r = (Uint8) (0);
		colors[12].g = (Uint8) (255);
		colors[12].b = (Uint8) (0);

		colors[13].r = (Uint8) (255);
		colors[13].g = (Uint8) (0);
		colors[13].b = (Uint8) (0);

		colors[14].r = (Uint8) (255);
		colors[14].g = (Uint8) (0);
		colors[14].b = (Uint8) (255);

		colors[15].r = (Uint8) (255);
		colors[15].g = (Uint8) (255);
		colors[15].b = (Uint8) (255);

		SDL_SetColors(video_overlay, colors, 0, 16);
		result = true;
	}
	else
	{
		printline("ERROR: Could not create badlands overlay");
	}

	return true;

}

// shuts down any video we might have initialized
void badlands::video_shutdown()
{
	SDL_FreeSurface(video_overlay);
}

// redraws our video on the screen (in case the screen needs to be redrawn)
void badlands::video_repaint()
{
	screen_needs_update = true;
	display_update();
}

// updates badlands's video
void badlands::display_update()
{
	if (screen_needs_update)
	{
		SDL_Rect dest;	// setup place on screen to blit badlands overlay
		dest.x = 0;	// upper left corner
		dest.y = 0;
		dest.w = (unsigned short) video_overlay->w;
		dest.h = (unsigned short) video_overlay->h;

		for (int charx = 6; charx < 47; charx++)
		{
			for (int chary = 2; chary < 32; chary++)
			{
				for (int x=0; x < 4; x++)
				{
					for (int y = 0; y < 8; y++)
					{
							Uint8 left_pixel = static_cast<Uint8>((character[cpumem[chary * 64 + charx + 0x4000]*32+x+4*y] & 0xf0) >> 4);
							Uint8 right_pixel = static_cast<Uint8>((character[cpumem[chary * 64 + charx + 0x4000]*32+x+4*y] & 0x0f));
							*((Uint8 *) video_overlay->pixels + (((chary - 2) * 8 + y) * BADLANDS_OVERLAY_W) + ((charx  - 6) * 8 + x * 2)) = left_pixel;
							*((Uint8 *) video_overlay->pixels + (((chary - 2) * 8 + y) * BADLANDS_OVERLAY_W) + ((charx  - 6) * 8 + x * 2 + 1)) = right_pixel;
					}
				}
			}
		}


		if (shoot_led)
		{
			char* t = "SHOOT!";
			draw_string(t, 8, 17);
		}


		dest.x = 0;	// upper left corner
		dest.y = 0;
		dest.w = (unsigned short) video_overlay->w;
		dest.h = (unsigned short) video_overlay->h;

		// we can't blit if we're using VLDP due to threading conflicts
		if (g_ldp->is_blitting_allowed())
		{
			SDL_BlitSurface(video_overlay, NULL, get_screen(), &dest);
//			SDL_Flip(get_screen());
			SDL_UpdateRects(get_screen(), 1, &dest);
		}
		screen_needs_update = false;
	}
}

// loads the Badlands ROMs
bool badlands::load_roms()
{
	return (load_rom("badlands/badlands.c13", &cpumem[0xc000], 0x2000) &&
		load_rom("badlands/badlands.a14", &cpumem[0xe000], 0x2000) &&

		load_rom("badlands/badlands.c8", &character[0x0000], 0x2000));
}

// this gets called when the user presses a key or moves the joystick
void badlands::input_enable(Uint8 move)
{
	switch (move)
	{
	case SWITCH_UP:
		break;
	case SWITCH_LEFT:
		break;
	case SWITCH_RIGHT:
		break;
	case SWITCH_DOWN:
		break;
	case SWITCH_BUTTON1: // '1' on keyboard
		banks[0] &= ~0x08; 
		break;
	case SWITCH_BUTTON2: // '2' on keyboard
		banks[0] &= ~0x10; 
		break;
	case SWITCH_BUTTON3: // space on keyboard
		banks[0] &= ~0x20; 
		break;
	case SWITCH_COIN1: 
		banks[0] &= ~0x01; 
		break;
	case SWITCH_COIN2: 
		banks[0] &= ~0x02; 
		break;
	case SWITCH_SERVICE: 
		banks[0] &= ~0x04; 
		break;
	case SWITCH_TEST: 
		break;
	default:
		printline("Error, bug in move enable");
		break;
	}
}  

// this gets called when the user releases a key or moves the joystick back to center position
void badlands::input_disable(Uint8 move)
{
	switch (move)
	{
	case SWITCH_UP:
		break;
	case SWITCH_LEFT:
		break;
	case SWITCH_RIGHT:
		break;
	case SWITCH_DOWN:
		break;
	case SWITCH_BUTTON1: // '1' on keyboard
		banks[0] |= 0x08; 
		break;
	case SWITCH_BUTTON2: // '2' on keyboard
		banks[0] |= 0x10; 
		break;
	case SWITCH_BUTTON3: // space on keyboard
		banks[0] |= 0x20; 
		break;
	case SWITCH_COIN1: 
		banks[0] |= 0x01; 
		break;
	case SWITCH_COIN2: 
		banks[0] |= 0x02; 
		break;
	case SWITCH_SERVICE: 
		banks[0] |= 0x04; 
		break;
	case SWITCH_TEST: 
		break;
	default:
		printline("Error, bug in move enable");
		break;
	}
}

// used to set dip switch values
bool badlands::set_bank(unsigned char which_bank, unsigned char value)
{
	bool result = true;
	
	switch (which_bank)
	{
	case 0:	// dipswitch 1
		banks[1] = (unsigned char) (value ^ 0xFF);	// dip switches are active low
		break;
	case 1:	// dipswitch 2
		banks[2] = (unsigned char) (value ^ 0xFF);	// switches are active low
		break;
	default:
		printline("ERROR: Bank specified is out of range!");
		result = false;
		break;
	}
	
	return result;
}

// returns a pointer to Badlands' video overlay surface
// also returns width, height, and bits per pixel of surface
SDL_Surface *badlands::get_video_overlay()
{
	return(video_overlay);
}

Uint8 badlands::get_transparent_color()
{
	return m_transparent_color;
}

SDL_Color *badlands::get_video_colors()
{
	return colors;
}
