/*
 * cliff.cpp
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// cliff.cpp

#include <stdio.h>
#include <string.h>
#include "../daphne.h"
#include "cliff.h"
#include "../io/conout.h"
#include "../sound/sound.h"
#include "../video/tms9128nl.h"
#include "../ldp-in/pr8210.h"
#include "game.h"
#include "../ldp-out/ldp.h"
#include Z80_INCLUDE

extern game *g_game;	// our global game class
extern ldp *g_ldp;

// cliff constructor
cliff::cliff()
{

	// initialize ldp_buf
	memset( (void *) &ldp_buf[0], 0, 10);	// fill ldp_buf with 0's
	ldp_buf_index = 0;
	nmi_enabled = false;
	bank_index = 0;

	banks[0] = 0xFF;  // bank 0 (dip switchs 1,2)
	banks[1] = 0; // anthony has these all on
	banks[2] = 0xFF & DIP22_UNKNOWN & DIP23_UNKNOWN; // bank 2 (dips 20-27)
	// this is how Anthony and Bo's machines are both configured

	banks[3] = 0xFF & DIP18_SHORTSCENES & DIP19_BUYIN; 

	banks[4] = 0xFF & DIP5_DIFFICULTY2; // bank 4 (dips 4-11) aka H11  (first bank for Anthony) all off
	banks[5] = 0xFF; // bank 5 (button data)
	banks[6] = 0xFF; // bank 6 (joystick data)
	banks[7] = 0xFF; // bank 7
	banks[8] = 0xFF; // bank 8
	banks[9] = 0xFF; // bank 9

	disc_fps = 29.97;
	game_type = GAME_CLIFF;
	cpu_type = CPU_Z80;
	cpu_hz = CLIFF_CPU_HZ;
	irq_period = CLIFF_IRQ_PERIOD;
	nmi_period = CLIFF_NMI_PERIOD;
	
	track_elapsed_cycles = true;	// needed to interpret PR-8210 signals

	has_issues = true;
	strcpy(game_issues, "The video overlay is totally broken.  You can, however, complete the game despite the problems.");

	num_sounds = 3;
	sound_name[S_C_CORRECT] = "cliff_correct.wav";
	sound_name[S_C_WRONG] = "cliff_wrong.wav";
	sound_name[S_C_STARTUP] = "cliff_startup.wav";
}

// goal to go constructor
gtg::gtg()
{
	game_type = GAME_GTG;
	has_issues = true;
	strcpy(game_issues, "There are various overlay glitches, and no move beeps, but the game is fully playable.");
	
	disc_side = 1;  //default to side 1 if no -preset is used

}

// notifies us that the PC of cliffy has changed due to a CALL or something like unto it
void cliff::update_pc(Uint32 new_pc)
{

//	Uint16 sp = 0;	// stack pointer

/*
	// potential blip of '1', must investigate
	if (new_pc == 0x1F86)
	{
		sp = (Uint16) Z80_GET_SP;	// we can figure out where we came from by looking at the stack

		// if 0x1F86 was called from 0x1F7F, we know we have got a blip of '1'
		// this is currently the only way I have found to reliably intercept this
		// we check for 0x1F82 because that is the instruction to be returned to after a call from 0x1F7F
		if ((cpumem[sp+1] == 0x1F) && (cpumem[sp] == 0x82))
		{
					ldp_buf[ldp_buf_index+1] = 1;
				// yes, buf+1, not buf (this is not a typo)
		}
	}
*/

}

// when z80 outputs to a port, this gets called
void cliff::port_write(Uint16 Port, Uint8 Value)
{

	char s[81] = { 0 };
	static unsigned int blip_counter = 0;

	Port &= 0xFF;	// strip off high byte

	switch(Port)
	{

	// this unsigned char is written to the screen at the specified coordinates (set by writing to port 0x54)
	case 0x44:
		// if an update is needed
		if (tms9128nl_setvidmem(Value))
		{
			tms9128nl_writechar(Value); // update the screen
		}
		break;

		// play sound, this is still a big hack
	case 0x46:
		if (Z80_GET_PC == 0xEA3)
		{
			sound_play(S_C_CORRECT);
		}
		if (Z80_GET_PC == 0xEB3)
		{
			sound_play(S_C_WRONG);
		}
		break;

		// outputting to port 0x54 is the same as outputting to
		// port #1 on the TMS chip (see reference)
	case 0x54:
		tms9128nl_setvidindex(Value);
		break;

	// get current frame number from the LDP
	case 0x57: //get frame from LDP
            sprintf(s,"Playing Frame: %d   \n",g_ldp->get_current_frame());
            tms9128nl_outcommand(s,43,23);
            break;
	case 0x60:

            // make sure that the bank request is in range
            if (Value < CLIFF_BANK_COUNT)
            {
                bank_index = Value;
            }

            // According to Jeff K's .ASM this selects no banks or something
            else if (Value == 0xF)
            {
            }

            else
            {
                sprintf(s, "A bank out of range was requested! %x", Value);
                printline(s);
            }
            break;

			// PR-8210 port that we can ignore with our current scheme
	case 0x64:
           break;

		   // another PR-8210 port that we can ignore
	case 0x66:
//			printf("0x66 written to at PC %x, value %x\n", Z80_GET_PC, Value);
			break;

	case 0x6A:
			// each blip sends 10 0x30's to port 6A which is how we detect a blip
			if (Value == 0x30)
			{
				blip_counter++;
				if (blip_counter == 10)
				{
					cliff_do_blip();
					blip_counter = 0;
				}
			}
			break;
	case 0x6E:
            //	outcommand("Led On");
//            		printline("LED On");
            break;
	case 0x6F:
            		//printline("LED Off");
            break;
	default:
            sprintf(s, "CLIFF: Unsupported Port Output-> %x : %x", Port, Value);
			printline(s);
            break;
	}
}


Uint8 cliff::port_read(Uint16 Port)
// Called whenever the emulator wants to read from a port
{

    char s[81] = { 0 };
    unsigned char result = 0;
	char frame[6] = { 0 };

	g_ldp->framenum_to_frame(g_ldp->get_current_frame(), frame);
	Port &= 0xFF;	// strip off high byte

    switch(Port)
    {

        // don't know what it is, but it's annoying =]
    case 0x39:
        break;

        // read from video memory
    case 0x45:
        result = tms9128nl_getvidmem();
        break;

    case 0x52: //return high byte of frame
        //if the upper 5 bits are not 0, the program will think
        //the LDP is busy, and not return the rest of the frame
        result = (unsigned char) (frame[0] & 0x0F);
		result |= 0xf8;
        break;
    case 0x51: //return mid byte of frame
		result = (unsigned char) ((frame[1] & 0xF) << 4);
		result |= ( (unsigned char) frame[2] & (unsigned char) 0x0F);
        break;
    case 0x50: //return frame low byte of frame
        result = (unsigned char) ((frame[3] & 0xF) << 4);
		result |= (frame[4] & 0xF);
        break;

    case 0x53:
        break;

		// something to do with video
    case 0x55:
        break;

        // read dip switches
    case 0x62:
        result = banks[bank_index];
        break;
    default:
        sprintf(s, "CLIFF: Unsupported Port Input-> %x (PC is %x)", Port, Z80_GET_PC);
        printline(s);
        break;
    }

    return(result);
}


void cliff::do_nmi()
{
	tms9128nl_update_overlay();	// redraw screen if necessary

	// if the TMS chip is producing interrupts, then we do an NMI
	if (tms9128nl_int_enabled())
	{
		  // SEND QUICK NMI PULSE HERE
		  Z80_ASSERT_NMI;
//		  z80_set_nmi_line(ASSERT_LINE);
//		  z80_execute(1);	// execute a quick instruction
//		  z80_set_nmi_line(CLEAR_LINE);
	}

}

void cliff::do_irq()
{
//	z80_set_irq_line(0, HOLD_LINE);	// assert IRQ line
	Z80_ASSERT_IRQ;
}


void outcommand(char *s) //dangit Matt, stop ripping out my useful routines =]
{ //this needs fixing anyway

	static int row=1,commandnum=0;

    int slength,x;
    char strtoprint[25],string2[25];
    //gotoxy(43,row++);
    if (row==21) row=1;
    slength=strlen(s);
    strcpy(strtoprint,s);
    for(x=slength;x<=18;x++)
        strcat(strtoprint," ");
    sprintf(string2,"%d) %s\n",commandnum++,strtoprint);
	tms9128nl_outcommand(string2,45,row);
}


// process a blip for the PR-8210
void cliff::cliff_do_blip()
{
	static Uint64 total_cycles = 0;

	Uint8 blip_value = 0;
	Uint64 ec = get_elapsed_cycles(total_cycles);
	total_cycles = get_total_cycles();

	// 12000 is our arbitrary upper limit for getting a valid blip (it looks like it's safe to go as low at 7000)
	// The elapsed cycles are greater than this number, we didn't get any blip at all
	if (ec < 12000)
	{
		// 4600 is our arbitrary boundary between a 0 and a 1.  If it's over this number, it's a 1, otherwise it's a 0
		if (ec > 4600)
		{
			blip_value = 1;
		}

		ldp_buf[ldp_buf_index] = blip_value;
		ldp_buf_index++;	// we got a blip so advance the pointer

		// if buffer is filled, send command
		if (ldp_buf_index > 9)
		{
			pr8210_command(ldp_buf);
			ldp_buf_index = 0;
		}
	} // if we got a blip at all
}


// set's service mode in Cliff Hanger
// enabled: 0 = off, 1 = on
void cliff::cliff_set_service_mode(int enabled)
{

	// turn on service mode
	if (enabled)
	{
		printline("Enabling service mode");
		banks[3] &= DIP12_SERVICE;
	}

	// turn off service mode
	else
	{
		printline("Disabling service mode");
		banks[3] |= ~DIP12_SERVICE;
	}

}

// sets switch test mode in Cliffy
// enabled: 0 = off, 1 = on
void cliff::cliff_set_test_mode(int enabled)
{
	if (enabled)
	{
		printline("Enabling test mode");
		banks[3] &= DIP13_SWITCHES;
	}
	else
	{
		printline("Disabling test mode");
		banks[3] |= ~DIP13_SWITCHES;
	}
}

// this gets called when the user presses a key or moves the joystick
void cliff::input_enable(Uint8 move)
{
	static unsigned char service_enabled = 0;	// start disabled
	static unsigned char test_enabled = 0;
	
	switch (move)
	{
	case SWITCH_UP:
		banks[6] &= ~1; // clear bit 0
		break;
	case SWITCH_LEFT:
		banks[6] &= ~8;
		break;
	case SWITCH_RIGHT:
		banks[6] &= ~2;
		break;
	case SWITCH_DOWN:
		banks[6] &= ~4; // clear bit 2
		break;
	case SWITCH_BUTTON1: // '1' on keyboard
		banks[5] &= ~4; // clear bit 2
		break;
	case SWITCH_BUTTON2: // '2' on keyboard
		banks[5] &= ~8;	// clear bit 3
		break;
	case SWITCH_BUTTON3: // spacebar
		banks[5] &= ~16;	// only press one of the hands buttons
								// if we press both, the high score section is ruined
		break;
	case SWITCH_COIN1:
		banks[5] &= ~1; // Coin 1 and Coin 2 might be reversed
		break;
	case SWITCH_COIN2:
		banks[5] &= ~2;  
		break;
	case SWITCH_SERVICE:
		service_enabled = (unsigned char) (service_enabled ^ 1);	// flip bit 0		
		cliff_set_service_mode(service_enabled);		
		break;
	case SWITCH_TEST:
		test_enabled = (unsigned char) (test_enabled ^ 1);
		cliff_set_test_mode(test_enabled);
		break;

	default:
		char s[81] = { 0 };
		sprintf(s, "Bug in Cliffy's input enable.  Input was %x", move);
		printline(s);
		break;
	}
}

// this gets called when the user releases a key or moves the joystick back to center position
void cliff::input_disable(Uint8 move)
{
	switch (move)
	{
	case SWITCH_UP:
		banks[6] |= 1;	// set bit 0
		break;
	case SWITCH_LEFT:
		banks[6] |= 8;
		break;
	case SWITCH_RIGHT:
		banks[6] |= 2;
		break;
	case SWITCH_DOWN:
		banks[6] |= 4;	// set bit 2
		break;
	case SWITCH_BUTTON1:
		banks[5] |= 4;
		break;
	case SWITCH_BUTTON2:
		banks[5] |= 8;
		break;
	case SWITCH_BUTTON3:
		banks[5] |= 16;
		break;
	case SWITCH_COIN1:	
		banks[5] |= 1;
		break;
	case SWITCH_COIN2:
		banks[5] |= 2;
		break;
	case SWITCH_SERVICE:
	case SWITCH_TEST:
		break;
	default:
		char s[81] = { 0 };
		sprintf(s, "Error, bug in Cliffy's input disable, input was %x", move);
		printline(s);
		break;
	}
}

// used to set dip switch values
bool cliff::set_bank(unsigned char which_bank, unsigned char value)
{
	bool result = true;
	
	switch (which_bank)
	{
	case 0:	// h11
		banks[4] = (unsigned char) (value ^ 0xFF);	// switches are active low
		break;
	case 1:	// g11
		banks[3] = (unsigned char) (value ^ 0xFF);	// switches are active low
		break;
	case 2:	// f11
		banks[2] = (unsigned char) (value ^ 0xFF);	// switches are active low
		break;
	case 3:	// e11
		banks[1] = (unsigned char) (value ^ 0xFF);	// switches are active low
		break;
	default:
		printline("ERROR: Bank specified is out of range!");
		result = false;
		break;
	}
	
	return result;
}

bool cliff::video_init()
{
	return(tms9128nl_display_init());
}

void cliff::video_shutdown()
{
	tms9128nl_display_shutdown();
}

void cliff::video_repaint()
{
	tms9128nl_display_repaint();
}

bool cliff::load_roms()
{
	bool result = (load_rom("cliff/cliff_u1.bin", &cpumem[0], 0x2000) &&
		load_rom("cliff/cliff_u2.bin", &cpumem[0x2000], 0x2000) &&
		load_rom("cliff/cliff_u3.bin", &cpumem[0x4000], 0x2000) &&
		load_rom("cliff/cliff_u4.bin", &cpumem[0x6000], 0x2000) &&
		load_rom("cliff/cliff_u5.bin", &cpumem[0x8000], 0x2000));

	// if a cheat has been requested, modify the ROM.  Note, this must be done after roms are loaded
	if (cheat_requested)
	{
		// infinite lives cheat
		cpumem[0xD36] = 0;	// NOP out code that decrements # of remaning lives :)
		cpumem[0xD37] = 0;
		cpumem[0xD38] = 0;

		printline("Cliff hanger infinite lives cheat enabled!");
	}

	return result;

}

// loads goal to go roms
bool gtg::load_roms()
{
	return (load_rom("gtg/gtg.rm0", &cpumem[0], 0x2000) &&
		load_rom("gtg/gtg.rm1", &cpumem[0x2000], 0x2000) &&
		load_rom("gtg/gtg.rm2", &cpumem[0x4000], 0x2000) &&
		load_rom("gtg/gtg.rm3", &cpumem[0x6000], 0x2000) &&
		load_rom("gtg/gtg.rm4", &cpumem[0x8000], 0x2000));
}

SDL_Surface *cliff::get_video_overlay()
{
	return (tms9128nl_get_overlay());
}

Uint8 cliff::get_transparent_color()
{
	return tms9128nl_get_transparent_color();
}

SDL_Color *cliff::get_video_colors()
{
	return tms9128nl_get_colors();
}

// returns how many pixels down to shift the video
int cliff::get_video_row_offset()
{
	return 48;
	// video overlay is 192 pixels high
	// that's 384 doubled
	// video area is 480 pixels high
	// therefore, to center it, we add 48 pixels so the overlay goes from 48 to 432
}

Uint8 gtg::cpu_mem_read(Uint16 addr)
{
 	Uint8 result = cpumem[addr];

	//mem hack for GtG disc side detection
	//(this address is filled by GtG's ISR when a chapter code is received from the
	//frame detect hardware.  The chapter code is '01' for side 1, and '77' for side 2)
 	if (addr == 0xe1ba)
 	{
 		static int a = 0;
 		a++;
 		//don't mess with the value the first few times the location is read -- it screws up the self-test
		if (a > 29) 
 		{
			if (disc_side==1)
				result = 0x01;
			else
				result = 0x77;
 		}
 		//printline("e1ba read");
 	}
 	return result;
}

void gtg::set_preset(int val)
{
	if (val==1) disc_side=1;
	else if (val==2) disc_side=2;
}


