/*
 * cputest.cpp
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// CPU TEST
// by Matt Ownby

// just loads a generic rom
// For CPU core development and testing

//#define USE_CPU_TEST 1

#ifdef USE_CPU_TEST

#include <stdio.h>
#include <stdlib.h>
#include "../daphne.h"
#include "cputest.h"
#include "../cpu/cpu-debug.h"
#include "../cpu/z80.h"
#include "../cpu/m80.h"
#include "../io/conout.h"
#include "../io/input.h"
#include "../timer/timer.h"

Uint32 g_active_cpu = 0;

//////////////////////////////////////////////////////////////////////////

Uint16 Z80_GET_pc()
{
	if (g_active_cpu == 0)
	{
		return Z80_GET_PC;
	}
	else
	{
		return m80_get_pc();
	}
}

void Z80_SET_pc(Uint16 val)
{
	if (g_active_cpu == 0)
	{
		Z80_SET_PC(val);
	}
	else
	{
		m80_set_pc(val);
	}
}

Uint16 Z80_GET_sp()
{
	if (g_active_cpu == 0)
	{
		return Z80_GET_SP;
	}
	else
	{
		return m80_get_sp();
	}
}

void Z80_SET_sp(Uint16 val)
{
	if (g_active_cpu == 0)
	{
		Z80_SET_SP(val);
	}
	else
	{
		m80_set_sp(val);
	}
}

Uint16 Z80_GET_bc()
{
	if (g_active_cpu == 0)
	{
		return Z80_GET_BC;
	}
	else
	{
		return m80_get_reg(M80_BC);
	}
}

Uint16 Z80_GET_de()
{
	if (g_active_cpu == 0)
	{
		return Z80_GET_DE;
	}
	else
	{
		return m80_get_reg(M80_DE);
	}
}

//////////////////////////////////////////////////////////////////////////

cputest::cputest()
{
	cpu_type = CPU_Z80;
	m_cpu_pc = 0x100;	// start the program counter at 0x100 for these tests
	m_compare_cpus = false;
}

Uint8 cputest::cpu_mem_read(Uint16 addr)
{
	if (g_active_cpu == 0)
	{
		return cpumem[addr];
	}
	else
	{
		return cpu2mem[addr];
	}
}

void cputest::cpu_mem_write(Uint16 addr, Uint8 value)
{
	if (g_active_cpu == 0)
	{
		cpumem[addr] = value;
		if ((m_compare_cpus) && (cpumem[addr] != cpu2mem[addr]))
		{
			char s[160];
			sprintf(s, "Memory write discrepancy!  At PC %x, target address %x, z80 value: %x, m80 value: %x, last jump: %x",
				Z80_GET_PC, addr, value, cpu2mem[addr], m_last_jump);
			printline(s);
#ifdef CPU_DEBUG
			set_cpu_trace(1);
#endif
		}
	}
	else
	{
		cpu2mem[addr] = value;
	}
}


Uint8 cputest::port_read(Uint16 port)
{
	return 0;
}

void cputest::update_pc(Uint32 new_pc)
{

	m_last_jump = new_pc;

	// this halts the tests
	if (new_pc == 0)
	{
		static int count = 0;
		int counter = 0;

		if (m_compare_cpus)
		{
			counter = 1;
		}

		if (count > counter)
		{
			outstr("PC went to 0 - ");
			set_quitflag();
		}
		count++;
	}

	// the MSX has a subroutine at 5 that can print messages to the screen (indicating pass or fail)
	else if (new_pc == 5)
	{
		Uint8 command = Z80_GET_bc() & 0x0F;
		Uint8 *membase = NULL;
		
		if (g_active_cpu == 0)
		{
			membase = cpumem;
		}
		else
		{
			membase = cpu2mem;
		}

		// if C is 9 when '5' is called, it means to print a $-terminated message to the screen
		if (command == 9)
		{
			char s[81] = { 0 };
			int i = 0;
			Uint16 srcmsg = Z80_GET_de();	// the source message is stored at DE
			Uint16 sp = Z80_GET_sp();
			Uint16 new_pc = membase[sp] | (membase[sp+1] << 8);	// calculate return value
			
			// strings are terminated with a $ oddly enough
			while ((membase[srcmsg+i] != '$') && (i < 80))
			{
				s[i] = membase[srcmsg+i];
				i++;
			}
			s[i] = 0;	// terminate string
			printline(s);	// and print it to the console
			
			Z80_SET_pc(new_pc);	// return from call
			Z80_SET_sp(sp+2);	// pop return address off stack
		}
		// if C is 0, it means to terminate the program (I think!)
		else if (command == 0)
		{
			printline("Got quit command!\n");
			set_quitflag();
		}
		// print just one character to the console
		else if (command == 2)
		{
			Uint16 sp = Z80_GET_sp();
			Uint16 new_pc = membase[sp] | (membase[sp+1] << 8);	// calculate return value
			outchr(Z80_GET_de() & 0xFF);
			Z80_SET_pc(new_pc);	// return from call
			Z80_SET_sp(sp+2);	// pop return address off stack			
		}
		else
		{
			printf("Warning, unknown command received at 5!\n");
			set_quitflag();
		}
	}

}

void cputest::set_preset(int val)
{
	bool result = false;
	
	switch (val)
	{
	case 0:
		result = load_rom("cputest/mathtest.bin", &cpumem[246], 166);
		break;
	case 1:
		result = load_rom("cputest/cpd.com", &cpumem[0x100], 256);
		break;
	case 2:
		result = load_rom("cputest/cpi.com", &cpumem[0x100], 256);
		break;
	case 3:
		result = load_rom("cputest/daa.com", &cpumem[0x100], 4224);
		break;
	case 4:
		result = load_rom("cputest/ind.com", &cpumem[0x100], 256);
		break;
	case 5:
		result = load_rom("cputest/ini.com", &cpumem[0x100], 256);
		break;
	case 6:
		result = load_rom("cputest/ldd.com", &cpumem[0x100], 256);
		break;
	case 7:
		result = load_rom("cputest/ldi.com", &cpumem[0x100], 256);
		break;
	case 8:
		result = load_rom("cputest/prelim.com", &cpumem[0x100], 1280);
		break;
	case 9:
		result = load_rom("cputest/savage.com", &cpumem[0x100], 14080);
		break;
	case 10:
		result = load_rom("cputest/zexall.com", &cpumem[0x100], 8704);
		break;
	case 11:
		result = load_rom("cputest/zexdoc.com", &cpumem[0x100], 8704);
		result = load_rom("cputest/zexdoc.com", &cpu2mem[0x100], 8704);
		m_compare_cpus = true;
		break;
	default:
		printline("Bad preset!");
		set_quitflag();
		break;
	} // end switch
	
	if (!result)
	{
		printline("ROM load failed.");
		set_quitflag();
	}
}


// initializes our cpu core
void cputest::cpu_init()
{
	// if our cpu is the z80
	if (cpu_type == CPU_Z80)
	{
		// only reset cpu if it is not initalized already
		if (!cpu_initialized)
		{
			OP_RAM = cpumem;	// set memory pointer to the beginning of our RAM
			OP_ROM = cpumem;
			if (m_compare_cpus)
			{
				m80_set_opcode_base(cpu2mem);	// put m80's cpu memory elsewhere
				m80_reset();
				m80_set_irq_callback(generic_m80_irq_callback);
				m80_set_nmi_callback(generic_m80_nmi_callback);
				m80_set_pc(m_cpu_pc);
				z80_reset(NULL);
				z80_set_irq_callback(irq_callback);
				z80_set_pc(m_cpu_pc);
			}
			else
			{
#ifdef USE_M80
				m80_set_opcode_base(cpumem);
				m80_reset();
				m80_set_irq_callback(generic_m80_irq_callback);
				m80_set_nmi_callback(generic_m80_nmi_callback);
				m80_set_pc(m_cpu_pc);
#else
				z80_reset(NULL);
				z80_set_irq_callback(irq_callback);
				z80_set_pc(m_cpu_pc);
#endif
			}
			cpu_initialized = true;
		}
	}
	
	// else we have no other cpus at the moment, so nothing to do
}

#ifdef CPU_DEBUG
#define CPU_BUG(reg)	\
	{	\
		char s[81] = { 0 };	\
		Uint16 m80pc = m80_get_pc();	\
		sprintf(s, "Discrepancy found at PC %x in register %s", z80_get_pc(), #reg);	\
		printline(s);	\
		sprintf(s, "Z80 REG is %x, M80 REG is %x", z80_get_reg(Z80 ## reg), m80_get_reg(M80 ## reg));	\
		printline(s);	\
		sprintf(s, "M80's last 4 bytes are %x %x %x %x",	\
			cpu2mem[m80pc-4], cpu2mem[m80pc-3], cpu2mem[m80pc-2], cpu2mem[m80pc-1]);	\
		printline(s);	\
		set_cpu_trace(1);	\
	}
#else
#define CPU_BUG(reg)
#endif

// our customized cpu execute routine which doesn't bother limiting speed,
// but runs the cpu as fast as possible, while checking for events periodically
void cputest::cpu_execute()
{
	char s[81] = { 0 };
	unsigned int execution_time = 0;
	Uint32 z80_cycles_used = 0, m80_cycles_used = 0;
	Uint16 old_pc = 0;
	
	SDL_Delay(1000);	// let threads get setup
	
	execution_time = refresh_ms_time();

	// if our cpu is the z80
	if (cpu_type == CPU_Z80)
	{
		// if we're running 2 cpus simultaneously
		if (m_compare_cpus)
		{
			int checker = 0;
			
			while (!get_quitflag())
			{
				// execute one instruction at a time and compare all registers
				// if a discrepancy is found, we've found a bug in our cpu core!
				g_active_cpu = 1;
				m80_cycles_used = m80_execute(1);
				g_active_cpu = 0;
				old_pc = z80_get_reg(Z80_PC);
				z80_cycles_used = z80_execute(1);
				
				if (m80_cycles_used != z80_cycles_used)
				{
					printline("cycle mismatch!");
					fprintf(stderr, "z80 cycles : %u, m80 cycles : %u\n", z80_cycles_used, m80_cycles_used);
					fprintf(stderr, "old PC was %x\n", old_pc);
#ifdef CPU_DEBUG
					set_cpu_trace(1);
#endif
				}
				
				if (z80_get_reg(Z80_AF) != m80_get_reg(M80_AF))
				{
					CPU_BUG(_AF);
				}
				if (z80_get_reg(Z80_BC) != m80_get_reg(M80_BC))
				{
					CPU_BUG(_BC);
				}
				if (z80_get_reg(Z80_DE) != m80_get_reg(M80_DE))
				{
					CPU_BUG(_DE);
				}
				if (z80_get_reg(Z80_HL) != m80_get_reg(M80_HL))
				{
					CPU_BUG(_HL);
				}
				if (z80_get_reg(Z80_IX) != m80_get_reg(M80_IX))
				{
					CPU_BUG(_IX);
				}
				if (z80_get_reg(Z80_IY) != m80_get_reg(M80_IY))
				{
					CPU_BUG(_IY);
				}
				if (z80_get_sp() != m80_get_sp())
				{
					CPU_BUG(_SP);
				}
				if (z80_get_pc() != m80_get_pc())
				{
					CPU_BUG(_PC);
				}

				// program runs a lot faster if we don't check input after cycle				
				if (checker > 5000)
				{
					SDL_check_input();
					checker = 0;
				}
				checker++;
				
			} // end while
		}
		else
		{
			// loop until the quit flag is set which means the user wants to quit the program
			while (!get_quitflag())
			{
				Z80_EXECUTE(100000);
				SDL_check_input();
			}
		}
	}

	sprintf(s, "Total execution time was %u ms", elapsed_ms_time(execution_time));
	printline(s);
}

#endif