/*
 * esh.cpp
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// esh.cpp
// by Matt Ownby and Mark Broadhead
//
// Memory map
// 0000 - 3fff Program roms
// e000 - e7ff Work ram - all of this appears to be backed up with a nicad battery
// f000 - f3ff video ram
// f400 - f7ff video control ram

#include <string.h>
#include "esh.h"
#include "../daphne.h"
#include Z80_INCLUDE
#include "../io/conout.h"
#include "../ldp-in/ldv1000.h"
#include "../ldp-out/ldp.h"
#include "../video/video.h"

extern ldp *g_ldp;

////////////////

esh::esh()
{
	memset(banks, 0xFF, 4);	// fill banks with 0xFF's
	
	game_type = GAME_ESH;
	disc_fps = 29.97;
	video_overlay = NULL;
	screen_needs_update = true;
	cpu_type = CPU_Z80;
	cpu_hz = 4000000;	// total guess
	nmi_period = (1000.0 / 60.0); // nmi from LD-V1000 command strobe (likely guess)
	irq_period = (1000.0 / 60.0); // irq from vblank (guess)
	blank_count = 0;
	m_transparent_color = 0;

	has_issues = true;
	strcpy(game_issues, "Game can be completed, but driver is very immature.  Various video/sound problems.");
}

// does anything special needed to send an IRQ
void esh::do_irq()
{
	// Redraws the screen (if needed) on interrupt
	display_update();
	blank_count++;
	if (blank_count >= 10)
	{
		blank_count = 0;
	}
	
	Z80_ASSERT_IRQ;
//	z80_set_irq_line(0, HOLD_LINE);	// assert IRQ line
	
}

// does anything special needed to send an NMI
void esh::do_nmi()
{
	Z80_ASSERT_NMI;
//	z80_set_nmi_line(ASSERT_LINE);
//	z80_execute(1);	// execute a quick instruction
//	z80_set_nmi_line(CLEAR_LINE);
}

Uint8 esh::cpu_mem_read(Uint16 addr)
{
	Uint8 result = cpumem[addr];


	/*
	if (addr > 0x4000 && addr < 0xe000)
	{
		char s[81];
		sprintf(s, "Gotcha!  WAIT started at PC %x\n", CPU_GET_PC);
		printline(s);
	}
	*/


	/*
	if ((addr >= 0x3F8C) && (addr <= 0x3FFF))
	{
		char s[81];
		sprintf(s, "Text string being read (%x) at PC %x\n", addr, CPU_GET_PC);
		printline(s);
	}
	*/

	return result;
}

void esh::cpu_mem_write(Uint16 addr, Uint8 value)
{
	cpumem[addr] = value;

	if (addr >= 0x4000 && addr < 0xe000) 
	{
		char s[81];
		sprintf(s, "Value of %x is being written to %x at PC %x\n", value, addr, Z80_GET_PC);
		printline(s);
	}
	if (addr > 0xf000 && addr <= 0xf7ff)
	{
		screen_needs_update = true;
	}
}

Uint8 esh::port_read(Uint16 port)
{
	char s[81];
	Uint8 result = 0;

	switch (port & 0xFF)
	{
	case 0xF0:	// switch 0?
		result = banks[0];
		break;
	case 0xF1:	// switch 1?
		result = banks[1];
		break;
	case 0xF2:	// switch 2?
		result = 0xff;
		break;
	case 0xF3:	// switch 3?
		result = 0xff;
		break;
	case 0xF4:
		result = read_ldv1000();
		sprintf(s, "%x read from LDV1000", result);
//		printline(s);
		break;
	default:
		sprintf(s, "Port %x being read at PC %x\n", port & 0xFF, Z80_GET_PC);
		printline(s);
		break;
	}

	return result;
}

void esh::port_write(Uint16 port, Uint8 value)
{
	char s[81];

	switch (port & 0xFF)
	{
	case 0xF4:
//		sprintf(s, "%x written to LDV1000", value);
//		printline(s);
		write_ldv1000(value);
		break;
	case 0xF5:	// controls background color?  this is just a guess but it seems related
		switch (value & 0xF)
		{
		case 0:	// transparent background?
			colors[0].r = colors[0].g = colors[0].b = 0;
			m_transparent_color = 0;
			break;
		case 2:	// black background?
			colors[0].r = colors[0].g = colors[0].b = 0;
			m_transparent_color = 31;
			break;
		case 0xC:	// blue background?
			colors[0].r = 0;
			colors[0].g = 0;
			colors[0].b = 255;
			m_transparent_color = 31;	// set it to an unused color
			break;
		case 0xE:	// another blue'ish background?
			colors[0].r = 50;	// make it a little different for the heck of it
			colors[0].g = 0;
			colors[0].b = 255;
			break;
		default:
			{
				char s[81];
				sprintf(s, "Unknown background color requested : %x", value);
				printline(s);
			}
			break;
		}
		SDL_SetColors(video_overlay, colors, 0, 1);	// set color #0 (background) to new color value
		screen_needs_update = true;
		break;
	case 0xF8:
	case 0xFE:
	case 0xFF:
		// 0 and 1 are rapidly written here, not sure why
		break;
	default:
		sprintf(s, "Port %x being written at PC %x with a value of %x\n", port & 0xFF, Z80_GET_PC, value);
		printline(s);
		break;
	}
}

// does any video initialization we might need
// returns 'true' if the initialization was successful, false if it failed
bool esh::video_init()
{
	bool result = false;

	video_overlay = SDL_CreateRGBSurface(SDL_SWSURFACE, ESH_OVERLAY_W, ESH_OVERLAY_H, 8, 0, 0, 0, 0); // create an 8-bit surface
	
	if (video_overlay)
	{
		// make a grayscale color palette because we don't know what the real palette is yet
		for (int i = 0; i < ESH_COLOR_COUNT; i++)
		{
			colors[i].r = (Uint8) (i << 3);
			colors[i].g = (Uint8) (i << 3);
			colors[i].b = (Uint8) (i << 3);
		}

		colors[0].r = 0;	// background color, this is supposed to be blue sometimes
		colors[0].b = 0;
		colors[0].g = 0;

		colors[1].r = 255;	// red (color of the heart)
		colors[1].g = 0;
		colors[1].b = 0;

		colors[2].r = 100;	// the heart's shoes, not sure what color it should be
		colors[2].g = 100;
		colors[2].b = 150;

		colors[3].r = 230;	// off whitish (the heart's hat)
		colors[3].g = 230;
		colors[3].b = 230;

		colors[4].r = 209;	// tannish/orange (arms and legs of heart)
		colors[4].g = 206;
		colors[4].b = 123;

		colors[7].r = 255;	// foreground color, making it white
		colors[7].g = 255;
		colors[7].b = 255;

		SDL_SetColors(video_overlay, colors, 0, ESH_COLOR_COUNT);
		result = true;
	}
	else
	{
		printline("ERROR: Could not create esh overlay");
	}

	return true;

}

// shuts down any video we might have initialized
void esh::video_shutdown()
{
	SDL_FreeSurface(video_overlay);
}

// redraws our video on the screen (in case the screen needs to be redrawn)
void esh::video_repaint()
{
	screen_needs_update = true;
	display_update();
}

// updates esh's video
void esh::display_update()
{
//	if (screen_needs_update)
	{
		SDL_Rect dest;	// setup place on screen to blit esh overlay
		dest.x = 0;	// upper left corner
		dest.y = 0;
		dest.w = (unsigned short) video_overlay->w;
		dest.h = (unsigned short) video_overlay->h;

		for (int charx = 0; charx < 32; charx++)
		{
			for (int chary = 0; chary < 32; chary++)
			{
				for (int y = 0; y < 8; y++)
				{

					// Memory region F400-F7FF controls the video
					//
					// Low nibble:
					// palette
					//
					// High nibble:
					// bit 4	use tile set 2 (tiles 256-511)
					// bit 5	unknown
					// bit 6	blink every other line
					// bit 7	blink the whole character
					
					if ((cpumem[chary * 32 + charx + 0xf400] !=0) && (cpumem[chary * 32 + charx + 0xf400] !=0x19) && (cpumem[chary * 32 + charx + 0xf400] !=0x89) && (cpumem[chary * 32 + charx + 0xf400] !=0x09) && (cpumem[chary * 32 + charx + 0xf400] !=0x49)) 
					{
						char s[81];
						sprintf(s, "Video control value %x", (cpumem[chary * 32 + charx + 0xf400]));
//						printline(s);
					}
					

					// the following is a very bad of doing things - FIX ME!
					Uint8 pixel1, pixel2, pixel3, pixel4, pixel5, pixel6, pixel7, pixel8; 
					
					// this is just an approximation of the character blinking
					if ((!((cpumem[chary * 32 + charx + 0xf400] >> 0x07) & 0x01) || blank_count < 7) && ((!(((cpumem[chary * 32 + charx + 0xf400] >> 0x06) & 0x01) && y%2 == 0) || blank_count < 3)) && ((cpumem[chary * 32 + charx + 0xf400]) & 0x0f))
					{
						// tile set 1
						if (!((cpumem[chary * 32 + charx + 0xf400] >> 0x04) & 0x01))
						{
							pixel1 = static_cast<Uint8>(((character[cpumem[chary * 32 + charx + 0xf000]*8+y] & 0x80) >> 7) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x1000] & 0x80) >> 6) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x2000] & 0x80) >> 5));
							pixel2 = static_cast<Uint8>(((character[cpumem[chary * 32 + charx + 0xf000]*8+y] & 0x40) >> 6) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x1000] & 0x40) >> 5) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x2000] & 0x40) >> 4));
							pixel3 = static_cast<Uint8>(((character[cpumem[chary * 32 + charx + 0xf000]*8+y] & 0x20) >> 5) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x1000] & 0x20) >> 4) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x2000] & 0x20) >> 3));
							pixel4 = static_cast<Uint8>(((character[cpumem[chary * 32 + charx + 0xf000]*8+y] & 0x10) >> 4) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x1000] & 0x10) >> 3) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x2000] & 0x10) >> 2));
							pixel5 = static_cast<Uint8>(((character[cpumem[chary * 32 + charx + 0xf000]*8+y] & 0x08) >> 3) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x1000] & 0x08) >> 2) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x2000] & 0x08) >> 1));
							pixel6 = static_cast<Uint8>(((character[cpumem[chary * 32 + charx + 0xf000]*8+y] & 0x04) >> 2) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x1000] & 0x04) >> 1) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x2000] & 0x04) >> 0));
							pixel7 = static_cast<Uint8>(((character[cpumem[chary * 32 + charx + 0xf000]*8+y] & 0x02) >> 1) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x1000] & 0x02) >> 0) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x2000] & 0x02) << 1));
							pixel8 = static_cast<Uint8>(((character[cpumem[chary * 32 + charx + 0xf000]*8+y] & 0x01) >> 0) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x1000] & 0x01) << 1) | ((character[cpumem[chary * 32 + charx + 0xf000]*8+y+0x2000] & 0x01) << 2));
						}
						// tile set 2
						else
						{
							pixel1 = static_cast<Uint8>(((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y] & 0x80) >> 7) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x1000] & 0x80) >> 6) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x2000] & 0x80) >> 5));
							pixel2 = static_cast<Uint8>(((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y] & 0x40) >> 6) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x1000] & 0x40) >> 5) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x2000] & 0x40) >> 4));
							pixel3 = static_cast<Uint8>(((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y] & 0x20) >> 5) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x1000] & 0x20) >> 4) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x2000] & 0x20) >> 3));
							pixel4 = static_cast<Uint8>(((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y] & 0x10) >> 4) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x1000] & 0x10) >> 3) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x2000] & 0x10) >> 2));
							pixel5 = static_cast<Uint8>(((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y] & 0x08) >> 3) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x1000] & 0x08) >> 2) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x2000] & 0x08) >> 1));
							pixel6 = static_cast<Uint8>(((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y] & 0x04) >> 2) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x1000] & 0x04) >> 1) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x2000] & 0x04) >> 0));
							pixel7 = static_cast<Uint8>(((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y] & 0x02) >> 1) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x1000] & 0x02) >> 0) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x2000] & 0x02) << 1));
							pixel8 = static_cast<Uint8>(((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y] & 0x01) >> 0) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x1000] & 0x01) << 1) | ((character[(cpumem[chary * 32 + charx + 0xf000]+0x100)*8+y+0x2000] & 0x01) << 2));
						}
					}
					else
					{
						pixel1 = pixel2 = pixel3 = pixel4 = pixel5 = pixel6 = pixel7 = pixel8 = 0;
					}
					*((Uint8 *) video_overlay->pixels + ((chary * 8 + y) * ESH_OVERLAY_W) + (charx * 8 + 0)) = pixel1;
					*((Uint8 *) video_overlay->pixels + ((chary * 8 + y) * ESH_OVERLAY_W) + (charx * 8 + 1)) = pixel2;
					*((Uint8 *) video_overlay->pixels + ((chary * 8 + y) * ESH_OVERLAY_W) + (charx * 8 + 2)) = pixel3;
					*((Uint8 *) video_overlay->pixels + ((chary * 8 + y) * ESH_OVERLAY_W) + (charx * 8 + 3)) = pixel4;
					*((Uint8 *) video_overlay->pixels + ((chary * 8 + y) * ESH_OVERLAY_W) + (charx * 8 + 4)) = pixel5;
					*((Uint8 *) video_overlay->pixels + ((chary * 8 + y) * ESH_OVERLAY_W) + (charx * 8 + 5)) = pixel6;
					*((Uint8 *) video_overlay->pixels + ((chary * 8 + y) * ESH_OVERLAY_W) + (charx * 8 + 6)) = pixel7;
					*((Uint8 *) video_overlay->pixels + ((chary * 8 + y) * ESH_OVERLAY_W) + (charx * 8 + 7)) = pixel8;
				}
			}
		}

		dest.x = 0;	// upper left corner
		dest.y = 0;
		dest.w = (unsigned short) video_overlay->w;
		dest.h = (unsigned short) video_overlay->h;

		// we can't blit if we're using VLDP due to threading conflicts
		if (g_ldp->is_blitting_allowed())
		{
			SDL_BlitSurface(video_overlay, NULL, get_screen(), &dest);
//			SDL_Flip(get_screen());
			SDL_UpdateRects(get_screen(), 1, &dest);
		}
		screen_needs_update = false;
	}
}

// loads the Esh's ROMs
bool esh::load_roms()
{
	
	return (load_rom("esh/h8_is1.bin", &cpumem[0], 0x2000) &&
		load_rom("esh/f8_is2.bin", &cpumem[0x2000], 0x2000) &&
		load_rom("esh/m3_a.bin", &character[0x0000], 0x1000) &&
		load_rom("esh/l3_b.bin", &character[0x1000], 0x1000) &&
		load_rom("esh/k3_c.bin", &character[0x2000], 0x1000));

}

// this gets called when the user presses a key or moves the joystick
void esh::input_enable(Uint8 move)
{
	switch (move)
	{
	case SWITCH_UP:
		banks[1] &= ~0x01;
		break;
	case SWITCH_LEFT:
		banks[1] &= ~0x04;
		break;
	case SWITCH_RIGHT:
		banks[1] &= ~0x08;
		break;
	case SWITCH_DOWN:
		banks[1] &= ~0x02;
		break;
	case SWITCH_BUTTON1: // '1' on keyboard
		banks[0] &= ~0x04;
		break;
	case SWITCH_BUTTON2: // '2' on keyboard
		break;
	case SWITCH_BUTTON3: // space on keyboard
		banks[1] &= ~0x10;
		break;
	case SWITCH_COIN1: 
		banks[0] &= ~0x01;
		break;
	case SWITCH_COIN2: 
		banks[0] &= ~0x02;
		break;
	case SWITCH_SERVICE: 
		break;
	case SWITCH_TEST: 
		banks[0] &= ~0x10;
		break;
	default:
		printline("Error, bug in move enable");
		break;
	}
}  

// this gets called when the user releases a key or moves the joystick back to center position
void esh::input_disable(Uint8 move)
{
	switch (move)
	{
	case SWITCH_UP:
		banks[1] |= 0x01;
		break;
	case SWITCH_LEFT:
		banks[1] |= 0x04;
		break;
	case SWITCH_RIGHT:
		banks[1] |= 0x08;
		break;
	case SWITCH_DOWN:
		banks[1] |= 0x02;
		break;
	case SWITCH_BUTTON1: // '1' on keyboard
		banks[0] |= 0x04;
		break;
	case SWITCH_BUTTON2: // '2' on keyboard
		break;
	case SWITCH_BUTTON3: // space on keyboard
		banks[1] |= 0x10;
		break;
	case SWITCH_COIN1: 
		banks[0] |= 0x01;
		break;
	case SWITCH_COIN2: 
		banks[0] |= 0x02;
		break;
	case SWITCH_SERVICE: 
		break;
	case SWITCH_TEST: 
		banks[0] |= 0x10;
		break;
	default:
		printline("Error, bug in move enable");
		break;
	}
}

// returns a pointer to Esh's video overlay surface
// also returns width, height, and bits per pixel of surface
SDL_Surface *esh::get_video_overlay()
{
	return(video_overlay);
}

Uint8 esh::get_transparent_color()
{
	return m_transparent_color;
}

SDL_Color *esh::get_video_colors()
{
	return colors;
}
