/*
 * firefox.cpp
 *
 * Copyright (C) 2001 Mark Broadhead
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// firefox.cpp
// by Mark Broadhead
//

#include <string.h>
#include "firefox.h"
#include "../daphne.h"
#include "../cpu/mc6809.h"
#include "../io/conout.h"
#include "../ldp-in/ldv1000.h"
#include "../ldp-out/ldp.h"
#include "../video/video.h"


extern ldp *g_ldp;

////////////////

firefox::firefox()
{
	memset(banks, 0xFF, 4);	// fill banks with 0xFF's
	banks[1] = 0x1f;
//	game_type = GAME_BADLANDS;
	disc_fps = 29.97;
	video_overlay = NULL;
	screen_needs_update = true;
	cpu_type = CPU_M6809;
	cpu_hz = 14318180 / 8;	
	nmi_period = 1000.0 / (60.0);
	irq_period = 1000.0 / (60.0 * 6.0);
	m_transparent_color = 0;
	current_bank = 0x0000;
}

void firefox::do_irq()
{
	mc6809_irq = 1;
}

// we cheat here... we call this once per hblank to update the display
void firefox::do_nmi()
{
	display_update();
}

// does anything special needed to send an FIRQ
void firefox::do_firq()
{
}

Uint8 firefox::cpu_mem_read(Uint16 addr)
{
//	char s[81] = {0};

	Uint8 result = cpumem[addr];

	// Inputs 1
	if (addr == 0x4100)
	{
		result = banks[0];
	}

	// Inputs 2
	else if (addr == 0x4101)
	{
		result = banks[1];
	}

	// Option Sw 1 
	else if (addr == 0x4103)
	{
//		result = banks[2];
	}

	// Option Sw 1 
	else if (addr == 0x4104)
	{
//		result = banks[3];
	}

	// Table Rom
	else if (addr >= 0x3000 && addr <= 0x3fff)
	{
		result = rombank[current_bank + (addr - 0x3000)];
//		sprintf(s, "current bank %x reading cpumem[%x], returning rombank[%x]", current_bank, addr, current_bank + (addr - 0x3000));
//		printline(s);
	}

	return result;
}

void firefox::cpu_mem_write(Uint16 addr, Uint8 value)
{
	char s[81] = {0};

	// Graphics/Alphanumerics RAM
	if (addr >= 0x1000 && addr <= 0x1fff)
	{
		screen_needs_update = true;
	}

	// Color Ram
	else if (addr >= 0x2800 && addr <= 0x2aff)
	{
		palette_modified = true;
	}
	
	// Rom paging @ 3000
	else if (addr == 0x4290)
	{
	switch(value & 0x1c)
	{
		case 0x00: /* T0 */
			current_bank = (0x1000 * (value & 0x3));
//			sprintf(s, "bank switch, T0: %x\n", value & 0x3);
//			printline(s);
			break;
		case 0x04: /* T1 */
			current_bank = (0x4000 + (0x1000 * (value & 0x3)));
//			sprintf(s, "bank switch, T1: %x\n", value & 0x3);
//			printline(s);
			break;
		case 0x10: /* T4 */
			current_bank = (0x8000 + (0x1000 * (value & 0x3)));
//			sprintf(s, "bank switch, T4: %x\n", value & 0x3);
//			printline(s);
			break;
		case 0x14: /* T5 */
			current_bank = (0xc000 + (0x1000 * (value & 0x3)));
//			sprintf(s, "bank switch, T5: %x\n", value & 0x3);
//			printline(s);
			break;
		default:
			sprintf(s,"Invalid bank switch, %x", value);
			printline(s);
			break;
	}
	}

	cpumem[addr] = value;
}

// does any video initialization we might need
// returns 'true' if the initialization was successful, false if it failed
bool firefox::video_init()
{
	bool result = false;

	video_overlay = SDL_CreateRGBSurface(SDL_SWSURFACE, FIREFOX_OVERLAY_W, FIREFOX_OVERLAY_H, 8, 0, 0, 0, 0); // create an 8-bit surface
	
	if (video_overlay)
	{
		result = true;
	}
	else
	{
		printline("ERROR: Could not create firefox overlay");
	}

	return true;

}

void firefox::recalc_palette()
{
	for (int x = 0; x < 256; x++)
	{
		colors[x].r = cpumem[0x2800 + x];
		colors[x].g = cpumem[0x2900 + x];
		colors[x].b = cpumem[0x2a00 + x];
	}

	SDL_SetColors(video_overlay, colors, 0, 256);
}

// shuts down any video we might have initialized
void firefox::video_shutdown()
{
	SDL_FreeSurface(video_overlay);
}

// redraws our video on the screen (in case the screen needs to be redrawn)
void firefox::video_repaint()
{
	screen_needs_update = true;
	display_update();
}

// updates firefox's video
void firefox::display_update()
{
	if (palette_modified)
	{
		recalc_palette();
	}

	if (screen_needs_update)
	{
		SDL_Rect dest;	// setup place on screen to blit firefox overlay
		dest.x = 0;	// upper left corner
		dest.y = 0;
		dest.w = (unsigned short) video_overlay->w;
		dest.h = (unsigned short) video_overlay->h;

		for (int charx = 10; charx < 50; charx++)
		{
			for (int chary = 0; chary < 60; chary++)
			{
				for (int x=0; x < 4; x++)
				{
					for (int y = 0; y < 8; y++)
					{
						Uint8 left_pixel = static_cast<Uint8>((character[cpumem[chary * 64 + charx + 0x1000]*32+x+4*y] & 0xf0) >> 4);
						Uint8 right_pixel = static_cast<Uint8>((character[cpumem[chary * 64 + charx + 0x1000]*32+x+4*y] & 0x0f));
//						left_pixel = (((left_pixel & 0x08) >> 3) | ((left_pixel & 0x04) >> 1) | ((left_pixel & 0x02) << 1)  | ((left_pixel & 0x01) << 3));  
//						right_pixel = (((right_pixel & 0x08) >> 3) | ((right_pixel & 0x04) >> 1) | ((right_pixel & 0x02) << 1)  | ((right_pixel & 0x01) << 3));  
						*((Uint8 *) video_overlay->pixels + (((chary) * 8 + y) * FIREFOX_OVERLAY_W) + ((charx - 10) * 16 + x * 4)) = left_pixel;
						*((Uint8 *) video_overlay->pixels + (((chary) * 8 + y) * FIREFOX_OVERLAY_W) + ((charx - 10) * 16 + x * 4 + 1)) = left_pixel;
						*((Uint8 *) video_overlay->pixels + (((chary) * 8 + y) * FIREFOX_OVERLAY_W) + ((charx - 10) * 16 + x * 4 + 2)) = right_pixel;
						*((Uint8 *) video_overlay->pixels + (((chary) * 8 + y) * FIREFOX_OVERLAY_W) + ((charx - 10) * 16 + x * 4 + 3)) = right_pixel;
					}
				}
			}
		}

		dest.x = 0;	// upper left corner
		dest.y = 0;
		dest.w = (unsigned short) video_overlay->w;
		dest.h = (unsigned short) video_overlay->h;

		// we can't blit if we're using VLDP due to threading conflicts
		if (g_ldp->is_blitting_allowed())
		{
			SDL_BlitSurface(video_overlay, NULL, get_screen(), &dest);
//			SDL_Flip(get_screen());
			SDL_UpdateRects(get_screen(), 1, &dest);
		}
		screen_needs_update = false;
	}
}

// loads the firefox ROMs
bool firefox::load_roms()
{
		// main program
	return (load_rom("firefox/136026.109", &cpumem[0x4000], 0x4000) &&
		load_rom("firefox/136026.110", &cpumem[0x8000], 0x4000) &&
		load_rom("firefox/136026.111", &cpumem[0xc000], 0x4000) &&

		// rom banks
		load_rom("firefox/136026.101", &rombank[0x0000], 0x4000) &&
		load_rom("firefox/136026.102", &rombank[0x4000], 0x4000) &&
		load_rom("firefox/136026.105", &rombank[0x8000], 0x4000) &&
		load_rom("firefox/136026.106", &rombank[0xc000], 0x4000) &&

		// characters
		load_rom("firefox/136026.125", &character[0x0000], 0x2000));
}

// this gets called when the user presses a key or moves the joystick
void firefox::input_enable(Uint8 move)
{
	switch (move)
	{
	case SWITCH_UP:
		break;
	case SWITCH_LEFT:
		break;
	case SWITCH_RIGHT:
		break;
	case SWITCH_DOWN:
		break;
	case SWITCH_BUTTON1: // '1' on keyboard
		banks[0] &= ~0x08; 
		break;
	case SWITCH_BUTTON2: // '2' on keyboard
		banks[0] &= ~0x10; 
		break;
	case SWITCH_BUTTON3: // space on keyboard
		banks[0] &= ~0x20; 
		break;
	case SWITCH_COIN1: 
		banks[0] &= ~0x01; 
		break;
	case SWITCH_COIN2: 
		banks[0] &= ~0x02; 
		break;
	case SWITCH_SERVICE: 
		banks[0] &= ~0x04; 
		break;
	case SWITCH_TEST: 
		banks[1] &= 0x10; 
		break;
	default:
		printline("Error, bug in move enable");
		break;
	}
}  

// this gets called when the user releases a key or moves the joystick back to center position
void firefox::input_disable(Uint8 move)
{
	switch (move)
	{
	case SWITCH_UP:
		break;
	case SWITCH_LEFT:
		break;
	case SWITCH_RIGHT:
		break;
	case SWITCH_DOWN:
		break;
	case SWITCH_BUTTON1: // '1' on keyboard
		banks[0] |= 0x08; 
		break;
	case SWITCH_BUTTON2: // '2' on keyboard
		banks[0] |= 0x10; 
		break;
	case SWITCH_BUTTON3: // space on keyboard
		banks[0] |= 0x20; 
		break;
	case SWITCH_COIN1: 
		banks[0] |= 0x01; 
		break;
	case SWITCH_COIN2: 
		banks[0] |= 0x02; 
		break;
	case SWITCH_SERVICE: 
		banks[0] |= 0x04; 
		break;
	case SWITCH_TEST: 
		banks[1] |= 0x10; 
		break;
	default:
		printline("Error, bug in move enable");
		break;
	}
}

// used to set dip switch values
bool firefox::set_bank(unsigned char which_bank, unsigned char value)
{
	bool result = true;
	
	switch (which_bank)
	{
	case 0:	// dipswitch 1
		banks[2] = (unsigned char) (value ^ 0xFF);	// dip switches are active low
		break;
	case 1:	// dipswitch 2
		banks[3] = (unsigned char) (value ^ 0xFF);	// switches are active low
		break;
	default:
		printline("ERROR: Bank specified is out of range!");
		result = false;
		break;
	}
	
	return result;
}

// returns a pointer to firefox' video overlay surface
// also returns width, height, and bits per pixel of surface
SDL_Surface *firefox::get_video_overlay()
{
	return(video_overlay);
}

Uint8 firefox::get_transparent_color()
{
	return m_transparent_color;
}

SDL_Color *firefox::get_video_colors()
{
	return colors;
}
