/*
 * game.cpp
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// game.cpp
// by Matt Ownby

#include <stdio.h>
#include "../daphne.h"
#include "../io/conout.h"
#include "game.h"
#include Z80_INCLUDE
#include "../cpu/mamewrap.h"
#include "../cpu/6809infc.h"
#include "../cpu/mc6809.h"
#include "../timer/timer.h"
#include "../io/input.h"

///////////////////////////////////////////////////////////

#ifdef CPU_DEBUG

// a generic example of how to setup debug names
struct addr_name game_addr_names[] =
{
	{ "Start", 0x0000 },
	{ NULL, 0 }
};
#endif

// breaks cpu execution into trace mode from whichever point the program is at
// This function is executed by typing "break" at the daphne console
// This can't be part of the class because I can't figure out how to make it work =]
void g_cpu_break (char *s)
{

	// I put this here simply to get rid of warning messages
	if (s)
	{
	}

#ifdef CPU_DEBUG
	set_cpu_trace(1);
#else
	printline("You have to compile with CPU_DEBUG defined to use the debugger!");
#endif

}


///////////////////////////////////////////////////////////

game::game()
{

	disc_fps = 0.0;
	game_type = GAME_UNDEFINED;
	cpu_type = CPU_UNDEFINED;
	cpu_initialized = false;
	cpu_hz = 0;
	irq_period = 0.0;
	nmi_period = 0.0;
	m_cpu_pc = 0;
	has_issues = false;	// assume game has no issues unless we specify some
#ifdef CPU_DEBUG
	addr_names = game_addr_names;
#endif
	
	cheat_requested = false;	// no cheats until specifically requested
	track_elapsed_cycles = true;
	m_tracked_elapsed_cycles = 0;
}

// does anything special needed to send an IRQ
void game::do_irq()
{

}

// does anything special needed to send an NMI
void game::do_nmi()
{

}

// reads a byte from the cpu's memory
Uint8 game::cpu_mem_read(Uint16 addr)
{
	return cpumem[addr];
}

// writes a byte to the cpu's memory
void game::cpu_mem_write(Uint16 addr, Uint8 value)
{
	cpumem[addr] = value;
}

// reads a byte from the cpu's port
Uint8 game::port_read(Uint16 port)
{
	char s[81] = { 0 };

	port &= 0xFF;	
	sprintf(s, "ERROR: CPU port %x read requested, but this function is unimplemented!", port);
	printline(s);

	return(0);
}

// writes a byte to the cpu's port
void game::port_write(Uint16 port, Uint8 value)
{
	char s[81] = { 0 };

	port &= 0xFF;
	sprintf(s, "ERROR: CPU port %x write requested (value %x) but this function is unimplemented!", port, value);
	printline(s);

}

// notifies us of the new Program Counter (which most games usually don't care about)
void game::update_pc(Uint32 new_pc)
{
	// I put this here to get rid of warnings
	if (new_pc)
	{
	}
}

void game::input_enable(Uint8 input)
{
	// get rid of warnings
	if (input)
	{
	}

	printline("Warning: generic input_enable function called, does nothing");

}

void game::input_disable(Uint8 input)
{
	// get rid of warnings
	if (input)
	{
	}

	printline("Warning: generic input_disable function called, does nothing");

}


// does any video initialization we might need
// returns 'true' if the initialization was successful, false if it failed
bool game::video_init()
{
	return(true);
}

// shuts down any video we might have initialized
void game::video_shutdown()
{
}

// redraws our video on the screen (in case the screen needs to be redrawn)
void game::video_repaint()
{
}

// generic preset function, does nothing
void game::set_preset(int preset)
{
	// get rid of warnings
	if (preset)
	{
	}

	printline("NOTE: There are no presets defined for the game you have chosen!");
}

// returns false if there was an error trying to set the bank value (ie if the user requested an illegal bank)
bool game::set_bank(unsigned char which_bank, unsigned char value)
{
	bool result = false;
	
	printline("NOTE: The ability to set bank values is not supported in this game");

	// get rid of warnings
	if (value && which_bank)
	{
	}
	
	return result;
}

// routine to load roms
// returns true if there were no errors
bool game::load_roms()
{
	printline("NOTE: ROM loading is not supported with this game");
	
	return(true);
}

// initializes our cpu core
void game::cpu_init()
{
	// if our cpu is the z80
	if (cpu_type == CPU_Z80)
	{
		// only reset cpu if it is not initalized already
		if (!cpu_initialized)
		{
#ifdef USE_M80
			m80_set_opcode_base(cpumem);
			m80_reset();
			m80_set_irq_callback(generic_m80_irq_callback);
			m80_set_nmi_callback(generic_m80_nmi_callback);
			m80_set_pc(m_cpu_pc);
#else
			OP_RAM = cpumem;	// set memory pointer to the beginning of our RAM
			OP_ROM = cpumem;
			z80_reset(NULL);
			z80_set_irq_callback(generic_z80_irq_callback);
			z80_set_pc(m_cpu_pc);
#endif
			cpu_initialized = true;
		}
	}
	if (cpu_type == CPU_M6809)
	{
		if (!cpu_initialized)
		{
			initialize_m6809();

			cpu_initialized = true;
		}
	}
	// else we have no other cpus at the moment, so nothing to do
}




// executes our cpu core.  this function only returns when the program exits
void game::cpu_execute()
{
	cycles_per_ms = cpu_hz / 1000;	// get # of cycles per ms instead of s
	cycles_per_irq = (Uint32) (cycles_per_ms * irq_period);
	cycles_per_nmi = (Uint32) (cycles_per_ms * nmi_period);

	cpu_timer = refresh_ms_time();	// so the cpu doesn't run too quickly when we first start
	flush_timers();				// so we we don't get flooded by NMI's/IRQ's when we first start

	// if our cpu is the z80
	if (cpu_type == CPU_Z80)
	{
		// loop until the quit flag is set which means the user wants to quit the program
		while (!get_quitflag())
		{
			cpu_loop(Z80_EXECUTE(cycles_per_ms));
			SDL_check_input();
		}
	}
	// if our cpu is the z80
	else if (cpu_type == CPU_M6809)
	{
		// loop until the quit flag is set which means the user wants to quit the program
		while (!get_quitflag())
		{
			cpu_loop(mc6809_StepExec(cycles_per_ms));
			SDL_check_input();
		}
	}
	else
	{
		printline("You have not defined which CPU you are using.");
		printline("DAPHNE will exit.");
	}
}

//#define CPU_DIAG

// called at regular intervals to service interrupt requests and slow the CPU down to the proper speed
void game::cpu_loop(Uint32 elapsed_cycles)
{

	unsigned int actual_elapsed_ms = 0;
	bool nmi_asserted = false;

	// this chunk of code tests to make sure the CPU is running
	// at the proper speed
#ifdef CPU_DIAG
	char s[160] = { 0 };
	static unsigned int t_cycle_count = 0;	// for speed test
	static unsigned int old_time = refresh_ms_time(); // " " "
	static unsigned int irq_count = 0;
	static unsigned int nmi_count = 0;
	static double compounded_mhz = 0.0;	// sum of all mhz ratings so we can get an average
	static int report_count = 0;	// how many reports have been given

	t_cycle_count += elapsed_cycles;

#define CPU_DIAG_ACCURACY 24000000

	// the bigger the #, the more accurate the result
	if (t_cycle_count >= CPU_DIAG_ACCURACY)
	{
		double cur_mhz = ((double) t_cycle_count / (double) elapsed_ms_time(old_time)) * 0.001;
		compounded_mhz += cur_mhz;
		report_count++;

		sprintf(s,"cycle_count = %d, time = %d ms, MHz = %f, avg MHz = %f",
			t_cycle_count, elapsed_ms_time(old_time),
			cur_mhz, (compounded_mhz / report_count));
		printline(s);
		sprintf(s, "IRQ's = %d, NMI's = %d", irq_count, nmi_count);
		printline(s);
		old_time = refresh_ms_time();
		t_cycle_count -= CPU_DIAG_ACCURACY;
		irq_count = 0;
		nmi_count = 0;
	}
#endif

	// if we are to keep track of elapsed cycles, then do so
	if (track_elapsed_cycles)
	{
		add_elapsed_cycles(elapsed_cycles);
	}
	s_cycle_count += elapsed_cycles;
	irq_cycle_count += elapsed_cycles;
	nmi_cycle_count += elapsed_cycles;

	// BEGIN FORCING EMULATOR TO RUN AT PROPER SPEED

	// if we've run 1 ms worth of cycles, wait for 1 ms to pass
	if (s_cycle_count >= cycles_per_ms)
	{
		expected_elapsed_ms++;
		actual_elapsed_ms = elapsed_ms_time(cpu_timer);

		// if not enough time has elapsed, slow down
		while (expected_elapsed_ms > actual_elapsed_ms)
		{
			make_delay(1);
			actual_elapsed_ms = elapsed_ms_time(cpu_timer);
		}

#define CPU_TIMER_LIMIT 100000

		// reset our expected elapsed timer regularly so
		// it doesn't get too big and overflow
		if (expected_elapsed_ms > CPU_TIMER_LIMIT)
		{
			expected_elapsed_ms -= CPU_TIMER_LIMIT;
			cpu_timer += CPU_TIMER_LIMIT;
		}

		s_cycle_count -= cycles_per_ms;
	}
	// END FORCING CPU TO RUN AT PROPER SPEED

	// BEGIN CHECK TO SEE IF WE NEED DO AN NMI
	
	// if NMI's are enabled
	if (nmi_period > 0.0)
	{
		// if we need to do an NMI
		if (nmi_cycle_count > (Sint32) cycles_per_nmi)
		{
			nmi_cycle_count -= cycles_per_nmi;
			do_nmi();
			nmi_asserted = true;
	#ifdef CPU_DIAG
			nmi_count++;
	#endif
		}
	}
	// END CHECK TO SEE IF WE NEED TO DO AN NMI

	// CHECK TO SEE IF WE NEED TO DO AN IRQ

	// if irq period is 0.0 it means we don't do one at all
	if (irq_period > 0.0)
	{
		if (irq_cycle_count > (Sint32) cycles_per_irq)
		{
			// don't do an NMI and an IRQ at the same time just to be safe
			if (!nmi_asserted)
			{
				irq_cycle_count -= cycles_per_irq;
				do_irq();
			#ifdef CPU_DIAG
				irq_count++;
			#endif
			}
			// else wait for the next time around
		}
	} // end if
	// END CHECK FOR IRQ
}


// does any clean-up necessary to shut the cpu down
void game::cpu_shutdown()
{

	// if our cpu is the z80
	if (cpu_type == CPU_Z80)
	{
		// only exit if cpu has been initialized
		if (cpu_initialized)
		{
			Z80_EXIT;
			cpu_initialized = false;
		}
	}
	else if (cpu_type == CPU_M6809)
	{
		if (cpu_initialized)
		{
			cpu_initialized = false;
		}
	}
	else
	{
		// add any other shutdown code here
	}
}

// flushes any timers used by emulator
// This makes it so if a laserdisc seek takes a long time, a flood of
//   interrupts aren't generated after the seek is completed.
// This should be called after any laserdisc command that might not be
//   instantaneous.
void game::flush_timers()
{
	irq_cycle_count = 0;
	nmi_cycle_count = 0;
	cpu_timer = refresh_ms_time();	// so the cpu doesn't run too quickly when we first start
	expected_elapsed_ms = 0;
	s_cycle_count = 0;	// for speed regulation
}

void game::add_elapsed_cycles(Uint32 cycles)
{	
	m_tracked_elapsed_cycles += cycles;
}

// returns the total # of cycles that have elapsed 
// This is very useful in determining how much "time" has elapsed for time critical things like controlling the PR-8210
// laserdisc player
Uint64 game::get_total_cycles()
{
	Uint64 result = m_tracked_elapsed_cycles;

	if (cpu_type == CPU_Z80)
	{
		result += Z80_GET_ELAPSED_CYCLES;	// tack on whatever cycles have elapsed most recently
	}
	
	return result;
}

Uint64 game::get_elapsed_cycles(Uint64 previous_cycles)
{
	Uint64 result = m_tracked_elapsed_cycles + Z80_GET_ELAPSED_CYCLES - previous_cycles;

	return result;
}

// returns a possible 8-bit video overlay (NULL if we have none)
SDL_Surface *game::get_video_overlay()
{
	return(NULL);
}

// returns which color in the 8-bit video overlay palette is to be transparent
Uint8 game::get_transparent_color()
{
	return 0;
}

// returns the color palette of the game's video overlay (if it exists)
SDL_Color *game::get_video_colors()
{
	return (NULL);
}

// how many pixels down to shift video overlay
int game::get_video_row_offset()
{
	return 0;
}

// how many pixels to the right to shift video overlay
int game::get_video_col_offset()
{
	return 0;
}

// enables a cheat (any cheat) in the current game.  Unlimited lives is probably the most common.
void game::enable_cheat()
{
	cheat_requested = true;
}

double game::get_disc_fps()
{
	return disc_fps;
}

Uint8 game::get_game_type()
{
	return game_type;
}

Uint8 game::get_cpu_type()
{
	return cpu_type;
}

Uint32 game::get_cpu_hz()
{
	return cpu_hz;
}

double game::get_irq_period()
{
	return irq_period;
}

// overrides default IRQ period setting and forces a new one
// this is useful for using Dragon's Lair NTSC roms with PAL discs which play at a different speed
void game::set_irq_period(double period)
{
	irq_period = period;
}

double game::get_nmi_period()
{
	return nmi_period;
}

Uint32 game::get_num_sounds()
{
	return num_sounds;
}

char *game::get_sound_name(int whichone)
{
	return sound_name[whichone];
}

// returns a string of text that explains the problems with the game OR else null if the game has no problems :)
char *game::get_issues()
{
	char *result = NULL;

	if (has_issues)
	{
		result = game_issues;
	}

	return result;
}

#ifdef CPU_DEBUG

// returns either a name for the address or else NULL if no name exists
// Used for debugging to improve readability (to assign function and variable names)
char *game::get_address_name(Uint16 addr)
{

	char *name = NULL;
	int i = 0;

	// loop until we've hit the end of our memory names or until we've found a match
	for (;;)
	{
		// if we're at the end of the list
		if (addr_names[i].name == NULL)
		{
			break;
		}
		else if (addr_names[i].address == addr)
		{
			name = addr_names[i].name;
			break;
		}
		i++;
	}

	return(name);
}

#endif
