/*
 * game.h
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// game.h
// by Matt Ownby

// declares a generic game class structure

// To add a new game to DAPHNE, make a new class that inherits from the game class
// You must supply input_enable and input_disable functions, but you do not need to supply
//  any other functions if you don't want to (but your game won't work if you don't hehe)
// See Dragon's Lair (no port functions) and Cliff Hanger (no memory functions) for examples

#ifndef GAME_H
#define GAME_H

#include <SDL.h>
#include "../sound/sound.h"

#define CPU_MEM_SIZE	0x10000	// 64k of cpu memory

//#define CPU_DIAG	// uncomment this to verify the CPU speed is correct

// structure for the cpu debugger... a memory address and its corresponding name
// This is so when debugging, you can have function names instead of numbers (it is easier to read)
struct addr_name
{
	char *name;	// arbitrary name assigned to that address
	Uint16 address;	// memory address
};

void g_cpu_break(char *);

class game
{
public:
	game();

	virtual void do_irq();		// does an IRQ tick
	virtual void do_nmi();		// does an NMI tick
	virtual Uint8 cpu_mem_read(Uint16 addr);			// memory read routine
	virtual void cpu_mem_write(Uint16 addr, Uint8 value);		// memory write routine
	virtual Uint8 port_read(Uint16 port);		// read from port
	virtual void port_write(Uint16 port, Uint8 value);		// write to a port
	virtual void update_pc(Uint32 new_pc);		// update the PC
	virtual void input_enable(Uint8);
	virtual void input_disable(Uint8);
	virtual bool video_init();	// function to initalize video
	virtual void video_shutdown();	// function to shutdown video
	virtual void video_repaint();	// function to repaint video
	virtual void set_preset(int);	// set up dip switches/rom names, etc with prepared values
	virtual bool set_bank (unsigned char, unsigned char);	// set dip switch values

	virtual bool load_roms();
	virtual void cpu_init();
	virtual void cpu_execute();
	virtual void cpu_loop(Uint32);
	virtual void cpu_shutdown();
	void flush_timers();
	void add_elapsed_cycles(Uint32);
	Uint64 get_total_cycles();
	Uint64 get_elapsed_cycles(Uint64);
	virtual SDL_Surface *get_video_overlay();
	virtual Uint8 get_transparent_color();
	virtual SDL_Color *get_video_colors();
	virtual int get_video_row_offset();
	virtual int get_video_col_offset();
	
	void enable_cheat();
	double get_disc_fps();	// return FPS of the laserdisc
	Uint8 get_game_type();
	Uint8 get_cpu_type();	// returns type of cpu
	Uint32 get_cpu_hz();	// returns speed of cpu
	double get_irq_period();
	void set_irq_period(double);	// change default IRQ period
	double get_nmi_period();
	Uint32 get_num_sounds();
	char *get_sound_name(int);
	char *get_issues();	// does the game have any performance issues the user should know about?
#ifdef CPU_DEBUG
	char *get_address_name(Uint16 addr);	// get a potential name for a memory address
											// (very useful for debugging)
#endif

	Uint8 cpumem[CPU_MEM_SIZE]; // generic cpu memory (it needs to be public for speed purposes)
	Uint8 cpu2mem[CPU_MEM_SIZE];	// generic cpu memory for a second cpu
	int (*irq_callback)(int);	// used by the z80 core

protected:
	double disc_fps;	// frames per second of the game's laserdisc;
	Uint8 game_type;	// which game it is
	Uint8 cpu_type;	// which CPU the hardware uses
	Uint32 cpu_hz;	// the speed of the main cpu
	Uint32 num_sounds;	// how many samples the game has to load
	char *sound_name[MAX_NUM_SOUNDS];	// names for each sound file
	Uint16 m_cpu_pc;	// the program counter that the cpu starts executing code on
	Uint32 m_active_cpu;	// which cpu is currently active (0 is the first cpu)
	double irq_period;	// how often the IRQ ticks (in milliseconds)
	double nmi_period;	// how often the NMI ticks (in milliseconds)
	bool cpu_initialized;	// whether the cpu has been initialized yet
	Uint32 s_cycle_count;	// for CPU speed regulation
	Uint32 expected_elapsed_ms;	// how many ms we expect to have elapsed (for CPU speed regulation)
	Uint32 cpu_timer;	// used to keep the CPU running at the proper speed
	Uint32 cycles_per_ms;	// how many cycles per millisecond the CPU runs
	Uint32 cycles_per_irq;	// how many cycles per IRQ tick
	Sint32 irq_cycle_count;	// used to keep track of when we need to do an IRQ
	Uint32 cycles_per_nmi;	// how many cycles per NMI tick
	Sint32 nmi_cycle_count;	// used to keep track of when we need to do an NMI
	bool track_elapsed_cycles;	// whether we should keep track of elapsed cycles for the game (for PR-8210 emulation)
	Uint64 m_tracked_elapsed_cycles;	// any cycles we've tracked so far
	bool has_issues;	// does game have any issues?
	char game_issues[110];	// description of any issues the game has that user should know 
	bool cheat_requested;	// whether user has requested any cheats to be enabled

#ifdef CPU_DEBUG
	// Here is an example of how to do it
	// make sure last entry is NULL
	struct addr_name *addr_names;
#endif

};

#endif
