// interstellar.cpp
// by Mark Broadhead
//
// Interstellar ((C) 1983 Funai/ESP)
//
// Interstellar uses 3 Z80s. 
// The main Z80 uses roms rom2.top to rom6.top
// The second Z80 seems to be used for ld-v1000 communication (it sends ld-v1000 commands
// to ports 1 and 2) and uses rom rom1.top
// The third Z80 is on the graphics board and uses rom rom11.bot. It's unclear if this
// processor is used for graphics.
//
// there are 3 graphics roms, rom7.bot - rom9.bot. these are in the 8x8 tile format with each
// rom a different bit plane
// 
// there are also 3 different proms for color
//
// It's currently unknown exactly what each processor's function is and how they communicate.

#include <string.h>
#include "interstellar.h"
#include "../daphne.h"
#include "../io/conout.h"
#include "../ldp-in/ldv1000.h"
#include "../ldp-out/ldp.h"
#include "../video/video.h"
#include "../cpu/m80.h"


interstellar::interstellar()
{
	disc_fps = 29.97; 
	game_type = GAME_INTERSTELLAR;
	cpu_type = CPU_Z80;
	cpu_hz = 4000000; // unverified
//	irq_period = 16.6; 
	nmi_period = 16.6; 
}

// does anything special needed to send an IRQ
void interstellar::do_irq()
{
//	Z80_ASSERT_IRQ;
}

// does anything special needed to send an NMI
void interstellar::do_nmi()
{
//	Z80_ASSERT_NMI;
}

// reads a byte from the cpu's memory
Uint8 interstellar::cpu_mem_read(Uint16 addr)
{
	if (addr <= 0x9fff)	// rom
	{
		return cpumem[addr];
	}
	else if (addr >= 0xa000 && addr < 0xb000 ) // ram
	{
		return cpumem[addr];
	}
	else if (addr >= 0xb000 && addr < 0xb200 ) // video ram?
	{
		return cpumem[addr];
	}
	else
	{
		char s[81] = { 0 };

		sprintf(s, "ERROR: CPU memory %x read requested, but this function is unimplemented!", addr);
		printline(s);
		return cpumem[addr];
	}
}

// writes a byte to the cpu's memory
void interstellar::cpu_mem_write(Uint16 addr, Uint8 value)
{
	if (addr <= 0x9fff)	// rom	
	{
		cpumem[addr] = value;
	}
	else if (addr >= 0xa000 && addr < 0xb000 ) // ram
	{
		cpumem[addr] = value;
	}
	else if (addr >= 0xb000 && addr < 0xb200 ) // video ram
	{
		cpumem[addr] = value;
	}
	else
	{
		char s[81] = { 0 };

		sprintf(s, "ERROR: CPU memory %x write requested (value %x) but this function is unimplemented!", addr, value);
		printline(s);
		cpumem[addr] = value;
	}	
}

// reads a byte from the cpu's port
Uint8 interstellar::port_read(Uint16 port)
{
	char s[81] = { 0 };

    unsigned char result = 0;

	port &= 0xFF;	// strip off high byte

    switch(port)
    {
	case 0x00:
		result = read_ldv1000();
		break;
    default:
//        sprintf(s, "INTERSTELLAR: Unsupported Port Input-> %x (PC is %x)", port, Z80_GET_PC);
        printline(s);
        break;
    }

    return(result);
}

// writes a byte to the cpu's port
void interstellar::port_write(Uint16 port, Uint8 value)
{
	char s[81] = { 0 };

	port &= 0xFF;
    switch(port)
	{
	case 0x01:
		write_ldv1000(value);
		break;
	default:
//        sprintf(s, "INTERSTELLAR: Unsupported Port Output-> %x:%x (PC is %x)", port, value, Z80_GET_PC);
        printline(s);
        break;
	}
}

bool interstellar::load_roms()
{
	return (load_rom("interstellar/rom1.top", &cpumem[0x0000], 0x2000) &&
		load_rom("interstellar/rom3.top", &cpumem[0x2000], 0x2000) &&
		load_rom("interstellar/rom4.top", &cpumem[0x4000], 0x2000) &&
		load_rom("interstellar/rom5.top", &cpumem[0x6000], 0x2000) &&
		load_rom("interstellar/rom6.top", &cpumem[0x8000], 0x2000));
}

// does any video initialization we might need
// returns 'true' if the initialization was successful, false if it failed
bool interstellar::video_init()
{
	return 1;
}

// shuts down any video we might have initialized
void interstellar::video_shutdown()
{
}

// redraws our video on the screen (in case the screen needs to be redrawn)
void interstellar::video_repaint()
{
}

// this gets called when the user presses a key or moves the joystick
void interstellar::input_enable(Uint8 move)
{
	switch (move)
	{
	}
}  

// this gets called when the user releases a key or moves the joystick back to center position
void interstellar::input_disable(Uint8 move)
{
	switch (move)
	{
	}
}
