/*
 * lair.cpp
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// Dragon's Lair/Space Ace driver for DAPHNE
// started by Matt Ownby
// contributions made by Mark Broadhead, Robert DiNapoli, Jeff Kulczycki
// If you don't see your name here, feel free to add it =]

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../daphne.h"
#include "lair.h"
#include "../ldp-in/ldv1000.h"
#include "../ldp-out/ldp.h"
#include "../video/scoreboard.h"
#include "../video/led.h"
#include "../io/conout.h"
#include "../sound/sound.h"

#ifdef USE_M80
#include "../cpu/m80.h"
#else
#include "../cpu/z80.h"
#endif

extern ldp *g_ldp;	// our global ldp class

//////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////

// lair class constructor (default the rev F2 roms)
lair::lair()
{
	memset(cpumem, 0, CPU_MEM_SIZE);
	switchA = 0x22;
	switchB = 0xD8;
	joyskill_val = 0xFF;	// all input cleared
	misc_val = 0x7F;	// bit 7 and 6 alternated, and bits 0-3 all set (bits 4-5 are unused)
	irq_period = LAIR_IRQ_PERIOD;
	nmi_period = (1000.0 / 60.0);	// dragon's lair has no NMI.  We use this to time the LD-V1000 strobes since it's convenient
	disc_fps = 23.976;
	game_type = GAME_LAIR;
	cpu_type = CPU_Z80;
	cpu_hz = LAIR_CPU_HZ;
	ldv1000_enable_instant_seeking();	// make the LD-V1000 perform instantaneous seeks because we can
	track_elapsed_cycles = true;
	m_status_strobe_timer = 0;
	m_time_strobes = false;	// since the Ld-V1000 strobes are new and experimental, we have them off by default for now

	num_sounds = 3;
	sound_name[S_DL_CREDIT] = "dl_credit.wav";
	sound_name[S_DL_ACCEPT] = "dl_accept.wav";
	sound_name[S_DL_BUZZ] = "dl_buzz.wav";
}

// Space Ace class constructor
ace::ace()
{
	game_type = GAME_ACE;	
	switchA = 0x3D;	// 2 coins/credit, attract mode audio always on
	switchB = 0xFE;	// 5 lives	
}


// dragon's lair enhanced class constructor
dle::dle()
{
}

///////////////////////////////////////////////////////////

// raise IRQ line
void lair::do_irq()
{
	Z80_ASSERT_IRQ;
}

// Dragon's Lair has no NMI
// We just use the timer as a convenience to keep track of the LD-V1000 status strobe
void lair::do_nmi()
{
	m_status_strobe_timer = get_total_cycles();
}

void lair::cpu_mem_write(Uint16 Addr, Uint8 Value)
// Called whenever the Z80 emulator wants to write to memory
{

	char s[81] = { 0 };

	// If the program is trying to write to hardware outputs
	if (Addr >= 0xE000)
	{
		// if the program is trying to write outside of its boundaries
	  if (Addr > CPU_MEM_SIZE)
	  {
		  printline("*** FATAL: MEM SIZE exceeded!");
	  }

	  else
	  {
			  switch (Addr)
			  {
			  	// real sound chip data is sent here
			  case 0xE000:
			  	break;

				  // I believe this controlled whether the bus was in input or output mode but it is not important
				  // for emulation
			  case 0xE008:
				  // it seems to be safe to totally ignore this =]
				  break;

			  // sound chip I/O (also controls dip switches)
			  case 0xE010:
				  switch (Value)
				  {
				  case 0x0E:
					// read dip-switch 1 and put result at 0xC000
					cpumem[0xC000] = switchA;
					break;
				  case 0x0F:
					// read dip-switch 2 and put result at 0xC000
					cpumem[0xC000] = switchB;
					break;
				  default:
					  break;
				  }
				  break;

			  // laserdisc control
			  case 0xE020:
				  write_ldv1000(Value);
				  break;

			  // E030-E03F are all LED's on the Dragon's Lair scoreboard
			  case 0xE030:
				  update_player_score(0, (Value & 0x0F), 1);
				  break;
			  case 0xE031:
				  update_player_score(1, (Value & 0x0F), 1);
				  break;
			  case 0xE032:
				  update_player_score(2, (Value & 0x0F), 1);
				  break;
			  case 0xE033:
				  update_player_score(3, (Value & 0x0F), 1);
				  break;
			  case 0xE034:
				  update_player_score(4, (Value & 0x0F), 1);
				  break;
			  case 0xE035:
				  update_player_score(5, (Value & 0x0F), 1);
				  break;

				  // 10's position of how many credits are deposited
			  case 0xE036:
				  update_credits(0, Value & 0x0F);
				  break;
				  // 1's position of how many credits are deposited
			  case 0xE037:
				  update_credits(1, Value & 0x0F);
				  break;

			  case 0xE038:
				  update_player_score(0, (Value & 0x0F), 0);
				  break;
			  case 0xE039:
				  update_player_score(1, (Value & 0x0F), 0);
				  break;
			  case 0xE03A:
				  update_player_score(2, (Value & 0x0F), 0);
				  break;
			  case 0xE03B:
				  // Space Ace Skill LED
			  	if (Value == 0xCC)
			  	{
					change_led(false, false, true);
			  	}
			  	else
			  	{
				  	update_player_score(3, (Value & 0x0F), 0);
				  }
				  break;
			  case 0xE03C:
				  update_player_score(4, (Value & 0x0F), 0);
				  break;
			  case 0xE03D:
				  // Captain Skill LED
			  	if (Value == 0xCC)
			  	{
					change_led(false, true, false);
			  	}
			  	else
			  	{
				  	update_player_score(5, (Value & 0x0F), 0);
				}
				break;

				  // lives of player 1
			  case 0xE03E:
			  	// show cadet skill LED
			  	if (Value == 0xCC)
			  	{
					change_led(true, false, false);
			  	}
			  	else
			  	{
				  	update_player_lives(Value & 0x0F, 0);
				}
				break;
				// lives of player 2
			  case 0xE03F:
			  	// not sure what this does?
			  	if (Value == 0xCC)
			  	{
//			  		printline("Clears LEDs but is not needed for emulation");
			  	}
			  	else
			  	{
				  	update_player_lives(Value & 0x0F, 1);
				  }
				  break;
			  default:
			  	sprintf(s, "Unknown hardware output at %x, value of %x, PC %x", Addr, Value, Z80_GET_PC);
			  	printline(s);
			  	break;
			  } // end switch
	  } // end if we're not writing above RAM limit
	} // end if we're writing to 0xE000 region

	// else if we're writing to RAM
	else if ((Addr >= 0xA000) && (Addr <= 0xAFFF))
	{
		switch(Addr)
		{
				  // We do not emulate the sound chip correctly.  Instead we do a cheat which gives us the benefit
				  // of less CPU cycles and also better quality sound.
				  // A01C gets written to every time a beep is supposed to be played, and A01D gets the memory location where
				  // the beep data is, so whenever A01C gets a 1 written to it, we check the value of A01D to see which type
				  // of beep we are supposed to play.  If anyone ever re-assembled the Dragon's Lair ROM, this scheme would
				  // likely break.
		case 0xA01C:
				  if (Value & 1)
				  {
					  int index = cpumem[0xA01D] | (cpumem[0xA01E] << 8);	// pointer to audio data

					  // only the 'accept' sound data has a D5 as the second byte
					  if (cpumem[index+1] == 0xD5)
					  {
						  sound_play(S_DL_ACCEPT);
					  }
					  // only the 'credit' sound data has a 0x66 as the second byte
					  else if (cpumem[index+1] == 0x66)
					  {
						  sound_play(S_DL_CREDIT);
					  }
					  // only the 'buzz' sound data has a 0x99 as its third byte
					  else if (cpumem[index+1] == 0x99)
					  {
						  sound_play(S_DL_BUZZ);
					  }
					  // else unknown sound, play an error
					  else
					  {
						printline("WARNING : Unknown dragon's lair sound!");
					  }
				  }
				  break;
		} // end switch

		cpumem[Addr] = Value; // always store to RAM if we write to it

	}	// end writing to RAM (>= 0xA000)

	// if we are trying to write below 0xA000, it means we are trying to write to ROM
	else
	{
		sprintf(s, "Error, program attempting to write to ROM (%x), PC is %x", Addr, Z80_GET_PC);
		printline(s);
	}
}


// Called whenever the Z80 emulator wants to read from memory
Uint8 lair::cpu_mem_read(Uint16 Addr)
{

	Uint8 result = cpumem[Addr];

	// the memory range of 0xC000 to somewhere less than 0xE000 is used by Lair/Ace to read external input
	// such as the joystick, the dip switches, or the laserdisc player
	if (Addr >= 0xC000)
	{
		// if we're within memory limits (this is kind of a pointless check since we can't ever exceed 16-bits anyway)
		if (Addr < CPU_MEM_SIZE)
		{
			  switch(Addr)
			  {
			  case 0xC008:	// joystick/spaceace skill query ...
				  result = joyskill_val;
				  break;
			  case 0xC010:
				  result = read_C010();
				  break;
			  case 0xC020:
				  result = read_ldv1000();
				  break;
			  default:
				  break;
			  }	// end switch
		}	// end else
	} // end if addr > 0xC000 ...

	return(result);

}

// initializes the scoreboard
bool lair::video_init()
{
	return scoreboard_init();
}

// frees any images we loaded in, etc
void lair::video_shutdown()
{
	scoreboard_shutdown();
}

// redraws the scoreboard on the screen
void lair::video_repaint()
{
	scoreboard_repaint();
}

// basically 'preset' is a macro to set a bunch of other options; useful as a good shortcut
void lair::set_preset(int preset)
{
	if (preset == 1)
	{
		printline("Experimental LD-V1000 strobes activated!");
		m_time_strobes = true;
	}

}

// used to set dip switch values
bool lair::set_bank(unsigned char which_bank, unsigned char value)
{
	bool result = true;
	
	switch (which_bank)
	{
	case 0:	// bank A
		switchA = (unsigned char) (value ^ 0xFF);	// in dragon's lair, dip switches are active low
		break;
	case 1:	// bank B
		switchB = (unsigned char) (value ^ 0xFF);	// switches are active low
		break;
	default:
		printline("ERROR: Bank specified is out of range!");
		result = false;
		break;
	}
	
	return result;
}

// loads the roms to be used with Dragon's Lair
bool lair::load_roms()
{	
	return (load_rom("lair/dl_f2_u1.bin", &cpumem[0], 0x2000) &&
		load_rom("lair/dl_f2_u2.bin", &cpumem[0x2000], 0x2000) &&
		load_rom("lair/dl_f2_u3.bin", &cpumem[0x4000], 0x2000) &&
		load_rom("lair/dl_f2_u4.bin", &cpumem[0x6000], 0x2000));

}

// loads the roms to be used with Space Ace
bool ace::load_roms()
{
	if (cheat_requested)
	{
		switchB &= 0xDF;	// set B5 ON (unlimited lives)
	}
	
	return (load_rom("ace/sa_a2_u1.bin", &cpumem[0], 0x2000) &&
		load_rom("ace/sa_a2_u2.bin", &cpumem[0x2000], 0x2000) &&
		load_rom("ace/sa_a2_u3.bin", &cpumem[0x4000], 0x2000) &&
		load_rom("ace/sa_a2_u4.bin", &cpumem[0x6000], 0x2000) &&
		load_rom("ace/sa_a2_u5.bin", &cpumem[0x8000], 0x2000));

}

// loads the roms to be used with Dragon's Lair Enhanced
bool dle::load_roms()
{
	bool result1 = false, result2 = false;
	Uint8 *readme_test = NULL;
	FILE *readme_file = NULL;

	result1 = (load_rom("dle/dle11u1l.bin", &cpumem[0], 0x2000) &&
		load_rom("dle/dle11u2l.bin", &cpumem[0x2000], 0x2000) &&
		load_rom("dle/dle11u3l.bin", &cpumem[0x4000], 0x2000) &&
		load_rom("dle/dle11u4l.bin", &cpumem[0x6000], 0x2000));

	// NOW check to make sure the DLE readme file is present and unaltered
	// Dave Hallock requested this check to make sure sites don't remove his readme file
	// and distribute DLE.

	// 22338 is the size of the readme file in bytes

	readme_file = fopen("roms/dle/readme11.txt", "rb");
	// if file exists, check its length
	if (readme_file)
	{
		readme_test = static_cast<Uint8 *>(malloc(22338));
		if (readme_test)
		{
			fread(readme_test, 1, 22338, readme_file);
			// if the DLE readme file has been unaltered, allow user to continue
			if (compute_checksum(readme_test, 22338) == 0x358A1CD7)
			{
				result2 = true;
			}
			// if checksum failed, it means readme file is altered
			else
			{
				printerror("DLE readme11.txt file is present but has been altered.");
				printerror("Please get the original readme file from www.d-l-p.com, thx.");
			}
			free(readme_test);
			readme_test = NULL;
		}
		fclose(readme_file);
	}
	else
	{
		printerror("Cannot find readme11.txt file which is to be distributed with DLE.");
		printerror("Please get this file from www.d-l-p.com");
	}

	return (result1 && result2);

}


Uint8 lair::read_C010()
// emulates the C010 memory location
// bit 0 = player 1 button, bit 1 = player 2 button (clear = button is depressed)
// bit 2 = coin 1 inserted, bit 3 = coin 2 inserted
// bit 6 = status strobe, bit 7 = command strobe
{

	// MATT : Since timing the strobes breaks DLE (causes overruns), I've decided to make it optional for now
	// until we have more time to try to fix VLDP and the other laserdisc players

	// if user chooses to use experimental strobe timer
	if (m_time_strobes)
	{
		Uint64 elapsed_cycles = get_elapsed_cycles(m_status_strobe_timer);	// how many CPU cycles have elapsed since status strobe

		// I made these constants here to force the compiler to compute them at compile time, just in case the compiler was too stupid to do so :)
		// 0.5 is added to avoid truncation errors

		const Uint32 until_status_end = (Uint32) (((LAIR_CPU_HZ / 1000000.0) * 26.0) + 0.5);
		// how many cycles per 26 uS.  Status strobe lasts 26 uS

		const Uint32 until_command_start = (Uint32) (((LAIR_CPU_HZ / 1000000.0) * (26.0 + 54.0)) + 0.5);
		// how many cycles until command strobe begins, relative to status strobe begin (26 uS for status, 54 between them)

		const Uint32 until_command_end = (Uint32) (((LAIR_CPU_HZ / 1000000.0) * (26.0 + 54.0 + 25.0)) + 0.5);
		// how many cycles until command strobe ends, relative to status strobe begin (26 uS for status, 54 in between, 25 for command)

		// if status strobe hasn't ended yet
		if (elapsed_cycles < until_status_end) // status strobe is 26 uS
		{
			misc_val &= 0x3f; //clear the strobes
			misc_val |= 0x40; //set status strobe
		}
		// if status strobe has ended, but command strobe has not begun, then clear both strobes
		else if (elapsed_cycles < until_command_start) // status to command strobe is 54 uS
		{
			misc_val &= 0x3f; //clear the strobes
		}
		// if command strobe has begun and hasn't ended yet
		else if (elapsed_cycles < until_command_end) // command strobe is 25 uS
		{
			misc_val &= 0x3f; //clear the strobes
			misc_val |= 0x80; //set command strobe
		}
		// if neither command nor status strobe has enabled, clear strobe values
		else 
		{
			misc_val &= 0x3f; //clear the strobes
		}
	}

	// if we want to use the old, reliable method
	else
	{
		misc_val ^= 0xC0;	// flip top two bits to alternate strobe
	}

	return(misc_val);
	
}



void lair::input_enable(Uint8 move)
{

		switch(move)
		{
		case SWITCH_UP:
			joyskill_val &= (unsigned char) ~1;	// clear bit 0
			break;
		case SWITCH_DOWN:
			joyskill_val &= (unsigned char) ~2;	// clear bit 1
			break;
		case SWITCH_LEFT:
			joyskill_val &= (unsigned char) ~4;	// clear bit 2
			break;
		case SWITCH_RIGHT:
			joyskill_val &= (unsigned char) ~8;	// clear bit 3
			break;
		case SWITCH_BUTTON1: // PLAYER 1
			misc_val &= ~1;	// clear bit 0 ...
			break;
		case SWITCH_BUTTON2:
			misc_val &= ~2;	// clear bit 1
			break;
		case SWITCH_BUTTON3: // SWORD
			joyskill_val &= (unsigned char) ~0x10;	// clear bit 4
			break;
		case SWITCH_COIN1: 
			misc_val &= (unsigned char) ~0x04;	
			break;
		case SWITCH_COIN2: 
			misc_val &= (unsigned char) ~0x08;	
			break;
		case SWITCH_SKILL1:	// cadet
			joyskill_val &= (unsigned char) ~0x20;	// clear bit 5
			break;
		case SWITCH_SKILL2:	// captain
			joyskill_val &= (unsigned char) ~0x40;	// clear bit 6
			break;
		case SWITCH_SKILL3:	// space ace
			joyskill_val &= (unsigned char) ~0x80;	// clear bit 7
			break;
		default:
			printline("Error, bug in Dragon's Lair's input enable");
			break;
		}
}


void lair::input_disable(Uint8 move)
{

		switch(move)
		{
		case SWITCH_UP:
			joyskill_val |= 1;	// set bit 0
			break;
		case SWITCH_DOWN:
			joyskill_val |= 2;	// set bit 1
			break;
		case SWITCH_LEFT:
			joyskill_val |= 4;	// set bit 2
			break;
		case SWITCH_RIGHT:
			joyskill_val |= 8;	// set bit 3
			break;
		case SWITCH_BUTTON1:
			misc_val |= 1; // PLAYER 1
			break;
		case SWITCH_BUTTON2:
			misc_val |= 2;	// set bit 1
			break;
		case SWITCH_BUTTON3: // SWORD
			joyskill_val |= 0x10;	// set bit 4
			break;
		case SWITCH_COIN1: 
			misc_val |= 0x04;	
			break;
		case SWITCH_COIN2: 
			misc_val |= 0x08;	
			break;
		case SWITCH_SKILL1:	// cadet
			joyskill_val |= (unsigned char) 0x20;	// set bit 5
			break;
		case SWITCH_SKILL2:	// captain
			joyskill_val |= (unsigned char) 0x40;	// set bit 6
			break;
		case SWITCH_SKILL3:	// space ace
			joyskill_val |= (unsigned char) 0x80;	// set bit 7
			break;
		default:
			printline("Error, bug in Dragon's Lair's move disable");
			break;
		}
}
