/*
 * superd.cpp
 *
 * Copyright (C) 2001 Mark Broadhead
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// superd.cpp
// by Mark Broadhead

#include <stdio.h>
#include <string.h>
#include "superd.h"
#include "../daphne.h"
#include Z80_INCLUDE
#include "../ldp-in/ldv1000.h"
#include "../ldp-out/ldp.h"
#include "../io/conout.h"
#include "../sound/sound.h"
#include "../timer/timer.h"
#include "../video/video.h"

extern ldp *g_ldp;

/////////////////////////////////////////////////////////////////////

#ifdef CPU_DEBUG

// super don function/variable names for debugging
struct addr_name superd_addr_names[] =
{
	{ "Start", 0x0000 },
	{ "SdqStart", 0x0210 },
	{ "MainGameLoop", 0x02A0 },
	{ "CheckExpectedJoystick", 0xF1E },
	{ "CheckExpectedButtons", 0xF32 },
	{ "CheckUnexpectedJoystick", 0xF76 },
	{ "EnableSkipFrames", 0x1083 },
	{ "MakeFrameNum", 0x112E },
	{ "LDV1000izeDigit", 0x115A },
	{ "HandleBadMove", 0x11DD },
	{ "LDV1000_Reject", 0x1426 },
	{ "LDV1000_Play2", 0x142B },
	{ "LDV1000_Play1", 0x1435 },
	{ "LDV1000_1X", 0x143F },
	{ "LDV1000_Stop", 0x1444 },
	{ "LDV1000_SetDisp", 0x1448 },
	{ "LDV1000_DispEnable", 0x1459 },
	{ "LDV1000_Send1AndDisp", 0x145E },
	{ "Send_to_LDV1000", 0x14AE },
	{ "Send2_to_LDV1000", 0x14DB },
	{ "MakeDelay", 0x1512 },
	{ "LDV1000_Search", 0x151B },
	{ "GetFrameDigit", 0x151F },
	{ "LDV1000_Clear", 0x1566 },
	{ "LDV1000_SkipFrames", 0x157D },
	{ "GetFrameNumber", 0x15AB },
	{ "Upate_P1_Score", 0x1698 },
	{ "Update_Score", 0x16A8 },
	{ "(DE)=(DE)+(HL)", 0x1D6B },
	{ "ISR2_Index", 0x4000 },
	{ "Stack0Ptr", 0x4002 },
	{ "Stack1Ptr", 0x4004 },
	{ "Stack2Ptr", 0x4006 },
	{ "Stack3Ptr", 0x4008 },
	{ "Stack4Ptr", 0x400A },
	{ "ExpectedFrame", 0x4050 },
	{ "NeedSearch?", 0x40D1 },
	{ "SkipFrames?", 0x40D2 },
	{ "FrameCalcPtr", 0x40E2 },
	{ "Joystick0", 0x40F0 },	// here is where info from input port 0 (joystick) get stored
	{ "Joystick1", 0x40F1 },
	{ "JoyChanged?1", 0x40F2 },
	{ "JoyChanged?2", 0x40F3 },
	{ "ButtonsCoin0", 0x40F4 }, // 4-8
	{ "Dip1_0", 0x40F8 },		// 8-B
	{ "Dip2_0", 0x40FC },		// C-F
	{ "Move_Accepted?", 0x4086 },
	{ "LDV1000Frame", 0x4100 },
	{ "ApproxCurFrame", 0x4110 },
	{ "Old2Frame", 0x4120 },
	{ "FrameDifference", 0x4130 },
	{ "DeathFrame", 0x4140 },
	{ "Frame2", 0x4120 },
	{ "FrameSubtract", 0x4130 },
	{ "P1Score0", 0x4190 },
	{ "P1Score1", 0x4191 },
	{ "P1Score2", 0x4192 },
	{ "P2Score0", 0x4198 },
	{ "P2Score1", 0x4199 },
	{ "P2Score2", 0x419A },
	{ NULL, 0 }
};

#endif


/////////////////////////////////////////////////////////////////////

// we do not want to clear the IRQ when it gets enabled
int superd_irq_callback(int irqline)
{
	// get rid of warnings
	if (irqline)
	{
	}
	return 0xff;
}


// superdon constructor
superd::superd()
{
	memset(banks, 0xFF, 4);	// fill banks with 0xFF's
	memset(colors, 0, sizeof(colors));	// fill color palette with 0's
	video_overlay = SDL_CreateRGBSurface(SDL_HWSURFACE, SUPERD_OVERLAY_W, SUPERD_OVERLAY_H, 8, 0, 0, 0, 0); // create an 8-bit surface
	screen_needs_update = true;

	disc_fps = 29.97;
	game_type = GAME_SUPERD;
	cpu_type = CPU_Z80;
	cpu_hz = SUPERD_CPU_HZ;
	irq_period = SUPERD_IRQ_PERIOD;
#ifdef CPU_DEBUG
	addr_names = superd_addr_names;
#endif
	ldp_output_latch = 0xff;

	irq_callback = superd_irq_callback;	// we have to do our own IRQ callback for Super Don
	
//	ldv1000_apply_superdon_hack();	// regulate the frame numbers so we don't get lost during gameplay
	
	// CHEAT : set super don to unlimited lives
//	banks[2] = 0xBF;

	ldv1000_enable_instant_seeking();	// enable instantaneous seeks because we can

	has_issues = true;
	strcpy(game_issues, "Death scenes are visible between each correct move");

	num_sounds = 5;
	sound_name[S_SD_COIN] = "sd_coin.wav";
	sound_name[S_SD_SUCCEED] = "sd_succeed.wav";
	sound_name[S_SD_FAIL] = "sd_fail.wav";
	sound_name[S_SDA_SUCCESS_LO] = "sda_success_lo.wav";
	sound_name[S_SDA_SUCCESS_HI] = "sda_success_hi.wav";

}



void superd::do_irq()
{
	// Redraws the screen (if needed) on interrupt
	display_update();
	ldp_input_latch = read_ldv1000();
	write_ldv1000(ldp_output_latch);
//	z80_set_irq_line(0, HOLD_LINE);	// assert IRQ line
	Z80_ASSERT_IRQ;
}


void superd::cpu_mem_write(Uint16 Addr, Uint8 Value)
// Called whenever the Z80 emulator wants to write to memory
{	
	cpumem[Addr] = Value;

	// if the cpu writes to video memory, update the screen on the next interrupt
	if (Addr >= 0x5c00 && Addr <= 0x5fff)
	{
		screen_needs_update = true;
	}

}

void superd::port_write(Uint16 Port, Uint8 Value)
// Called whenever the emulator wants to output to a port
{

    char s[81] = { 0 };
	static int snd_succeed_count = 0;	// a cheat to make sure we only play the succeed sample once
	static int snd_coin_count = 0;	// a cheat to make sure we only play the coin sample once

    Port &= 0xFF;	// strip off high byte

    switch(Port)
	{
	
	case 0x00:	//	Send command to LVD-1000
		ldp_output_latch = Value;
		break;

// super don's sound chip is the sn76496
		
		// OK, the sound chip code, like other sound code in DAPHNE, is a cheat and not
		// true emulation.
		// The sound chip sends a bunch of data to port 4 to control the sound chip.
		// All we do here is look for unique data to figure out which sample we should play.
		// It turns out that only 8's are sent for the coin sound, only 11's are sent for the
		// successful move sound, and only 0xC1 is send for the bad move sound.
		// Also, about eight 0x11's are sent for the successful move, so we make sure we only
		// play a sound for 1 out of every eight 0x11's we get.  Same type of deal for the coin
		// sound.
	case 0x04:	//	Send data to sound chip
		if (Value == 0x08)
		{
			if (snd_coin_count == 0)
			{
				sound_play(S_SD_COIN);
			}

			snd_coin_count++;
			if (snd_coin_count >= 2)
			{
				snd_coin_count = 0;
			}
		}
		if (Value == 0x11)
		{
			if (snd_succeed_count == 0)
			{
				sound_play(S_SD_SUCCEED);
			}

			snd_succeed_count++;
			if (snd_succeed_count >= 8)
			{
				snd_succeed_count = 0;
			}
		}
		if (Value == 0xc1)
		{
			printline("Playing fail sound");
			sound_play(S_SD_FAIL);
		}
		
		//alternate sounds for alternate ROM set (same counter should be okay)
		//you get more points and a higher-pitched sound if your reaction is quick
		if (Value == 0x12)
		{
			if (snd_succeed_count == 0)
			{
				sound_play(S_SDA_SUCCESS_LO);
			}

			snd_succeed_count++;
			if (snd_succeed_count >= 8)
			{
				snd_succeed_count = 0;
			}
		}
		
		if (Value == 0x0F)
		{
			if (snd_succeed_count == 0)
			{
				sound_play(S_SDA_SUCCESS_HI);
			}

			snd_succeed_count++;
			if (snd_succeed_count >= 8)
			{
				snd_succeed_count = 0;
			}
		}
		
		
		
		break;
		

	case 0x08:	//	Latch for outputs to vid hardware and coin counters

		// bit 6 is set, clear the IRQ
		if ((Value >> 6) & 0x01)
		{
			Z80_CLEAR_IRQ;
//			z80_set_irq_line(0, CLEAR_LINE);
		}
		break;

	case 0x0c:	//	Output to HD46505
		break;

	case 0x0d:	//	Output to HD46505
		break;

	default:
        sprintf(s, "SUPERDON: Unsupported Port Output-> %x:%x (PC is %x)", Port, Value, Z80_GET_PC);
        printline(s);
        break;

	}		
}

Uint8 superd::port_read(Uint16 Port)
// Called whenever the emulator wants to read from a port
{

    char s[81] = { 0 };
    unsigned char result = 0;

	Port &= 0xFF;	// strip off high byte

    switch(Port)
    {

	case 0x00:	//	Joystick
		result = banks[0];
		break;

	case 0x01:	//	Buttons/Coins
		result = banks[1];
		break;

	case 0x02:	//	Dipswitch 1
		result = banks[2];
		break;

	case 0x03:	//	Dipswitch 2
		result = banks[3];
		break;

	case 0x04:	//	Read ldp-1000 status
		result = ldp_input_latch;
		break;

    default:
        sprintf(s, "SUPERD: Unsupported Port Input-> %x (PC is %x)", Port, Z80_GET_PC);
        printline(s);
        break;
    }

    return(result);

}

// used to set dip switch values
bool superd::set_bank(unsigned char which_bank, unsigned char value)
{
	bool result = true;
	
	switch (which_bank)
	{
	case 0:	// bank A
		banks[2] = (unsigned char) (value ^ 0xFF);	// dip switches are active low
		break;
	case 1:	// bank B
		banks[3] = (unsigned char) (value ^ 0xFF);	// switches are active low
		break;
	default:
		printline("ERROR: Bank specified is out of range!");
		result = false;
		break;
	}
	
	return result;
}

bool superd::load_roms()
{
	
	return (load_rom("sdq/sdq-prog.bin", &cpumem[0], 16384) &&
		load_rom("sdq/sdq-cprm.bin", &color_prom[0], sizeof(color_prom)) &&
		load_rom("sdq/sdq-char.bin", &character[0], sizeof(character)));

}


// superdon (short scenes) constructor
sdqshort::sdqshort()
{
}



bool sdqshort::load_roms()
{
	
	return (load_rom("sdqshort/sdq_c45.rom", &cpumem[0], 16384) &&
		load_rom("sdq/sdq-cprm.bin", &color_prom[0], sizeof(color_prom)) &&
		load_rom("sdqshort/sdq_a8.rom", &character[0], sizeof(character)));

}


// initializes superdon's video
bool superd::video_init()
{

	bool result = false;

	//Convert palette rom into a useable palette
	for (int i = 0; i < 32; i++)
	{
		Uint8 bit0,bit1,bit2;

		/* red component */
		bit0 = static_cast<Uint8>((color_prom[i] >> 5) & 0x01);
		bit1 = static_cast<Uint8>((color_prom[i] >> 6) & 0x01);
		bit2 = static_cast<Uint8>((color_prom[i] >> 7) & 0x01);
		colors[i].r = static_cast<Uint8>((0x23 * bit0) + (0x4b * bit1) + (0x91 * bit2));
		/* green component */
		bit0 = static_cast<Uint8>((color_prom[i] >> 2) & 0x01);
		bit1 = static_cast<Uint8>((color_prom[i] >> 3) & 0x01);
		bit2 = static_cast<Uint8>((color_prom[i] >> 4) & 0x01);
		colors[i].g = static_cast<Uint8>((0x23 * bit0) + (0x4b * bit1) + (0x91 * bit2));
		/* blue component */
		bit0 = static_cast<Uint8>(0);
		bit1 = static_cast<Uint8>((color_prom[i] >> 0) & 0x01);
		bit2 = static_cast<Uint8>((color_prom[i] >> 1) & 0x01);
		colors[i].b = static_cast<Uint8>((0x23 * bit0) + (0x4b * bit1) + (0x91 * bit2));
	}
	
	if (video_overlay)
	{
		SDL_SetColors(video_overlay, colors, 0, 32);
		result = true;
	}
	else
	{
		printline("ERROR: Could not create super don overlay");
	}

	return result;
}

// shuts down superdon's video
void superd::video_shutdown()
{
	SDL_FreeSurface(video_overlay);
}

void superd::video_repaint()
{
	screen_needs_update = true;
	display_update();	// MATT: added this so screen would repaint while I debugged
}

// updates superdon's video
void superd::display_update()
{
	if (screen_needs_update)
	{
		SDL_Rect dest;	// setup place on screen to blit superd overlay
		dest.x = 0;	// upper left corner
		dest.y = 0;
		dest.w = (unsigned short) video_overlay->w;
		dest.h = (unsigned short) video_overlay->h;

		for (int charx = 0; charx < 32; charx++)
		{
			for (int chary = 0; chary < 32; chary++)
			{
				for (int x=0; x < 4; x++)
				{
					for (int y = 0; y < 8; y++)
					{
							Uint8 left_pixel = static_cast<Uint8>((character[cpumem[chary * 32 + charx + 0x5c00]*32+x+4*y] & 0xf0) >> 4);
							Uint8 right_pixel = static_cast<Uint8>((character[cpumem[chary * 32 + charx + 0x5c00]*32+x+4*y] & 0x0f));
							*((Uint8 *) video_overlay->pixels + ((chary * 8 + y) * SUPERD_OVERLAY_W) + (charx * 8 + x * 2)) = left_pixel;
							*((Uint8 *) video_overlay->pixels + ((chary * 8 + y) * SUPERD_OVERLAY_W) + (charx * 8 + x * 2 + 1)) = right_pixel;
					}
				}
			}
		}


//		SDL_BlitSurface(video_overlay, NULL, 

		dest.x = 0;	// upper left corner
		dest.y = 0;
		dest.w = (unsigned short) video_overlay->w;
		dest.h = (unsigned short) video_overlay->h;

		// SMPEG doesn't allow blitting, which is why we have to check
		if (g_ldp->is_blitting_allowed())
		{
			SDL_BlitSurface(video_overlay, NULL, get_screen(), &dest);
			SDL_Flip(get_screen());
		}
		screen_needs_update = false;
	}
	
	return;
}

// returns a pointer to Super Don's video overlay surface
// also returns width, height, and bits per pixel of surface
SDL_Surface *superd::get_video_overlay()
{
	return(video_overlay);
}

// returns which color in the palette is transparent.  In this case it's 0xF
Uint8 superd::get_transparent_color()
{
	return 0xF;
}

// returns a pointer to Super Don's video overlay color palette
SDL_Color *superd::get_video_colors()
{
	return colors;
}

// this gets called when the user presses a key or moves the joystick
void superd::input_enable(Uint8 move)
{
	switch (move)
	{
	case SWITCH_UP:
		banks[0] &= ~0x80; 
		break;
	case SWITCH_LEFT:
		banks[0] &= ~0x02;
		break;
	case SWITCH_RIGHT:
		banks[0] &= ~0x20;
		break;
	case SWITCH_DOWN:
		banks[0] &= ~0x08; 
		break;
	case SWITCH_BUTTON1: // '1' on keyboard
		banks[1] &= ~0x80; 
		break;
	case SWITCH_BUTTON2: // '2' on keyboard
		banks[1] &= ~0x40;
		break;
	case SWITCH_BUTTON3: // space on keyboard
		banks[1] &= ~0x20;
		break;
	case SWITCH_COIN1: 
		banks[1] &= ~0x08;
		break;
	case SWITCH_COIN2: 
		banks[1] &= ~0x04;
		break;
	case SWITCH_SERVICE: 
		banks[1] &= ~0x01;
		break;
	case SWITCH_TEST: 
		banks[1] &= ~0x02;
		break;
	default:
		printline("Error, bug in move enable");
		break;
	}
}  

// this gets called when the user releases a key or moves the joystick back to center position
void superd::input_disable(Uint8 move)
{
	switch (move)
	{
	case SWITCH_UP:
		banks[0] |= 0x80; 
		break;
	case SWITCH_LEFT:
		banks[0] |= 0x02;
		break;
	case SWITCH_RIGHT:
		banks[0] |= 0x20;
		break;
	case SWITCH_DOWN:
		banks[0] |= 0x08; 
		break;
	case SWITCH_BUTTON1: // '1' on keyboard
		banks[1] |= 0x80; 
		break;
	case SWITCH_BUTTON2: // '2' on keyboard
		banks[1] |= 0x40;
		break;
	case SWITCH_BUTTON3: // space on keyboard
		banks[1] |= 0x20;
		break;
	case SWITCH_COIN1: 
		banks[1] |= 0x08;
		break;
	case SWITCH_COIN2: 
		banks[1] |= 0x04;
		break;
	case SWITCH_SERVICE: 
		banks[1] |= 0x01;
		break;
	case SWITCH_TEST: 
		banks[1] |= 0x02;
		break;
	default:
		printline("Error, bug in move enable");
		break;
	}
}
