/*
 * thayers.cpp
 *
 * Copyright (C) 2001 Mark Broadhead
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// thayers.cpp
// by Mark Broadhead 
// based, in part, on information by Jeff Kulczycki

#include <stdio.h>
#include <string.h>
#include "thayers.h"
#include "../daphne.h"
#include "../io/conout.h"
#include "../ldp-in/ldv1000.h"
#include "../timer/timer.h"
#include "../video/scoreboard.h"
#include "../sound/ssi263.h"
#include "../ldp-out/ldp.h"
#include "../video/video.h"


#include Z80_INCLUDE

extern ldp *g_ldp;

thayers::thayers()
{
	old_result = 0xfb; // Temp hack
	memset(banks, 0xFF, 2);	// fill banks with 0xFF's
	banks[0] = 0x0;	// put it on free play
	disc_fps = 29.97;
	game_type = GAME_THAYERS;
	cpu_type = CPU_Z80;
	cpu_hz = THAYERS_CPU_HZ;
	irq_period = 0.0; 
	nmi_period = 32.10 / 2; // this should be 32.678(timer from the COP420) but that causes scene overruns
	int_status = 0x3f;
	ldv1000_enable_instant_seeking();
	has_issues = true;
	strcpy(game_issues, "Game can be completed, but synthesized speech is subtitled");
}

// does anything special needed to send an IRQ
void thayers::thayers_irq()
{
//	z80_set_irq_line(0, PULSE_LINE);	// assert IRQ line
	Z80_ASSERT_IRQ;
}

// Thayer's doesn't use the NMI so we'll issue the IRQ's that the cop420 should be doing
void thayers::do_nmi()
{
//	char s[81] = {0};
	
	message_timer++;
	// print the speech string if a new one is available
	if (new_string)
	{
		char* t = ((char *) (&cpumem[0xa000]));

		draw_string(t, 0, 17);

		new_string = false;
		message_timer = 0;
	}
	
	// clear the message after a few seconds
	if (message_timer == 200)
	{
		char t[54] = {0};

		memset(t, 0x20, 53);	// set the string to a bunch of blanks
		
		draw_string(t, 0, 17);
	}

	// reset the int status
	int_status |= 0x04;
	
	// we only issue the timer interrupt every other time this is called
	static int m = 0;
	if (m == 0)
	{
		int_status &= ~0x10;
		m = 1;
	}
	else 
	{
		m = 0;
	}
	thayers_irq();
}

// reads a byte from the cpu's memory
Uint8 thayers::cpu_mem_read(Uint16 addr)
{
	if (addr == 0xbe17) // CHEAT - We need this since the CPU isn't recieving status from the COP420
		return 0x01;
	return cpumem[addr];
}

// writes a byte to the cpu's memory
void thayers::cpu_mem_write(Uint16 addr, Uint8 value)
{
	// any writes to a500 indicate we have a new text string
	if (addr == 0xa500)
	{
		new_string = true;
	}

	cpumem[addr] = value;
}


// reads a byte from the cpu's port
Uint8 thayers::port_read(Uint16 port)
{
	port &= 0xFF;	// strip off high byte

	Uint8 result = 0;

	switch (port)
	{
	case 0x40:	// Bits 3 always high, bit 2 SSI263 Request data, bit 4 /TIMER INT, bit 5 /DATA RDY INT, bit 6 /CART PRES, others are unused
		result = int_status;

		// if we're ready to receive a phoneme, then inform the ISR
		if (ssi263_phoneme_requested())
		{
			result &= ~4;	// clear bit 2 (active low)
		}
		break;
	case 0x80:	// Read data from COP421 - keyboard data should appear here
		result = cop_write_latch;
		break;
	case 0xe0:	// Doesn't really exist... not in schematics
		break;
	case 0xf0:	// Read Data from LD-V1000
		result = read_ldv1000();
		break;
	case 0xf1:	// Bits 0-3 Swith B, bit 4 coin 1, bit 5 coin 2, bit 6 LD Enter, bit 7 LD Ready
		// we keep flipping these bits to simulate command strobe from the LD-V1000
		old_result = static_cast<Uint8>(0xc0 ^ old_result);
		result = static_cast<Uint8>(old_result & banks[1]);
		break;
	case 0xf2:	// Switch A
		result = banks[0]; 
		break;
	
	default:
		char s[81] = { 0 };
	
		sprintf(s, "ERROR: CPU port %x read requested, but this function is unimplemented!", port);
		printline(s);
	}
	return(result);
}

// writes a byte to the cpu's port
void thayers::port_write(Uint16 port, Uint8 value)
{
	char s[81] = { 0 };
	port &= 0xFF;	// strip off high byte

	switch (port)
	{
	case 0x00:	// Ports 00-04 output to speech chip
		ssi263_reg0(value);
		break;
	case 0x01:
		ssi263_reg1(value);
		break;
	case 0x02:
		ssi263_reg2(value);
		break;
	case 0x03:	
		ssi263_reg3(value);
		break;
	case 0x04:
		ssi263_reg4(value);
		break;
	case 0x20:	// Bits 5-7 write to COP421 ports G0-G2, Bit 2 /BANKSELT
		if (value == 0x20)
		{
			cop_write_latch = 0xfa;	
		}
		break;
	case 0x40:	//	Bit 0 Latch for /ENCARTDET, Bit 1 Latch for RESOS, Other bits look unused....
		break;
//	case 0x80:	// Write data to COP421
//		cop_read_latch = value;
//		break;
	case 0xa0:	// TIMER_INT_L 
		int_status |= 0x10;
		break;
	case 0xc0:	// DATA_READY_INT_L 
		int_status |= 0x20;
		cop_write_latch = 0x00;
		break;
	case 0xe0:	// Doesn't really exist? Not in schematics
		break;
	case 0xf3:	// Interrupt Triger
		thayers_irq(); 
//		printline("Got self INT");
		break;
	case 0xf4:	// Write data to LD-V1000
		write_ldv1000(value);
//		sprintf(s, "THAYERS: Write to LD-V1000: %x", value);
//		printline(s);
		break;
	case 0xf5:	// Latch, only bits 4-7 are used, bit 4 coin counter, bit 5 enable writes to LDP, bit 6 LD Enter, bit 7 INT/EXT(LD-V1000)
		break;
	case 0xf6:	// DEN 1 - To scoreboard
		write_scoreboard(static_cast<Uint8>((value & 0x70) >> 4), static_cast<Uint8>(value & 0x0f), 0);
		break;
	case 0xf7:	// DEN 2 - To scoreboard
		write_scoreboard(static_cast<Uint8>((value & 0x70) >> 4), static_cast<Uint8>(value & 0x0f), 1);
		break;
	default:
		sprintf(s, "ERROR: CPU port %x write requested (value %x) at pc %x", port, value, Z80_GET_PC);
		printline(s);
		break;
	}

}
 

// does any video initialization we might need
// returns 'true' if the initialization was successful, false if it failed
bool thayers::video_init()
{
	bool result = true;

	video_overlay = SDL_CreateRGBSurface(SDL_SWSURFACE, THAYERS_OVERLAY_W, THAYERS_OVERLAY_H, 8, 0, 0, 0, 0); // create an 8-bit surface
	
	// fill color palette with schlop because we only use colors 0 and 0xFF for now
	for (int i = 0; i < 256; i++)
	{
		colors[i].r = (unsigned char) i;
		colors[i].g = (unsigned char) i;
		colors[i].b = (unsigned char) i;
	}

	if (video_overlay)
	{
		if (SDL_SetColors(video_overlay, colors, 0, 256))
		{
			result = true;
		}
	}
	else
	{
		printline("Thayers : Could not create RGB surface!");
	}

//	return result;

	
	return scoreboard_init();
}

// shuts down any video we might have initialized
void thayers::video_shutdown()
{
	scoreboard_shutdown();
}

// redraws our video on the screen (in case the screen needs to be redrawn)
void thayers::video_repaint()
{
	scoreboard_repaint();
}

bool thayers::load_roms()
{
	return (
		load_rom("tq/tq_u33.bin", &cpumem[0x0000], 0x8000) &&
		load_rom("tq/tq_u1.bin", &cpumem[0xc000], 0x2000) &&
		load_rom("tq/tq_cop.bin", &coprom[0x000], 0x400)
	);
}

// this might be an older version - it contains secret 'warps' like pressing '1' when druce asks you where to begin
bool thayersalt::load_roms()
{
	return (
		load_rom("tq/tq_u33.bin", &cpumem[0x0000], 0x8000) &&
		load_rom("tqalt/thayers.u1", &cpumem[0xc000], 0x2000) &&
		load_rom("tq/tq_cop.bin", &coprom[0x000], 0x400)
	);
}

// this handles _any_ keypress on the keyboard
void thayers::process_keydown(SDLKey key)
{
	key_press = true;

	// Check to see if value of key is a letter between the a and z keys
	// (there are no SDLK_A or SDLK_Z keys.  the keys are returned raw without checking to see
	// if shift or caps lock is enabled)
	if (key >= SDLK_a && key <= SDLK_z)
	{
		cop_write_latch = static_cast<Uint8>(key - SDLK_SPACE);	// convert lowercase keys to uppercase
		int_status &= ~0x20;
		thayers_irq();
	}

	// check to see if key is a number on the top row of the keyboard (not keypad)
	else if (key >= SDLK_0 && key <= SDLK_9)
	{
		cop_write_latch = static_cast<Uint8>(key);
		int_status &= ~0x20;
		thayers_irq();
	}

	// escape quits daphne =]
	else if (key == SDLK_ESCAPE)
	{
		set_quitflag();
	}

	// blue button 1
	else if (key == SDLK_F1)
	{
		cop_write_latch = 0x80;
		int_status &= ~0x20;
		thayers_irq();
	}

	// blue button 2
	else if (key == SDLK_F2)
	{
		cop_write_latch = 0x81;
		int_status &= ~0x20;
		thayers_irq();
	}

	// blue button 3
	else if (key == SDLK_F3)
	{
		cop_write_latch = 0x82;
		int_status &= ~0x20;
		thayers_irq();
	}

	// blue button 4
	else if (key == SDLK_F4)
	{
		cop_write_latch = 0x83;
		int_status &= ~0x20;
		thayers_irq();
	}
	
	// insert coin left chute
	else if (key == SDLK_F5)
	{
		banks[1] &= ~0x10;
	}

	// insert coin right chute
	else if (key == SDLK_F6)
	{
		banks[1] &= ~0x20;
	}

	// take a screen shot
	else if (key == SDLK_F12)
	{
		g_ldp->request_screenshot();
	}

	// else we recognized no keys so print an error
	else
	{
		char s[81] = { 0 };

		sprintf(s, "THAYERS: Unhandled keypress: %x", key);
		printline(s);
	}
}

// this is called when the user releases a key
void thayers::process_keyup(SDLKey key)
{
	key_press = false;

	// left coin chute release
	if (key == SDLK_F5)
	{
		banks[1] |= 0x10;
	}

	// right coin chute release
	if (key == SDLK_F6)
	{
		banks[1] |= 0x20;
	}
}

// used to set dip switch values
bool thayers::set_bank(unsigned char which_bank, unsigned char value)
{
	bool result = true;
	
	switch (which_bank)
	{
	case 0:	
		banks[0] = (unsigned char) (value ^ 0xFF);	// dip switches are active low
		break;
	case 1:	
		banks[1] = (unsigned char) ((((value ^ 0xFF) & 0x0f)) | (banks[1] & 0xf0));	// dip switches are active low
		break;
	default:
		printline("ERROR: Bank specified is out of range!");
		result = false;
		break;
	}
	
	return result;
}

// Thayer's uses the scoreboard a little different than DL/SA so we need to unpack the data
void thayers::write_scoreboard(Uint8 address, Uint8 data, int den)
{
	if (address <= 5)
	{
		update_player_score(address, data, den);
	}	
	else if (den == 0 && address <= 7)
	{
		update_player_lives(data, address - 6);
	}
	else if (den == 1 && address <= 7)
	{
		update_credits(address - 6 , data);
	}
	else
	{
		char s[81] = { 0 };

		sprintf(s, "THAYERS: Unsupported write to scoreboard: Address %x Data %x ", address, data);
		printline(s);
	}
}

/* These functions are used with the COP cpu core which currently is disabled
void thayers::write_d_port(unsigned char d)
{
	// Data Ready INT
	if (!(d & 0x2))
	{
		int_status &= ~0x20;
	}
	// Timer INT
	if (!(d & 0x1))
	{
		int_status &= ~0x10;
//		printline("Got timer INT");
	}
	// Do interrupt if either of the above are active
	if ((d & 0x3) != 0x03)
	{
		thayers_irq();
	}
}

void thayers::write_l_port(unsigned char l)
{
	if (l)
	{
	}
}

unsigned char thayers::read_l_port()
{
	return cop_read_latch;
}

unsigned char thayers::read_g_port()
{
	return cop_g_read_latch;
}

void thayers::write_g_port(unsigned char) 
{
	printf("Not implemented\n");
}
*/

// returns a pointer to Thayers's video overlay surface
// also returns width, height, and bits per pixel of surface
SDL_Surface *thayers::get_video_overlay()
{
	return(video_overlay);
}

Uint8 thayers::get_transparent_color()
{
	return 0;
}

SDL_Color *thayers::get_video_colors()
{
	return colors;
}

