/*
 * cmdline.cpp
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// Parses our command line and sets our variables accordingly
// by Matt Ownby

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "cmdline.h"
#include "conout.h"
#include <SDL_getenv.h>	// for getenv and putenv support
#include "../video/video.h"
#include "../video/led.h"
#include "../daphne.h"
#include "../game/lair.h"
#include "../game/cliff.h"
#include "../cpu/cpu-debug.h"
#include "../game/game.h"
#include "../game/superd.h"
#include "../game/thayers.h"
#include "../game/speedtest.h"
#include "../game/seektest.h"
#include "../game/cputest.h"
#include "../game/firefox.h"
#include "../game/ffr.h"
#include "../game/astron.h"
#include "../game/esh.h"
#include "../game/laireuro.h"
#include "../game/badlands.h"
#include "../ldp-out/ldp.h"
#include "../ldp-out/sony.h"
#include "../ldp-out/pioneer.h"
#include "../ldp-out/ld-v6000.h"
#include "../ldp-out/hitachi.h"
#include "../ldp-out/philips.h"
#include "../ldp-out/ldp-combo.h"
#include "../ldp-out/ldp-vldp.h"

extern game *g_game;	// our global game class
extern ldp *g_ldp;		// our global ldp class

// Win32 doesn't use strcasecmp, it uses stricmp (lame)
#ifdef WIN32
#define strcasecmp stricmp
#endif


// parses the game type from the command line and allocates g_game
// returns true if successful or false of failed
bool parse_game_type(char *cmdline)
{
	bool result = true;
	char s[81] = { 0 };

	// first thing we need to get from the command line is the game type
	get_next_word(s, cmdline);

	if (strcasecmp(s, "lair")==0)
	{
		g_game = new lair();
	}
	else if (strcasecmp(s, "ace") == 0)
	{
		g_game = new ace();
	}
	else if (strcasecmp(s, "dle") == 0)
	{
		g_game = new dle();
	}
	else if (strcasecmp(s, "cliff")==0)
	{
		g_game = new cliff();
	}
	else if (strcasecmp(s, "gtg") == 0)
	{
		g_game = new gtg();
	}
	else if (strcasecmp(s, "superdon")==0)  //left in for old times' sake
	{
		g_game = new superd();
	}
	else if (strcasecmp(s, "sdq")==0)
	{
		g_game = new superd();
	}
	else if (strcasecmp(s, "sdqshort")==0)
	{
		g_game = new sdqshort();
	}
	else if (strcasecmp(s, "tq")==0)
	{
		g_game = new thayers();
	}
	else if (strcasecmp(s, "tqalt")==0)
	{
		g_game = new thayersalt();
	}
	else if (strcasecmp(s, "speedtest")==0)
	{
		g_game = new speedtest();
	}
	else if (strcasecmp(s, "seektest")==0)
	{
		g_game = new seektest();
	}
//	else if (strcasecmp(s, "cputest")==0)
//	{
//		g_game = new cputest();
//	}
	else if (strcasecmp(s, "firefox")==0)
	{
		g_game = new firefox();
	}
	else if (strcasecmp(s, "ffr")==0)
	{
		g_game = new ffr();
	}
	else if (strcasecmp(s, "astronp")==0)
	{
		g_game = new astron();
	}
	else if (strcasecmp(s, "esh")==0)
	{
		g_game = new esh();
	}
	else if (strcasecmp(s, "astron")==0)
	{
		g_game = new astronh();
	}
	else if (strcasecmp(s, "galaxy")==0)
	{
		g_game = new galaxy();
	}
	else if (strcasecmp(s, "laireuro")==0)
	{
		g_game = new laireuro();
	}
	else if (strcasecmp(s, "blazer")==0)
	{
		g_game = new blazer();
	}
	else if (strcasecmp(s, "cobraab")==0)
	{
		g_game = new cobraab();
	}
	else if (strcasecmp(s, "badlands")==0)
	{
		g_game = new badlands();
	}
	else
	{
		printline("ERROR: Unknown game type specified!");
		result = false;
	}

	// safety check
	if (!g_game)
	{
		result = false;
	}

	return(result);
}


// parses the LDP type from the command line and allocates g_ldp
// returns true if successful or false if failed
bool parse_ldp_type(char *cmdline)
{
	bool result = true;
	char s[81] = { 0 };

	get_next_word(s, cmdline);

	if (strcasecmp(s, "sony")==0)
	{
		g_ldp = new sony();
	}
	else if (strcasecmp(s, "pioneer") == 0)
	{
		g_ldp = new pioneer();
	}
	else if (strcasecmp(s, "v6000")==0)
	{
		g_ldp = new v6000();
	}
	else if (strcasecmp(s, "hitachi")==0)
	{
		g_ldp = new hitachi();
	}
	else if (strcasecmp(s, "philips")==0)
	{
		g_ldp = new philips();
	}
	else if (strcasecmp(s, "vldp")==0)
	{
		g_ldp = new ldp_vldp();
	}
	else if (strcasecmp(s, "combo")==0)
	{
		g_ldp = new combo();
	}
	else if (strcasecmp(s, "noldp")==0)
	{
		g_ldp = new ldp();	// generic interface
	}
	else
	{
		printline("ERROR: Unknown laserdisc player type specified");
		result = false;
	}

	// safety check
	if (!g_ldp)
	{
		result = false;
	}

	return(result);
}



// Parses the command line which is stored in 'cmdline' and sets up the system accordingly
// I chose to have the entire command line stored in 1 string to make it easier for Windows and Linux compatibility
//  UPDATE: however since windows now no longer uses a Winmain function, this is kind of pointless =(
// Returns 1 on success or 0 on failure
bool parse_cmd_line(char *cmdline)
{

	bool result = true;	// assume success
	char s[81] = { 0 };
	int i = 0;

	//////////////////////////////////////////////////////////////////////////////////////

	// if we know our game type
	if (result)
	{
  	  // while we have stuff left in the command line to parse
	  for (;;)
	  {
		get_next_word(s, cmdline);

		// if there is nothing left for us to parse, break out of the while loop
		if (s[0] == 0)
		{
			break;
		}

		// if they are pointing to a framefile to be used by VLDP
		else if (strcasecmp(s, "-framefile")==0)
		{
			ldp_vldp *cur_ldp = dynamic_cast<ldp_vldp *>(g_ldp);	// see if the currently selected LDP is VLDP
			get_next_word(s, cmdline);
			// if it is a vldp, then this option has meaning and we can use it
			if (cur_ldp)
			{
				cur_ldp->set_framefile(s);
			}
			else
			{
				printline("You can only set a framefile using VLDP as your laserdisc player!");
				result = 0;
			}
		}

		else if (strcasecmp(s, "-nosound")==0)
		{
			set_sound_enabled(0);
			printline("Disabling sound...");
		}
		else if (strcasecmp(s, "-scoreboard")==0)
		{
			set_scoreboard(1);
			printline("Enabling external scoreboard...");
		}
		else if (strcasecmp(s, "-scoreport")==0)
		{
			get_next_word(s, cmdline);
			i = atoi(s);
			set_scoreboard_port((unsigned char) i);
			sprintf(s, "Setting scoreboard port to %d", i);
			printline(s);
		}
		else if (strcasecmp(s, "-port")==0)
		{
			get_next_word(s, cmdline);
			i = atoi(s);
			set_serial_port((unsigned char) i);
			sprintf(s, "Setting serial port to %d", i);
			printline(s);
		}
		else if (strcasecmp(s, "-baud")==0)
		{
			get_next_word(s, cmdline);
			i = atoi(s);
			set_baud_rate(i);
			sprintf(s, "Setting baud rate to %d", i);
			printline(s);
		}

		// used to modify the dip switch settings of the game in question
		else if (strcasecmp(s, "-bank")==0)
		{
			get_next_word(s, cmdline);
			i = s[0] - '0';	// bank to modify.  We convert to a number this way to catch more errors
			
			get_next_word(s, cmdline);
			unsigned char value = (unsigned char) (strtol(s, NULL, 2));	// value to be set is in base 2 (binary)

			result = g_game->set_bank((unsigned char) i, (unsigned char) value);
		}

		else if (strcasecmp(s, "-latency")==0)
		{
			get_next_word(s, cmdline);
			i = atoi(s);
			g_ldp->set_search_latency(i);
			sprintf(s, "Setting Search Latency to %d milliseconds", i);
			printline(s);
		}
		else if (strcasecmp(s, "-cheat")==0)
		{
			g_game->enable_cheat();	// enable any cheat we have available :)
		}

		// enable keyboard LEDs for use with games such as Space Ace
		else if (strcasecmp(s, "-enable_leds")==0)
		{
			enable_leds();
		}

		// run daphne in fullscreen mode
		else if (strcasecmp(s, "-fullscreen")==0)
		{
			set_fullscreen(true);
		}
		
		// disable YUV hardware acceleration
		else if (strcasecmp(s, "-nohwaccel")==0)
		{
#ifdef WIN32
			putenv("SDL_VIDEO_YUV_HWACCEL=0");
#else
			setenv("SDL_VIDEO_YUV_HWACCEL", "0", 1);
#endif
		}
		else if (strcasecmp(s, "-pal_dl")==0)
		{
			set_frame_modifier(MOD_PAL_DL);
			printline("Setting DAPHNE up for the PAL Dragon's Lair disc");
			g_game->set_irq_period(LAIR_IRQ_PERIOD * (23.976/25.0));
			// DL PAL runs at a different speed so we need to overclock the IRQ slightly
		}
		else if (strcasecmp(s, "-pal_sa")==0)
		{
			set_frame_modifier(MOD_PAL_SA);
			printline("Setting DAPHNE up for the PAL Space Ace disc");
		}
		else if (strcasecmp(s, "-pal_dl_sc") == 0)
		{
			set_frame_modifier(MOD_PAL_DL_SC);
			printline("Setting DAPHNE up for the PAL Dragon's Lair Software Corner disc");
			g_game->set_irq_period(LAIR_IRQ_PERIOD * (23.976/25.0));
			// DL Amiga runs at a different speed so we need to overclock the IRQ slightly
		}
		else if (strcasecmp(s, "-pal_sa_sc") == 0)
		{
			set_frame_modifier(MOD_PAL_SA_SC);
			printline("Setting DAPHNE up for the PAL Space Ace Software Corner disc");
		}
		else if (strcasecmp(s, "-spaceace91")==0)
		{
			set_frame_modifier(MOD_SA91);
			printline("Setting DAPHNE to play a Space Ace '91 disc");
		}
		else if (strcasecmp(s, "-preset")==0)
		{
			get_next_word(s, cmdline);
			i = atoi(s);
			g_game->set_preset(i);
		}
	else if (strcasecmp(s, "-x")==0)
		{
			get_next_word(s, cmdline);
			i = atoi(s);
			set_video_width((Uint16)i);
			sprintf(s, "Setting screen width to %d", i);
			printline(s);
		}
		else if (strcasecmp(s, "-y")==0)
		{
			get_next_word(s, cmdline);
			i = atoi(s);
			set_video_height((Uint16)i);
			sprintf(s, "Setting screen height to %d", i);
			printline(s);
		}
		else if (strcasecmp(s, "-trace")==0)
		{
#ifdef CPU_DEBUG
			printline("CPU tracing enabled");
			set_cpu_trace(1);
#else
			printline("DAPHNE needs to be compiled in debug mode for this to work");
			result = 0;
#endif
		}

		else
		{
			printline("Unknown command line parameter:");
			printline(s);
			result = false;
		}
	  } // end for
	} // end if we know our game type

	return(result);

}


// copies a single word from cmdline into result (a word is any group of characters separate on both sides by whitespace)
// if there are no more words to be copied, result will contain a null string
// uses a static index to remember the last word it parsed from cmdline
// The reason I decided to use a static index was because I am assuming the command line will only be parsed once
//  and a static index created an easy way for me to implement this function
void get_next_word(char *result, char *cmdline)
{

	static int cmdline_index = 0;
	int result_index = 0;

	// copy a word from the cmdline to the result
	while ((cmdline[cmdline_index] != ' ') && (cmdline[cmdline_index] != 0))
	{
		result[result_index] = cmdline[cmdline_index];
		cmdline_index = cmdline_index + 1;
		result_index = result_index + 1;
	}

	// advance our index until we either get to the next word or else the end of the string
	while (cmdline[cmdline_index] == ' ')
	{
		cmdline_index = cmdline_index + 1;
	}

	result[result_index] = 0;	// terminate our result string

}
