/*
 * pr8210.cpp
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// pr8210.cpp
// by Matt Ownby
// translates PR-8210 blips into commands
// this should be useful for any game that uses the PR-8210

#include <stdio.h>
#include "../daphne.h"
#include "../io/conout.h"
#include "pr8210.h"
#include "../ldp-out/ldp.h"

extern ldp *g_ldp;	// our global ldp class

int g_ignore_count = 0;	// how many consecutive blips to ignore
int g_pr8210_playing = 0;	// whether frames on PR8210 are advancing at
						// usual rate (ie, is PR8210 playing?)
int g_pr8210_seek_received = 0;	// whether we've received a seek command
static char frame[6] = {0};

// returns the value of the g_pr8210_playing flag
// so we don't have to use extern statements to get it
int get_pr8210_playing()
{
	return(g_pr8210_playing);
}

// processes 10 blips into a PR-8210 command
void pr8210_command(unsigned char *blips)
{
	int command = 0;	// the blips without the header+footer
	static int which_command = 0; // to keep track of 1st, 2nd, or 3rd command (it gets sent 3 times)

	char s[81] = {0};
	for (int x = 0; x < 10; x++)
	{
		sprintf(s, "Blip %x: %x", x, blips[x]);
	//	printline(s);
	}

	// if we are supposed to parse the next blip, do so!
	if (g_ignore_count == 0)
	{
	  // only pay attention to first command since we get 3 in a row
	  if (which_command == 0)
	  {
		// verify header and trailer
		if ((blips[0] == 0) && (blips[1] == 0) && (blips[2] == 1)
		   && (blips[8] == 0) && (blips[9] == 0))
		{
			
			// convert command into a 5-digit hex number
			command = (blips[3] << 16) |
					(blips[4] << 12) |
					(blips[5] << 8) |
					(blips[6] << 4) |
					(blips[7]);
					
			switch (command)
			{
				case 0x10100:	// play
					pr8210_play();
					break;
				case 0x00100:	// step forward 
				case 0x01010:	// pause
					pr8210_pause();
					break;
				case 0x01110:	// audio 1/L
					pr8210_audio1();
					break;
				case 0x10110:	// audio 2/R
					pr8210_audio2();
					break;
				case 0x11010:	// seek
					pr8210_seek();
					break;
				case 0x00001:	// 0
					pr8210_add_digit('0');
					break;
				case 0x10001:	// 1
					pr8210_add_digit('1');
					break;
				case 0x01001:	// 2
					pr8210_add_digit('2');
					break;
				case 0x11001:	// 3
					pr8210_add_digit('3');
					break;
				case 0x00101:	// 4
					pr8210_add_digit('4');
					break;
				case 0x10101:	// 5
					pr8210_add_digit('5');
					break;
				case 0x01101:	// 6
					pr8210_add_digit('6');
					break;
				case 0x11101:	// 7
					pr8210_add_digit('7');
					break;
				case 0x00011:	// 8
					pr8210_add_digit('8');
					break;
				case 0x10011:	// 9
					pr8210_add_digit('9');
					break;
				case 0x00000:	// empty filler, just ignore
//					printline("PR8210 Filler");
					break;
				default:
					sprintf(s, "PR8210: Unknown command %x", command);
					printline(s);
					break;
			} // end switch
		} // end valid header and footer
		else
		{
			printline("PR8210 Error: Bad header and footer");
		}
	  } // end if which command is 0

	  // else ignore command since it is 2nd or 3rd

	} // end if we aren't suposed to ignore

	// we are to ignore this one
	else
	{
		g_ignore_count--;
	}
	
	which_command++;
	
	// if this is the 3rd command, get ready for 1st command again
	if (which_command > 2)
	{
		which_command = 0;
	}
	
}

void pr8210_play()
{
//	printline("PR8210 Play");
	g_ldp->pre_play();
	g_pr8210_playing = 1;	// virtual PR8210 is now playing
}

void pr8210_pause()
{
//	printline("PR8210 Pause");
	g_ldp->pre_pause();
	g_pr8210_playing = 0;
}

// Audio 1/L
void pr8210_audio1()
{
	printline("PR8210 Audio 1/L (unsupported command!)");
}

// Audio 2/R
void pr8210_audio2()
{
	printline("PR8210 Audio 2/R (unsupported command!)");
}

void pr8210_seek()
{
//	printline("PR8210 Seek");
	g_pr8210_seek_received = 1;
}

// ch is the digit to add and it's in ASCII format
void pr8210_add_digit(char ch)
{

	static int digit_count = 0;

	// make sure we are getting digits after we've gotten a seek
	if (g_pr8210_seek_received)
	{
//		g_ldp->add_digit(ch);	// add digit to real laserdisc player buffer
		frame[digit_count] = ch;

		digit_count++;
		if (digit_count >= 5)
		{
			g_pr8210_seek_received = 0;
			digit_count = 0;
			g_ldp->pre_search(frame);	// perform seek
			g_pr8210_playing = 0;
		}
	}
	
	// if we are getting digits without first getting a seek, something
	// is wrong
	else
	{
		printline("PR8210 error: digit received without seek command");
	}	
		
}
