/*
 * ldp-vldp.cpp
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// ldp-vldp.c
// by Matt Ownby

// pretends to be an LDP, but uses the VLDP library for output
// (for people who have no laserdisc player)

#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "../io/conout.h"
//#include "../io/conin.h"	// REMOVE ME
#include "../io/error.h"
#include "../video/video.h"
#include "../timer/timer.h"
#include "../daphne.h"
#include "../io/input.h"
#include <SDL_mixer.h>
#include "../game/game.h"
#include "rgb2yuv.h"
#include "ldp-vldp.h"

extern game *g_game;	// global game class (needed to query disc FPS)
void ldp_vldp_audio_callback(void *unused, Uint8 *stream, int len);	// declaration for callback in other function

// video overlay stuff
SDL_Surface *g_gamevid;	// pointer to the game's video overlay
SDL_Color *g_gamecolors;	// pointer to the game's color palette

bool g_take_screenshot = false;	// if true, a screenshot will be taken at the next available opportunity

/////////////////////////////////////////

// We have to dynamically load the .DLL/.so file due to incompatibilities between MSVC++ and mingw32 library files
// These pointers and typedefs assist us in doing so.

typedef int (*initproc)(SDL_Surface *);
typedef int (*openproc)(char *);
typedef int (*intproc)();
typedef int (*searchproc)(Uint16);
typedef void (*yuvproc)(void (*callback)(SDL_Overlay *cur_overlay));
typedef Uint16 (*frameproc)();
typedef double (*fpsproc)();
typedef Uint8 (*fieldsproc)();

initproc pvldp_init;
openproc pvldp_open_mpeg;
searchproc pvldp_search;
intproc pvldp_play, pvldp_pause, pvldp_shutdown;
yuvproc pvldp_set_yuv_callback;
frameproc pvldp_get_current_frame;
fpsproc pvldp_get_fps;
fieldsproc pvldp_uses_fields;

/////////////////////////////////////////

ldp_vldp::ldp_vldp()
{
	blitting_allowed = false;	// Absolutely no video functions are allowed when using VLDP!  this is due to thread issues
	m_target_mpegframe = 0;	// mpeg frame # we are seeking to
	memset( (void *) &m_mpeg_path[0], 0, sizeof(m_mpeg_path));
	memset(m_cur_mpeg_filename, 0, sizeof(m_cur_mpeg_filename));
	m_file_index = 0; // # of mpeg files in our list
	strcpy(m_framefile, "undefined");	// user has to set the name of the framefile
	m_audio_file_opened = false;
	enable_audio1();	// both channels should be on by default
	enable_audio2();
}

// called when daphne starts up
bool ldp_vldp::init_player()
{
	bool result = false;
	bool need_to_parse = false;	// whether we need to parse all video

	if (read_frame_conversions())
	{
		// just a sanity check to make sure their frame file is correct
		if (first_video_file_exists())
		{

			// if the last video file has not been parsed, assume none of them have been
			// This is safe because if they have been parsed, it will just skip them
			if (!last_video_file_parsed())
			{
				printnotice("Press any key to parse your video file(s). This may take a while. Press ESC if you'd rather quit.");
				need_to_parse = true;
			}
			
			if (audio_init() && !get_quitflag())
			{
				if (load_vldp_lib())
				{
					if ((pvldp_init)(get_screen()))
					{
						result = true;

						// if we need to parse all the video
						if (need_to_parse)
						{
							parse_all_video();
						}

						// check to see if we need to draw a YUV video overlay	
						g_gamevid = g_game->get_video_overlay();
						g_gamecolors = g_game->get_video_colors();

						// if this game uses a video overlay, then set the YUV callback.
						if (g_gamevid && g_gamecolors)
						{
							(pvldp_set_yuv_callback)(yuv_callback);
						}
						// else don't set it because performance is better without it
					
						Mix_HookMusic(ldp_vldp_audio_callback, NULL); // hook into VLDP audio callback
						
					} // end if reading the frame conversion file worked
					else
					{
						printline("Could not initialize VLDP!");
					}
				} // end if VLDP loaded ok
				else
				{
					printline("Could not load VLDP dynamic library!!!");
				}
			} // if audio init succeeded
			else
			{
				// only report an audio problem if there is one
				if (!get_quitflag())
				{
					printline("Could not initialize VLDP audio!");
				}

				// otherwise report that they quit
				else
				{
					printline("VLDP : Quit requested, shutting down!");
				}
			} // end if audio init failed or if user opted to quit instead of parse
		} // end if first file was present (sanity check)
		// else if first file was not found, we do just quit because an error is printed elsewhere

	} // end if framefile was read in properly
	else
	{
		printline("Could not open frame file!");
	}

	return result;
}

void ldp_vldp::shutdown_player()
{
	if (pvldp_shutdown)
	{
		(pvldp_shutdown)();
	}
	Mix_HookMusic(NULL, NULL); // unhook from sdl mixer
	free_vldp_lib();
	audio_shutdown();
}

bool ldp_vldp::search(char *frame)
{
	
	bool result = false;
	char filename[81] = { 0 };
	char oggname[81] = { 0 };
	Uint16 target_ld_frame = (Uint16) atoi(frame);
	double fps = 0;
	Uint8 uses_fields = 0;
	double audio_target_pos = 0.0;	// position in audio to seek to

	audio_pause();	// pause the audio before we seek so we don't have overrun

	// if we haven't opened our first file yet
/*	if (fps == 0.0)
	{
		printline("NOTE : FPS not yet available, assuming mpeg matches up with disc for now");
		fps = g_game->get_disc_fps();
	}
*/
	m_target_mpegframe = mpeg_info(filename, target_ld_frame); // try to get a filename
	audio_target_pos = m_target_mpegframe / g_game->get_disc_fps();	// # of seconds to seek to in the audio stream

	// if we can convert target frame into a filename, do it!
	if (filename[0] != 0)
	{
		result = true;	// now we assume success unless we fail
				
		// if the file to be opened is different from the one we have opened
		// OR if we don't yet have a file open ...
		// THEN open the file! :)
		if (strcmp(filename, m_cur_mpeg_filename) != 0)
		{
			// if we were able to successfully open the video file
			if ((pvldp_open_mpeg)(filename))
			{
				result = true;
				strcpy(m_cur_mpeg_filename, filename); // make video file our new current file

				// try to open an optional audio file to go along with video
				oggize_path(oggname, filename);
				m_audio_file_opened = open_audio_stream(oggname);
			}
			else
			{
				outstr("LDP-VLDP.CPP : Could not open video file ");
				printline(filename);
			}
		}

		// if we're ok so far, try the search
		if (result)
		{
			fps = (pvldp_get_fps)();
			uses_fields = (pvldp_uses_fields)();

			// if we're using fields, we have to modify the mpeg target frame so it is the mpeg target field instead
			if (uses_fields)
			{
				m_target_mpegframe = m_target_mpegframe << 1;	// multiply by two since two fields == one frame
			}

			// if the mpeg's FPS and the disc's FPS differ, we need to adjust the mpeg frame
			// NOTE: AVOID this if you can because it makes seeking less accurate
			if (g_game->get_disc_fps() != fps)
			{
				char s[160] = { 0 };
				sprintf(s, "NOTE: converting FPS from %f to %f. This may be less accurate.", g_game->get_disc_fps(), fps);
				printline(s);
				m_target_mpegframe = (Uint16) (((m_target_mpegframe / g_game->get_disc_fps()) * fps) + 0.5);
				// add 0.5 to round to nearest whole number when truncating		
			}

			// try to search to the requested frame
			if ((pvldp_search)((Uint16) m_target_mpegframe))
			{
				// if we have an audio file opened, do an audio seek also
				if (m_audio_file_opened)
				{
					result = seek_audio(audio_target_pos);
				}
			}
			else
			{
				printline("LDP-VLDP.CPP : Search failed in video or audio file");
			}
		}
		// else opening the file failed
	}

	return(result);
}

void ldp_vldp::play()
{
	char full_path[320];
	char ogg_path[320];
	int ok = 1;	// whether it's ok to issue the play command
	
	// if we haven't opened any mpeg file yet, then do so now
	if (!m_cur_mpeg_filename[0])
	{
		strcpy(full_path, m_mpeg_path);
		strcat(full_path, m_mpeginfo[0].name);
		ok = (pvldp_open_mpeg)(full_path);	// get the first mpeg available in our list
		if (ok)
		{
			strcpy(m_cur_mpeg_filename, full_path);
//			printf("LDP-VLDP : PLAY, set filename to %s\n", m_cur_mpeg_filename);
			
			// try to open an optional audio file to go along with video
			oggize_path(ogg_path, full_path);
			m_audio_file_opened = open_audio_stream(ogg_path);
		}

		else
		{
			outstr("LDP-VLDP.CPP : in play() function, could not open mpeg file ");
			printline(full_path);
		}
	}

	// don't issue play if the mpeg open command failed	
	if (ok)
	{
		ok = (pvldp_play)();
		audio_play();
	}
	
	if (!ok)
	{
		printline("VLDP ERROR : play command failed!");
	}
}

void ldp_vldp::pause()
{
	(pvldp_pause)();
	audio_pause();
}

/*
Uint16 ldp_vldp::get_current_frame()
{
	// since the mpeg's beginning does not correspond to the laserdisc's beginning, we add the offset
	Sint32 result = m_cur_ldframe_offset + (pvldp_get_current_frame)();
	double fps = (pvldp_get_fps)();
	
	if (result <= 0)
	{
		result = 0;
	}

	else
	{
		// FIXME : THIS CODE HAS BUGS IN IT, I HAVEN'T TRACKED THEM DOWN YET HEHE
		// if the mpeg's FPS and the disc's FPS differ, we need to adjust the frame that we return
		// NOTE: AVOID this if you can because it makes seeking less accurate
		if (g_game->get_disc_fps() != fps)
		{
//			printf("Converting frame %d to ", result);
			result = (Sint32) (((result / fps) * g_game->get_disc_fps()) + 0.5);
			// add 0.5 to round to nearest whole number when truncating
//			printf("%d\n", result);
		}
	}

	return (Uint16) result;
}
*/

// takes a screenshot of the current frame + any video overlay
void ldp_vldp::request_screenshot()
{
	g_take_screenshot = true;
}

// sets the name of the frame file
void ldp_vldp::set_framefile(char *filename)
{
	strcpy(m_framefile, filename);
}

//////////////////////////////////

// loads the VLDP dynamic library, returning true on success
bool ldp_vldp::load_vldp_lib()
{
	bool result = false;

    m_dll_instance = M_LOAD_LIB(vldp);	// load VLDP.DLL or libvldp.so

    // If the handle is valid, try to get the function address. 
    if (m_dll_instance) 
    {
		pvldp_init = (initproc) M_GET_SYM(m_dll_instance, "vldp_init");
		pvldp_open_mpeg = (openproc) M_GET_SYM(m_dll_instance, "vldp_open_mpeg");
		pvldp_search = (searchproc) M_GET_SYM(m_dll_instance, "vldp_search");
		pvldp_play = (intproc) M_GET_SYM(m_dll_instance, "vldp_play");
		pvldp_pause = (intproc) M_GET_SYM(m_dll_instance, "vldp_pause");
		pvldp_shutdown = (intproc) M_GET_SYM(m_dll_instance, "vldp_shutdown");
		pvldp_set_yuv_callback = (yuvproc) M_GET_SYM(m_dll_instance, "vldp_set_yuv_callback");
		pvldp_get_current_frame = (frameproc) M_GET_SYM(m_dll_instance, "vldp_get_current_frame");
		pvldp_get_fps = (fpsproc) M_GET_SYM(m_dll_instance, "vldp_get_fps");
		pvldp_uses_fields = (fieldsproc) M_GET_SYM(m_dll_instance, "vldp_uses_fields");

		// if all functions were found, then return success
		if (pvldp_init && pvldp_open_mpeg && pvldp_search && pvldp_play && pvldp_pause && pvldp_shutdown
			&& pvldp_set_yuv_callback && pvldp_get_current_frame && pvldp_get_fps && pvldp_uses_fields)
		{
			result = true;
		}
		else
		{
			printline("VLDP LOAD ERROR : some DLL functions could not be loaded");
		}
	}
	else
	{
		printline("ERROR: could not open the VLDP dynamic library");
	}

	return result;
}

// frees the VLDP dynamic library if we loaded it in
void ldp_vldp::free_vldp_lib()
{
	// don't free if the library was never opened
	if (m_dll_instance)
	{
		M_FREE_LIB(m_dll_instance);
	}
}




// read frame conversions in from LD-frame to mpeg-frame data file
bool ldp_vldp::read_frame_conversions()
{

	FILE *fileConvert = NULL;
	char s[81] = { 0 };
	int cAssigned;
	char frame_string[81] = { 0 };
	Sint32 frame;
	bool result = false;

	fileConvert = fopen(m_framefile, "r");

	// if the file was not found in relative directory, try looking for it in a framefile directory
	if (!fileConvert)
	{
		char newpath[160];
		strcpy(newpath, "framefile/");
		strcat(newpath, m_framefile);
		fileConvert = fopen(newpath, "r");
	}

	if (fileConvert)
	{
		fscanf(fileConvert, "%s", m_mpeg_path); // read in path

		// Clean up after the user if they goofed up on the path
		// Windows paths use \, linux uses /
#ifdef WIN32
		if (m_mpeg_path[strlen(m_mpeg_path)-1] != '\\')
		{
			printline("VLDP path missing \\ on the end.  I'm adding it.");
			strcat(m_mpeg_path, "\\");
		}
#else
		if (m_mpeg_path[strlen(m_mpeg_path)-1] != '/')
		{
			printline("VLDP path missing / on the end.  I'm adding it.");
			strcat(m_mpeg_path, "/");
		}
#endif

		// read entire file
		while(!feof(fileConvert))
		{

			cAssigned = fscanf(fileConvert, "%s %s", frame_string, s);
			frame = atoi(frame_string);
				               
			// if we read 2 entries like we expected, store them
			if(cAssigned == 2)
			{
				m_mpeginfo[m_file_index].frame = (Sint32) frame;
				strcpy(m_mpeginfo[m_file_index].name, s);
				m_file_index = m_file_index + 1;
			}
		}

		result = true;	
		fclose(fileConvert);
	}
	else
	{
		sprintf(s, "Could not open mpeg frame conversion file : %s", m_framefile);
		printline(s);
	}
	
	return result;
}

// if file does not exist, we print an error message
bool ldp_vldp::first_video_file_exists()
{
	char full_path[160] = { 0 };
	char s[160] = { 0 };
	FILE *F = NULL;
	bool result = false;
	
	// if we have at least one file
	if (m_file_index)
	{
		strcpy(full_path, m_mpeg_path);
		strcat(full_path, m_mpeginfo[0].name);	// first video file
		F = fopen(full_path, "rb");	// test to see if file exists
		
		// if we could open the file, then file exists
		if (F)
		{
			result = true;
			fclose(F);
		}
		else
		{
			sprintf(s, "Could not open file : %s", full_path);
			printerror(s);
		}
	}

	return result;
}

// returns true if the last video file has been parsed
// This is so we don't parse_all_video if all files are already parsed
bool ldp_vldp::last_video_file_parsed()
{
	char full_path[160] = { 0 };
	FILE *F = NULL;
	bool result = false;
	
	// if we have at least one file
	if (m_file_index)
	{
		strcpy(full_path, m_mpeg_path);
		strcat(full_path, m_mpeginfo[m_file_index-1].name);
		strcpy(full_path + strlen(full_path) - 3, "dat");	// change suffix
		F = fopen(full_path, "rb");	// test to see if file exists
		
		// if we could open the file, then file exists
		if (F)
		{
			result = true;
			fclose(F);
		}
	}
	// else there is a problem with the frame file so return false
	
	return result;
}

// opens (and closes) all video files, forcing any unparsed video files to get parsed
void ldp_vldp::parse_all_video()
{
	int i = 0;
	char full_path[160] = { 0 };
	
	for (i = 0; i < m_file_index; i++)
	{
		strcpy(full_path, m_mpeg_path);	// create a full pathname to file
		strcat(full_path, m_mpeginfo[i].name);	// add current filename
		printf("Opening file %s ...\n", full_path);
		(pvldp_open_mpeg)(full_path);
		(pvldp_search)(0);	// search to frame 0 to render it so the user has something to watch while he/she waits
//		con_getkey();	// wait for keypress after rendering frame (REMOVE ME, JUST FOR TESTING)
	}
}

// returns # of frames to seek into file, and mpeg filename
// if there's an error, filename is NULL
// WARNING: This assumes the mpeg and disc are running at exactly the same FPS
// If they aren't, you will need to calculate the actual mpeg frame to seek to
// The reason I don't return time here instead of frames is because it is more accurate to
//  return frames if they are at the same FPS (which hopefully they are hehe)
Uint16 ldp_vldp::mpeg_info (char *filename, Uint16 ld_frame)
{
	int index = 0;
	Uint16 mpeg_frame = 0;	// which mpeg frame to seek (assuming mpeg and disc have same FPS)

	// find the mpeg file that has the LD frame inside of it
	while ((index+1 < m_file_index) && (ld_frame >= m_mpeginfo[index+1].frame))
	{
		index = index + 1;
	}

	// make sure we have a filename at this spot
	if (m_mpeginfo[index].name[0])
	{
		strcpy(filename, m_mpeg_path);
		strcat(filename, m_mpeginfo[index].name);
		mpeg_frame = (Uint16) (ld_frame - m_mpeginfo[index].frame);
		m_cur_ldframe_offset = m_mpeginfo[index].frame;
//		printf("setting m_cur_ldframe_offset %d\n", m_cur_ldframe_offset);
	}
	else
	{
		printline("VLDP error, no filename found");
		filename[0] = 0;	// ERROR
	}
	
	return(mpeg_frame);
}

////////////////////////////////////////////////////////////////////////////////////////








// Draws game video overlay on top of laserdisc YUV video
// It is assumed the the Overlay is _already locked_
// NOTE : this function could be optimized--it's called every frame on games w/ video overlay
void yuv_callback(SDL_Overlay *dst)
{

	// we assume that g_gamevid, and g_gamecolors are all valid at this point

	Uint8 transparent_val = g_game->get_transparent_color();
	int vertical_padding = g_game->get_video_row_offset();
	Uint8 *Y = dst->pixels[0] + (dst->pitches[0] * vertical_padding);	// the Y channel
	Uint8 *V = dst->pixels[1] + ((dst->w >> 2) * vertical_padding);	// V channel
	Uint8 *U = dst->pixels[2] + ((dst->w >> 2) * vertical_padding);	// U channel
	struct yuvcolors8 yuv_components = { 0, 0, 0 };	// the YUV components of the game video pixels
	int row = 0;
	int col = 0;

	Uint8 *gamevid_pixels = (Uint8 *) g_gamevid->pixels;
	Uint8 *temp = NULL;	// temporary pointer
	Uint8 cur_color = 0;
	int last_color = -1;	// the last color we converted to YUV
	// made an integer so I could set it to -1 and not risk overlapping
	// any legitimate 8-bit color

	// go through each row of the overlay and copy it to the YUV stream
	for (row = 0; (row < g_gamevid->h) && ((row << 1) < dst->h); row ++)
	{
		for (col = 0; col < g_gamevid->w; col++)
		{
			cur_color = *gamevid_pixels;
			
			// if current game video overlay pixel is not transparent
			// then write it to the YUV stream
			if (cur_color != transparent_val)
			{
				// RGB2YUV conversion is expensive so we only want to do it if it
				// necessary
				if (cur_color != last_color)
				{
					// convert RGB to YUV since our mpeg stream is in YUV format
					yuv_components = rgb2yuv8(g_gamecolors[cur_color].r,
						g_gamecolors[cur_color].g,
						g_gamecolors[cur_color].b);
					last_color = cur_color;
				}

				// we double size of pixel so we need to write
				// the value in 4 places
				
				// this plots a 2x2 square (a 1x1 pixel doubled in size)
				*Y = yuv_components.y;
				*(Y+1) = yuv_components.y;
				temp = Y + dst->w;
				*temp = yuv_components.y;
				*(temp+1) = yuv_components.y;

				// U and V are half the size of Y so we do not need to enlarge
				*V = yuv_components.v;
				*U = yuv_components.u;
			}
			gamevid_pixels++;	// advance to next pixel
			Y += 2;	// advance two pixels over since game video is doubled
			U++;
			V++;
		}
		Y += dst->w;
	}

	// if we've been instructed to take a screenshot, do so now that the overlay is in place
	if (g_take_screenshot)
	{
		g_take_screenshot = false;
		take_screenshot(dst);
	}
}
