
/*
 * ldp.cpp
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LDP.CPP
// by Matt Ownby

// The code in this file translates general LDP functions into LDP-specific functions
// Part of the Daphne emulator

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../daphne.h"
#include "../io/serial.h"
#include "ldp.h"
#include "../timer/timer.h"
#include "../io/conout.h"
#include "framemod.h"
#include "../game/game.h"
#include "../game/boardinfo.h"
#include Z80_INCLUDE

extern game *g_game;	// global name class

// generic ldp constructor
ldp::ldp()
{
	blitting_allowed = true;
	need_serial = false;
	serial_initialized = false;	// assume we don't need serial
	player_initialized = false;
	last_frame = 0;
	play_time = g_game->get_total_cycles();
	is_playing = false;
	search_latency = 0;
	stop_on_quit = false;	// stopping on quit is not cool for testing
	skipping_supported = false;
	skip_instead_of_search = false;
	max_skippable_frames = 0;
}

// Initializes and "pings" the LDP/VLDP and returns true if the player is online and detected
// or false if the LDP could not be initialized
bool ldp::init ()
{

	bool result = true;	// assume everything works until we find out otherwise
	bool temp = true;	// needed to make && work the way we want below

	// If we are controlling a real LDP,
	// or if we are controlling a combo and DVD initialized properly,
	// then initialize the serial port here
	if (need_serial)
	{
		serial_initialized = serial_init(get_serial_port(), get_baud_rate());
		temp = serial_initialized;
	}
	
	player_initialized = init_player();
	result = temp && player_initialized;
	
	return(result);

}

// initializes the player after serial has been initialized
// this is so that the main init function can initialize the serial without having each LDP
// initialize its own serial
bool ldp::init_player()
{

	return(true);

}

// STOPs the player, shuts down the serial if it's open, then calls shutdown_player for player-specific stuff
void ldp::shutdown()
{
	// if player has been initialized, stop the player from playing
	if (player_initialized && stop_on_quit)
	{
		pre_stop();
	}
	
	// if serial has been initialized, shut it down now
	if (serial_initialized)
	{
		serial_close();
		serial_initialized = false;
	}

	shutdown_player();
	player_initialized = false;
}

// does any player specific stuff  
void ldp::shutdown_player()
{
}

// prepares to search
// searches to the frame inside the "frame" array
// frame_index should be set to FRAME_SIZE before this function is called
// (this happens naturally if add_digit is called 5 times)
bool ldp::pre_search(char* frame)
{

	Uint16 frame_number = 0;	// frame to search to
	bool result = false;
	char s1[81] = { 0 };

	frame[5] = 0;	// terminate the string ...
	frame_number = (Uint16) atoi(frame);

	// If we need to alter the frame # before searching
	if (need_frame_conversion())
	{
		Uint16 unadjusted_frame = frame_number;
		frame_number = (Uint16) do_frame_conversion(frame_number);
		framenum_to_frame(frame_number, frame);
		sprintf(s1, "Search to %d (formerly %d) received - ", frame_number, unadjusted_frame);
	}
	else
	{
		sprintf(s1, "Search to %d received", frame_number);
	}

	outstr(s1);

	// if it's Dragon's Lair/Space Ace, print the board we are on
	if (g_game->get_game_type() <= GAME_ACE)
	{
		outstr(" - ");
		print_board_info(g_game->cpumem[0xA00E], g_game->cpumem[0xA00F], g_game->cpumem[Z80_GET_IY]);
	}
	else
	{
		newline();
	}

	// If the user requested a delay before seeking, make it now
	if (search_latency > 0)
	{
		make_delay(get_search_latency());
	}

	// HERE IS WHERE THE SEARCH ACTUALLY TAKES PLACE

	// if the player supports skipping AND user has requested the we skip instead of searching when possible
	if (skipping_supported & skip_instead_of_search)
	{
		int difference = frame_number - get_current_frame();	// how many frames we'd have to skip

		// if we can skip forward instead of seeking, do it!
		if ((difference <= max_skippable_frames) && (difference > 1))
		{
			result = pre_skip_forward((Uint16) difference);
		}
		// if it's too far to skip, search instead
		else
		{
			result = search(frame);
		}
	}
	// otherwise do a regular search
	else
	{
		result = search(frame);
	}

	if (result)
	{
		last_frame = (Uint16) atoi(frame);
	}
	g_game->flush_timers();
	is_playing = false;

	return(result);
}

// does the actual search
// frame is in ASCII format, and is always 5 digits
bool ldp::search(char *new_frame)
{
	// I did this to get rid of compiler warnings
	if (new_frame)
	{
	}

	return(true);
}

// prepare to skip forward a certain # of frames and continue playing
// Call this function, do not call skip_forward directly
bool ldp::pre_skip_forward(Uint16 frames_to_skip)
{
	bool result = false;
	char s[81] = { 0 };

	sprintf(s, "Skipping forward %d frames", frames_to_skip);
	printline(s);	

	if (is_playing)
	{
		result = skip_forward(frames_to_skip);
	}
	else
	{
		printline("LDP ERROR: Skip forward command was called when the disc wasn't playing");
	}

	g_game->flush_timers();	// !!! WILL THIS FIX SUPER DON ?????? (Matt being excited)

	return(result);
}

// skips forward a certain number of frames and continues playing
// Do not call this function directly!  Call pre_skip_forward instead
bool ldp::skip_forward(Uint16 frames_to_skip)
{

	Uint16 cur_frame = get_current_frame();
	bool result = false;
	
	char frame[6] = {0};

	cur_frame = (Uint16) (cur_frame + frames_to_skip);
	sprintf(frame, "%05d", cur_frame);
	//frame_index = 5;	// index must be here because pre_search terminates string according to index

	result = pre_search(frame);
	pre_play();

	return(result);
}

// prepares to play the disc
void ldp::pre_play()
{

	printline("Play");

	// we only want to refresh the frame calculation stuff if the disc is
	// not playing
	if (!is_playing)
	{
		play();	// it's important for SMPEG that a play be only sent if the disc is not playing
		is_playing = true;
		play_time = g_game->get_total_cycles();
	}
	else
	{
		printline("(LDP : disc is already playing, play command ignored)");
	}
	g_game->flush_timers();

}

// starts playing the laserdisc
void ldp::play()
{
}

// sets the speed at which we playback at (1 is standard speed)
void ldp::set_play_speed(Uint8 speed)
{
	char s[81] = { 0 };

	sprintf(s, "LDP: Playback speed of %d was requested (ignored)", speed);
	printline(s);
}

// prepares to pause
void ldp::pre_pause()
{
	// only send pause command if disc is playing
	// some games (Super Don) repeatedly flood with a pause command and this doesn't work well with the Hitachi
	if (is_playing)
	{
		printline("Pause");
		last_frame = get_current_frame();
		is_playing = false;
		pause();
	}
	else
	{
		printline("LDP : Received pause while disc was stopped, ignoring");
	}
}

// pauses the disc
void ldp::pause()
{
}

// prepares to stop the disc
// "stop" is defined as going to frame 0 and stopping the motor so that
// the player has to spin up again to begin playing
void ldp::pre_stop()
{
	printline("Stop");
	last_frame = 0;
	stop();
	is_playing = false;
}

// stops the disc
void ldp::stop()
{
}

// returns the current frame number that the disc is on
// this is a generic function which computes the current frame number using the elapsed time
// and the framerate of the disc.  Obviously querying the laserdisc player would be preferable
// if possible (some laserdisc players don't like to be queried too often)
Uint16 ldp::get_current_frame()
{
	Uint16 result = 0;
//	double elapsed_secs = ((double) elapsed_ms_time(play_time)) / 1000.0; // compute elapsed seconds
	
	// it seems kind of silly to cast to Uint32 then to double, but Uint64 can't go straight to double
	double elapsed_secs = ((double)((Uint32)(g_game->get_elapsed_cycles(play_time))) / (g_game->get_cpu_hz()));

	// if the disc is playing, compute which frame # we're on
	if (is_playing)
	{
		result = (Uint16) (last_frame + (Uint16) ((elapsed_secs * g_game->get_disc_fps())+0.5));
		// I add +0.5 to round to the nearest whole number instead of truncate
	}
	// otherwise the disc is idle, so return the frame
	else
	{
		result = last_frame;
	}

	return(result);
}

// steps froward 1 frame
// (I added this for debugging purposes)
void ldp::pre_step_forward()
{
	// this assumes the player is paused
	last_frame ++;
}

// returns value of blitting_allowed.  SMPEG does not allow blitting
bool ldp::is_blitting_allowed()
{
	return (blitting_allowed);
}

bool ldp::is_disc_playing()
{
	return is_playing;
}

// converts an integer frame number into ASCII (with leading zeroes)
void ldp::framenum_to_frame(Uint16 num, char *f)
{

    sprintf(f, "%05d", num);

}

Uint32 ldp::get_search_latency()
{
	return(search_latency);
}

void ldp::set_search_latency(Uint32 value)
{
	search_latency = value;
}

void ldp::enable_audio1()
{
	printline("Audio1 enable received (ignored)");
}

void ldp::enable_audio2()
{
	printline("Audio2 enable received (ignored)");
}

void ldp::disable_audio1()
{
	printline("Audio1 disable received (ignored)");
}

void ldp::disable_audio2()
{
	printline("Audio2 disable received (ignored)");
}

// asks LDP to take a screenshot if that's possible
// it's only possible with SMPEG as of this time
void ldp::request_screenshot()
{
	printline("NOTE: current laserdisc player does not support taking screenshots, sorry");
}
