/*
 * scoreboard.cpp
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


// Handles the scoreboard for Dragon's Lair / Space Ace

#include "../daphne.h"
#include "video.h"
#include "../io/parallel.h"
#include "../io/conout.h"
#include "../timer/timer.h"
#include "../io/error.h"

// Arrays that hold the current value of the LED's so we don't display the same number over and over again
// (Dragon's Lair will constantly refresh the credits LED's, this just wastes CPU power to display it constantly)
// This is also used when the LED's need to be redrawn

int player1_score_values[6] = { 0xf, 0xf, 0xf, 0xf, 0xf, 0xf };	// 0xF is a blank LED
int player1_lives_value = 0xf;
int player2_score_values[6] = { 0xf, 0xf, 0xf, 0xf, 0xf, 0xf };
int player2_lives_value = 0xf;
int credits_values[2] = { 0xf, 0xf };

unsigned char rsb_value = 0;	// real scoreboard value

// Graphical coordinates of where all these LED's should go

#define LED_WIDTH 40
#define LED_HEIGHT 53
#define TITLE_HEIGHT 20

const int player_title_x = 217;
const int player_score_x = 200;
const int player1_score_y = 73;
const int player1_title_y = player1_score_y - 27;
const int player1_lives_y = player1_score_y + LED_HEIGHT + 14;
const int player_lives_x = 320 - LED_WIDTH + 15;
const int player1_lives_title_y = player1_lives_y + (LED_HEIGHT/2) - (TITLE_HEIGHT/2);	// centered with LED
const int player_lives_title_x = player_lives_x + LED_WIDTH + 21;
const int player2_score_y = player1_lives_y + LED_HEIGHT + 45;
const int player2_title_y = player2_score_y - 27;
const int player2_lives_y = player2_score_y + LED_HEIGHT + 14;
const int player2_lives_title_y = player2_lives_y + (LED_HEIGHT/2) - (TITLE_HEIGHT/2);
const int credits_x = player_lives_x - LED_WIDTH;
const int credits_y = player2_lives_y + LED_HEIGHT + 23;
const int credits_title_y = credits_y + 18;


// digit = which LED to update (range 0-5)
// LED 0 is the left-most LED
// value = which value to put there (range 0-F)
// returns 1 on success, 0 on failure
// player = which player, 0 is player 1, anything else is player 2
int update_player_score(int digit, int value, int player)
{

	int success = 0;
	int coord_x = 0;
	int coord_y = 0;
	int already_displayed = 1;	// whether the number is already displayed on the scoreboard
	int rsb_port = get_scoreboard_port();

	// make sure all parameters are within the proper range and we aren't in debug mode
	if (((digit >=0) && (digit <= 5))
		&& ((value >=0) && (value <= 0x0F)))
	{
		coord_x = (digit * LED_WIDTH) + player_score_x;
		if (player == 0)
		{
			// if number is not already displayed
			if (player1_score_values[digit] != value)
			{
				coord_y = player1_score_y;
				player1_score_values[digit] = value;
				already_displayed = 0;

				// if we are driving a real scoreboard
				if (get_scoreboard())
				{
					rsb_value = (unsigned char) ((digit << 4) | value);
					par_base2(rsb_port, 0);
					par_base0(rsb_port, rsb_value);
					par_base2(rsb_port, 2);
				}
			}
		}
		else
		{
			// if number is not already displayed
			if (player2_score_values[digit] != value)
			{
				coord_y = player2_score_y;
				player2_score_values[digit] = value;
				already_displayed = 0;

				// if we are driving a real scoreboard
				if (get_scoreboard())
				{
					rsb_value = (unsigned char) ((digit << 4) | value);
					par_base2(rsb_port, 0);
					par_base0(rsb_port, rsb_value);
					par_base2(rsb_port, 8);
				}
			}
		}

		// if the number we want to draw is not already being displayed
		if (!already_displayed)
		{
			success = draw_led(value, coord_x, coord_y);
		}
	}
	// else the parameters are out of range

	else
	{
		char s[81] = { 0 };

		sprintf(s, "SCOREBOARD: Unsupported write to scoreboard: Digit %x Value %x ", digit, value);
		printline(s);
	}

	return(success);

}

// value = which value to put in the single LED (range 0-F)
// player = which player, 0 is player 1, anything else is player 2
// returns 1 on success
int update_player_lives(int value, int player)
{

	int success = 0;
	int coord_x = 0, coord_y = 0;
	int already_displayed = 1;
	int rsb_port = get_scoreboard_port();

	if ((value >= 0) && (value <= 0x0F))
	{
		coord_x = player_lives_x;
		if (player == 0)
		{
			if (player1_lives_value != value)
			{
				coord_y = player1_lives_y;
				player1_lives_value = value;
				already_displayed = 0;

				if (get_scoreboard())
				{
					rsb_value=6; //p1 lives
					rsb_value=(unsigned char) (rsb_value<<4); //shift to address space
					rsb_value |= value; //turn on digit value
					par_base2(rsb_port, 0);
					par_base0(rsb_port, rsb_value);
					par_base2(rsb_port, 2);
				}
			}
		}
		else
		{
			if (player2_lives_value != value)
			{
				coord_y = player2_lives_y;
				player2_lives_value = value;
				already_displayed = 0;

				if (get_scoreboard())
				{
					rsb_value = 7; //p2 lives
					rsb_value = (unsigned char) (rsb_value << 4); //shift to address space
					rsb_value |= value; //turn on digit value
					par_base2(rsb_port, 0);
					par_base0(rsb_port, rsb_value);
					par_base2(rsb_port, 2);
				}
			}
		}

		if (!already_displayed)
		{
			success = draw_led(value, coord_x, coord_y);
		}
	}

	return(success);

}

// digit = which LED to update (0 or 1)
// value = which value to put there (range 0-F)
// returns 1 on success
int update_credits(int digit, int value)
{

	int success = 0;
	int coord_x = 0;
	int rsb_port = get_scoreboard_port();

	if (((digit >= 0) && (digit <= 1))
		&& ((value >= 0) && (value <= 0x0F)))
	{
		// if the number to be displayed is not already displayed
		if (credits_values[digit] != value)
		{
			coord_x = (LED_WIDTH * digit) + credits_x;
			credits_values[digit] = value;
			success = draw_led(value, coord_x, credits_y);

			if (get_scoreboard())
			{
				if (digit == 0)
				{
					rsb_value = 6;
				}
				else
				{
					rsb_value = 7;
				}
				rsb_value = (unsigned char) (rsb_value << 4);
				rsb_value |= value;
				par_base2(rsb_port, 0);
				par_base0(rsb_port, rsb_value);
				par_base2(rsb_port, 8);
			} // end real scoreboard
		}
	}

	return(success);

}



// repaints the scoreboard including all of the words such as "PLAYER 1 SCORE"
// returns 1 if successful, 0 if failed
void scoreboard_repaint()
{

	int digit_index = 0;
	int value1 = 0, value2 = 0;	// values to be displayed in each LED

	draw_othergfx(B_DL_PLAYER1, player_title_x, player1_title_y);
	draw_othergfx(B_DL_PLAYER2, player_title_x, player2_title_y);
	draw_othergfx(B_DL_LIVES, player_lives_title_x, player1_lives_title_y);
	draw_othergfx(B_DL_LIVES, player_lives_title_x, player2_lives_title_y);
	draw_othergfx(B_DL_CREDITS, player_lives_title_x, credits_title_y);

	// Redraw each player's score
	for (digit_index = 0; digit_index < 6; digit_index ++)
	{
		value1 = player1_score_values[digit_index];	// grab the value before we alter it
		value2 = player2_score_values[digit_index];
		player1_score_values[digit_index] ^= 0x0F;	// arbitrary number that forces a redraw because it is different
		player2_score_values[digit_index] ^= 0x0F;
		update_player_score(digit_index, value1, 0);
		update_player_score(digit_index, value2, 1);
	}

	// Redraw each player's lives
	value1 = player1_lives_value;
	value2 = player2_lives_value;
	player1_lives_value ^= 0x0F;	// arbitrary number that forces a redraw because it is different
	player2_lives_value ^= 0x0F;
	update_player_lives(value1, 0);
	update_player_lives(value2, 1);

	// Redraw the credits
	value1 = credits_values[0];
	value2 = credits_values[1];
	credits_values[0] ^= 0x0F;	// arbitrary number that forces a redraw because it is different
	credits_values[1] ^= 0x0F;
	update_credits(0, value1);
	update_credits(1, value2);

}



// initializes the scoreboard, including
//  setting up parallel port, if need be
// returns 1 if successful or 0 if not successful
bool scoreboard_init()
{

	bool result = true;

	// for now we're just doing a quick hack!
	if (get_scoreboard())
	{
		result = par_init(get_scoreboard_port());
	}
	scoreboard_repaint();

	if (!result)
	{
		printerror("Scoreboard initialization failed!");
	}

	return(result);

}

void scoreboard_shutdown()
{

	if (get_scoreboard())
	{
		par_close(get_scoreboard_port());
	}

}
