/*
 * tms9128nl.cpp
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <SDL.h>
#include "SDL_DrawText.h"
//#include "z80.h"
#include "tms9128nl.h"
#include "video.h"
#include "../daphne.h"
#include "../io/conout.h"
#include "../ldp-out/ldp.h"
#include <stdio.h>
#include <string.h>

extern ldp *g_ldp;

static unsigned char vidmem[32767] = { 0 };	// video memory
static unsigned char lowbyte = 0;
static unsigned char highbyte = 0;
static unsigned int rvidindex;
static unsigned int wvidindex;
static int toggleflag = 0; //keep track of low / high byte
static int vidmode = 0;  //keep track of video mode
static int viddisp = 1; //enable video display 0=off 1=on
static int vidreg[8] = { 0 }; //registers 0-7
static int rowdiv = 40; //text mode

SDL_Surface *g_tms9128nl_overlay = NULL;	// video overlay frame buffer
SDL_Color g_colors[256] = { { 0 } };	// color palette
Uint8 g_transparent_color = 0x7F;	// which color on the color palette is transparent (if any
unsigned char g_tms_pnt_addr = 0;	// pattern name table address
unsigned char g_tms_ct_addr = 0;	// color table address
unsigned char g_tms_pgt_addr = 0;	// pattern generation table address
unsigned char g_tms_sat_addr = 0;	// sprite attribute table address
unsigned char g_tms_sgt_addr = 0;	// sprite generation table address
unsigned char g_tms_foreground_color = 0xF;	// white (3 bit)
unsigned char g_tms_background_color = 0;	// black (3 bit)

bool g_tms_interrupt_enabled = false;	// whether NMI is on or off
unsigned char g_tms9128nl_needs_update = 0;		// whether screen needs to be updated

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

bool tms9128nl_int_enabled()
{
	return(g_tms_interrupt_enabled);
}

void tms9128nl_writechar(unsigned char value)
//write a character on the screen, using 40*24 mode emulation
// this is the same as writing to register #0 on a TMS chip
{
    unsigned int row = 0,col = 0;
    int base = 0;

    if (viddisp==0) return; //video display is off
    //if (vidmode!=1) return; //only text mode 1 for now

	// MODE 1 (bitmapped text)
	if (vidmode == 1)
	{
		if ((wvidindex >= 0) && (wvidindex <= 0x3c0))
		{
			base=0;
			rowdiv=40; //40*24

			row = (wvidindex-base-1) / rowdiv;
			col = (wvidindex-base-1) % rowdiv;

		    if ((col==31) && (rowdiv==32)) return; //problems with col31? or bug in fancyclearscreen routine?

		    tms9128nl_drawchar(value, col, row);
		}
	} // end mode 1

	// MODE 2 (Cliff Hanger graphical logo)
	else if (vidmode == 2)
	{
		if ((wvidindex>=0x3c00) && (wvidindex<=0x3f00)) //pattern name table - graphics mode
		{
			base=0x3c01;
			rowdiv=32; //32*24

			row = (wvidindex-base-1) / rowdiv;
			col = (wvidindex-base-1) % rowdiv;

		    if ((col==31) && (rowdiv==32)) return; //problems with col31? or bug in fancyclearscreen routine?

		    tms9128nl_drawchar(value, col, row);
		}
	}

} //end writechar

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
unsigned char tms9128nl_getvidmem(void)
{
    return vidmem[rvidindex++];
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
void tms9128nl_setvidindex(unsigned char value)
//port 54 is the index into video memory. First a low byte is sent, then a highbyte
// this is the same as writing to register #1 for a TMS chip (the same rules apply)
{
    static int tempindex = 0;
	char s[81] = { 0 };	// just a temp string

    if (toggleflag==0)
    {
        wvidindex=0;
        rvidindex=0;
        lowbyte=value;
        //    printf("lowbyte: %d\n",lowbyte);
    }
    else
    {
        highbyte=value;

		// if bit 7 is set, it means a TMS register is being written to
        if (highbyte & 0x80)
        {
			int which_reg = highbyte & 0x7F;	// the register that is being written to

            vidreg[which_reg] = lowbyte; // set the register in memory in case we need to refer to it later

			// handle register access
            switch (which_reg)
            {
				// register 0 (select screen mode 2 or Black & White)
            case 0:
				// if only bit 1 is set, it means to go to graphics mode 2
                if (lowbyte & 2)
				{
					printline("TMS: mode 2");
                    vidmode = 2;
				}

				// if the B&W bit is set, flag an error
				else if (lowbyte & 1)
				{
//					printline("TMS: B&W bit set, unsupported");
				}
                break;

				// register 1 (4/16K, blank screen, GINT, mode1, mode3, sprite mode, sprite scale)
            case 1:
				// if bit 0 is set, MAG
				if (lowbyte & 1)
				{
//					printline("TMS: double sprite size not supported");
				}

				// if bit 1 is set, 16x16 sprites is what we use, else 8x8
				if (lowbyte & 2)
				{
//					printline("TMS: 16x16 sprites requested");
				}
				else
				{
//					printline("TMS: 8x8 sprites requested");
					// don't print anything since this is what we expect
				}


				// bit 3 set == mode 3
                if (lowbyte & 0x08)
				{
					printline("TMS: mode 3");
    				vidmode=3;
				}
				// bit 4 set == mode 1
        		else if (lowbyte & 0x10) //bit 4 set - text mode 1
				{
					printline("TMS: mode 1");
					vidmode=1;
				}
				// if all vid bits have been cleared, then set vidmode to 0
				// we need to check this somewhere, here is as good a place as any I suppose
		        else if (!(vidreg[0]&0x02)) //no vid bits set
				{
					printline("TMS: vidmode 0 special");
		            vidmode=0;
				}

				// bit 5 set == generate interrupts
				if (lowbyte & 0x20)
				{
//					printline("TMS: Generate interrupts enabled");
					// don't print anything since this is what we expect
					g_tms_interrupt_enabled = 1;
				}
				else
				{
//					printline("TMS: Generate interrupts disabled");
					g_tms_interrupt_enabled = 0;
				}

				// don't blank screen
                if (lowbyte & 0x40)
				{
                    viddisp=1; // enable video display
				}

				// blank screen, leaving only backdrop
                else
				{
					int i = 0;

                    viddisp=0; // disable video display (blank screen)

					// erase all pixels on overlay screen and update screen
					for (i = 0; i < TMS9128NL_OVERLAY_W * TMS9128NL_OVERLAY_H; i++)
					{
						*((Uint8 *) g_tms9128nl_overlay->pixels + i) = 0x0;
					}
					g_tms9128nl_needs_update = 1;
//					tms9128nl_update_overlay();

//					printline("TMS: VIDEO DISPLAY TO BE BLANKED!");
				}

				// bit 7 set == select 16K ram, else 4K ram
				if (lowbyte & 0x80)
				{
//					printline("TMS: 16K ram selected");
					// don't print anything since this is what we expect
				}
				else
				{
//					printline("TMS: 4k ram selected");
				}
                break;

				// register #2 sets address for pattern name table
			case 2:
				g_tms_pnt_addr = (unsigned char) (lowbyte & 0xF); // only lowest 4 bits
				sprintf(s, "TMS: Pattern Name Table Address set to %x", g_tms_pnt_addr);
//				printline(s);
				break;
				// register #3 sets address for color table
			case 3:
				g_tms_ct_addr = lowbyte;
				sprintf(s, "TMS: Color Table Address set to %x", g_tms_ct_addr);
//				printline(s);
				break;

				// register #4 sets address for pattern generation table
			case 4:
				g_tms_pgt_addr = (unsigned char) (lowbyte & 7); // only lowest 3 bits
				sprintf(s, "TMS: Pattern Generation Table set to %x", g_tms_pgt_addr);
//				printline(s);
				break;

				// register #5 sets address for sprite attribute table
			case 5:
				g_tms_sat_addr = (unsigned char) (lowbyte & 0x7F); // discard high bit
				sprintf(s, "TMS: Sprite Attribute Table address set to %x", g_tms_sat_addr);
//				printline(s);
				break;
				
				// register #6 sets address for sprite generation table
			case 6:
				g_tms_sgt_addr = (unsigned char) (lowbyte & 0x7);	// only lowest 3 bits
				sprintf(s, "TMS: Sprite Generator Table address set to %x", g_tms_sgt_addr);
//				printline(s);
				break;

				// register #7 sets foreground and background colors
			case 7:
				g_tms_foreground_color = (unsigned char) ((lowbyte & 0xF0) >> 4);
				g_tms_background_color = (unsigned char) (lowbyte & 0x0F);

				// if the foreground color is transparent, make a note of it
				if (g_tms_foreground_color == 0)
				{
					g_transparent_color = 0xFF;
				}
				// if the background color is transparent, make a note of it
				else if (g_tms_background_color == 0)
				{
					g_transparent_color = 0;
				}
				// otherwise set it to a value that never gets used
				else
				{
					g_transparent_color = 0x7F;
				}

				sprintf(s, "TMS: Foreground color set to %x, background set to %x",
					g_tms_foreground_color, g_tms_background_color);

				{
					SDL_Color fore, back;

					tms9128nl_convert_color(g_tms_foreground_color, &fore);
					tms9128nl_convert_color(g_tms_background_color, &back);

					SDL_SetColors(g_tms9128nl_overlay, &back, 0, 1);		// set new background
					SDL_SetColors(g_tms9128nl_overlay, &fore, 255, 256); // set new foreground

					g_colors[0] = back;
					g_colors[255] = fore;	// update color palette with new colors

					g_tms9128nl_needs_update = 1;
					tms9128nl_update_overlay();	// redraw
				}

//				printline(s);
				break;

			default:
				sprintf(s,"TMS: Register %d was written to (unsupported)", which_reg);
//				printline(s);
				break;
            } //end switch
        } //end if

        else
        { //we're writing to the screen
            highbyte=(unsigned char) (highbyte&0x3f); //strip top 2 bits
            tempindex=(highbyte*256)+lowbyte;
            wvidindex=tempindex;
            rvidindex=tempindex;
        } //endif not highbyte
    } //endif toggleflag
    toggleflag=toggleflag^1; //flip bit 1
    //printf("WVidindex: %d\n",wvidindex);
    //printf("RVidindex: %d\n",rvidindex);
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

// writes a byte to the video memory
// returns 1 if that value is different from what was already there (ie, we need to update the screen)
int tms9128nl_setvidmem(unsigned char Value)
//return a 1 if the screen needs updating
{
	int result = 0;

	// if we find a value that wasn't already there
    if (vidmem[wvidindex] != Value)
    {
        vidmem[wvidindex] = Value;
        result = 1;
    }

	wvidindex++;	// the index always advances when we write
	return(result);

}








void tms9128nl_convert_color(unsigned char color_src, SDL_Color *color)
{

	char s[81] = { 0 };

	switch(color_src)
	{
	case 0: // transparent
		color->r = 31;
		color->g = 31;
		color->b = 31;
		break;
	case 1: // black
		color->r = 0;
		color->g = 0;
		color->b = 0;
		break;
	case 2: // medium green;
		color->r = 36;
		color->g = 219;
		color->b = 36;
		break;
	case 3: // light green
		color->r = 109;
		color->g = 255;
		color->b = 109;
		break;
	case 4: // dark blue
		color->r = 36;
		color->g = 36;
		color->b = 255;
		break;
	case 5: // light blue
		color->r = 73;
		color->g = 109;
		color->b = 255;
		break;
	case 6:	// dark red
//		color->r = 182;
//		color->g = 36;
//		color->b = 36;

		// these colors aren't correct for the TMS chip but they make Cliffy look a bit more
		// authentic (the move prompts are more purplish than red)
		color->r = 125;
		color->g = 0;
		color->b = 50;
		break;
	case 7: // light cyan
		color->r = 73;
		color->g = 219;
		color->b = 255;
		break;
	case 8:	// medium red
		color->r = 255;
		color->g = 36;
		color->b = 36;
		break;
	case 9: // light red
		color->r = 255;
		color->g = 109;
		color->b = 109;
		break;
	case 10:	// dark yellow
		color->r = 219;
		color->g = 219;
		color->b = 36;
		break;
	case 11: // light yellow
		color->r = 219;
		color->g = 219;
		color->b = 146;
		break;
	case 12: // dark green
		color->r = 36;
		color->g = 146;
		color->b = 36;
		break;
	case 13:	// magenta
		color->r = 219;
		color->g = 73;
		color->b = 182;
		break;
	case 14: // grey
		color->r = 182;
		color->g = 182;
		color->b = 182;
		break;
	case 15: // white
		color->r = 255;
		color->g = 255;
		color->b = 255;
		break;
	default:
		sprintf(s, "UNSUPPORTED COLOR passed into convert color : %d", color_src);
		printline(s);
		break;
	}

}

// draws a character to the Cliff video display
// 40 columns by 24 rows
// uses Cliffy's video memory to retrieve 8x8 character bitmap
void tms9128nl_drawchar(unsigned char ch, int col, int row)
{

	const int char_width = 8;

	int bmp_index = (ch * 8) + (g_tms_pgt_addr << 11); // index in vid mem where bmp data is located
//	int bmp_index = (ch * 8) + 0x3800;
	int i = 0, j = 0;	// temp indices
	int x = col * char_width;
	int y = row * 8;
	unsigned char line = 0;
	SDL_Color foreground, background;

	// colors are special in vidmode 2, we don't handle them properly yet
	if (vidmode == 2)
	{
		// Add the line below to get custom chars for the CH Logo 
		if (row <= 11)
		{
			bmp_index = (ch * 8) + 8*256 * (row > 7); 
		}
		else
		{
			bmp_index = (ch * 8) + 0x3800; 
		}

		g_tms_foreground_color = 0xF;	// temporary hack for now
		g_tms_background_color = 4;
		tms9128nl_convert_color(g_tms_foreground_color, &foreground);
		tms9128nl_convert_color(g_tms_background_color, &background);
		SDL_SetColors(g_tms9128nl_overlay, &background, 0, 1);		// set new background
		SDL_SetColors(g_tms9128nl_overlay, &foreground, 255, 256); // set new foreground
	}
	else
	{
		tms9128nl_convert_color(g_tms_foreground_color, &foreground);
		tms9128nl_convert_color(g_tms_background_color, &background);
	}

	// draw each line of character into new surface
	for (i = 0; i < 8; i++)
	{
		line = vidmem[bmp_index + i];	// get a line

		// handle each pixel across
		for (j = char_width - 1; j >= 0; j--)
		{
			// if rightmost bit is 1, it means draw the pixel
			if (line & 1)
			{
				*((Uint8 *) g_tms9128nl_overlay->pixels + ((y+i) * TMS9128NL_OVERLAY_W) + (x+j)) = 0xFF;
			}
			// else don't draw the pixel
			else
			{
				*((Uint8 *) g_tms9128nl_overlay->pixels + ((y+i) * TMS9128NL_OVERLAY_W) + (x+j)) = 0x0;
			}
			line = (unsigned char) (line >> 1);
		}
	} // end for loop

//	tms9128nl_update_overlay();
	g_tms9128nl_needs_update = 1;

}

void tms9128nl_outcommand(char *s,int col,int row)
{
    SDL_Rect dest;
    //char s[255] = { 0 };

    dest.x = (short) ((col*6) + 200);
    dest.y = (short) ((row*13) + 100);
    dest.w = (unsigned short) (6 * strlen(s)); // width of rectangle area to draw (width of font * length of string)
    dest.h = 13;	// height of area (height of font)

    
    // smpeg freaks out if it's not the only thing drawing to the screen
    if (g_ldp->is_blitting_allowed())
    {
	    SDL_FillRect(get_screen(), &dest, 0); // erase anything at our destination before we print new text
	    DrawText(s, get_screen(), FONT_SMALL, dest.x, dest.y);
		SDL_UpdateRects(get_screen(), 1, &dest);	
    }
}




// initializes cliff "scoreboard" or whatever
bool tms9128nl_display_init()
{

	bool result = false;
	int i = 0;

	// fill color palette with schlop because we only use colors 0 and 0xFF for now
	for (i = 0; i < 256; i++)
	{
		g_colors[i].r = (unsigned char) i;
		g_colors[i].g = (unsigned char) i;
		g_colors[i].b = (unsigned char) i;
	}

	g_tms9128nl_overlay = SDL_CreateRGBSurface(SDL_SWSURFACE, TMS9128NL_OVERLAY_W, TMS9128NL_OVERLAY_H, 8, 0, 0, 0, 0); // create an 8-bit surface

	if (g_tms9128nl_overlay)
	{
		if (SDL_SetColors(g_tms9128nl_overlay, g_colors, 0, 256))
		{
			result = true;
		}
	}
	else
	{
		printline("TMS9128NL : Could not create RGB surface for TMS video!");
	}

	return result;

}

// shuts down cliff "scoreboard"
void tms9128nl_display_shutdown()
{
	SDL_FreeSurface(g_tms9128nl_overlay);
}

// redraws cliff "scoreboard" after screen has been erased to black
void tms9128nl_display_repaint()
{

//	draw_othergfx(B_tms9128nl_LOGO, 100, 10); // just to show this is independent of lair

}

// blits cliff overlay onto screen
void tms9128nl_update_overlay()
{

	SDL_Rect dest;

	if (g_tms9128nl_needs_update)
	{
	  dest.x = 0;
	  dest.y = 0;	        
	  dest.w = (unsigned short) g_tms9128nl_overlay->w;
	  dest.h = (unsigned short) g_tms9128nl_overlay->h;

		// SMPEG doesn't allow blitting, which is why we have to check
		if (g_ldp->is_blitting_allowed())
		{

		  SDL_BlitSurface(g_tms9128nl_overlay, NULL, get_screen(), &dest);
		  SDL_Flip(get_screen());
		}
	}
	g_tms9128nl_needs_update = 0;

}

SDL_Surface *tms9128nl_get_overlay()
{
	return g_tms9128nl_overlay;
}

Uint8 tms9128nl_get_transparent_color()
{
	return g_transparent_color;
}

SDL_Color *tms9128nl_get_colors()
{
	return g_colors;
}
