/*
 * vldp.h
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of VLDP, a virtual laserdisc player.
 *
 * VLDP is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * VLDP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// API for the mpeg2 virtual laserdisc player--
// by Matt Ownby, July 24th, 2001

#ifdef __cplusplus
extern "C" {
#endif

#include <SDL.h>

enum { STAT_ERROR, STAT_BUSY, STAT_STOPPED, STAT_PLAYING, STAT_PAUSED };

// initializes the player, passing in the SDL surface that the video will be
// displayed on.  Returns 1 on success, or 0 on failure.
// Once the player has been initialized, nothing else should write to the
// SDL_Surface (absolutely nothing! hehe).
int vldp_init(SDL_Surface *screen);

// shuts down the player, de-allocating any memory that was allocated
void vldp_shutdown();

// Sets an optional callback which is allowed to modify the YUV image before
//  it is blitted to the screen.
// This is useful for drawing video overlay on top of the YUV image for
//  laserdisc games that support it.
// Setting the callback pointer to NULL disables it.
void vldp_set_yuv_callback(void (*callback)(SDL_Overlay *cur_overlay));

// the audio callback you need to hook into with SDL audio or SDL mixer
void vldp_audio_callback(void *unused, Uint8 *stream, int len);

// opens (and parses if necessary) an mpeg1 or mpeg2 file specified by
// 'filename'.  I haven't decided it this file will be a system stream or just
// a video stream.  Returns 1 on success or 0 on failure.
// This also returns immediately, but you have to wait for the current status to change to READY
// before you issue any other commands.  The only command you can issue if the status is not READY is the
// shutdown command.
int vldp_open_mpeg(char *filename);

// returns the framerate that the mpeg is supposed to run at (NOT its current actual framerate!)
// This is useful for emulating 23.976 fps discs using 29.97 fps mpegs (or vice versa)
// NOTE : The mpeg must be opened for the result to mean anything.  Otherwise, the result is random.
double vldp_get_fps();

// returns 0 if the video stream uses frames or non-zero if it uses fields
// NOTE : the mpeg must be opened for the result to mean anything
Uint8 vldp_uses_fields();

// returns the current VLDP status.  You should only really need to use this after issuing an open command.
// If you have to use this at any other time, perhaps VLDP ought to be redesigned.
int vldp_get_status();

// begins playing the mpeg.  Function returns the instant that the mpeg begins
// playing and not before.
int vldp_play();

// searches to, and _renders_ an mpeg frame (imitating the behavior of a laser
// disc player).  The first frame is frame 0 (the first frame on a laserdisc
// is frame 1).
// You can search to I, P or B frames.  The search is exact,
// not approximate.  It also should be VERY FAST (ie, search completes in < 1 second).  This is the one feature
// that no other mpeg player I have encountered has.
// returns 1 when the search is complete and the frame has been displayed or
// 0 if there was a problem.
int vldp_search(Uint16 frame);

// pauses the playback of the video, showing a still frame.  If the video
// is already paused, this does nothing.
int vldp_pause();

// similar to pause except this function also steps forward 1 frame
// Notice there is no step_backward.  This is to alert the user the stepping backward is expensive and if you must use do it, use search instead.
// returns 1 on success
int vldp_step_forward();

// mimics the behavior of a laserdisc player by drawing a black screen,
// and resetting the current frame to the first frame.
int vldp_stop();

// returns the current mpeg frame # that is being displayed on the screen
// The first frame is frame # 0
Uint16 vldp_get_current_frame();

// Enables audio in either the left (0) or right (1) channel.  If audio is
// already enabled, this does nothing.
int vldp_enable_audio(int channel);

// Disables audio in either the left (0) or right (1) channel. If audio is
// already disabled, this does nothing.
int vldp_disable_audio(int channel);

#ifdef __cplusplus
}
#endif
