// main.c

#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>
#include "include/vldp.h"

#ifdef WIN32
#define MPEG_NAME "d:/mpeg2/dl_640_hq_10k.m2v"
#else
#define MPEG_NAME "/dosf/mpeg2/dldvdrom/dls00.vob.m2v"
#endif

typedef int (*initproc)(SDL_Surface *);
typedef int (*openproc)(char *);
typedef int (*intproc)();
typedef int (*searchproc)(Uint16);
typedef Uint16 (*frameproc)();
typedef void (*audiohook)(void *, Uint8 *, int);

initproc pvldp_init;
openproc pvldp_open_mpeg;
searchproc pvldp_search;
intproc pvldp_play, pvldp_pause, pvldp_shutdown, pvldp_step_forward;
frameproc pvldp_get_current_frame;
audiohook pvldp_audio_callback;

#ifdef WIN32
#include <windows.h>

HINSTANCE g_dll_instance;

// loads the DLL
int load_dll()
{
	int result = 0;

    // Get a handle to the DLL module.

    g_dll_instance = LoadLibrary("vldp"); 

    // If the handle is valid, try to get the function address. 
    if (g_dll_instance != NULL) 
    {
		pvldp_init = (initproc) GetProcAddress(g_dll_instance, "vldp_init");
		pvldp_open_mpeg = (openproc) GetProcAddress(g_dll_instance, "vldp_open_mpeg");
		pvldp_search = (searchproc) GetProcAddress(g_dll_instance, "vldp_search");
		pvldp_play = (intproc) GetProcAddress(g_dll_instance, "vldp_play");
		pvldp_pause = (intproc) GetProcAddress(g_dll_instance, "vldp_pause");
		pvldp_shutdown = (intproc) GetProcAddress(g_dll_instance, "vldp_shutdown");
		result = 1;
	}

	return result;
}

void free_dll()
{
	if (g_dll_instance)
	{
		FreeLibrary(g_dll_instance);
	}
}
#endif

/////////////////////////////////

#ifdef LINUX

#include <dlfcn.h>

void *g_dll_instance = NULL;

// loads the DLL
int load_dll()
{
	int result = 0;
	
    // Get a handle to the DLL module.

	g_dll_instance = dlopen("libvldp.so", RTLD_NOW);

    // If the handle is valid, try to get the function address. 
    if (g_dll_instance != NULL) 
    {
		pvldp_init = (initproc) dlsym(g_dll_instance, "vldp_init");
		pvldp_open_mpeg = (openproc) dlsym(g_dll_instance, "vldp_open_mpeg");
		pvldp_search = (searchproc) dlsym(g_dll_instance, "vldp_search");
		pvldp_play = (intproc) dlsym(g_dll_instance, "vldp_play");
		pvldp_pause = (intproc) dlsym(g_dll_instance, "vldp_pause");
		pvldp_shutdown = (intproc) dlsym(g_dll_instance, "vldp_shutdown");
		pvldp_get_current_frame = (frameproc) dlsym(g_dll_instance, "vldp_get_current_frame");
		pvldp_step_forward = (intproc) dlsym(g_dll_instance, "vldp_step_forward");
		if (pvldp_init && pvldp_open_mpeg && pvldp_search && pvldp_play && pvldp_pause && pvldp_shutdown && pvldp_get_current_frame && pvldp_step_forward)
		{
			result = 1;
		}
		else
		{
			printf("Could not load all the syms!\n");
		}
	}
	else
	{
		printf("ERROR : %s\n", dlerror());
	}

	return result;
}

void free_dll()
{
	if (g_dll_instance)
	{
		printf("Freeing .SO library ...\n");
		dlclose(g_dll_instance);
	}
}
#endif


//////////////////////////////////

void my_callback(SDL_Overlay *yuv)
{
	Uint8 *ptr = NULL;
	
	SDL_LockYUVOverlay(yuv);
	if (yuv == 0)
	{
		ptr = yuv->pixels[0];	// get Y
		*ptr++ = 0xFF;
		*ptr++ = 0xFF;
		*ptr++ = 0xFF;
		*ptr++ = 0xFF;
	}
	SDL_UnlockYUVOverlay(yuv);
}

int main(int argc, char **argv)
{
	SDL_Surface *surface = NULL;
	const SDL_VideoInfo *vidinfo = NULL;
	Uint8 suggested_bpp = 0;
	
	if (!load_dll())
	{
		fprintf(stderr, "load_dll failed, shutting down!\n");
		exit(1);
	}

#ifdef LINUX
    setenv("SDL_VIDEO_YUV_HWACCEL", "1", 1);
//	setenv("MPEG2_DEBUG", "1", 1);
//    setenv("SDL_VIDEO_X11_NODIRECTCOLOR", "1", 1);
#endif

    if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_NOPARACHUTE))
	{
        fprintf (stderr, "sdl video/audio initialization failed.\n");
        return 1;
    }

	vidinfo = SDL_GetVideoInfo();
	suggested_bpp = vidinfo->vfmt->BitsPerPixel;

	surface = SDL_SetVideoMode(640, 480, suggested_bpp, SDL_SWSURFACE);
		
	SDL_ShowCursor(SDL_DISABLE);	// hide mouse

	(pvldp_init) (surface);

	if ((pvldp_open_mpeg)(MPEG_NAME))
	{
		(pvldp_play)();
		getchar();
		(pvldp_search)(0);
		getchar();
		(pvldp_play)();
		getchar();
	}
	else
	{
		printf("Open mpeg failed!\n");
	}

	(pvldp_shutdown)();

	// free the SDL surface and shut down
    SDL_FreeSurface (surface);

	SDL_QuitSubSystem(SDL_INIT_VIDEO | SDL_INIT_AUDIO);

    SDL_Quit();

	free_dll();

	return 0;
}
