/*
 * vldp.c
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of VLDP, a virtual laserdisc player.
 *
 * VLDP is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * VLDP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// see vldp.h for documentation on how to use the API

#include <stdio.h>
#include <string.h>
//#include <inttypes.h>
#include "vldp.h"
#include "vldp_common.h"

/////////////////////////////////////////////////////////////////////

#define STRSIZE 81

//////////////////////////////////////////////////////////////////////////////////////

int vldp_wait_for_status(int stat);

//////////////////////////////////////////////////////////////////////////////////////

SDL_mutex *g_vldp_mutex;	// we lock this mutex before reading/writing
				// any of these global variables.  Very
				// important!

SDL_Thread *private_thread = NULL;

int g_searched_yet = 0;	// have we issued any search commands yet?

int p_initialized = 0;	// whether VLDP has been initialized

int g_req_quit = 0;	// if this is set, we quit graciously or ungraciously, it doesn't matter ...
int g_ack_quit = 0;	// indicates private thread has responded and is shutting down

int g_req_open = 0;
int g_ack_open = 0;

char g_req_file[STRSIZE];	// requested mpeg filename
//char g_ack_file[STRSIZE];	// actual mpeg filename

// req yuv callback and actual yuv callback here later

int g_req_search = 0;		// requests mpeg to search to frame specified
int g_ack_search = 0;		// acknowledgement of search request
Uint16 g_req_frame = 0;		// requested frame to search to
Uint16 g_ack_frame = 0;		// actual current frame

int g_req_play = 0;		// requests mpeg status to change to playing
int g_ack_play = 0;
int g_req_pause = 0;		// requests mpeg status to change to paused
int g_ack_pause = 0;
int g_req_step_forward = 0;	// optional request to step forward 1 frame.. pause must also also be requested for this to get noticed!
int g_ack_step_forward = 0;
int g_req_stop = 0;		// requests mpeg status to change to stopped
int g_ack_stop = 0;

int g_ack_status = 0;		// actual mpeg playing status (see enums)

double g_fps = 0.0;	// the framerate of the mpeg
Uint8 g_uses_fields = 0;	// does the video stream use fields instead of frames?

void (*g_yuv_callback)(SDL_Overlay *cur_overlay) = NULL;	// an optional callback that gets called right before we blit
	// this allows the user to add things to the YUV overlay (such as video overlay)

/////////////////////////////////////////////////////////////////////

int vldp_init(SDL_Surface *screen)
{
	p_initialized = 0;

	g_vldp_mutex = SDL_CreateMutex();
	if (g_vldp_mutex)
	{
		private_thread = SDL_CreateThread(idle_handler, screen);	// start out handler going
		if (private_thread)
		{
			p_initialized = 1;
		}
		else
		{
			SDL_DestroyMutex(g_vldp_mutex);
		}
	}

	return p_initialized;
}

void vldp_shutdown()
{
	// only shutdown if we have previous initialized
	if (p_initialized)
	{
		VLDP_REQUEST(quit);
		VLDP_WAIT_FOR_ACK(quit);
		SDL_WaitThread(private_thread, NULL);	// wait for private thread to terminate
		SDL_DestroyMutex(g_vldp_mutex);
	}
	p_initialized = 0;
}


void vldp_set_yuv_callback(void (*callback)(SDL_Overlay *cur_overlay))
{
	if (p_initialized)
	{
		VLDP_LOCK;
		g_yuv_callback = callback;
		VLDP_UNLOCK;
	}
}

// requests that we open an mpeg file
int vldp_open_mpeg(char *filename)
{
  int result = 2;	// assume we're busy so the loop below works the first time
  FILE *F = NULL;
	
  if (p_initialized)
  {
	F = fopen(filename, "rb");
	// if file exists, we can open it
	if (F)
	{
		fclose(F);
		VLDP_LOCK;
		strcpy(g_req_file, filename);
		g_req_open = 1;
		VLDP_UNLOCK;

		VLDP_WAIT_FOR_ACK(open);
		
		// loop while the mpeg is parsing
		while (result == 2)
		{
			result = vldp_wait_for_status(STAT_STOPPED);	// wait until we get stopped or an error
			
			// if we're busy
/*			if (result == 2)
			{
				printf("VLDP OPEN : Still busy opening m2v file, so we're waiting again (endlessly)\n");
			}	*/
		}
	}
  }
	return result;

}

double vldp_get_fps()
{
	double result = 0.0;

	if (p_initialized)
	{
		VLDP_LOCK;
		result = g_fps;
		VLDP_UNLOCK;
	}
	return result;
}

Uint8 vldp_uses_fields()
{
	Uint8 result = 0;

	if (p_initialized)
	{
		VLDP_LOCK;
		result = g_uses_fields;
		VLDP_UNLOCK;
	}
	return result;
}

int vldp_play()
{
	int result = 0;
		
	if (p_initialized)
	{
		// if we are playing without searching, seek to frame 0 to avoid the 'green effect'
		if (!g_searched_yet)
		{
			vldp_search(0);
		}
		VLDP_REQUEST(play);
		VLDP_WAIT_FOR_ACK(play);
		result = vldp_wait_for_status(STAT_PLAYING);
	}
	return(result);
}

int vldp_search(Uint16 frame)
{
	int result = 0;

	if (p_initialized)
	{
		VLDP_LOCK;
		g_searched_yet = 1;
		g_req_frame = frame;
		g_req_search = 1;
		VLDP_UNLOCK;
		VLDP_WAIT_FOR_ACK(search);
		result = vldp_wait_for_status(STAT_PAUSED);	// wait until search is complete
	}
	return result;
}

int vldp_pause()
{
	int result = 0;
	if (p_initialized)
	{
		VLDP_REQUEST(pause);
		VLDP_WAIT_FOR_ACK(pause);
		result = vldp_wait_for_status(STAT_PAUSED);
	}
	return result;
}

int vldp_step_forward()
{
	int result = 0;
		
	if (p_initialized)
	{
		VLDP_REQUEST(step_forward);
		VLDP_REQUEST(pause);	// must must always be requested along with step forward for the private thread to take notice
		VLDP_WAIT_FOR_ACK(step_forward);
		VLDP_WAIT_FOR_ACK(pause);
		result = vldp_wait_for_status(STAT_PAUSED);
	}
	return result;
}

int vldp_stop()
{	
	return 0;	// this is a useless function which I haven't finished yet anyway, so don't use it :)
	/*
	int result = 0;
	VLDP_REQUEST(stop);
	VLDP_WAIT_FOR_ACK(stop);
	vldp_wait_for_status(STAT_STOPPED);
	return result;
	*/
}

Uint16 vldp_get_current_frame()
{
	Uint16 result = 0;

	if (p_initialized)
	{
		VLDP_LOCK;
		result = g_ack_frame;
		VLDP_UNLOCK;
	}

	return result;
}

///////////////////////////////////////////////////////////////////////////

// waits until the disc status is 'stat'
// if the stat we want is received, return 1 (success)
// if we time out, or if we get an error stat, return 0 (error)
// if we are legitimately busy, we return 2 (busy)
int vldp_wait_for_status(int stat)
{
	int result = 0;	// assume error unless we explicitly
	int done = 0;
	Uint32 cur_time = SDL_GetTicks();
	while (!done && (SDL_GetTicks() - cur_time < VLDP_TIMEOUT))
	{
		SDL_Delay(1);
		VLDP_LOCK;
		if (g_ack_status == stat)
		{
			done = 1;
			result = 1;
		}
		else if (g_ack_status == STAT_ERROR)
		{
			done = 1;
		}		
		// else keep waiting
		VLDP_UNLOCK;
	}

	// if we timed out but are busy, indicate that
	if (g_ack_status == STAT_BUSY)
	{
		result = 2;
	}
	
	// else if we timed out
	else if (SDL_GetTicks() - cur_time >= VLDP_TIMEOUT)
	{
		fprintf(stderr, "VLDP ERROR!!!!  Timed out with getting our expected response!\n");
	}

	return result;
}
