/*
 * vldp_internal.c
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of VLDP, a virtual laserdisc player.
 *
 * VLDP is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * VLDP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// contains the function for the VLDP private thread

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "inttypesreplace.h"
#include <SDL.h>

#include "vldp.h"
#include "vldp_common.h"
#include "vldp_internal.h"
#include "mpegscan.h"

#include "video_out.h"
#include "video_out_sdl.h"
#include "mpeg2.h"
#include "mpeg2_internal.h"
#include "mm_accel.h"

/////////////////////////////

// NOTICE : these variables should only be used by the private thread !!!!!!!!!!!!

int s_paused = 0;	// whether the video is to be paused
int s_step_forward = 0;	// whether to step 1 frame forward
int s_blanked = 0;	// whether the mpeg video is to be blanked
int s_frames_to_skip = 0;	// how many frames to skip before rendering the next frame (used for P and B frames seeking)
int s_skip_all = 0;	// if we are to skip any frame to be displayed (to avoid seeing frames we shouldn't)
int s_overlay_allocated = 0;	// whether the SDL overlays have been allocated

Uint32 s_timer = 0;	// FPS timer used by the blitting code to run at the right speed
Uint32 s_elapsed_frames = 0;	// how many frames have elapsed since we reset the counter

SDL_Overlay *s_hw_overlay = NULL;	// hardware overlay used to display YUV data to screen

#define MAX_LDP_FRAMES 60000

FILE *g_mpeg_handle = NULL;	// mpeg file we currently have open
static mpeg2dec_t g_mpeg_data;	// structure to hold our mpeg2 data
vo_instance_t *s_video_output = NULL;
SDL_Surface *s_surface = NULL;	// surface we are blitting to
Uint32 g_frame_position[MAX_LDP_FRAMES] = { 0 };	// the file position of each I frame
Uint16 g_totalframes = 0;	// total # of frames in the current mpeg

#define BUFFER_SIZE 262144
static Uint8 g_buffer[BUFFER_SIZE];	// buffer to hold mpeg2 file as we read it in

////////////////////////////////////////////////

// this is our video thread which gets called
int idle_handler(void *surface)
{
	int done = 0;

	s_surface = surface;
	s_video_output = vo_sdl_open((SDL_Surface *) surface);	// setup SDL YUV overlay
	if (s_video_output)
	{
		mpeg2_init (&g_mpeg_data, mm_accel(), s_video_output);	// setup libmpeg2 decoder
	}
	else
	{
		fprintf(stderr, "VLDP : Error opening SDL for video output!\n");
		done = 1;
	}

	// unless we are drawing video to the screen, we just sit here
	// and listen for orders from the parent thread
	while (!done)
	{
		SDL_Delay(1);	// don't hog CPU while idling
		VLDP_LOCK;
		if (g_req_quit)
		{
			done = 1;
		}
		
		if (g_req_open)
		{
			idle_handler_open();
			idle_handler_reset();
		}
		else if (g_req_play)
		{
			// shouldn't need to reset since we are not changing position in the mpeg file
			idle_handler_play();
		}
		else if (g_req_search)
		{
			idle_handler_reset();
			idle_handler_search();
		}
		else if (g_req_pause)
		{
			// shouldn't need to reset since we are not changing position in the mpeg file
			idle_handler_pause();
		}
		else if (g_req_stop)
		{
			idle_handler_reset();
			idle_handler_stop();
		}
		VLDP_UNLOCK;
	} // end while
			
	// if we have a file open, close it
	if (g_mpeg_handle)
	{
		fclose(g_mpeg_handle);
		g_mpeg_handle = 0;
	}

	g_ack_status = STAT_ERROR;
	mpeg2_close(&g_mpeg_data);	// shutdown libmpeg2
	vo_close(s_video_output);		// shutdown SDL YUV overlay

	VLDP_ACK(quit);

	return 0;

}

// the handler we call if we're paused... called from video_out_sdl
// globals are unlocked
void paused_handler()
{

	SDL_Delay(1);	// don't hog CPU while frame is paused
	VLDP_LOCK;

	g_ack_status = STAT_PAUSED;

	// if they've send a play request, then unpause and start playing
	if (g_req_play)
	{
		respond_req_play();
	}
	// if we get any of these commands, we have to skip the remaining buffered frames and reset
	else if (g_req_stop | g_req_quit | g_req_open | g_req_search)
	{
		s_skip_all = 1;
		// we don't acknowledge these requests here, we let the idle handler take care of them
	}
	// if they request a pause when we're already paused
	else if (g_req_pause)
	{
		// if they've also requested a step forward
		if (g_req_step_forward)
		{
			s_step_forward = 1;
			VLDP_ACK(step_forward);
		}
		else
		{			
			fprintf(stderr, "VLDP WARNING: redundant pause received, ignored\n");
		}
		VLDP_ACK(pause);
	}
	// else, keep pausing

	VLDP_UNLOCK;
	
}

// the handler we call if we're playing
// globals are unlocked
void play_handler()
{	
	VLDP_LOCK;
	
	// if we get a pause request while playing
	if (g_req_pause)
	{
		// if they've also requested a step forward
		if (g_req_step_forward)
		{
			s_step_forward = 1;
			VLDP_ACK(step_forward);
		}
		VLDP_ACK(pause);
		s_paused = 1;
		s_blanked = 0;
	}
	
	// if we are definitely playing
	else
	{
		// if someone makes a play request after we are already playing, just acknowledge it and do nothing else
		if (g_req_play)
		{
			VLDP_ACK(play);
			fprintf(stderr, "VLDP WARNING: Redundant play command received, ignoring\n");
		}
		else if (g_req_stop | g_req_quit | g_req_open | g_req_search)
		{
			s_skip_all = 1;	// bail out, handle these requests in the idle handler
		}
		
		// the moment we start playing, we need to change our status, and reset the FPS timer
		if (g_ack_status != STAT_PLAYING)
		{
			g_ack_status = STAT_PLAYING;
			s_timer = SDL_GetTicks();	// since we start playing now, we refresh the timer
			s_elapsed_frames = 0;
		}
	}
	VLDP_UNLOCK;
}

// resets libmpeg2 as if it were being run for the first time
// (needed when seeking within the file or opening a new file)
// assumes globals are locked
void idle_handler_reset()
{	
	// only reset if we have a file open
	if (g_mpeg_handle)
	{
	    g_mpeg_data.shift = 0xffffff00;
	    g_mpeg_data.is_sequence_needed = 1;
	    g_mpeg_data.drop_flag = 0;
	    g_mpeg_data.drop_frame = 0;
	    g_mpeg_data.in_slice = 0;
	    g_mpeg_data.chunk_ptr = g_mpeg_data.chunk_buffer;
	    g_mpeg_data.code = 0xb4;

	    memset (g_mpeg_data.picture, 0, sizeof (picture_t));

	    header_state_init (g_mpeg_data.picture);

		vldp_process_sequence_header();	// we need to process the sequence header before we can jump around the file for frames
	}
	else
	{
		fprintf(stderr, "VLDP ERROR : no mpeg file has been opened, can't reset\n");
	}
}

// assumes global variables are locked
void ivldp_set_framerate(Uint8 frame_rate_code)
{
	// now to compute the framerate

	switch (frame_rate_code)
	{
	case 1:
		g_fps = 23.976;
		break;
	case 2:
		g_fps = 24.0;
		break;
	case 3:
		g_fps = 25.0;
		break;
	case 4:
		g_fps = 29.97;
		break;
	case 5:
		g_fps = 30.0;
		break;
	case 6:
		g_fps = 50.0;
		break;
	case 7:
		g_fps = 59.94;
		break;
	case 8:
		g_fps = 60.0;
		break;
	default:
		// else we got an invalid frame rate code
		fprintf(stderr, "ERROR : Invalid frame rate code!\n");
		g_fps = 1.0;	// to avoid divide by 0 error
		break;
	} // end switch
}

// feeds mpeg2dec the beginning of the file up to the first Group of Picture
// It needs this info to get setup, and we probably don't want to start playing from the beginning of the file
// (assumes globals are locked)
void vldp_process_sequence_header()
{
	Uint32 val = 0;
	int index = 0;

#define BYTES_TO_READ 0x200

	rewind(g_mpeg_handle);	// start at the beginning
	fread(&g_buffer, 1, BYTES_TO_READ, g_mpeg_handle);	// assume that we must be the first frame in 0x200 bytes
		// if not, we'll have to increase the number
		
	// go until we have found the first frame or we run out of data
	while (val != 0x000001B8)
	{
		val = val << 8;
		val |= g_buffer[index];	// add newest byte to bottom of val
		index++;	// advance the end pointer
		if (index > BYTES_TO_READ)
		{
			fprintf(stderr, "VLDP : Could not find first frame in 0x%x bytes.  Modify source code to increase buffer!\n", BYTES_TO_READ);
			break;
		}
	}

	mpeg2_decode_data (&g_mpeg_data, g_buffer, g_buffer + index - 4);	// decode the sequence header

}

// opens a new mpeg2 file (assumes globals are locked and that file to open exists)
void idle_handler_open()
{
	VLDP_ACK(open);

	g_ack_status = STAT_ERROR;	// assume error unless we meet all conditions

	// if we have previously opened an mpeg, we need to close it and reset
	if (g_mpeg_handle)
	{
//		printf("OPEN: a previous file was open, so we're closing it\n");
		fclose(g_mpeg_handle);
		g_mpeg_handle = 0;
		vo_close(s_video_output);	// we have to de-allocate the SDL overlay because the new mpeg2 might have different dimensions
		// close doesn't really shut down SDL, it just de-allocates the SDL overlays and marks them as de-allocated
		// that means next time sdl_setup is called, they will be re-allocated.
	}

	// check to make sure it's a video stream and also get framerate
	g_mpeg_handle = fopen(g_req_file, "rb");
	if (g_mpeg_handle)
	{
		Uint8 small_buf[8];
		fread(&small_buf, 1, 8, g_mpeg_handle);	// 1st 8 bytes reveals much
		
		// if we find the proper mpeg2 video header at the beginning of the file
		if (((small_buf[0] << 24) | (small_buf[1] << 16) | (small_buf[2] << 8) | small_buf[3]) == 0x000001B3)
		{
			ivldp_set_framerate(small_buf[7] & 0xF);	// set the framerate
			fclose(g_mpeg_handle);	// close file since the next function opens it again

			g_ack_status = STAT_BUSY;	// assume that we might be parsing for a while, in which case we should be busy

			// load/parse all the frame locations in the file for super fast seeking
			if (ivldp_get_mpeg_frame_offsets(g_req_file))
			{
				g_mpeg_handle = fopen(g_req_file, "rb");	// open the new mpeg file
				g_ack_status = STAT_STOPPED;	// now that the file is open, we're ready to play
			}
			else
			{
				fprintf(stderr, "VLDP PARSE ERROR : Is the video stream damaged?\n");
				fclose(g_mpeg_handle);
				g_mpeg_handle = NULL;
				g_ack_status = STAT_ERROR;	// change from BUSY to ERROR
			}
		}
		
		// if the file had a bad header
		else
		{
			fprintf(stderr, "VLDP ERROR : Did not find expected header.  Is this mpeg stream demultiplexed??\n");
			fclose(g_mpeg_handle);
			g_mpeg_handle = NULL;
		}
	} // end if file exists
	else
	{
		fprintf(stderr, "ERROR : Could not open file!\n");
	}
	
}

// starts playing the mpeg from its current position
// assumes globals are locked
void idle_handler_play()
{
	respond_req_play();

	// when the frame is actually blitted is when we set the status to STAT_PLAYING

	VLDP_UNLOCK;	// notice, we were locked when we entered the function
	ivldp_render();
	VLDP_LOCK;	// we were locked when we came in, best to leave locked as well
}

// responds to play request
// assumes globals are locked
void respond_req_play()
{
	VLDP_ACK(play);
	g_ack_status = STAT_BUSY;	// it may take a few moments to start playing, so we're busy for now
	s_paused = 0;	// we to not want to pause on 1 frame
	s_blanked = 0;	// we want to see the video
	s_frames_to_skip = 0;	// skip no frames, just play from current position
	s_timer = SDL_GetTicks();	// we reset this mainly so we don't drop any frames when we begin playing
	s_elapsed_frames = 0;
		// this value is reset again as soon as we confirm that we are playing
}

// displays 1 or more frames to the screen, according to the state variables.
// This function can be used to do both still frames and moving video.  Play and search both use this function.
// assumes GLOBALS ARE UNLOCKED
void ivldp_render()
{
    Uint8 *end = NULL;
	int render_finished = 0;

	// check to make sure a file has been opened
	if (!g_mpeg_handle)
	{
		render_finished = 1;
		fprintf(stderr, "VLDP RENDER ERROR : we tried to render an mpeg but none was open!\n");
		g_ack_status = STAT_ERROR;
	}
		
    while (!render_finished)
    {
		end = g_buffer + fread (g_buffer, 1, BUFFER_SIZE, g_mpeg_handle);	// read chunk of video stream
		mpeg2_decode_data (&g_mpeg_data, g_buffer, end);	// display it to the screen

		VLDP_LOCK;
		
		render_finished = g_req_quit;	// I would've checked this in the while loop, but we weren't locked
		
		// if we've read to the end of the mpeg2 file, then we can't play anymore, so we pause on last frame
		if (end != g_buffer + BUFFER_SIZE)
		{
			printf("NOTE : Reached end of video file, pausing!\n");
			g_ack_status = STAT_PAUSED;
			render_finished = 1;
			s_skip_all = 0;	// in case we had to stop displaying frames, turn them back on now
			g_ack_frame = 0;
			rewind(g_mpeg_handle);	// go to beginning of file in case someone foolishly calls play again
			// NOTE : this may be an unoptimal solution
		}

		// else if we need to stop rendering due to any of these reasons, do so
		else if (g_req_open | g_req_search | g_req_stop | g_req_quit)
		{
			g_ack_status = STAT_BUSY;
			s_skip_all = 0;	// if we had to hide frames, re-enable them now
			render_finished = 1;
		}
		
		VLDP_UNLOCK;
    } // end while
}

// searches to any arbitrary frame, be it I, P, or B, and renders it
// assumes globals are locked
void idle_handler_search()
{
	Uint32 proposed_pos = 0;
	Uint16 actual_frame = g_req_frame;
	int skipped_I = 0;

	VLDP_ACK(search);
	g_ack_status = STAT_BUSY;

	// do a bounds check
	if (g_req_frame < g_totalframes)
	{		
		proposed_pos = g_frame_position[g_req_frame];	// get the proposed position

		s_frames_to_skip = 1;	// first frame in mpeg2dec is nothing, so we always skip at least 1

		while (1)
		{
		  // if the frame we want is not an I frame, go backward until we find an I frame, and increase # of frames to skip forward
		  while ((proposed_pos == 0xFFFFFFFF) && (actual_frame > 0))
		  {
			s_frames_to_skip++;
			actual_frame--;
			proposed_pos = g_frame_position[actual_frame];
		  }
		  skipped_I++;

			// if we are only 2 frames away from an I frame, we will get a corrupted image and need to go back to
			// the I frame before this one
		  if ((skipped_I < 2) && (s_frames_to_skip < 3) && (actual_frame > 0))
		  {
		  	proposed_pos = 0xFFFFFFFF;
		  }
		  else
		  {
		  	break;
		  }
		}
//		printf("frames_to_skip is %d, skipped_I is %d\n", s_frames_to_skip, skipped_I);
//		printf("position in mpeg2 stream we are seeking to : %x\n", proposed_pos);

		fseek(g_mpeg_handle, proposed_pos, SEEK_SET);	// go to the place in the stream where the I frame begins

		g_ack_frame = g_req_frame - 1;	// this will be adjusted in the player
	
		s_paused = 1;	// we do want to pause on the frame we search to
		s_blanked = 0;	// we want to see the frame
		s_timer = SDL_GetTicks();	// reset timer so framerate is correct
		s_elapsed_frames = 0;

		VLDP_UNLOCK;	
		ivldp_render();
		VLDP_LOCK;
	}
	else
	{
		fprintf(stderr, "SEARCH ERROR : frame %u was requested, but it is out of bounds\n", g_req_frame);
		g_ack_status = STAT_ERROR;
	}
}

// pauses the playback
// assumes globals are locked
void idle_handler_pause()
{
	// does nothing since we need to check for pauses during playback, not during idle time
	if (g_req_step_forward)
	{
		fprintf(stderr, "PAUSE ERROR : step forward called when disc was not playing\n");
		VLDP_ACK(step_forward);
	}
	else
	{
		fprintf(stderr, "PAUSE ERROR : pause called when video was stopped\n");
	}
	VLDP_ACK(pause);
	
	g_ack_status = STAT_ERROR;
}

// 'stops' the playback
// assumes globals are locked
void idle_handler_stop()
{
	VLDP_ACK(stop);
	g_ack_status = STAT_STOPPED;

	rewind(g_mpeg_handle);	// go to first of stream when stopping
	s_frames_to_skip = 0;	// skip no frames
	s_paused = 1;	// we do want to pause on the frame we search to
	s_blanked = 1;	// we want a blank screen when 'stopped'
	s_timer = SDL_GetTicks();	// reset timer so framerate is correct
	s_elapsed_frames = 0;

	VLDP_UNLOCK;	
	ivldp_render();
	VLDP_LOCK;

}

// copies the contents of src into dst
// assumes overlays are locked and *IMPORTANT* assumes overlays are the same size
/*
void ivldp_copy_overlay(SDL_Overlay *dst, SDL_Overlay *src)
{
	Uint8 *src_ptr = src->pixels[0];	// start with Y
	Uint8 *dst_ptr = dst->pixels[0];
	Uint32 size = src->w * src->h;
	memcpy(dst_ptr, src_ptr, size);
	src_ptr = src->pixels[1];	// V
	dst_ptr = dst->pixels[1];
	memcpy(dst_ptr, src_ptr, size >> 2);
	src_ptr = src->pixels[2];	// U
	dst_ptr = dst->pixels[2];
	memcpy(dst_ptr, src_ptr, size >> 2);
}
*/

// parses an mpeg video stream to get its frame offsets, or if the parsing had taken place earlier
// assumes globals are LOCKED
int ivldp_get_mpeg_frame_offsets(char *mpeg_name)
{
	char datafilename[320] = { 0 };
	int mpeg_datafile_good = 0;
	FILE *data_file = NULL;
	FILE *mpeg_file = NULL;
	int result = 1;
	Uint32 mpeg_size = 0;
	struct stat filestats;	// used to get the size of a file
	struct dat_header header;

	// GET LENGTH OF ACTUAL FILE
	mpeg_file = fopen(mpeg_name, "rb");
	if (mpeg_file)
	{
		fstat(fileno(mpeg_file), &filestats);	// get stats for file to get file length
		mpeg_size = filestats.st_size;
	}
	else
	{
		fprintf(stderr, "ERROR: video file could not be found!\n");
		g_req_quit = 1;
	}

	// change extension of file to be dat instead of (presumably) m2v
	strcpy(datafilename, mpeg_name);
	strcpy(&datafilename[strlen(mpeg_name)-3], "dat");

	// loop until we get a good datafile or until user quits
	while (!mpeg_datafile_good && !g_req_quit)
	{
		// keep trying to make the datafile until we succeed of the user quits
		while (!data_file && !g_req_quit)
		{
			data_file = fopen(datafilename, "rb");	// check to see if datafile exists

			// if file does not exist, we have to create it
			if (!data_file)
			{
				result = ivldp_parse_mpeg_frame_offsets(datafilename, mpeg_size, mpeg_file);
				data_file = NULL;	// need to set to NULL so we load the file on the next loop
			}
		} // end of inner while that tries to open the datafile

		// if mpeg parsing worked properly (or if it was not needed)
		if (result)
		{
			// now that file exists and we have it open, we have to read it

			fseek(data_file, 0L, SEEK_SET);
			fread(&header, sizeof(header), 1, data_file);	// read .DAT file header

			// if version, file size, or finished are wrong, the dat file is no good and has to be regenerated
			if ((header.length != mpeg_size) || (header.version != DAT_VERSION) || (header.finished != 1))
			{
//				printf("*** Alleged mpeg size is %u, actual size is %u\n", header.length, mpeg_size);
//				printf("Finished flag is %x\n", header.finished);
//				printf("DAT version is %x\n", header.version);
				printf("NOTICE : MPEG data file has to be created again!\n");
				fclose(data_file);
				data_file = NULL;
				unlink(datafilename);	// delete file and create it
			}
			else
			{
				g_uses_fields = header.uses_fields;
				mpeg_datafile_good = 1;	// escape the loop
			}
		} // end if mpeg parsing worked (or if it was not necessary)
	} // end while we don't have a good datafile
	
	// if we didn't exit the loop because someone quit, then we need to read the datafile
	// data_file is still open
	if (!g_req_quit && result && data_file)
	{
		g_totalframes = 0;

		// read all the frame positions
		// if we don't read 4 bytes, it means we've hit the EOF and we're done
		while (fread(&g_frame_position[g_totalframes], 4, 1, data_file) == 1)
		{
			g_totalframes++;
		}
//		printf("*** g_totalframes is %u\n", g_totalframes);
	}

	if (data_file)
	{
		fclose(data_file);
	}
	if (mpeg_file)
	{
		fclose(mpeg_file);
	}
	
	return result;
}

// assume globals are locked
int ivldp_parse_mpeg_frame_offsets(char *datafilename, Uint32 mpeg_size, FILE *mpeg_file)
{
	int result = 1;
	FILE *data_file = fopen(datafilename, "wb");	// create file
	struct dat_header header;	// header to put inside .DAT file

	// if we could create the file successfully, then we need to populate it
	if (data_file)
	{
	  SDL_Surface *progress = SDL_CreateRGBSurface(SDL_SWSURFACE, s_surface->w, 1, 32, 0xFF, 0xFF00, 0xFF0000, 0);
	  // progress indicator surface

	  if (progress)
	  {
		Uint32 pos = 0;	// position in the file
		int count = 0;
		int parse_result = 0;
		SDL_Rect dest;

		// prepare to blit our progress indicator to the screen by making a destination blit area
		dest.x = (short) 0;	// start at left
		dest.y = (short) s_surface->h / 2;	// put it halfway down the screen
		dest.w = (unsigned short) progress->w;
		dest.h = (unsigned short) progress->h;

		printf("MPEG frame data does not exist and must be generated.\n");
		printf("Please be patient, this could take a few minutes...\n");

		header.version = DAT_VERSION;
		header.finished = 0;
		header.uses_fields = 0;
		header.length = mpeg_size;
		fwrite(&header, sizeof(header), 1, data_file);
		// first thing that goes in the file is the .DAT header
		// That way we can re-use the file another time with confidence that it's
		// the right one

		init_mpegscan();

		// keep reading the file while there is a file left to be read
		do
		{
#define PARSE_CHUNK 200000

			parse_result = parse_video_stream(mpeg_file, data_file, PARSE_CHUNK);
			pos += PARSE_CHUNK;

			// we want to give the user updates but don't want to flood them
			if (count > 10)
			{
				count = 0;
				ivldp_update_progress_indicator(progress, (double) ((double) pos / mpeg_size));
				SDL_BlitSurface(progress, NULL, s_surface, &dest);
				SDL_UpdateRects(s_surface, 1, &dest);
			}
			count++;
			VLDP_UNLOCK;	// give parent thread a chance to shut us down if necessary
			VLDP_LOCK;

			// if the user has requested a premature quit, then do so immediately and return an error
			if (g_req_quit)
			{
				parse_result = P_ERROR;
			}
		} while (parse_result == P_IN_PROGRESS);

		// if parse finished, then we have to update the header
		if (parse_result != P_ERROR)
		{
			header.finished = 1;
			header.uses_fields = 0;
			if (parse_result == P_FINISHED_FIELDS)
			{
				header.uses_fields = 1;
			}
			fseek(data_file, 0L, SEEK_SET);
			fwrite(&header, sizeof(header), 1, data_file);	// save changes
		}

		// we have to close data file because it's write-only
		// and it needs to be re-opened read-only
		fclose(data_file);
		data_file = NULL;

		// if the mpeg did not finish parsing gracefully, we've got problems
		// NOTE : I separated this from the other if above to guarantee that the file gets closed
		if (parse_result == P_ERROR)
		{
			fprintf(stderr, "There was an error parsing the MPEG file.\n");
			fprintf(stderr, "Either there is a bug in the parser or the MPEG file is corrupt.\n");
			fprintf(stderr, "OR the user aborted the decoding process :)\n");
			result = 0;
			unlink(datafilename);
		}
		// erase line on screen so it doesn't show up between YUV overlays
		if (SDL_LockSurface(s_surface) == 0)
		{
			SDL_FillRect(s_surface, NULL, 0);
			SDL_Flip(s_surface);
			SDL_UnlockSurface(s_surface);
		}
		SDL_FreeSurface(progress);	// free surface after we're done using it
	  } // end if progress was created successfully
	  else
	  {
	  	fprintf(stderr, "could not create SDL surface, wtf?\n");
	  }

	} // end if we could create the file successfully

	// we couldn't create data file which means no write permission probably
	// this is probably a good time to shut VLDP down =]
	else
	{
		fprintf(stderr, "Could not create file %s\n", datafilename);
		fprintf(stderr, "This probably means you don't have permission to create the file\n");
		result = 0;
	}

	return result;
}

// updates our progress indicator to show how close the mpeg parsing it to being complete
// percentage complete is between 0.0 and 1.0
void ivldp_update_progress_indicator(SDL_Surface *indicator, double percentage_completed)
{
	Uint32 width = 0;
	static Uint32 cur_pixel = (Uint32) -1;	// remember our old position so we can start there again
	static Uint32 *pos = NULL;	// start with red
	Uint32 cur_color = 0xFF;

	// protect against potential overflow
	if (percentage_completed > 1.0)
	{
		percentage_completed = 1.0;
	}

	width = (Uint32) (indicator->w * percentage_completed);

	// if we have started over, then reset the current pixel to 0
	if (width < cur_pixel)
	{
		pos = indicator->pixels;
		cur_pixel = 0;
	}

	for (; cur_pixel < width; cur_pixel++)
	{
		cur_color = 0xFF | (((Uint8) (0xFF * percentage_completed)) << 8) | (((Uint8) (0xFF * percentage_completed)) << 16);
		// grow from red to white
		*pos = cur_color;
		pos++;
	}
}
