/*
 * m80.h
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef M80_H
#define M80_H

// m80.h

/* if we are integrating with daphne, define this */
#define INTEGRATE 1

#ifdef INTEGRATE
#include <SDL.h>
#include "mamewrap.h"
#include "cpu-debug.h"
#endif

#define M80_LITTLE_ENDIAN	1

// only works with GCC compiler (no MSVC++)
//#define GCC_X86_ASM	1

//////////////////////////

#ifndef Uint32
#define Uint32 unsigned int
#endif

#ifndef Uint16
#define Uint16 unsigned short
#endif

#ifndef Uint8
#define Uint8 unsigned char
#endif

#ifndef Sint8
#define Sint8 signed char
#endif

#ifndef CLEAR_LINE
#define CLEAR_LINE 0
#endif

#ifndef ASSERT_LINE
#define ASSERT_LINE 1
#endif

/////////////////////////

typedef enum
{
	M80_PC, M80_SP, M80_AF, M80_AFPRIME, M80_HL, M80_HLPRIME, M80_DE, M80_DEPRIME,
	M80_BC, M80_BCPRIME, M80_IX, M80_IY, M80_RI, M80_REG_COUNT
} regnames;

#define C_FLAG	1
#define N_FLAG	2
#define P_FLAG	4
#define V_FLAG	4
#define U3_FLAG	8
#define H_FLAG	16
#define U5_FLAG	32
#define Z_FLAG	64
#define S_FLAG	128

// I _hate_ macros but these just plain make the code look better =)
#define FLAGS	m80_regs[M80_AF].b.l
#define A		m80_regs[M80_AF].b.h
#define AF		m80_regs[M80_AF].w
#define AFPRIME	m80_regs[M80_AFPRIME].w
#define B		m80_regs[M80_BC].b.h
#define C		m80_regs[M80_BC].b.l
#define BC		m80_regs[M80_BC].w
#define BCPRIME	m80_regs[M80_BCPRIME].w
#define D		m80_regs[M80_DE].b.h
#define E		m80_regs[M80_DE].b.l
#define DE		m80_regs[M80_DE].w
#define DEPRIME	m80_regs[M80_DEPRIME].w
#define H		m80_regs[M80_HL].b.h
#define L		m80_regs[M80_HL].b.l
#define HL		m80_regs[M80_HL].w
#define HLPRIME	m80_regs[M80_HLPRIME].w
#define PC		m80_regs[M80_PC].w
#define SP		m80_regs[M80_SP].w
#define R		m80_regs[M80_RI].b.h
#define I		m80_regs[M80_RI].b.l
#define IXh	m80_regs[M80_IX].b.h
#define IXl	m80_regs[M80_IX].b.l
#define IX	m80_regs[M80_IX].w
#define IYh	m80_regs[M80_IY].b.h
#define IYl	m80_regs[M80_IY].b.l
#define IY	m80_regs[M80_IY].w

#ifdef CPU_DEBUG
#define M80_ERROR(MESSAGE)	printf(MESSAGE); printf("Opcode %x at PC %x\n", opcode_base[PC-1], PC-1)
#else
#define M80_ERROR(MESSAGE)
#endif

// keeps out emulated hardware advised of our current PC (only updated on branches or calls)
#define M80_CHANGE_PC(newpc)	\
	change_pc16(newpc)

// returns a byte from memory
#define M80_READ_BYTE(addr)	\
	cpu_readmem16(addr)	\
/*	opcode_base[addr] */

// read Z80 memory into 16-bit z80 register
#define M80_READ_WORD(addr, reg_index)	\
	m80_regs[reg_index].b.l = M80_READ_BYTE(addr);	\
	m80_regs[reg_index].b.h = M80_READ_BYTE(addr+1)

// writes an 8-bit byte into z80 memory
// addr is where to write, val is which value to write
#define M80_WRITE_BYTE(addr, val)	\
	cpu_writemem16(addr, val)	\
/*	opcode_base[addr] = val */

// write 16-bit z80 reg into Z80 memory
#define M80_WRITE_WORD(addr, reg_index)	\
	M80_WRITE_BYTE(addr, m80_regs[reg_index].b.l);	\
	M80_WRITE_BYTE(addr+1, m80_regs[reg_index].b.h)

#define M80_GET_ARG opcode_base[PC++]
#define M80_PEEK_ARG opcode_base[PC]

// get a word and advance PC
// Little Endian has a distinct direct advantage to this problem by being
// able to use a single memory read, while
// Big Endian has to use two 8-bit memory reads
#ifdef M80_LITTLE_ENDIAN
#define M80_GET_WORD 	*( (Uint16 *) ( (Uint8 *) (opcode_base + PC))); PC += 2
#else
#define M80_GET_WORD	(opcode_base[PC++] || (opcode_base[PC++] << 8))
#endif

// peek a word but don't advance PC
#ifdef M80_LITTLE_ENDIAN
#define M80_PEEK_WORD	*( (Uint16 *) ( (Uint8 *) (opcode_base + PC)))
#else
#define M80_PEEK_WORD	(opcode_base[PC] || (opcode_base[PC+1] << 8))
#endif

// Interrupt check macro
#define CHECK_INTERRUPT	\
	if (g_nmi_state == ASSERT_LINE) /* NMI takes priority over IRQ so check it first */	\
	{	\
		m80_activate_nmi();	\
	}	\
	else if (g_irq_state == ASSERT_LINE)	\
	{	\
		/* we can only do an IRQ if IFF1 flipflop is set */	\
		if (g_IFF1)	\
		{	\
			m80_activate_irq();	\
		}	\
	}

#define M80_EAT_EXTRA_DD_FD	\
/* any extra DD's or FD's in front of a DD/FD instruction is ignored */	\
/* But in order to avoid a recursive loop, we eat up any extras right now */	\
	while ((opcode_base[PC] == 0xDD) || (opcode_base[PC] == 0xFD))	\
	{	\
		PC++;	\
		M80_INC_R;	/* each extra DD or FD costs an R */	\
		g_cycles_executed += op_cycles[0]; /* each extra DD or FD costs a NOP cycle count */	\
	}


// Enable Maskable Interrupt macro
#define M80_EI	\
	/* IFF1 and IFF2 are simultaneously set by EI */	\
	g_IFF1 = 1;	\
	g_IFF2 = 1;	\
	g_got_EI = 1;	/* the next instruction must not be interrupted */

// Disable Maskable Interrupt macro
#define M80_DI	\
	g_IFF1 = 0;	\
	g_IFF2 = 0;	\
/*	g_got_EI = 1; */	/* it's DI, but we still need the same action */

// FIXME: I removed the above g_got_EI in order to compare exactly with mame's core
// but I believe mame has a bug in this regard and that the g_got_EI should be uncommencted
// once we're done debugging

// Macro to go into HALT mode
#define M80_START_HALT	\
	g_m80_halted = 1;	\
	PC--;	\
	/* move PC so it's pointing back to this HALT statement, so next time instruction executes, */	\
	/* it comes to this HALT again */	\
	/* This way we don't have to waste time doing any checking at the beginning */	\
	/* If we didn't just barely get an EI */	\
	/* instruction, then interrupts won't be set this time around and we */	\
	/* can safely use up the rest of the cycles quickly. */	\
	if (!g_got_EI)	\
	{	\
		int remaining = g_cycles_to_execute - g_cycles_executed;	\
		/* if we still have cycles that need to be used up */	\
		if (remaining > 0)	\
		{	\
			g_cycles_executed += (((remaining >> 2) + 1) << 2);	\
			/* use up remaining cycles we're committed to, in multiples of 4 */	\
			/* so that we exceed cycles_to_execute */	\
		}	\
	}

// macro to check to see if we're halted and if so, fix the PC and unhalt us
#define M80_STOP_HALT	\
	if (g_m80_halted)	\
	{	\
		g_m80_halted = 0;	\
		PC++;	\
	}

// Branch macro
#define M80_BRANCH(pc_offset) PC = PC + ((Sint8) pc_offset); M80_CHANGE_PC(PC)
// Conditional branch macro that adds the extra cycles
#define M80_BRANCH_COND(condition)	\
	if (condition)	\
	{	\
		g_cycles_executed += 5; /* 5 extra cycles if branch is taken */	\
		M80_BRANCH(M80_GET_ARG);	\
	}	\
	/* else if the branch is not taken, we advance 1 to get to next instruction */	\
	else { PC++; }

// Jump macro
#define M80_JUMP \
	PC = M80_PEEK_WORD; M80_CHANGE_PC(PC)

// Conditional jump macro that adds the extra cycles
#define M80_JUMP_COND(condition)	\
	if (condition)	\
	{	\
		M80_JUMP;	\
	}	\
	/* if condition isn't true, we have to advance the PC by 2 to skip over the address */	\
	else	\
	{	\
		PC += 2;	\
	}

// Call macro
#define M80_CALL	\
	{	\
	if (M80_PEEK_WORD == 0x465) printf("at PC %x we're calling 0x465\n", PC);	\
		Uint16 destination = M80_GET_WORD;	\
		M80_PUSH16(M80_PC);	\
		PC = destination;	\
		M80_CHANGE_PC(PC);	\
	}

// Conditional call macro that adds the extra cycles
#define M80_CALL_COND(condition)	\
	if (condition)	\
	{	\
		g_cycles_executed += 7;	/* always 7 extra cycles for conditional calls */	\
		M80_CALL;	\
	}	\
	/* if condition isn't true, we have to advance the PC by 2 to skip over address */	\
	else	\
	{	\
		PC += 2;	\
	}

// return macro
#define M80_RET	M80_POP16(M80_PC)
// conditional return macro that adds the extra cycles
#define M80_RET_COND(condition)	\
	if (condition)	\
	{	\
		g_cycles_executed += 6;	/* always 6 extra cycles for conditional returns */	\
		M80_RET;	\
	}	\
	/* else don't advance PC, it's already in the right spot */

// Reset macro
#define M80_RST(addr) M80_PUSH16(M80_PC); PC = addr; M80_CHANGE_PC(PC)

// push macro, NOTE, you have to use the register index, not the value
#define M80_PUSH16(reg_index) SP -= 2; M80_WRITE_WORD(SP, reg_index)

// pop macro, NOTE, you have to use the register index, not the value
#define M80_POP16(reg_index) M80_READ_WORD(SP, reg_index); SP += 2;

// when incrementing a z80 register, here are the rules
// Carry flag is preserved
// N flag is always cleared
// if old_value & 0xF == 0xF then H flag will be set (half carry)
// if new_value == 0x80 then V flag will be set (2's compliment overflow)
// if new_value & 0x80 then S flag will be set (negative number)
// if new_value == 0 then Z flag will be set (zero flag)
#define M80_INC_REG8(which_reg)	\
	which_reg++;	\
	FLAGS = (FLAGS & C_FLAG); /* preserve carry */	\
	FLAGS |= m80_inc_flags[which_reg]

// when decrementing a z80 register, here are the rules
// Carry flag is preserved
// N flag is always set
// if old_value & 0xF == 0x00 then H flag will be set (half carry)
// if new_value == 0x7F then V flag will be set (2's compliment overflow)
// if new_value & 0x80 then S flag will be set (negative number)
// if new_value == 0 then Z flag will be set (zero flag)	
#define M80_DEC_REG8(which_reg)	\
	which_reg--;	\
	FLAGS = (FLAGS & C_FLAG);	\
	FLAGS |= m80_dec_flags[which_reg]

// macro to increment R
// this is the same as a regular increment except bit 7 never changes
#define M80_INC_R	\
	R = (R + 1) & 0x7F | (R & 0x80)

// Rotate Left Accumulator - flags are different from RL A
#define M80_RLA	\
{	\
	/* use a temp variable greater than 8 bits so we can find out if there was a carry */	\
	Uint32 result = ((A << 1) | (FLAGS & C_FLAG));	\
	FLAGS &= (S_FLAG | Z_FLAG | P_FLAG);	\
	FLAGS |= (result & (U5_FLAG | U3_FLAG));	\
	FLAGS |= (result >> 8);	/* set carry flag if bit 8 of result is set */	\
	A = (Uint8) result;	\
}

// Rotate Left - Hi bit goes into carry, carry goes into bit 0
#define M80_RL(which_reg)	\
	{	\
		Uint32 temp = (which_reg << 1) | (FLAGS & C_FLAG);	\
		which_reg = (Uint8) temp;	\
		FLAGS = m80_sz53p_flags[which_reg];	/* set carry to 0 and get S, Z and P flags */	\
		FLAGS |= (temp >> 8); /* set carry to the value of the hi bit */	\
	}

/* does an RL on (IXIY + d) and on a register (undocumented instruction) */
#define M80_RL_COMBO(reg, addr)	\
{	\
	Uint8 val = M80_READ_BYTE(addr);	\
	M80_RL(val);	\
	M80_WRITE_BYTE(addr, val);	\
	reg = val;	\
}

// Rotate Right Accumulator - flags are different from RR A
#define M80_RRA	\
{	\
	Uint8 result = (A >> 1) | (FLAGS << 7);	\
	FLAGS &= (S_FLAG | Z_FLAG | P_FLAG);	\
	FLAGS |= (A & 1);	\
	FLAGS |= (result & (U5_FLAG | U3_FLAG));	\
	A = (Uint8) result;	\
}

// Rotate Right - Bit 0 goes into carry, carry goes into Hi bit
#define M80_RR(which_reg)	\
	{	\
		Uint8 new_carry = which_reg & C_FLAG;	\
		which_reg = (which_reg >> 1) | (FLAGS << 7);	\
		FLAGS = m80_sz53p_flags[which_reg];	/* set carry to 0 and get S, Z and P flags */	\
		FLAGS |= new_carry;	\
	}

/* does an RR on (IXIY + d) and on a register (undocumented instruction) */
#define M80_RR_COMBO(reg, addr)	\
{	\
	Uint8 val = M80_READ_BYTE(addr);	\
	M80_RR(val);	\
	M80_WRITE_BYTE(addr, val);	\
	reg = val;	\
}

// Rotate Left Accumulator - Flags are different from the RLC A instruction
#define M80_RLCA	\
	A = (A >> 7) | (A << 1);	\
	FLAGS &= (S_FLAG | Z_FLAG | P_FLAG);	/* preserve S, Z and P flags, clear H and N flags */	\
	FLAGS |= (A & (U5_FLAG | U3_FLAG | C_FLAG));	/* carry contains the the previous hi-bit which is now bit 0 */

// Rotate Left Circular - Carry gets hi-bit but is otherwise irrelevant to this operation
#define M80_RLC(src_reg)	\
	FLAGS = src_reg >> 7; /* put hi-bit into carry flag */	\
	src_reg = ((src_reg << 1) | FLAGS); /* and also into bit 0 of new value */	\
	FLAGS |= m80_sz53p_flags[src_reg];

/* does an RLC on (IXIY + d) and on a register (undocumented instruction) */
#define M80_RLC_COMBO(reg, addr)	\
{	\
	Uint8 val = M80_READ_BYTE(addr);	\
	M80_RLC(val);	\
	M80_WRITE_BYTE(addr, val);	\
	reg = val;	\
}

// macro for RRCA - flags are different from RRC A
#define M80_RRCA	\
	FLAGS &= (S_FLAG | Z_FLAG | P_FLAG);	\
	FLAGS |= (A & C_FLAG);	\
	A = (A << 7) | (A >> 1);	\
	FLAGS |= (Uint8) (A & (U5_FLAG | U3_FLAG));	\

// Rotate Right Circular - Carry gets low-bit but is otherwise irrelevant to this operation
#define M80_RRC(src_reg)	\
	FLAGS = src_reg & C_FLAG;	/* put low-bit into carry flag */	\
	src_reg = ((src_reg >> 1) | (src_reg << 7)); /* low-bit becomes hi-bit */	\
	FLAGS |= m80_sz53p_flags[src_reg]

/* does an RRC on (IXIY + d) and on a register (undocumented instruction) */
#define M80_RRC_COMBO(reg, addr)	\
{	\
	Uint8 val = M80_READ_BYTE(addr);	\
	M80_RRC(val);	\
	M80_WRITE_BYTE(addr, val);	\
	reg = val;	\
}

// Shift Left Arithmetic/Logical - Carry gets hi-bit, a 0 gets put in at bit 0
#define M80_SLA(reg)	\
	FLAGS = reg >> 7;	/* get value for new carry flag */	\
	reg = reg << 1;	\
	FLAGS |= m80_sz53p_flags[reg]

/* does an SLA on (IXIY + d) and on a register (undocumented instruction) */
#define M80_SLA_COMBO(reg, addr)	\
{	\
	Uint8 val = M80_READ_BYTE(addr);	\
	M80_SLA(val);	\
	M80_WRITE_BYTE(addr, val);	\
	reg = val;	\
}

// Shift left inverted arithmetic - Carry gets hi-bit, a 1 gets put in at bit 0
#define M80_SLIA(reg)	\
	FLAGS = reg >> 7;	/* get value for new carry */	\
	reg = (reg << 1) | 1;	\
	FLAGS |= m80_sz53p_flags[reg]

/* does an SLIA on (IXIY + d) and on a register (undocumented instruction) */
#define M80_SLIA_COMBO(reg, addr)	\
{	\
	Uint8 val = M80_READ_BYTE(addr);	\
	M80_SLIA(val);	\
	M80_WRITE_BYTE(addr, val);	\
	reg = val;	\
}

// Shift Right Arithmetic (preserves sign) - Carry gets bit 0, the old bit 7 goes to the new bit 7
#define M80_SRA(reg)	\
	FLAGS = reg & C_FLAG;	/* put low bit into carry */	\
	reg = (reg >> 1) | (reg & 0x80);	/* shift and preserve hi bit */	\
	FLAGS |= m80_sz53p_flags[reg]

#define M80_SRA_COMBO(reg, addr)	\
{	\
	Uint8 val = M80_READ_BYTE(addr);	\
	M80_SRA(val);	\
	M80_WRITE_BYTE(addr, val);	\
	reg = val;	\
}

// shift right logical - Carry gets bit 0, bit 7 gets 0
#define M80_SRL(reg)	\
	FLAGS = reg & C_FLAG;	\
	reg = reg >> 1;	\
	FLAGS |= m80_sz53p_flags[reg]

/* does an SRL on (IXIY + d) and on a register (undocumented instruction) */
#define M80_SRL_COMBO(reg, addr)	\
{	\
	Uint8 val = M80_READ_BYTE(addr);	\
	M80_SRL(val);	\
	M80_WRITE_BYTE(addr, val);	\
	reg = val;	\
}


// tests the status of a bit in the register
#define M80_BIT(bit, reg)	\
	FLAGS = (FLAGS & C_FLAG); /* preserve C flag, clear N flag */	\
	FLAGS |= m80_szhp_bit_flags[(1 << bit) & reg];	/* update the SZHP flags based on result */	\
	FLAGS |= (reg & 0x28);	/* set 3/5 flags based on value of register */	\
	/* (using YAZE as authoritative since there is some dispute about how BIT works */

// tests the status of a bit in the register
// clears bits 3 and 5 (for the BIT x, (HL) instructions)
#define M80_BIT_NO35(bit, reg)	\
	FLAGS = (FLAGS & C_FLAG); /* preserve C flag, clear N, 5, 3 flags */	\
	FLAGS |= m80_szhp_bit_flags[(1 << bit) & reg];	/* update the SZHP flags based on result */	\

#define M80_BIT_OP(opcode, bit, reg)	\
	if (reg & (1 << bit)) AF = (AF & ~0xfe) | H_FLAG | (((opcode & 0x38) == 0x38) << 7);	\
	else AF = (AF & ~0xfe) | 0x54;	/* Z, H, and P flag */	\
	if ((opcode & 7) != 6) AF |= (reg & 0x28); /* 5 and 3 flags */

// clear a bit in the register
#define M80_RES(bit, reg)	reg &= ~(1 << bit)

// a combo version of the reset instruction, for IXIY + d
#define M80_RES_COMBO(bit, reg, addr)	\
{	\
	Uint8 val = M80_READ_BYTE(addr);	\
	M80_RES(bit, val);	\
	M80_WRITE_BYTE(addr, val);	\
	reg = val;	\
}

// set a bit in the register
#define M80_SET(bit, reg)	reg |= (1 << bit)

/* an IXIY combo version of the set instruction */
#define M80_SET_COMBO(bit, reg, addr)	\
{	\
	Uint8 val = M80_READ_BYTE(addr);	\
	M80_SET(bit, val);	\
	M80_WRITE_BYTE(addr, val);	\
	reg = val;	\
}


// macros for 16-bit register addition

#ifdef GCC_X86_ASM
#define M80_ADD_REGS16(dest_reg, source_reg)	\
	asm(	\
	"addb	%%bl, %%cl\n\t"	/* the actual addition is here */ \
	/* pipeline stall, can anyone fix? */	\
	"adcb	%%bh, %%ch\n\t"	/* 16-bit add does NOT work because Aux Carry (HC) flag is wrong */ \
	"lahf	\n\t"	/* put flags into ah register */	\
	"movb	%%ch, %%bh\n\t"	/* prepare to get flags 3/5 (we can't modify %ch) */	\
	"andb	$0x11, %%ah\n\t"	/* isolate carry and auxiliary carry flags */	\
	"andb	$0x28, %%bh\n\t"	/* get bits 3 and 5 from upper byte ot result */	\
	"andb	$196, %%al\n\t"	/* preserve S, Z, and V flags */	\
	"orb	%%ah, %%al\n\t"	/* put carry and auxiliary carry into Z80 flags */	\
	"orb	%%bh, %%al\n\t"	/* adds bits 3 and 5 to the FLAGS register */	\
	:"=c" (dest_reg), "=a" (FLAGS)	\
	:"c" (dest_reg), "a" (FLAGS), "b" (source_reg)	\
	);
#else
#define M80_ADD_REGS16(dest_reg, source_reg) \
{	\
	Uint32 result = dest_reg + source_reg;	\
	Uint32 cbits = (dest_reg ^ source_reg ^ result) >> 8;	\
	dest_reg = (Uint16) result;	\
	AF = (AF & (0xFF00 | S_FLAG | Z_FLAG | V_FLAG)) |	\
		((result >> 8) & (U3_FLAG | U5_FLAG)) |	\
		(cbits & H_FLAG) |	\
		((cbits >> 8) & C_FLAG);	\
}
#endif

#ifdef GCC_X86_ASM
// We unfortunately could not use a 16-bit subtraction because the x86 Aux Carry flag would
// be different from the z80's.  However, with 8-bit, it is the same.  This subtraction is
// further complicated because we have to change the Z flag, which means we have to AND
// the Z flag result from both 8-bit addtions.
// This assembly is optimized for pipe-lining.  See if you can do better (the last 3 lines will stall)
#define M80_SBC_REGS16(dest_reg, source_reg)	\
	asm(	\
	"shrb	$1, %%al\n\t"	/* put z80 carry flag into x86 carry flag, we don't need to preserve flags */	\
	"sbbb	%%bl, %%cl\n\t"	/* subtract lower bytes */ \
	"lahf	\n\t"	/* we need to capture the Z flag from the lower result */	\
	"sbbb	%%bh, %%ch\n\t"	/* subtract upper bytes */ \
	"movb	%%ah, %%bl\n\t"	/* store lower flags into bl (which we can safely clobber) */	\
	"lahf	\n\t"	/* put upper x86 flags into ah register */	\
	"movb	%%ch, %%bh\n\t"	/* prepare to get 3/5 flags, we can't modify %ch */	\
	"setob	%%al\n\t"	/* if overflow is true, set it (we're starting new flags here) */	\
	"andb	$0x40, %%bl\n\t"	/* isolate Z flag from lower operation */	\
	"shlb	$2, %%al\n\t"	/* shift overflow flag to its proper spot */	\
	"orb	$0xBF, %%bl\n\t"	/* set all the other lower bits to prepare for an AND merge with the upper */	\
	"andb	$0xD1, %%ah\n\t"	/* isolate S/Z/H and C x86 flags from upper operation */	\
	"andb	$0x28, %%bh\n\t"	/* get bits 3 and 5 from upper byte ot result */	\
	"orb	$2, %%al\n\t"	/* set N flag, which is always set in subtraction */	\
	"andb	%%ah, %%bl\n\t"	/* merge lower bits and upper, taking the AND of the Z flag */	\
	"orb	%%bh, %%al\n\t"	/* adds bits 3 and 5 to the FLAGS register */	\
	"orb	%%bl, %%al\n\t"	/* merge bits and V_FLAG */	\
	:"=c" (dest_reg), "=a" (FLAGS)	\
	:"c" (dest_reg), "a" (FLAGS), "b" (source_reg)	\
	);
#else
// Subtract with carry for two 16-bit registers (dest_reg = dest_reg - source-reg - Carry)
#define M80_SBC_REGS16(dest_reg, source_reg)	\
{	\
	Uint32 result = dest_reg - source_reg - (AF & C_FLAG);	\
	Uint32 cbits = (dest_reg ^ source_reg ^ result) >> 8;	\
	dest_reg = result;	\
	AF = (AF & ~0xff) | ((result >> 8) & 0xa8) |	\
		(((result & 0xffff) == 0) << 6) |	\
		(((cbits >> 6) ^ (cbits >> 5)) & 4) |	\
		(cbits & 0x10) | 2 | ((cbits >> 8) & 1);	\
}
#endif

// Add with carry for two 16-bit registers (dest_reg = dest_reg + source_reg + Carry)
#ifdef GCC_X86_ASM
// We unfortunately could not use a 16-bit addition because the x86 Aux Carry flag would
// be different from the z80's.  However, with 8-bit, it is the same.  This addition is
// further complicated because we have to change the Z flag, which means we have to AND
// the Z flag result from both 8-bit addtions.
#define M80_ADC_REGS16(dest_reg, source_reg)	\
	asm(	\
	"shrb	$1, %%al\n\t"	/* put z80 carry flag into x86 carry flag, we don't need to preserve flags */	\
	"adcb	%%bl, %%cl\n\t"	/* add lower bytes */ \
	"lahf	\n\t"	/* we need to capture the Z flag from the lower result */	\
	"adcb	%%bh, %%ch\n\t"	/* add upper bytes */ \
	"movb	%%ah, %%bl\n\t"	/* store lower flags into bl (which we can safely clobber) */	\
	"lahf	\n\t"	/* put upper x86 flags into ah register */	\
	"movb	%%ch, %%bh\n\t"	/* prepare to get 3/5 flags, we can't modify %ch */	\
	"setob	%%al\n\t"	/* if overflow is true, set it (we're starting new flags here) */	\
	"andb	$0x40, %%bl\n\t"	/* isolate Z flag from lower operation */	\
	"shlb	$2, %%al\n\t"	/* shift overflow flag to its proper spot */	\
	"orb	$0xBF, %%bl\n\t"	/* set all the other lower bits to prepare for an AND merge with the upper */	\
	"andb	$0xD1, %%ah\n\t"	/* isolate S/Z/H and C x86 flags from upper operation */	\
	"andb	$0x28, %%bh\n\t"	/* get bits 3 and 5 from upper byte ot result */	\
	"andb	%%ah, %%bl\n\t"	/* merge lower bits and upper, taking the AND of the Z flag */	\
	"orb	%%bh, %%al\n\t"	/* adds bits 3 and 5 to the FLAGS register */	\
	"orb	%%bl, %%al\n\t"	/* merge bits and V_FLAG */	\
	:"=c" (dest_reg), "=a" (FLAGS)	\
	:"c" (dest_reg), "a" (FLAGS), "b" (source_reg)	\
	);
#else
#define M80_ADC_REGS16(dest_reg, source_reg) \
{	\
	Uint32 result = dest_reg + source_reg + (AF & C_FLAG);	\
	Uint32 cbits = (dest_reg ^ source_reg ^ result) >> 8;	\
	dest_reg = result;	\
	AF = (AF & ~0xff) | ((result >> 8) & 0xa8) |	\
		(((result & 0xffff) == 0) << 6) |	/* Z_FLAG */	\
		(((cbits >> 6) ^ (cbits >> 5)) & 4) |	\
		(cbits & 0x10) | ((cbits >> 8) & 1);	\
}
#endif

#define X	printf("Instruction executed that is not implemented yet!\n")

#ifdef GCC_X86_ASM
// NOTE: I optimized this assembly for pipe-lining.  see if you can improve it!
#define M80_ADD_TO_A(which_reg)	\
	asm (	\
	"addb %2,%0	\n\t" /* do the addition */	\
	"lahf	\n\t" /* put intel flags (which are almost the same as z80) into ah */	\
	"setob %1	\n\t" /* set flags reg to 1 if intel overflow flag is set */	\
	"andb $0xd1,%%ah	\n\t" /* strip out non-common intel flags, preserve S,Z,HC, and C */ \
	"shlb $2,%1 	\n\t"	/* shift left twice so overflow flag is in correct position (1 cycle on Pentium) */	\
	"movb %0,%2	\n\t" /* move addition result into %2 so we can get X/Y z80 flags */	\
	"orb %%ah,%1	\n\t" /* merge overflow bit with the rest of the intel flags */	\
	"andb $0x28,%2	\n\t" /* isolate X/Y flags */	\
	"orb %2,%1	\n\t" /* merge them in with the other flags */	\
	:"=r" (A), "=r" (FLAGS)	\
	:"q" (which_reg), "1" (FLAGS), "0" (A)	\
	: "%ah"	/* ah was clobbered */ \
	);
#else
#define M80_ADD_TO_A(which_reg)	\
{	\
	Uint32 sum = which_reg + A;	\
		/* Overflow can only occur in addition if the signs of the operands are the same */	\
		/* If new A's sign is different from old A's, and if the signs of the operands are the same, set V */	\
	FLAGS = ((((sum ^ A) & (which_reg ^ A ^ 0x80)) >> 5) & V_FLAG);	/* clear N flag */ \
	FLAGS |= ((sum ^ A ^ which_reg) & H_FLAG);	/* H_FLAG is set if 1 or all of the bit 4's are set */	\
	FLAGS |= ((sum >> 8) & C_FLAG); /* if bit 8 of sum is set, set carry flag */	\
	A = (Uint8) sum;	\
	FLAGS |= m80_sz53_flags[A];	/* set S, Z, 5 and 3 flags */	\
}
#endif


#ifdef GCC_X86_ASM
// optimized for pipelining, see if you can improve it!
#define M80_ADC_TO_A(which_reg)	\
	asm (	\
	"shrb  $1,%1	\n\t"	/* copy Z80 carry flag into Intel carry flag */	\
	"adcb %2,%0	\n\t" /* do the addition, including the carry */	\
	"lahf	\n\t" /* put intel flags (which are almost the same as z80) into ah */	\
	"setob %1	\n\t" /* set flags reg to 1 if intel overflow flag is set */	\
	"andb $0xd1,%%ah	\n\t" /* strip out non-common intel flags, preserve S,Z,HC, and C */ \
	"shlb $2,%1 	\n\t"	/* shift left twice so overflow flag is in correct position (1 cycle on Pentium) */	\
	"movb %0,%2	\n\t" /* move addition result into %2 so we can get X/Y z80 flags */	\
	"orb %%ah,%1	\n\t" /* merge overflow bit with the rest of the intel flags */	\
	"andb $0x28,%2	\n\t" /* isolate X/Y flags */	\
	"orb %2,%1	\n\t" /* merge them in with the other flags */	\
	:"=r" (A), "=r" (FLAGS)	\
	:"q" (which_reg), "1" (FLAGS), "0" (A)	\
	: "%ah"	/* ah was clobbered */ \
	);
#else
#define M80_ADC_TO_A(which_reg)	\
{	\
	Uint32 sum = which_reg + A + (FLAGS & C_FLAG);	\
		/* Overflow can only occur in addition if the signs of the operands are the same */	\
		/* If new A's sign is different from old A's, and if the signs of the operands are the same, set V */	\
	FLAGS = ((((sum ^ A) & (which_reg ^ A ^ 0x80)) >> 5) & V_FLAG);	/* clear N flag */ \
	FLAGS |= ((sum ^ A ^ which_reg) & H_FLAG);	/* H_FLAG is set if 1 or all of the bit 4's are set */	\
	FLAGS |= ((sum >> 8) & C_FLAG); /* if bit 8 of sum is set, set carry flag */	\
	A = (Uint8) sum;	\
	FLAGS |= m80_sz53_flags[A];	/* set S, Z, Y and X flags */	\
}
#endif



#ifdef GCC_X86_ASM
#define M80_SUB_FROM_A(which_reg)	\
	asm (	\
	"subb %2,%1	\n\t" /* do the subtraction (1 cycle) */	\
	"lahf	\n\t" /* put intel flags (which are almost the same as z80) into ah (2 cycles) */	\
	"setob %0	\n\t" /* set flags reg to 1 if intel overflow flag is set (1 or 2 cycles) */	\
	"andb $0xd1,%%ah	\n\t" /* strip out non-common intel flags (1 cycle) */ \
	"shlb $2,%0 	\n\t"	/* shift left twice so overflow flag is in correct position (1 cycle on Pentium) */	\
	"movb %1,%2	\n\t" /* move subtraction result into available reg for X/Y z80 flags (1 cycle) */	\
	"orb $2, %0	\n\t" /* set N_FLAG in the flags register (1 cycle) */	\
	"andb $0x28,%2	\n\t" /* isolate X/Y flags (1 cycle) */	\
	"orb %%ah,%0	\n\t" /* merge overflow bit with the rest of the intel flags (1 cycle) */	\
	/* stall */	\
	"orb %2,%0	\n\t" /* merge them in with the other flags (1 cycle) */	\
	:"=r" (FLAGS), "=r" (A)	\
	:"r" (which_reg), "0" (FLAGS), "1" (A)	\
	: "%ah"	/* ah was clobbered */ \
	);
#else
#define M80_SUB_FROM_A(which_reg)	\
{	\
	Uint32 diff = A - which_reg;	\
	FLAGS = N_FLAG;	/* N_FLAG always gets set when subtracting */	\
		/* Overflow can only occur in subtraction if the signs of the operands are different */	\
		/* If new A's sign is different from old A's, and if the signs of the operands are different, set V */	\
	FLAGS |= ((((diff ^ A) & (which_reg ^ A)) >> 5) & V_FLAG);	\
	FLAGS |= ((diff ^ A ^ which_reg) & H_FLAG);	/* H_FLAG is set if 1 or all of the bit 4's are set */	\
	FLAGS |= ((diff >> 8) & C_FLAG); /* if bit 8 of diff is set, set carry flag */	\
	A = (Uint8) diff;	\
	FLAGS |= m80_sz53_flags[A];	/* set S, Z, Y and X flags */	\
}
#endif


#ifdef GCC_X86_ASM
#define M80_SBC_FROM_A(which_reg)	\
	asm (	\
	"shrb $1,%0	\n\t" /* copy z80 carry flag into intel carry flag */	\
	"sbbb %2,%1	\n\t" /* do the subtraction w/ carry (1 cycle) */	\
	"lahf	\n\t" /* put intel flags (which are almost the same as z80) into ah (2 cycles) */	\
	"setob %0	\n\t" /* set flags reg to 1 if intel overflow flag is set (1 or 2 cycles) */	\
	"andb $0xd1,%%ah	\n\t" /* strip out non-common intel flags (1 cycle) */ \
	"shlb $2,%0 	\n\t"	/* shift left twice so overflow flag is in correct position (1 cycle on Pentium) */	\
	"movb %1,%2	\n\t" /* move subtraction result into available reg for X/Y z80 flags (1 cycle) */	\
	"orb $2, %0	\n\t" /* set N_FLAG in the flags register (1 cycle) */	\
	"andb $0x28,%2	\n\t" /* isolate X/Y flags (1 cycle) */	\
	"orb %%ah,%0	\n\t" /* merge overflow bit with the rest of the intel flags (1 cycle) */	\
	/* stall */	\
	"orb %2,%0	\n\t" /* merge them in with the other flags (1 cycle) */	\
	:"=r" (FLAGS), "=r" (A)	\
	:"r" (which_reg), "0" (FLAGS), "1" (A)	\
	: "%ah"	/* ah was clobbered */ \
	);
#else
#define M80_SBC_FROM_A(which_reg)	\
{	\
	Uint32 diff = A - which_reg - (FLAGS & C_FLAG);	\
	FLAGS = N_FLAG;	/* N_FLAG always gets set when subtracting */	\
		/* Overflow can only occur in subtraction if the signs of the operands are different */	\
		/* If new A's sign is different from old A's, and if the signs of the operands are different, set V */	\
	FLAGS |= ((((diff ^ A) & (which_reg ^ A)) >> 5) & V_FLAG);	\
	FLAGS |= ((diff ^ A ^ which_reg) & H_FLAG);	/* H_FLAG is set if 1 or all of the bit 4's are set */	\
	FLAGS |= ((diff >> 8) & C_FLAG); /* if bit 8 of diff is set, set carry flag */	\
	A = (Uint8) diff;	\
	FLAGS |= m80_sz53_flags[A];	/* set S, Z, 5 and 3 flags */	\
}
#endif

#ifdef GCC_X86_ASM
#define M80_COMPARE_WITH_A(which_reg)	\
	asm (	\
	"cmpb %%al,%%cl	\n\t" /* do the comparison */	\
	"lahf	\n\t" /* put intel flags (which are almost the same as z80) into ah (2 cycles) */	\
	"setob %0	\n\t" /* set flags reg to 1 if intel overflow flag is set (1 or 2 cycles) */	\
	"andb $0xd1,%%ah	\n\t" /* strip out non-common intel flags (1 cycle) */ \
	"shlb $2,%0 	\n\t"	/* shift left twice so overflow flag is in correct position (1 cycle on Pentium) */	\
	"andb $0x28,%%al	\n\t" /* isolate X/Y flags from operand (1 cycle) */	\
	"orb $2, %0	\n\t" /* set N_FLAG in the flags register (1 cycle) */	\
	/* stall */	\
	"orb %%ah,%0	\n\t" /* merge overflow bit with the rest of the intel flags (1 cycle) */	\
	/* stall */	\
	"orb %%al,%0	\n\t" /* merge them in with the other flags (1 cycle) */	\
	:"=r" (FLAGS)	\
	:"a" (which_reg), "c" (A), "0" (FLAGS)	\
	);
#else
#define M80_COMPARE_WITH_A(which_reg)	\
{	\
	Uint32 sum = A - which_reg;	\
	Uint32 cbits = A ^ which_reg ^ sum;	\
		AF = (AF & ~0xff) | (sum & 0x80) |	\
			(((sum & 0xff) == 0) << 6) | (which_reg & 0x28) |	\
			(((cbits >> 6) ^ (cbits >> 5)) & 4) | 2 |	\
			(cbits & 0x10) | ((cbits >> 8) & 1);	\
}
#endif

// FIXME: I think the above COMPARE is broken.  According to Sean Young, flags 5 and 3 are
// from the operand, not the result.  Of course this is very minor, but it's still a bug that needs fixing.

#define M80_AND_WITH_A(which_reg)	\
	A &= which_reg;	\
	/* clear N and C, set H, and fill in S, Z, and P appropriately */	\
	FLAGS = m80_sz53p_flags[A] | H_FLAG;

#define M80_XOR_WITH_A(which_reg)	\
	A ^= which_reg;	\
	/* clear N,C, and H, and fill in S, Z, and P appropriately */	\
	FLAGS = m80_sz53p_flags[A];

#define M80_OR_WITH_A(which_reg)	\
	A |= which_reg;	\
	/* clear N,C, and H, and fill in S, Z, and P appropriately */	\
	FLAGS = m80_sz53p_flags[A];

// macro to swap AF with AF'
#define M80_EX_AFS	\
{	\
	Uint16 tmp = AFPRIME;	\
	AFPRIME = AF;	\
	AF = tmp;	\
}

// macro to swap HL and DE
#define M80_EX_DEHL	\
{	\
	Uint16 temp;	\
	temp = HL;	\
	HL = DE;	\
	DE = temp;	\
}

// compliment A
#define M80_CPL	\
	A = (Uint8) (A ^ 0xFF);	\
	FLAGS &= (Uint8) (S_FLAG | Z_FLAG | P_FLAG | C_FLAG );	\
	FLAGS |= (Uint8) (H_FLAG | N_FLAG);	\
	FLAGS |= (Uint8) (A & (U5_FLAG | U3_FLAG));	\
	/* see Sean Young's z80 doc for more info about how CPL behaves */

// set carry flag
#define M80_SCF	\
	FLAGS &= (Uint8) (S_FLAG | Z_FLAG | P_FLAG);	\
	FLAGS |= C_FLAG;	\
	FLAGS |= (Uint8) (A & (U5_FLAG | U3_FLAG));

// compliment carry flag
#define M80_CCF	\
	FLAGS &= (S_FLAG | Z_FLAG | P_FLAG | C_FLAG);	\
	FLAGS |= (FLAGS & C_FLAG) << 4;	/* H flag gets the old value of C */ 	\
	FLAGS ^= C_FLAG;	\
	FLAGS |= (A & (U5_FLAG | U3_FLAG));

// Exchange most of the registers (EXX)
#define M80_EXX	\
{	\
	Uint16 temp = BC;	\
	BC = BCPRIME;	\
	BCPRIME = temp;	\
	temp = DE;	\
	DE = DEPRIME;	\
	DEPRIME = temp;	\
	temp = HL;	\
	HL = HLPRIME;	\
	HLPRIME = temp;	\
}


// performs fast DAA using a lookup table
// Thx to Sean young for the idea and Juergen Buchmueller for the implementation :)
#define M80_DAA	\
{	\
	int index = A;	\
	if ( FLAGS & C_FLAG )	\
	{	\
		index |= 0x100;	\
	}	\
	if (FLAGS & H_FLAG)	\
	{	\
		index |= 0x200;	\
	}	\
	if (FLAGS & N_FLAG)	\
	{	\
		index |= 0x400;	\
	}	\
	AF = daa_table[index];	\
}


// writes A into a port.  We add A*256 to the port according to Sean Young's document
#define M80_OUT_A(port)	\
	cpu_writeport16(port | (A << 8), A)

// outputs a register to a port
#define M80_OUT16(port, value)	\
	cpu_writeport16(port, value)

// inputs from a port (8-bits) into A (no flags are affected by this)
// adds A*256 to the 8-bit port according to Sean Young's document
#define M80_IN_A(port)	\
	A = cpu_readport16(port | (A << 8))

// inputs from a port (16-bits) into a register
#define M80_IN16(port, reg)	\
	reg = cpu_readport16(port);	\
	FLAGS = FLAGS & C_FLAG;	/* preserve only carry */	\
	FLAGS |= m80_sz53p_flags[reg]

// block instructions here

// some of the the repeating block instructions have this code in common
#define M80_BLOCK_REPEAT(base)	\
	base;	\
	if (BC)	\
	{	\
		PC -= 2;	\
		g_cycles_executed += 5;	\
	}

// repeat until BC is 0 OR until A = (HL)  (which is true when Z flag is set)
#define M80_BLOCK2_REPEAT(base)	\
	base;	\
	if ((BC != 0) && !(FLAGS & Z_FLAG))	\
	{	\
		PC -= 2;	\
		g_cycles_executed += 5;	\
	}

#define M80_LDI	\
	{	\
		Uint8 temp = M80_READ_BYTE(HL);	\
		M80_WRITE_BYTE(DE, temp);	\
		DE++;	\
		HL++;	\
		BC--;	\
		FLAGS &= S_FLAG | Z_FLAG | C_FLAG; /* preserve s, z, c, clear h and c */ \
		/* if BC is not 0, set P flag */	\
		if (BC)	\
		{	\
			FLAGS |= P_FLAG;	\
		}	\
		FLAGS |= ((A + temp) & 2) << 4;	/* U5_FLAG is bit 1 of A+(HL) */	\
		FLAGS |= ((A + temp) & U3_FLAG);	/* U3_FLAG is bit 3 of A+(HL) */	\
	}	\

// macro for the CPI instruction
#define M80_CPI \
	{	\
		Uint32 temp = M80_READ_BYTE(HL);	\
		Uint32 cbits, sum;	\
		++HL;	\
		sum = A - temp;	\
		cbits = A ^ temp ^ sum;	\
		AF = (AF & ~0xfe) | (sum & 0x80) | (!(sum & 0xff) << 6) |	\
			(((sum - ((cbits&16)>>4))&2) << 4) | (cbits & 16) |	\
			((sum - ((cbits >> 4) & 1)) & 8) |	\
			((--BC & 0xffff) != 0) << 2 | 2;	\
		if ((sum & 15) == 8 && (cbits & 16) != 0)	\
			AF &= ~8;	\
	}

#define M80_CPIR	\
{	\
	Uint32 acu = A;	\
	Uint32 temp, op, sum, cbits;	\
	BC &= 0xffff;	\
	do {	\
		temp = M80_READ_BYTE(HL); ++HL;	\
		op = --BC != 0;	\
		sum = acu - temp;	\
	} while (op && sum != 0);	\
	cbits = acu ^ temp ^ sum;	\
	AF = (AF & ~0xfe) | (sum & 0x80) | (!(sum & 0xff) << 6) |	\
		(((sum - ((cbits&16)>>4))&2) << 4) |	\
		(cbits & 16) | ((sum - ((cbits >> 4) & 1)) & 8) |	\
		op << 2 | 2;	\
	if ((sum & 15) == 8 && (cbits & 16) != 0)	\
		AF &= ~8;	\
}

#define M80_CPD {													\
	Uint8 val = M80_READ_BYTE(HL);										\
	Uint8 res = A - val;										\
	HL--; BC--;												\
	FLAGS = (FLAGS & C_FLAG) | (m80_sz53_flags[res] & ~(U5_FLAG|U3_FLAG))	\
		| ((A ^ val ^ res) & H_FLAG) | N_FLAG;  \
	if( FLAGS & H_FLAG ) res -= 1; 									\
	if( res & 0x02 ) FLAGS |= U5_FLAG; /* bit 1 -> flag 5 */			\
	if( res & 0x08 ) FLAGS |= U3_FLAG; /* bit 3 -> flag 3 */			\
	if( BC ) FLAGS |= V_FLAG; 										\
}

#define M80_CPDR	\
{	\
	Uint32 temp;	\
	Uint32 op;	\
	Uint32 sum;	\
	Uint32 cbits;	\
	do {	\
		temp = M80_READ_BYTE(HL);	\
		--HL;	\
		op = --BC != 0;	\
		sum = A - temp;	\
	} while (op && sum != 0);	\
	cbits = A ^ temp ^ sum;	\
	AF = (AF & ~0xfe) | (sum & 0x80) | (!(sum & 0xff) << 6) |	\
		(((sum - ((cbits&16)>>4))&2) << 4) |	\
		(cbits & 16) | ((sum - ((cbits >> 4) & 1)) & 8) |	\
		op << 2 | 2;	\
	if ((sum & 15) == 8 && (cbits & 16) != 0)	\
		AF &= ~8;	\
}

// macro for LDD instruction
#define M80_LDD	\
{	\
	Uint8 acu = M80_READ_BYTE(HL);	\
	--HL;	\
	M80_WRITE_BYTE(DE, acu);	\
	--DE;	\
	acu += A;	\
	AF = (AF & ~0x3e) | (acu & 8) | ((acu & 2) << 4) |	\
		(((--BC & 0xffff) != 0) << 2);	\
}

// macro for LDDR instruction
#define M80_LDDR	\
{	\
	Uint8 acu;	\
	BC &= 0xffff;	\
	do	\
	{	\
		acu = M80_READ_BYTE(HL);	\
		--HL;	\
		M80_WRITE_BYTE(DE, acu);	\
		--DE;	\
	} while (--BC);	\
	acu += A;	\
	AF = (AF & ~0x3e) | (acu & 8) | ((acu & 2) << 4);	\
}

// puts together an IXIY + d and advances PC, useful for most instructions
#define IXIY_OFFSET(IXIY)	IXIY + ((Sint8) M80_GET_ARG)

// puts together an IXIY + d but does not advance PC.  Needed for math operations that use this
// value more than once
#define IXIY_OFFSET_PEEK(IXIY)	IXIY + ((Sint8) M80_PEEK_ARG)

///////////////////

typedef union
{
#ifdef M80_LITTLE_ENDIAN
	struct
	{
		Uint8 l,h;
	} b;
#else
	struct
	{
		Uint8 h, l;
	} b;
#endif
	Uint16 w;
} pair;

//////////////////////////////////////////////////////////

void m80_set_opcode_base(Uint8 *address);
void m80_reset();
Uint32 m80_execute(Uint32 cycles_to_execute);
static __inline__ void m80_exec_cb();
void m80_set_nmi_line(Uint8);
void m80_set_irq_line(Uint8);
void m80_activate_nmi();
void m80_activate_irq();
Uint16 m80_get_pc();
void m80_set_pc(Uint16);
Uint16 m80_get_sp();
void m80_set_sp(Uint16);
Uint16 m80_get_reg(int index);
void m80_set_reg(int index, Uint16 val);
void m80_set_irq_callback(Uint32 (*callback)());
void m80_set_nmi_callback(Uint32 (*callback)());
Uint32 m80_get_cycles_executed();
///////////////////////////////////////////////////////////

// Special Thanx to Juergen Buchmueller for this table
static Uint16 daa_table[0x800] =
{
  (0x00<<8)   +Z_FLAG         +V_FLAG      ,
  (0x01<<8)                        ,
  (0x02<<8)                        ,
  (0x03<<8)               +V_FLAG      ,
  (0x04<<8)                        ,
  (0x05<<8)               +V_FLAG      ,
  (0x06<<8)               +V_FLAG      ,
  (0x07<<8)                        ,
  (0x08<<8)            +U3_FLAG         ,
  (0x09<<8)            +U3_FLAG+V_FLAG      ,
  (0x10<<8)         +H_FLAG            ,
  (0x11<<8)         +H_FLAG   +V_FLAG      ,
  (0x12<<8)         +H_FLAG   +V_FLAG      ,
  (0x13<<8)         +H_FLAG            ,
  (0x14<<8)         +H_FLAG   +V_FLAG      ,
  (0x15<<8)         +H_FLAG            ,
  (0x10<<8)                        ,
  (0x11<<8)               +V_FLAG      ,
  (0x12<<8)               +V_FLAG      ,
  (0x13<<8)                        ,
  (0x14<<8)               +V_FLAG      ,
  (0x15<<8)                        ,
  (0x16<<8)                        ,
  (0x17<<8)               +V_FLAG      ,
  (0x18<<8)            +U3_FLAG+V_FLAG      ,
  (0x19<<8)            +U3_FLAG         ,
  (0x20<<8)      +U5_FLAG+H_FLAG            ,
  (0x21<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x22<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x23<<8)      +U5_FLAG+H_FLAG            ,
  (0x24<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x25<<8)      +U5_FLAG+H_FLAG            ,
  (0x20<<8)      +U5_FLAG               ,
  (0x21<<8)      +U5_FLAG      +V_FLAG      ,
  (0x22<<8)      +U5_FLAG      +V_FLAG      ,
  (0x23<<8)      +U5_FLAG               ,
  (0x24<<8)      +U5_FLAG      +V_FLAG      ,
  (0x25<<8)      +U5_FLAG               ,
  (0x26<<8)      +U5_FLAG               ,
  (0x27<<8)      +U5_FLAG      +V_FLAG      ,
  (0x28<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x29<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x30<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x31<<8)      +U5_FLAG+H_FLAG            ,
  (0x32<<8)      +U5_FLAG+H_FLAG            ,
  (0x33<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x34<<8)      +U5_FLAG+H_FLAG            ,
  (0x35<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x30<<8)      +U5_FLAG      +V_FLAG      ,
  (0x31<<8)      +U5_FLAG               ,
  (0x32<<8)      +U5_FLAG               ,
  (0x33<<8)      +U5_FLAG      +V_FLAG      ,
  (0x34<<8)      +U5_FLAG               ,
  (0x35<<8)      +U5_FLAG      +V_FLAG      ,
  (0x36<<8)      +U5_FLAG      +V_FLAG      ,
  (0x37<<8)      +U5_FLAG               ,
  (0x38<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x40<<8)         +H_FLAG            ,
  (0x41<<8)         +H_FLAG   +V_FLAG      ,
  (0x42<<8)         +H_FLAG   +V_FLAG      ,
  (0x43<<8)         +H_FLAG            ,
  (0x44<<8)         +H_FLAG   +V_FLAG      ,
  (0x45<<8)         +H_FLAG            ,
  (0x40<<8)                        ,
  (0x41<<8)               +V_FLAG      ,
  (0x42<<8)               +V_FLAG      ,
  (0x43<<8)                        ,
  (0x44<<8)               +V_FLAG      ,
  (0x45<<8)                        ,
  (0x46<<8)                        ,
  (0x47<<8)               +V_FLAG      ,
  (0x48<<8)            +U3_FLAG+V_FLAG      ,
  (0x49<<8)            +U3_FLAG         ,
  (0x50<<8)         +H_FLAG   +V_FLAG      ,
  (0x51<<8)         +H_FLAG            ,
  (0x52<<8)         +H_FLAG            ,
  (0x53<<8)         +H_FLAG   +V_FLAG      ,
  (0x54<<8)         +H_FLAG            ,
  (0x55<<8)         +H_FLAG   +V_FLAG      ,
  (0x50<<8)               +V_FLAG      ,
  (0x51<<8)                        ,
  (0x52<<8)                        ,
  (0x53<<8)               +V_FLAG      ,
  (0x54<<8)                        ,
  (0x55<<8)               +V_FLAG      ,
  (0x56<<8)               +V_FLAG      ,
  (0x57<<8)                        ,
  (0x58<<8)            +U3_FLAG         ,
  (0x59<<8)            +U3_FLAG+V_FLAG      ,
  (0x60<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x61<<8)      +U5_FLAG+H_FLAG            ,
  (0x62<<8)      +U5_FLAG+H_FLAG            ,
  (0x63<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x64<<8)      +U5_FLAG+H_FLAG            ,
  (0x65<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x60<<8)      +U5_FLAG      +V_FLAG      ,
  (0x61<<8)      +U5_FLAG               ,
  (0x62<<8)      +U5_FLAG               ,
  (0x63<<8)      +U5_FLAG      +V_FLAG      ,
  (0x64<<8)      +U5_FLAG               ,
  (0x65<<8)      +U5_FLAG      +V_FLAG      ,
  (0x66<<8)      +U5_FLAG      +V_FLAG      ,
  (0x67<<8)      +U5_FLAG               ,
  (0x68<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x69<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x70<<8)      +U5_FLAG+H_FLAG            ,
  (0x71<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x72<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x73<<8)      +U5_FLAG+H_FLAG            ,
  (0x74<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x75<<8)      +U5_FLAG+H_FLAG            ,
  (0x70<<8)      +U5_FLAG               ,
  (0x71<<8)      +U5_FLAG      +V_FLAG      ,
  (0x72<<8)      +U5_FLAG      +V_FLAG      ,
  (0x73<<8)      +U5_FLAG               ,
  (0x74<<8)      +U5_FLAG      +V_FLAG      ,
  (0x75<<8)      +U5_FLAG               ,
  (0x76<<8)      +U5_FLAG               ,
  (0x77<<8)      +U5_FLAG      +V_FLAG      ,
  (0x78<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x79<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x80<<8)+S_FLAG      +H_FLAG            ,
  (0x81<<8)+S_FLAG      +H_FLAG   +V_FLAG      ,
  (0x82<<8)+S_FLAG      +H_FLAG   +V_FLAG      ,
  (0x83<<8)+S_FLAG      +H_FLAG            ,
  (0x84<<8)+S_FLAG      +H_FLAG   +V_FLAG      ,
  (0x85<<8)+S_FLAG      +H_FLAG            ,
  (0x80<<8)+S_FLAG                     ,
  (0x81<<8)+S_FLAG            +V_FLAG      ,
  (0x82<<8)+S_FLAG            +V_FLAG      ,
  (0x83<<8)+S_FLAG                     ,
  (0x84<<8)+S_FLAG            +V_FLAG      ,
  (0x85<<8)+S_FLAG                     ,
  (0x86<<8)+S_FLAG                     ,
  (0x87<<8)+S_FLAG            +V_FLAG      ,
  (0x88<<8)+S_FLAG         +U3_FLAG+V_FLAG      ,
  (0x89<<8)+S_FLAG         +U3_FLAG         ,
  (0x90<<8)+S_FLAG      +H_FLAG   +V_FLAG      ,
  (0x91<<8)+S_FLAG      +H_FLAG            ,
  (0x92<<8)+S_FLAG      +H_FLAG            ,
  (0x93<<8)+S_FLAG      +H_FLAG   +V_FLAG      ,
  (0x94<<8)+S_FLAG      +H_FLAG            ,
  (0x95<<8)+S_FLAG      +H_FLAG   +V_FLAG      ,
  (0x90<<8)+S_FLAG            +V_FLAG      ,
  (0x91<<8)+S_FLAG                     ,
  (0x92<<8)+S_FLAG                     ,
  (0x93<<8)+S_FLAG            +V_FLAG      ,
  (0x94<<8)+S_FLAG                     ,
  (0x95<<8)+S_FLAG            +V_FLAG      ,
  (0x96<<8)+S_FLAG            +V_FLAG      ,
  (0x97<<8)+S_FLAG                     ,
  (0x98<<8)+S_FLAG         +U3_FLAG         ,
  (0x99<<8)+S_FLAG         +U3_FLAG+V_FLAG      ,
  (0x00<<8)   +Z_FLAG   +H_FLAG   +V_FLAG   +C_FLAG,
  (0x01<<8)         +H_FLAG         +C_FLAG,
  (0x02<<8)         +H_FLAG         +C_FLAG,
  (0x03<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x04<<8)         +H_FLAG         +C_FLAG,
  (0x05<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x00<<8)   +Z_FLAG         +V_FLAG   +C_FLAG,
  (0x01<<8)                     +C_FLAG,
  (0x02<<8)                     +C_FLAG,
  (0x03<<8)               +V_FLAG   +C_FLAG,
  (0x04<<8)                     +C_FLAG,
  (0x05<<8)               +V_FLAG   +C_FLAG,
  (0x06<<8)               +V_FLAG   +C_FLAG,
  (0x07<<8)                     +C_FLAG,
  (0x08<<8)            +U3_FLAG      +C_FLAG,
  (0x09<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x10<<8)         +H_FLAG         +C_FLAG,
  (0x11<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x12<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x13<<8)         +H_FLAG         +C_FLAG,
  (0x14<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x15<<8)         +H_FLAG         +C_FLAG,
  (0x10<<8)                     +C_FLAG,
  (0x11<<8)               +V_FLAG   +C_FLAG,
  (0x12<<8)               +V_FLAG   +C_FLAG,
  (0x13<<8)                     +C_FLAG,
  (0x14<<8)               +V_FLAG   +C_FLAG,
  (0x15<<8)                     +C_FLAG,
  (0x16<<8)                     +C_FLAG,
  (0x17<<8)               +V_FLAG   +C_FLAG,
  (0x18<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x19<<8)            +U3_FLAG      +C_FLAG,
  (0x20<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x21<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x22<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x23<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x24<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x25<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x20<<8)      +U5_FLAG            +C_FLAG,
  (0x21<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x22<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x23<<8)      +U5_FLAG            +C_FLAG,
  (0x24<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x25<<8)      +U5_FLAG            +C_FLAG,
  (0x26<<8)      +U5_FLAG            +C_FLAG,
  (0x27<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x28<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x29<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x30<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x31<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x32<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x33<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x34<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x35<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x30<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x31<<8)      +U5_FLAG            +C_FLAG,
  (0x32<<8)      +U5_FLAG            +C_FLAG,
  (0x33<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x34<<8)      +U5_FLAG            +C_FLAG,
  (0x35<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x36<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x37<<8)      +U5_FLAG            +C_FLAG,
  (0x38<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x40<<8)         +H_FLAG         +C_FLAG,
  (0x41<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x42<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x43<<8)         +H_FLAG         +C_FLAG,
  (0x44<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x45<<8)         +H_FLAG         +C_FLAG,
  (0x40<<8)                     +C_FLAG,
  (0x41<<8)               +V_FLAG   +C_FLAG,
  (0x42<<8)               +V_FLAG   +C_FLAG,
  (0x43<<8)                     +C_FLAG,
  (0x44<<8)               +V_FLAG   +C_FLAG,
  (0x45<<8)                     +C_FLAG,
  (0x46<<8)                     +C_FLAG,
  (0x47<<8)               +V_FLAG   +C_FLAG,
  (0x48<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x49<<8)            +U3_FLAG      +C_FLAG,
  (0x50<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x51<<8)         +H_FLAG         +C_FLAG,
  (0x52<<8)         +H_FLAG         +C_FLAG,
  (0x53<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x54<<8)         +H_FLAG         +C_FLAG,
  (0x55<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x50<<8)               +V_FLAG   +C_FLAG,
  (0x51<<8)                     +C_FLAG,
  (0x52<<8)                     +C_FLAG,
  (0x53<<8)               +V_FLAG   +C_FLAG,
  (0x54<<8)                     +C_FLAG,
  (0x55<<8)               +V_FLAG   +C_FLAG,
  (0x56<<8)               +V_FLAG   +C_FLAG,
  (0x57<<8)                     +C_FLAG,
  (0x58<<8)            +U3_FLAG      +C_FLAG,
  (0x59<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x60<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x61<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x62<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x63<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x64<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x65<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x60<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x61<<8)      +U5_FLAG            +C_FLAG,
  (0x62<<8)      +U5_FLAG            +C_FLAG,
  (0x63<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x64<<8)      +U5_FLAG            +C_FLAG,
  (0x65<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x66<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x67<<8)      +U5_FLAG            +C_FLAG,
  (0x68<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x69<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x70<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x71<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x72<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x73<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x74<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x75<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x70<<8)      +U5_FLAG            +C_FLAG,
  (0x71<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x72<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x73<<8)      +U5_FLAG            +C_FLAG,
  (0x74<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x75<<8)      +U5_FLAG            +C_FLAG,
  (0x76<<8)      +U5_FLAG            +C_FLAG,
  (0x77<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x78<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x79<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x80<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0x81<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0x82<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0x83<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0x84<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0x85<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0x80<<8)+S_FLAG                  +C_FLAG,
  (0x81<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0x82<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0x83<<8)+S_FLAG                  +C_FLAG,
  (0x84<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0x85<<8)+S_FLAG                  +C_FLAG,
  (0x86<<8)+S_FLAG                  +C_FLAG,
  (0x87<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0x88<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0x89<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0x90<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0x91<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0x92<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0x93<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0x94<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0x95<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0x90<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0x91<<8)+S_FLAG                  +C_FLAG,
  (0x92<<8)+S_FLAG                  +C_FLAG,
  (0x93<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0x94<<8)+S_FLAG                  +C_FLAG,
  (0x95<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0x96<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0x97<<8)+S_FLAG                  +C_FLAG,
  (0x98<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0x99<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0xA0<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xA1<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xA2<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xA3<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xA4<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xA5<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xA0<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xA1<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xA2<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xA3<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xA4<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xA5<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xA6<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xA7<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xA8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xA9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xB0<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xB1<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xB2<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xB3<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xB4<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xB5<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xB0<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xB1<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xB2<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xB3<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xB4<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xB5<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xB6<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xB7<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xB8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xB9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xC0<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0xC1<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0xC2<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0xC3<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0xC4<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0xC5<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0xC0<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0xC1<<8)+S_FLAG                  +C_FLAG,
  (0xC2<<8)+S_FLAG                  +C_FLAG,
  (0xC3<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0xC4<<8)+S_FLAG                  +C_FLAG,
  (0xC5<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0xC6<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0xC7<<8)+S_FLAG                  +C_FLAG,
  (0xC8<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0xC9<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0xD0<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0xD1<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0xD2<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0xD3<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0xD4<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0xD5<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0xD0<<8)+S_FLAG                  +C_FLAG,
  (0xD1<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0xD2<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0xD3<<8)+S_FLAG                  +C_FLAG,
  (0xD4<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0xD5<<8)+S_FLAG                  +C_FLAG,
  (0xD6<<8)+S_FLAG                  +C_FLAG,
  (0xD7<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0xD8<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0xD9<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0xE0<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xE1<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xE2<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xE3<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xE4<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xE5<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xE0<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xE1<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xE2<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xE3<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xE4<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xE5<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xE6<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xE7<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xE8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xE9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xF0<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xF1<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xF2<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xF3<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xF4<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xF5<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xF0<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xF1<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xF2<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xF3<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xF4<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xF5<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xF6<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xF7<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xF8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xF9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x00<<8)   +Z_FLAG   +H_FLAG   +V_FLAG   +C_FLAG,
  (0x01<<8)         +H_FLAG         +C_FLAG,
  (0x02<<8)         +H_FLAG         +C_FLAG,
  (0x03<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x04<<8)         +H_FLAG         +C_FLAG,
  (0x05<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x00<<8)   +Z_FLAG         +V_FLAG   +C_FLAG,
  (0x01<<8)                     +C_FLAG,
  (0x02<<8)                     +C_FLAG,
  (0x03<<8)               +V_FLAG   +C_FLAG,
  (0x04<<8)                     +C_FLAG,
  (0x05<<8)               +V_FLAG   +C_FLAG,
  (0x06<<8)               +V_FLAG   +C_FLAG,
  (0x07<<8)                     +C_FLAG,
  (0x08<<8)            +U3_FLAG      +C_FLAG,
  (0x09<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x10<<8)         +H_FLAG         +C_FLAG,
  (0x11<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x12<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x13<<8)         +H_FLAG         +C_FLAG,
  (0x14<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x15<<8)         +H_FLAG         +C_FLAG,
  (0x10<<8)                     +C_FLAG,
  (0x11<<8)               +V_FLAG   +C_FLAG,
  (0x12<<8)               +V_FLAG   +C_FLAG,
  (0x13<<8)                     +C_FLAG,
  (0x14<<8)               +V_FLAG   +C_FLAG,
  (0x15<<8)                     +C_FLAG,
  (0x16<<8)                     +C_FLAG,
  (0x17<<8)               +V_FLAG   +C_FLAG,
  (0x18<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x19<<8)            +U3_FLAG      +C_FLAG,
  (0x20<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x21<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x22<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x23<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x24<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x25<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x20<<8)      +U5_FLAG            +C_FLAG,
  (0x21<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x22<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x23<<8)      +U5_FLAG            +C_FLAG,
  (0x24<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x25<<8)      +U5_FLAG            +C_FLAG,
  (0x26<<8)      +U5_FLAG            +C_FLAG,
  (0x27<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x28<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x29<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x30<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x31<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x32<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x33<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x34<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x35<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x30<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x31<<8)      +U5_FLAG            +C_FLAG,
  (0x32<<8)      +U5_FLAG            +C_FLAG,
  (0x33<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x34<<8)      +U5_FLAG            +C_FLAG,
  (0x35<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x36<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x37<<8)      +U5_FLAG            +C_FLAG,
  (0x38<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x40<<8)         +H_FLAG         +C_FLAG,
  (0x41<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x42<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x43<<8)         +H_FLAG         +C_FLAG,
  (0x44<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x45<<8)         +H_FLAG         +C_FLAG,
  (0x40<<8)                     +C_FLAG,
  (0x41<<8)               +V_FLAG   +C_FLAG,
  (0x42<<8)               +V_FLAG   +C_FLAG,
  (0x43<<8)                     +C_FLAG,
  (0x44<<8)               +V_FLAG   +C_FLAG,
  (0x45<<8)                     +C_FLAG,
  (0x46<<8)                     +C_FLAG,
  (0x47<<8)               +V_FLAG   +C_FLAG,
  (0x48<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x49<<8)            +U3_FLAG      +C_FLAG,
  (0x50<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x51<<8)         +H_FLAG         +C_FLAG,
  (0x52<<8)         +H_FLAG         +C_FLAG,
  (0x53<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x54<<8)         +H_FLAG         +C_FLAG,
  (0x55<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x50<<8)               +V_FLAG   +C_FLAG,
  (0x51<<8)                     +C_FLAG,
  (0x52<<8)                     +C_FLAG,
  (0x53<<8)               +V_FLAG   +C_FLAG,
  (0x54<<8)                     +C_FLAG,
  (0x55<<8)               +V_FLAG   +C_FLAG,
  (0x56<<8)               +V_FLAG   +C_FLAG,
  (0x57<<8)                     +C_FLAG,
  (0x58<<8)            +U3_FLAG      +C_FLAG,
  (0x59<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x60<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x61<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x62<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x63<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x64<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x65<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x06<<8)               +V_FLAG      ,
  (0x07<<8)                        ,
  (0x08<<8)            +U3_FLAG         ,
  (0x09<<8)            +U3_FLAG+V_FLAG      ,
  (0x0A<<8)            +U3_FLAG+V_FLAG      ,
  (0x0B<<8)            +U3_FLAG         ,
  (0x0C<<8)            +U3_FLAG+V_FLAG      ,
  (0x0D<<8)            +U3_FLAG         ,
  (0x0E<<8)            +U3_FLAG         ,
  (0x0F<<8)            +U3_FLAG+V_FLAG      ,
  (0x10<<8)         +H_FLAG            ,
  (0x11<<8)         +H_FLAG   +V_FLAG      ,
  (0x12<<8)         +H_FLAG   +V_FLAG      ,
  (0x13<<8)         +H_FLAG            ,
  (0x14<<8)         +H_FLAG   +V_FLAG      ,
  (0x15<<8)         +H_FLAG            ,
  (0x16<<8)                        ,
  (0x17<<8)               +V_FLAG      ,
  (0x18<<8)            +U3_FLAG+V_FLAG      ,
  (0x19<<8)            +U3_FLAG         ,
  (0x1A<<8)            +U3_FLAG         ,
  (0x1B<<8)            +U3_FLAG+V_FLAG      ,
  (0x1C<<8)            +U3_FLAG         ,
  (0x1D<<8)            +U3_FLAG+V_FLAG      ,
  (0x1E<<8)            +U3_FLAG+V_FLAG      ,
  (0x1F<<8)            +U3_FLAG         ,
  (0x20<<8)      +U5_FLAG+H_FLAG            ,
  (0x21<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x22<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x23<<8)      +U5_FLAG+H_FLAG            ,
  (0x24<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x25<<8)      +U5_FLAG+H_FLAG            ,
  (0x26<<8)      +U5_FLAG               ,
  (0x27<<8)      +U5_FLAG      +V_FLAG      ,
  (0x28<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x29<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x2A<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x2B<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x2C<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x2D<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x2E<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x2F<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x30<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x31<<8)      +U5_FLAG+H_FLAG            ,
  (0x32<<8)      +U5_FLAG+H_FLAG            ,
  (0x33<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x34<<8)      +U5_FLAG+H_FLAG            ,
  (0x35<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x36<<8)      +U5_FLAG      +V_FLAG      ,
  (0x37<<8)      +U5_FLAG               ,
  (0x38<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x3A<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x3B<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x3C<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x3D<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x3E<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x3F<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x40<<8)         +H_FLAG            ,
  (0x41<<8)         +H_FLAG   +V_FLAG      ,
  (0x42<<8)         +H_FLAG   +V_FLAG      ,
  (0x43<<8)         +H_FLAG            ,
  (0x44<<8)         +H_FLAG   +V_FLAG      ,
  (0x45<<8)         +H_FLAG            ,
  (0x46<<8)                        ,
  (0x47<<8)               +V_FLAG      ,
  (0x48<<8)            +U3_FLAG+V_FLAG      ,
  (0x49<<8)            +U3_FLAG         ,
  (0x4A<<8)            +U3_FLAG         ,
  (0x4B<<8)            +U3_FLAG+V_FLAG      ,
  (0x4C<<8)            +U3_FLAG         ,
  (0x4D<<8)            +U3_FLAG+V_FLAG      ,
  (0x4E<<8)            +U3_FLAG+V_FLAG      ,
  (0x4F<<8)            +U3_FLAG         ,
  (0x50<<8)         +H_FLAG   +V_FLAG      ,
  (0x51<<8)         +H_FLAG            ,
  (0x52<<8)         +H_FLAG            ,
  (0x53<<8)         +H_FLAG   +V_FLAG      ,
  (0x54<<8)         +H_FLAG            ,
  (0x55<<8)         +H_FLAG   +V_FLAG      ,
  (0x56<<8)               +V_FLAG      ,
  (0x57<<8)                        ,
  (0x58<<8)            +U3_FLAG         ,
  (0x59<<8)            +U3_FLAG+V_FLAG      ,
  (0x5A<<8)            +U3_FLAG+V_FLAG      ,
  (0x5B<<8)            +U3_FLAG         ,
  (0x5C<<8)            +U3_FLAG+V_FLAG      ,
  (0x5D<<8)            +U3_FLAG         ,
  (0x5E<<8)            +U3_FLAG         ,
  (0x5F<<8)            +U3_FLAG+V_FLAG      ,
  (0x60<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x61<<8)      +U5_FLAG+H_FLAG            ,
  (0x62<<8)      +U5_FLAG+H_FLAG            ,
  (0x63<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x64<<8)      +U5_FLAG+H_FLAG            ,
  (0x65<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x66<<8)      +U5_FLAG      +V_FLAG      ,
  (0x67<<8)      +U5_FLAG               ,
  (0x68<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x69<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x6A<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x6B<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x6C<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x6D<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x6E<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x6F<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x70<<8)      +U5_FLAG+H_FLAG            ,
  (0x71<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x72<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x73<<8)      +U5_FLAG+H_FLAG            ,
  (0x74<<8)      +U5_FLAG+H_FLAG   +V_FLAG      ,
  (0x75<<8)      +U5_FLAG+H_FLAG            ,
  (0x76<<8)      +U5_FLAG               ,
  (0x77<<8)      +U5_FLAG      +V_FLAG      ,
  (0x78<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x79<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x7A<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x7B<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x7C<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x7D<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x7E<<8)      +U5_FLAG   +U3_FLAG+V_FLAG      ,
  (0x7F<<8)      +U5_FLAG   +U3_FLAG         ,
  (0x80<<8)+S_FLAG      +H_FLAG            ,
  (0x81<<8)+S_FLAG      +H_FLAG   +V_FLAG      ,
  (0x82<<8)+S_FLAG      +H_FLAG   +V_FLAG      ,
  (0x83<<8)+S_FLAG      +H_FLAG            ,
  (0x84<<8)+S_FLAG      +H_FLAG   +V_FLAG      ,
  (0x85<<8)+S_FLAG      +H_FLAG            ,
  (0x86<<8)+S_FLAG                     ,
  (0x87<<8)+S_FLAG            +V_FLAG      ,
  (0x88<<8)+S_FLAG         +U3_FLAG+V_FLAG      ,
  (0x89<<8)+S_FLAG         +U3_FLAG         ,
  (0x8A<<8)+S_FLAG         +U3_FLAG         ,
  (0x8B<<8)+S_FLAG         +U3_FLAG+V_FLAG      ,
  (0x8C<<8)+S_FLAG         +U3_FLAG         ,
  (0x8D<<8)+S_FLAG         +U3_FLAG+V_FLAG      ,
  (0x8E<<8)+S_FLAG         +U3_FLAG+V_FLAG      ,
  (0x8F<<8)+S_FLAG         +U3_FLAG         ,
  (0x90<<8)+S_FLAG      +H_FLAG   +V_FLAG      ,
  (0x91<<8)+S_FLAG      +H_FLAG            ,
  (0x92<<8)+S_FLAG      +H_FLAG            ,
  (0x93<<8)+S_FLAG      +H_FLAG   +V_FLAG      ,
  (0x94<<8)+S_FLAG      +H_FLAG            ,
  (0x95<<8)+S_FLAG      +H_FLAG   +V_FLAG      ,
  (0x96<<8)+S_FLAG            +V_FLAG      ,
  (0x97<<8)+S_FLAG                     ,
  (0x98<<8)+S_FLAG         +U3_FLAG         ,
  (0x99<<8)+S_FLAG         +U3_FLAG+V_FLAG      ,
  (0x9A<<8)+S_FLAG         +U3_FLAG+V_FLAG      ,
  (0x9B<<8)+S_FLAG         +U3_FLAG         ,
  (0x9C<<8)+S_FLAG         +U3_FLAG+V_FLAG      ,
  (0x9D<<8)+S_FLAG         +U3_FLAG         ,
  (0x9E<<8)+S_FLAG         +U3_FLAG         ,
  (0x9F<<8)+S_FLAG         +U3_FLAG+V_FLAG      ,
  (0x00<<8)   +Z_FLAG   +H_FLAG   +V_FLAG   +C_FLAG,
  (0x01<<8)         +H_FLAG         +C_FLAG,
  (0x02<<8)         +H_FLAG         +C_FLAG,
  (0x03<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x04<<8)         +H_FLAG         +C_FLAG,
  (0x05<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x06<<8)               +V_FLAG   +C_FLAG,
  (0x07<<8)                     +C_FLAG,
  (0x08<<8)            +U3_FLAG      +C_FLAG,
  (0x09<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x0A<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x0B<<8)            +U3_FLAG      +C_FLAG,
  (0x0C<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x0D<<8)            +U3_FLAG      +C_FLAG,
  (0x0E<<8)            +U3_FLAG      +C_FLAG,
  (0x0F<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x10<<8)         +H_FLAG         +C_FLAG,
  (0x11<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x12<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x13<<8)         +H_FLAG         +C_FLAG,
  (0x14<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x15<<8)         +H_FLAG         +C_FLAG,
  (0x16<<8)                     +C_FLAG,
  (0x17<<8)               +V_FLAG   +C_FLAG,
  (0x18<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x19<<8)            +U3_FLAG      +C_FLAG,
  (0x1A<<8)            +U3_FLAG      +C_FLAG,
  (0x1B<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x1C<<8)            +U3_FLAG      +C_FLAG,
  (0x1D<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x1E<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x1F<<8)            +U3_FLAG      +C_FLAG,
  (0x20<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x21<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x22<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x23<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x24<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x25<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x26<<8)      +U5_FLAG            +C_FLAG,
  (0x27<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x28<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x29<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x2A<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x2B<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x2C<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x2D<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x2E<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x2F<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x30<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x31<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x32<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x33<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x34<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x35<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x36<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x37<<8)      +U5_FLAG            +C_FLAG,
  (0x38<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x3A<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x3B<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x3C<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x3D<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x3E<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x3F<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x40<<8)         +H_FLAG         +C_FLAG,
  (0x41<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x42<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x43<<8)         +H_FLAG         +C_FLAG,
  (0x44<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x45<<8)         +H_FLAG         +C_FLAG,
  (0x46<<8)                     +C_FLAG,
  (0x47<<8)               +V_FLAG   +C_FLAG,
  (0x48<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x49<<8)            +U3_FLAG      +C_FLAG,
  (0x4A<<8)            +U3_FLAG      +C_FLAG,
  (0x4B<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x4C<<8)            +U3_FLAG      +C_FLAG,
  (0x4D<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x4E<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x4F<<8)            +U3_FLAG      +C_FLAG,
  (0x50<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x51<<8)         +H_FLAG         +C_FLAG,
  (0x52<<8)         +H_FLAG         +C_FLAG,
  (0x53<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x54<<8)         +H_FLAG         +C_FLAG,
  (0x55<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x56<<8)               +V_FLAG   +C_FLAG,
  (0x57<<8)                     +C_FLAG,
  (0x58<<8)            +U3_FLAG      +C_FLAG,
  (0x59<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x5A<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x5B<<8)            +U3_FLAG      +C_FLAG,
  (0x5C<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x5D<<8)            +U3_FLAG      +C_FLAG,
  (0x5E<<8)            +U3_FLAG      +C_FLAG,
  (0x5F<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x60<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x61<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x62<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x63<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x64<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x65<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x66<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x67<<8)      +U5_FLAG            +C_FLAG,
  (0x68<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x69<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x6A<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x6B<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x6C<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x6D<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x6E<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x6F<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x70<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x71<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x72<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x73<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x74<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x75<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x76<<8)      +U5_FLAG            +C_FLAG,
  (0x77<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x78<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x79<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x7A<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x7B<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x7C<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x7D<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x7E<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x7F<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x80<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0x81<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0x82<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0x83<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0x84<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0x85<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0x86<<8)+S_FLAG                  +C_FLAG,
  (0x87<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0x88<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0x89<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0x8A<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0x8B<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0x8C<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0x8D<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0x8E<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0x8F<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0x90<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0x91<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0x92<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0x93<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0x94<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0x95<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0x96<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0x97<<8)+S_FLAG                  +C_FLAG,
  (0x98<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0x99<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0x9A<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0x9B<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0x9C<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0x9D<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0x9E<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0x9F<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0xA0<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xA1<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xA2<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xA3<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xA4<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xA5<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xA6<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xA7<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xA8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xA9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xAA<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xAB<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xAC<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xAD<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xAE<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xAF<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xB0<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xB1<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xB2<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xB3<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xB4<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xB5<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xB6<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xB7<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xB8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xB9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xBA<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xBB<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xBC<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xBD<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xBE<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xBF<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xC0<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0xC1<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0xC2<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0xC3<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0xC4<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0xC5<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0xC6<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0xC7<<8)+S_FLAG                  +C_FLAG,
  (0xC8<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0xC9<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0xCA<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0xCB<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0xCC<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0xCD<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0xCE<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0xCF<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0xD0<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0xD1<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0xD2<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0xD3<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0xD4<<8)+S_FLAG      +H_FLAG   +V_FLAG   +C_FLAG,
  (0xD5<<8)+S_FLAG      +H_FLAG         +C_FLAG,
  (0xD6<<8)+S_FLAG                  +C_FLAG,
  (0xD7<<8)+S_FLAG            +V_FLAG   +C_FLAG,
  (0xD8<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0xD9<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0xDA<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0xDB<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0xDC<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0xDD<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0xDE<<8)+S_FLAG         +U3_FLAG+V_FLAG   +C_FLAG,
  (0xDF<<8)+S_FLAG         +U3_FLAG      +C_FLAG,
  (0xE0<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xE1<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xE2<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xE3<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xE4<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xE5<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xE6<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xE7<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xE8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xE9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xEA<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xEB<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xEC<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xED<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xEE<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xEF<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xf0<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xf1<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xf2<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xf3<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xf4<<8)+S_FLAG   +U5_FLAG+H_FLAG         +C_FLAG,
  (0xf5<<8)+S_FLAG   +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0xf6<<8)+S_FLAG   +U5_FLAG      +V_FLAG   +C_FLAG,
  (0xf7<<8)+S_FLAG   +U5_FLAG            +C_FLAG,
  (0xf8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xf9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xfA<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xfB<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xfC<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0xfD<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xfE<<8)+S_FLAG   +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0xfF<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x00<<8)   +Z_FLAG   +H_FLAG   +V_FLAG   +C_FLAG,
  (0x01<<8)         +H_FLAG         +C_FLAG,
  (0x02<<8)         +H_FLAG         +C_FLAG,
  (0x03<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x04<<8)         +H_FLAG         +C_FLAG,
  (0x05<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x06<<8)               +V_FLAG   +C_FLAG,
  (0x07<<8)                     +C_FLAG,
  (0x08<<8)            +U3_FLAG      +C_FLAG,
  (0x09<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x0A<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x0B<<8)            +U3_FLAG      +C_FLAG,
  (0x0C<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x0D<<8)            +U3_FLAG      +C_FLAG,
  (0x0E<<8)            +U3_FLAG      +C_FLAG,
  (0x0F<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x10<<8)         +H_FLAG         +C_FLAG,
  (0x11<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x12<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x13<<8)         +H_FLAG         +C_FLAG,
  (0x14<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x15<<8)         +H_FLAG         +C_FLAG,
  (0x16<<8)                     +C_FLAG,
  (0x17<<8)               +V_FLAG   +C_FLAG,
  (0x18<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x19<<8)            +U3_FLAG      +C_FLAG,
  (0x1A<<8)            +U3_FLAG      +C_FLAG,
  (0x1B<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x1C<<8)            +U3_FLAG      +C_FLAG,
  (0x1D<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x1E<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x1F<<8)            +U3_FLAG      +C_FLAG,
  (0x20<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x21<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x22<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x23<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x24<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x25<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x26<<8)      +U5_FLAG            +C_FLAG,
  (0x27<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x28<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x29<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x2A<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x2B<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x2C<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x2D<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x2E<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x2F<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x30<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x31<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x32<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x33<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x34<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x35<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x36<<8)      +U5_FLAG      +V_FLAG   +C_FLAG,
  (0x37<<8)      +U5_FLAG            +C_FLAG,
  (0x38<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x3A<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x3B<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x3C<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x3D<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x3E<<8)      +U5_FLAG   +U3_FLAG      +C_FLAG,
  (0x3F<<8)      +U5_FLAG   +U3_FLAG+V_FLAG   +C_FLAG,
  (0x40<<8)         +H_FLAG         +C_FLAG,
  (0x41<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x42<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x43<<8)         +H_FLAG         +C_FLAG,
  (0x44<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x45<<8)         +H_FLAG         +C_FLAG,
  (0x46<<8)                     +C_FLAG,
  (0x47<<8)               +V_FLAG   +C_FLAG,
  (0x48<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x49<<8)            +U3_FLAG      +C_FLAG,
  (0x4A<<8)            +U3_FLAG      +C_FLAG,
  (0x4B<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x4C<<8)            +U3_FLAG      +C_FLAG,
  (0x4D<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x4E<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x4F<<8)            +U3_FLAG      +C_FLAG,
  (0x50<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x51<<8)         +H_FLAG         +C_FLAG,
  (0x52<<8)         +H_FLAG         +C_FLAG,
  (0x53<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x54<<8)         +H_FLAG         +C_FLAG,
  (0x55<<8)         +H_FLAG   +V_FLAG   +C_FLAG,
  (0x56<<8)               +V_FLAG   +C_FLAG,
  (0x57<<8)                     +C_FLAG,
  (0x58<<8)            +U3_FLAG      +C_FLAG,
  (0x59<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x5A<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x5B<<8)            +U3_FLAG      +C_FLAG,
  (0x5C<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x5D<<8)            +U3_FLAG      +C_FLAG,
  (0x5E<<8)            +U3_FLAG      +C_FLAG,
  (0x5F<<8)            +U3_FLAG+V_FLAG   +C_FLAG,
  (0x60<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x61<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x62<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x63<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x64<<8)      +U5_FLAG+H_FLAG         +C_FLAG,
  (0x65<<8)      +U5_FLAG+H_FLAG   +V_FLAG   +C_FLAG,
  (0x00<<8)   +Z_FLAG         +V_FLAG+N_FLAG   ,
  (0x01<<8)                  +N_FLAG   ,
  (0x02<<8)                  +N_FLAG   ,
  (0x03<<8)               +V_FLAG+N_FLAG   ,
  (0x04<<8)                  +N_FLAG   ,
  (0x05<<8)               +V_FLAG+N_FLAG   ,
  (0x06<<8)               +V_FLAG+N_FLAG   ,
  (0x07<<8)                  +N_FLAG   ,
  (0x08<<8)            +U3_FLAG   +N_FLAG   ,
  (0x09<<8)            +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x04<<8)                  +N_FLAG   ,
  (0x05<<8)               +V_FLAG+N_FLAG   ,
  (0x06<<8)               +V_FLAG+N_FLAG   ,
  (0x07<<8)                  +N_FLAG   ,
  (0x08<<8)            +U3_FLAG   +N_FLAG   ,
  (0x09<<8)            +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x10<<8)                  +N_FLAG   ,
  (0x11<<8)               +V_FLAG+N_FLAG   ,
  (0x12<<8)               +V_FLAG+N_FLAG   ,
  (0x13<<8)                  +N_FLAG   ,
  (0x14<<8)               +V_FLAG+N_FLAG   ,
  (0x15<<8)                  +N_FLAG   ,
  (0x16<<8)                  +N_FLAG   ,
  (0x17<<8)               +V_FLAG+N_FLAG   ,
  (0x18<<8)            +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x19<<8)            +U3_FLAG   +N_FLAG   ,
  (0x14<<8)               +V_FLAG+N_FLAG   ,
  (0x15<<8)                  +N_FLAG   ,
  (0x16<<8)                  +N_FLAG   ,
  (0x17<<8)               +V_FLAG+N_FLAG   ,
  (0x18<<8)            +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x19<<8)            +U3_FLAG   +N_FLAG   ,
  (0x20<<8)      +U5_FLAG         +N_FLAG   ,
  (0x21<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x22<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x23<<8)      +U5_FLAG         +N_FLAG   ,
  (0x24<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x25<<8)      +U5_FLAG         +N_FLAG   ,
  (0x26<<8)      +U5_FLAG         +N_FLAG   ,
  (0x27<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x28<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x29<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG   ,
  (0x24<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x25<<8)      +U5_FLAG         +N_FLAG   ,
  (0x26<<8)      +U5_FLAG         +N_FLAG   ,
  (0x27<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x28<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x29<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG   ,
  (0x30<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x31<<8)      +U5_FLAG         +N_FLAG   ,
  (0x32<<8)      +U5_FLAG         +N_FLAG   ,
  (0x33<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x34<<8)      +U5_FLAG         +N_FLAG   ,
  (0x35<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x36<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x37<<8)      +U5_FLAG         +N_FLAG   ,
  (0x38<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG   ,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x34<<8)      +U5_FLAG         +N_FLAG   ,
  (0x35<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x36<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x37<<8)      +U5_FLAG         +N_FLAG   ,
  (0x38<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG   ,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x40<<8)                  +N_FLAG   ,
  (0x41<<8)               +V_FLAG+N_FLAG   ,
  (0x42<<8)               +V_FLAG+N_FLAG   ,
  (0x43<<8)                  +N_FLAG   ,
  (0x44<<8)               +V_FLAG+N_FLAG   ,
  (0x45<<8)                  +N_FLAG   ,
  (0x46<<8)                  +N_FLAG   ,
  (0x47<<8)               +V_FLAG+N_FLAG   ,
  (0x48<<8)            +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x49<<8)            +U3_FLAG   +N_FLAG   ,
  (0x44<<8)               +V_FLAG+N_FLAG   ,
  (0x45<<8)                  +N_FLAG   ,
  (0x46<<8)                  +N_FLAG   ,
  (0x47<<8)               +V_FLAG+N_FLAG   ,
  (0x48<<8)            +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x49<<8)            +U3_FLAG   +N_FLAG   ,
  (0x50<<8)               +V_FLAG+N_FLAG   ,
  (0x51<<8)                  +N_FLAG   ,
  (0x52<<8)                  +N_FLAG   ,
  (0x53<<8)               +V_FLAG+N_FLAG   ,
  (0x54<<8)                  +N_FLAG   ,
  (0x55<<8)               +V_FLAG+N_FLAG   ,
  (0x56<<8)               +V_FLAG+N_FLAG   ,
  (0x57<<8)                  +N_FLAG   ,
  (0x58<<8)            +U3_FLAG   +N_FLAG   ,
  (0x59<<8)            +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x54<<8)                  +N_FLAG   ,
  (0x55<<8)               +V_FLAG+N_FLAG   ,
  (0x56<<8)               +V_FLAG+N_FLAG   ,
  (0x57<<8)                  +N_FLAG   ,
  (0x58<<8)            +U3_FLAG   +N_FLAG   ,
  (0x59<<8)            +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x60<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x61<<8)      +U5_FLAG         +N_FLAG   ,
  (0x62<<8)      +U5_FLAG         +N_FLAG   ,
  (0x63<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x64<<8)      +U5_FLAG         +N_FLAG   ,
  (0x65<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x66<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x67<<8)      +U5_FLAG         +N_FLAG   ,
  (0x68<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG   ,
  (0x69<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x64<<8)      +U5_FLAG         +N_FLAG   ,
  (0x65<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x66<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x67<<8)      +U5_FLAG         +N_FLAG   ,
  (0x68<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG   ,
  (0x69<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x70<<8)      +U5_FLAG         +N_FLAG   ,
  (0x71<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x72<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x73<<8)      +U5_FLAG         +N_FLAG   ,
  (0x74<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x75<<8)      +U5_FLAG         +N_FLAG   ,
  (0x76<<8)      +U5_FLAG         +N_FLAG   ,
  (0x77<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x78<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x79<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG   ,
  (0x74<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x75<<8)      +U5_FLAG         +N_FLAG   ,
  (0x76<<8)      +U5_FLAG         +N_FLAG   ,
  (0x77<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x78<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x79<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG   ,
  (0x80<<8)+S_FLAG               +N_FLAG   ,
  (0x81<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x82<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x83<<8)+S_FLAG               +N_FLAG   ,
  (0x84<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x85<<8)+S_FLAG               +N_FLAG   ,
  (0x86<<8)+S_FLAG               +N_FLAG   ,
  (0x87<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x88<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x89<<8)+S_FLAG         +U3_FLAG   +N_FLAG   ,
  (0x84<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x85<<8)+S_FLAG               +N_FLAG   ,
  (0x86<<8)+S_FLAG               +N_FLAG   ,
  (0x87<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x88<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x89<<8)+S_FLAG         +U3_FLAG   +N_FLAG   ,
  (0x90<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x91<<8)+S_FLAG               +N_FLAG   ,
  (0x92<<8)+S_FLAG               +N_FLAG   ,
  (0x93<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x94<<8)+S_FLAG               +N_FLAG   ,
  (0x95<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x96<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x97<<8)+S_FLAG               +N_FLAG   ,
  (0x98<<8)+S_FLAG         +U3_FLAG   +N_FLAG   ,
  (0x99<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x34<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x35<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x36<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x37<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x38<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x40<<8)                  +N_FLAG+C_FLAG,
  (0x41<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x42<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x43<<8)                  +N_FLAG+C_FLAG,
  (0x44<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x45<<8)                  +N_FLAG+C_FLAG,
  (0x46<<8)                  +N_FLAG+C_FLAG,
  (0x47<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x48<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x49<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x44<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x45<<8)                  +N_FLAG+C_FLAG,
  (0x46<<8)                  +N_FLAG+C_FLAG,
  (0x47<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x48<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x49<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x50<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x51<<8)                  +N_FLAG+C_FLAG,
  (0x52<<8)                  +N_FLAG+C_FLAG,
  (0x53<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x54<<8)                  +N_FLAG+C_FLAG,
  (0x55<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x56<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x57<<8)                  +N_FLAG+C_FLAG,
  (0x58<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x59<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x54<<8)                  +N_FLAG+C_FLAG,
  (0x55<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x56<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x57<<8)                  +N_FLAG+C_FLAG,
  (0x58<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x59<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x60<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x61<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x62<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x63<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x64<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x65<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x66<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x67<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x68<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x69<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x64<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x65<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x66<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x67<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x68<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x69<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x70<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x71<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x72<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x73<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x74<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x75<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x76<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x77<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x78<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x79<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x74<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x75<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x76<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x77<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x78<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x79<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x80<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x81<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x82<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x83<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x84<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x85<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x86<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x87<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x88<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x89<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0x84<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x85<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x86<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x87<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x88<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x89<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0x90<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x91<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x92<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x93<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x94<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x95<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x96<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x97<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x98<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0x99<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x94<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x95<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x96<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x97<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x98<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0x99<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xA0<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xA1<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xA2<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xA3<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xA4<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xA5<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xA6<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xA7<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xA8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0xA9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xA4<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xA5<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xA6<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xA7<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xA8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0xA9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xB0<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xB1<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xB2<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xB3<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xB4<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xB5<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xB6<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xB7<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xB8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xB9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0xB4<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xB5<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xB6<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xB7<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xB8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xB9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0xC0<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xC1<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xC2<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xC3<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xC4<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xC5<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xC6<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xC7<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xC8<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0xC9<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xC4<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xC5<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xC6<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xC7<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xC8<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0xC9<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xD0<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xD1<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xD2<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xD3<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xD4<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xD5<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xD6<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xD7<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xD8<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xD9<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0xD4<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xD5<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xD6<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xD7<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xD8<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xD9<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0xE0<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xE1<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xE2<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xE3<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xE4<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xE5<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xE6<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xE7<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xE8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xE9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0xE4<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xE5<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xE6<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xE7<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xE8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xE9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0xf0<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xf1<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xf2<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xf3<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xf4<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xf5<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xf6<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xf7<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xf8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0xf9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xf4<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xf5<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xf6<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xf7<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xf8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0xf9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x00<<8)   +Z_FLAG         +V_FLAG+N_FLAG+C_FLAG,
  (0x01<<8)                  +N_FLAG+C_FLAG,
  (0x02<<8)                  +N_FLAG+C_FLAG,
  (0x03<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x04<<8)                  +N_FLAG+C_FLAG,
  (0x05<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x06<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x07<<8)                  +N_FLAG+C_FLAG,
  (0x08<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x09<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x04<<8)                  +N_FLAG+C_FLAG,
  (0x05<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x06<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x07<<8)                  +N_FLAG+C_FLAG,
  (0x08<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x09<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x10<<8)                  +N_FLAG+C_FLAG,
  (0x11<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x12<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x13<<8)                  +N_FLAG+C_FLAG,
  (0x14<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x15<<8)                  +N_FLAG+C_FLAG,
  (0x16<<8)                  +N_FLAG+C_FLAG,
  (0x17<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x18<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x19<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x14<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x15<<8)                  +N_FLAG+C_FLAG,
  (0x16<<8)                  +N_FLAG+C_FLAG,
  (0x17<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x18<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x19<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x20<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x21<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x22<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x23<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x24<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x25<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x26<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x27<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x28<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x29<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x24<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x25<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x26<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x27<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x28<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x29<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x30<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x31<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x32<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x33<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x34<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x35<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x36<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x37<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x38<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x34<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x35<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x36<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x37<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x38<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x40<<8)                  +N_FLAG+C_FLAG,
  (0x41<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x42<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x43<<8)                  +N_FLAG+C_FLAG,
  (0x44<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x45<<8)                  +N_FLAG+C_FLAG,
  (0x46<<8)                  +N_FLAG+C_FLAG,
  (0x47<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x48<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x49<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x44<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x45<<8)                  +N_FLAG+C_FLAG,
  (0x46<<8)                  +N_FLAG+C_FLAG,
  (0x47<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x48<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x49<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x50<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x51<<8)                  +N_FLAG+C_FLAG,
  (0x52<<8)                  +N_FLAG+C_FLAG,
  (0x53<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x54<<8)                  +N_FLAG+C_FLAG,
  (0x55<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x56<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x57<<8)                  +N_FLAG+C_FLAG,
  (0x58<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x59<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x54<<8)                  +N_FLAG+C_FLAG,
  (0x55<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x56<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x57<<8)                  +N_FLAG+C_FLAG,
  (0x58<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x59<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x60<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x61<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x62<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x63<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x64<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x65<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x66<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x67<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x68<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x69<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x64<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x65<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x66<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x67<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x68<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x69<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x70<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x71<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x72<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x73<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x74<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x75<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x76<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x77<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x78<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x79<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x74<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x75<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x76<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x77<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x78<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x79<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x80<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x81<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x82<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x83<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x84<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x85<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x86<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x87<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x88<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x89<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0x84<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x85<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x86<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x87<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x88<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x89<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0x90<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x91<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x92<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x93<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x94<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x95<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x96<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x97<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x98<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0x99<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x94<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x95<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x96<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x97<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x98<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0x99<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xfA<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0xfB<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0xfC<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0xfD<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0xfE<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0xfF<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x00<<8)   +Z_FLAG         +V_FLAG+N_FLAG   ,
  (0x01<<8)                  +N_FLAG   ,
  (0x02<<8)                  +N_FLAG   ,
  (0x03<<8)               +V_FLAG+N_FLAG   ,
  (0x04<<8)                  +N_FLAG   ,
  (0x05<<8)               +V_FLAG+N_FLAG   ,
  (0x06<<8)               +V_FLAG+N_FLAG   ,
  (0x07<<8)                  +N_FLAG   ,
  (0x08<<8)            +U3_FLAG   +N_FLAG   ,
  (0x09<<8)            +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x0A<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x0B<<8)         +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x0C<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x0D<<8)         +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x0E<<8)         +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x0F<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x10<<8)                  +N_FLAG   ,
  (0x11<<8)               +V_FLAG+N_FLAG   ,
  (0x12<<8)               +V_FLAG+N_FLAG   ,
  (0x13<<8)                  +N_FLAG   ,
  (0x14<<8)               +V_FLAG+N_FLAG   ,
  (0x15<<8)                  +N_FLAG   ,
  (0x16<<8)                  +N_FLAG   ,
  (0x17<<8)               +V_FLAG+N_FLAG   ,
  (0x18<<8)            +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x19<<8)            +U3_FLAG   +N_FLAG   ,
  (0x1A<<8)         +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x1B<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x1C<<8)         +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x1D<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x1E<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x1F<<8)         +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x20<<8)      +U5_FLAG         +N_FLAG   ,
  (0x21<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x22<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x23<<8)      +U5_FLAG         +N_FLAG   ,
  (0x24<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x25<<8)      +U5_FLAG         +N_FLAG   ,
  (0x26<<8)      +U5_FLAG         +N_FLAG   ,
  (0x27<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x28<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x29<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG   ,
  (0x2A<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x2B<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x2C<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x2D<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x2E<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x2F<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x30<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x31<<8)      +U5_FLAG         +N_FLAG   ,
  (0x32<<8)      +U5_FLAG         +N_FLAG   ,
  (0x33<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x34<<8)      +U5_FLAG         +N_FLAG   ,
  (0x35<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x36<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x37<<8)      +U5_FLAG         +N_FLAG   ,
  (0x38<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG   ,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x3A<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x3B<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x3C<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x3D<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x3E<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x3F<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x40<<8)                  +N_FLAG   ,
  (0x41<<8)               +V_FLAG+N_FLAG   ,
  (0x42<<8)               +V_FLAG+N_FLAG   ,
  (0x43<<8)                  +N_FLAG   ,
  (0x44<<8)               +V_FLAG+N_FLAG   ,
  (0x45<<8)                  +N_FLAG   ,
  (0x46<<8)                  +N_FLAG   ,
  (0x47<<8)               +V_FLAG+N_FLAG   ,
  (0x48<<8)            +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x49<<8)            +U3_FLAG   +N_FLAG   ,
  (0x4A<<8)         +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x4B<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x4C<<8)         +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x4D<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x4E<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x4F<<8)         +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x50<<8)               +V_FLAG+N_FLAG   ,
  (0x51<<8)                  +N_FLAG   ,
  (0x52<<8)                  +N_FLAG   ,
  (0x53<<8)               +V_FLAG+N_FLAG   ,
  (0x54<<8)                  +N_FLAG   ,
  (0x55<<8)               +V_FLAG+N_FLAG   ,
  (0x56<<8)               +V_FLAG+N_FLAG   ,
  (0x57<<8)                  +N_FLAG   ,
  (0x58<<8)            +U3_FLAG   +N_FLAG   ,
  (0x59<<8)            +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x5A<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x5B<<8)         +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x5C<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x5D<<8)         +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x5E<<8)         +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x5F<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x60<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x61<<8)      +U5_FLAG         +N_FLAG   ,
  (0x62<<8)      +U5_FLAG         +N_FLAG   ,
  (0x63<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x64<<8)      +U5_FLAG         +N_FLAG   ,
  (0x65<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x66<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x67<<8)      +U5_FLAG         +N_FLAG   ,
  (0x68<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG   ,
  (0x69<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x6A<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x6B<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x6C<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x6D<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x6E<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x6F<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x70<<8)      +U5_FLAG         +N_FLAG   ,
  (0x71<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x72<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x73<<8)      +U5_FLAG         +N_FLAG   ,
  (0x74<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x75<<8)      +U5_FLAG         +N_FLAG   ,
  (0x76<<8)      +U5_FLAG         +N_FLAG   ,
  (0x77<<8)      +U5_FLAG      +V_FLAG+N_FLAG   ,
  (0x78<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x79<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG   ,
  (0x7A<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x7B<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x7C<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x7D<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x7E<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x7F<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x80<<8)+S_FLAG               +N_FLAG   ,
  (0x81<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x82<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x83<<8)+S_FLAG               +N_FLAG   ,
  (0x84<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x85<<8)+S_FLAG               +N_FLAG   ,
  (0x86<<8)+S_FLAG               +N_FLAG   ,
  (0x87<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x88<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG   ,
  (0x89<<8)+S_FLAG         +U3_FLAG   +N_FLAG   ,
  (0x8A<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x8B<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x8C<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x8D<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x8E<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG   ,
  (0x8F<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG   ,
  (0x90<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x91<<8)+S_FLAG               +N_FLAG   ,
  (0x92<<8)+S_FLAG               +N_FLAG   ,
  (0x93<<8)+S_FLAG            +V_FLAG+N_FLAG   ,
  (0x34<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x35<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x36<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x37<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x38<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x3A<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x3B<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x3C<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x3D<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x3E<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x3F<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x40<<8)                  +N_FLAG+C_FLAG,
  (0x41<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x42<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x43<<8)                  +N_FLAG+C_FLAG,
  (0x44<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x45<<8)                  +N_FLAG+C_FLAG,
  (0x46<<8)                  +N_FLAG+C_FLAG,
  (0x47<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x48<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x49<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x4A<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x4B<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x4C<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x4D<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x4E<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x4F<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x50<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x51<<8)                  +N_FLAG+C_FLAG,
  (0x52<<8)                  +N_FLAG+C_FLAG,
  (0x53<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x54<<8)                  +N_FLAG+C_FLAG,
  (0x55<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x56<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x57<<8)                  +N_FLAG+C_FLAG,
  (0x58<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x59<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x5A<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x5B<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x5C<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x5D<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x5E<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x5F<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x60<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x61<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x62<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x63<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x64<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x65<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x66<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x67<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x68<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x69<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x6A<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x6B<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x6C<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x6D<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x6E<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x6F<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x70<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x71<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x72<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x73<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x74<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x75<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x76<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x77<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x78<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x79<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x7A<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x7B<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x7C<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x7D<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x7E<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x7F<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x80<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x81<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x82<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x83<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x84<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x85<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x86<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x87<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x88<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x89<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0x8A<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x8B<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x8C<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x8D<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x8E<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x8F<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x90<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x91<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x92<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x93<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x94<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x95<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x96<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x97<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x98<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0x99<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x9A<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x9B<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x9C<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x9D<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x9E<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x9F<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xA0<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xA1<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xA2<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xA3<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xA4<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xA5<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xA6<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xA7<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xA8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0xA9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xAA<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xAB<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xAC<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xAD<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xAE<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xAF<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xB0<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xB1<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xB2<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xB3<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xB4<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xB5<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xB6<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xB7<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xB8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xB9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0xBA<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xBB<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xBC<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xBD<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xBE<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xBF<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xC0<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xC1<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xC2<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xC3<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xC4<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xC5<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xC6<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xC7<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xC8<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0xC9<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xCA<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xCB<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xCC<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xCD<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xCE<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xCF<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xD0<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xD1<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xD2<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xD3<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xD4<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xD5<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xD6<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0xD7<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0xD8<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xD9<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0xDA<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xDB<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xDC<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xDD<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xDE<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xDF<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xE0<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xE1<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xE2<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xE3<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xE4<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xE5<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xE6<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xE7<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xE8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xE9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0xEA<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xEB<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xEC<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xED<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xEE<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xEF<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xf0<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xf1<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xf2<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xf3<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xf4<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xf5<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xf6<<8)+S_FLAG   +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0xf7<<8)+S_FLAG   +U5_FLAG         +N_FLAG+C_FLAG,
  (0xf8<<8)+S_FLAG   +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0xf9<<8)+S_FLAG   +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xfA<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xfB<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xfC<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0xfD<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xfE<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0xfF<<8)+S_FLAG   +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x00<<8)   +Z_FLAG         +V_FLAG+N_FLAG+C_FLAG,
  (0x01<<8)                  +N_FLAG+C_FLAG,
  (0x02<<8)                  +N_FLAG+C_FLAG,
  (0x03<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x04<<8)                  +N_FLAG+C_FLAG,
  (0x05<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x06<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x07<<8)                  +N_FLAG+C_FLAG,
  (0x08<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x09<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x0A<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x0B<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x0C<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x0D<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x0E<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x0F<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x10<<8)                  +N_FLAG+C_FLAG,
  (0x11<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x12<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x13<<8)                  +N_FLAG+C_FLAG,
  (0x14<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x15<<8)                  +N_FLAG+C_FLAG,
  (0x16<<8)                  +N_FLAG+C_FLAG,
  (0x17<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x18<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x19<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x1A<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x1B<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x1C<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x1D<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x1E<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x1F<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x20<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x21<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x22<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x23<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x24<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x25<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x26<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x27<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x28<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x29<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x2A<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x2B<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x2C<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x2D<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x2E<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x2F<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x30<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x31<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x32<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x33<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x34<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x35<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x36<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x37<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x38<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x39<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x3A<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x3B<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x3C<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x3D<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x3E<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x3F<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x40<<8)                  +N_FLAG+C_FLAG,
  (0x41<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x42<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x43<<8)                  +N_FLAG+C_FLAG,
  (0x44<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x45<<8)                  +N_FLAG+C_FLAG,
  (0x46<<8)                  +N_FLAG+C_FLAG,
  (0x47<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x48<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x49<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x4A<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x4B<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x4C<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x4D<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x4E<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x4F<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x50<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x51<<8)                  +N_FLAG+C_FLAG,
  (0x52<<8)                  +N_FLAG+C_FLAG,
  (0x53<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x54<<8)                  +N_FLAG+C_FLAG,
  (0x55<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x56<<8)               +V_FLAG+N_FLAG+C_FLAG,
  (0x57<<8)                  +N_FLAG+C_FLAG,
  (0x58<<8)            +U3_FLAG   +N_FLAG+C_FLAG,
  (0x59<<8)            +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x5A<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x5B<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x5C<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x5D<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x5E<<8)         +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x5F<<8)         +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x60<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x61<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x62<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x63<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x64<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x65<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x66<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x67<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x68<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x69<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x6A<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x6B<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x6C<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x6D<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x6E<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x6F<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x70<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x71<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x72<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x73<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x74<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x75<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x76<<8)      +U5_FLAG         +N_FLAG+C_FLAG,
  (0x77<<8)      +U5_FLAG      +V_FLAG+N_FLAG+C_FLAG,
  (0x78<<8)      +U5_FLAG   +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x79<<8)      +U5_FLAG   +U3_FLAG   +N_FLAG+C_FLAG,
  (0x7A<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x7B<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x7C<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x7D<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x7E<<8)      +U5_FLAG+H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x7F<<8)      +U5_FLAG+H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x80<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x81<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x82<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x83<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x84<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x85<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x86<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x87<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x88<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x89<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0x8A<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x8B<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x8C<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x8D<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x8E<<8)+S_FLAG      +H_FLAG+U3_FLAG+V_FLAG+N_FLAG+C_FLAG,
  (0x8F<<8)+S_FLAG      +H_FLAG+U3_FLAG   +N_FLAG+C_FLAG,
  (0x90<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x91<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x92<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x93<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x94<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x95<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x96<<8)+S_FLAG            +V_FLAG+N_FLAG+C_FLAG,
  (0x97<<8)+S_FLAG               +N_FLAG+C_FLAG,
  (0x98<<8)+S_FLAG         +U3_FLAG   +N_FLAG+C_FLAG,
  (0x99<<8)+S_FLAG         +U3_FLAG+V_FLAG+N_FLAG+C_FLAG
};

#endif

